Package: BinaryDosage
Title: Creates, Merges, and Reads Binary Dosage Files
Version: 2.0.0
Authors@R: 
    c(person(given = "John",
             family = "Morrison",
             role = c("aut", "cre"),
             email = "jmorr@usc.edu"),
      person(given = "NCI",
             role = "fnd",
             comment = "CA196559"),
      person(given = "NCI",
             role = "fnd",
             comment = "CA201407"),
      person(given = "NIEHS",
             role = "fnd",
             comment = "ES007048"),
      person(given = "NHLBI",
             role = "fnd",
             comment = "HL115606"))
Description: Tools to create binary dosage files from either VCF
    or GEN files, merge binary dosage files, and read
    binary dosage files. Binary dosage files tend to have quicker
    read times than VCF and GEN formats. There is a small increase
    in size compared to compressed VCF and GEN files.
License: GPL-3
Encoding: UTF-8
Suggests: knitr, rmarkdown, testthat (>= 2.1.0), vcfppR
VignetteBuilder: knitr
RoxygenNote: 7.3.3
LinkingTo: Rcpp
Imports: Rcpp, digest, prodlim
NeedsCompilation: yes
Packaged: 2026-04-24 19:52:03 UTC; jmorr
Author: John Morrison [aut, cre],
  NCI [fnd] (CA196559),
  NCI [fnd] (CA201407),
  NIEHS [fnd] (ES007048),
  NHLBI [fnd] (HL115606)
Maintainer: John Morrison <jmorr@usc.edu>
Repository: CRAN
Date/Publication: 2026-04-29 18:10:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-04-29 20:35:11 UTC; unix
Archs: BinaryDosage.so.dSYM

# Generated by roxygen2: do not edit by hand

export(bdapply)
export(bdmerge)
export(closebd5con)
export(genapply)
export(gentobd)
export(getaaf)
export(getbd5snp)
export(getbd5snp_buf)
export(getbd5snp_con)
export(getbdinfo)
export(getgeninfo)
export(getmaf)
export(getrsq)
export(getsnp)
export(getvcfinfo)
export(mergebd)
export(openbd5con)
export(subsetbd)
export(updatebd)
export(vcfapply)
export(vcftobd)
export(vcftobdlegacy)
importFrom(Rcpp,sourceCpp)
importFrom(digest,digest)
importFrom(prodlim,row.match)
importFrom(stats,var)
importFrom(utils,read.table)
useDynLib(BinaryDosage, .registration = TRUE)

BinaryDosage: Creates, Merges, and Reads Binary Dosage Files
================

<!-- badges: start -->

[![AppVeyor build
status](https://ci.appveyor.com/api/projects/status/github/USCbiostats/BinaryDosage?branch=master&svg=true)](https://ci.appveyor.com/project/USCbiostats/BinaryDosage)
[![Codecov test
coverage](https://codecov.io/gh/USCbiostats/BinaryDosage/branch/master/graph/badge.svg)](https://app.codecov.io/gh/USCbiostats/BinaryDosage?branch=master)
<!-- badges: end -->

# Binary Dosage Files

### Introduction

Genotype imputation is an essential tool in genomics, enabling
association testing with markers not directly genotyped, increasing
statistical power, and facilitating data pooling between studies that
employ different genotyping platforms. Two commonly used software
packages for imputation are
[minimac](https://genome.sph.umich.edu/wiki/Minimac) and
[Impute2](http://mathgen.stats.ox.ac.uk/impute/impute_v2.html).
Furthermore, services such as the [Michigan Imputation
Server](https://imputationserver.sph.umich.edu/index.html) have made
genotype imputation much more accessible and streamlined.

While a number of software options are available for analyses of imputed
data (e.g. PLINK, EPACTS), fewer are available for Genomewide Gene x
Environment Interaction Scan (GWIS). Furthermore, data management tasks
such as parsing, subsetting, and merging, while manageable in smaller
studies, quickly become unwieldy and prohibitively slow with very large
samples sizes. We aim to address these limitations by converting
imputation outputs into a binary dosage file. The benefits of a binary
format are two fold - decreased hard drive storage requirements
(compared to a VCF file), and speed of parsing/analyses. The
BinaryDosage package contains functions to convert VCF and Impute2
formatted files into binary dosage files, along with functions to merge
samples.

For GWAS/GWIS analysis of BinaryDosage files, please refer to the
[**GxEScanR**](https://github.com/USCbiostats/GxEScanR) package.

### Description

##### Binary dosage data sets contain the following information:

- Sample information
  - Family ID
  - Subject ID
- SNP information
  - Chromosome number\
  - SNP ID\
  - Location in base pairs\
  - Reference allele\
  - Alternate allele\
- Genetic information
  - Dosage values
  - Genotype probabilities, Pr(*g=0*), Pr(*g=1*), Pr(*g=2*)

There are 5 formats for a binary dosage data set. Data sets in formats
1, 2, and 3 have 3 files, a sample information file, a SNP information
file, and a genetic information file. Data sets in format 4 have just 1
file. Format 5 uses per-SNP gzip compression and stores metadata in a
companion RDS file (`.bdinfo`). This file contains all the information
listed above and may contain the following information.

**Note:** Format 5 is the recommended format for new data sets.

- Additional SNP information
  - Alternate allele frequency
  - Minor allele frequency
  - Average call rate
  - Imputation r squared
- Merging information
  - Number of data sets merged
  - Sample size of each data set merged

### Functions

#### Format 5 (recommended)

- **vcftobd** - Converts a bgzipped VCF file to a Format 5 binary dosage
  data set (requires vcfppR)
- **getbd5snp** - Reads a single SNP from a Format 5 file by index or
  ID; returns a list with dosage and genotype probabilities
- **getbd5snp_buf** - Like **getbd5snp** but writes results into
  pre-allocated vectors, avoiding repeated memory allocation in loops
- **openbd5con** - Opens a persistent connection to a Format 5 file for
  use with **getbd5snp_con**
- **closebd5con** - Explicitly closes a connection opened by
  **openbd5con**
- **getbd5snp_con** - Like **getbd5snp_buf** but reuses an open
  connection, eliminating file open/close overhead per SNP
- **updatebd** - Converts a legacy format (1–4) binary dosage file to
  Format 5
- **subsetbd** - Creates a new Format 5 file containing a subset of SNPs
  and/or subjects from any binary dosage file (formats 1–5)
- **mergebd** - Merges two or more Format 5 files into a single Format 5
  file

#### Legacy formats (1–4)

- **vcftobdlegacy** - Converts a VCF file to a legacy format (1–4)
  binary dosage data set (deprecated; use **vcftobd** instead)
- **gentobd** - Converts a GEN (impute2) file to a binary dosage data
  set
- **bdmerge** - Merges multiple legacy binary dosage data sets into a
  single data set
- **getbdinfo** - Creates an R list containing information about a
  binary dosage data set (required for **getsnp** and **bdapply**)
- **getvcfinfo** - Creates an R list containing information about a VCF
  file (required for **vcfapply**)
- **getgeninfo** - Creates an R list containing information about a GEN
  file (required for **genapply**)
- **bdapply** - Applies a function to the data for each SNP in a binary
  dosage file (requires list returned by **getbdinfo**)
- **vcfapply** - Applies a function to the data for each SNP in a VCF
  file (requires list returned by **getvcfinfo**)
- **genapply** - Applies a function to the data for each SNP in a GEN
  file (requires list returned by **getgeninfo**)
- **getsnp** - Returns dosage and genotype probabilities for a single
  SNP from a binary dosage file

# Installation

1.  Install the [devtools](https://github.com/r-lib/devtools) package
2.  Install the
    [BinaryDosage](https://github.com/USCbiostats/BinaryDosage) package
    directly from the USCbiostats repository on GitHub:

``` r
remove.packages("BinaryDosage")
devtools::install_github("https://github.com/USCbiostats/BinaryDosage")

library(BinaryDosage)
```

To install the package with vignettes built, use the `build_vignettes`
option:

``` r
devtools::install_github("https://github.com/USCbiostats/BinaryDosage", build_vignettes = TRUE)

library(BinaryDosage)
```

# Usage

#### General Workflow

The general workflow for using binary dosage data sets is as follows:

- Convert VCF or GEN files to a binary dosage data set
  - Note: When converting a VCF file to a binary dosage data set, the
    information file associated with the vcf can be used to add
    additional imputation information to the binary dosage data set
  - Note: When converting a GEN file to a binary dosage data set, the
    subject IDs can either be on the first line of the GEN file or in a
    separate sample file
- Merge binary dosage datasets into a single data set
- Apply a function to each SNP in the data set using bdapply
- Extract SNPs for further analysis

#### Examples

The examples below use the default values for the functions. More
information about the functions and their options can be found using the
help files and the vignettes.

##### Example files

In the examples below the input files are included with the binary
dosage package and the output files are written to R temporary files. In
normal use, the user would provide the names of the input and output
files.

###### Input files

Example datasets *set1a.vcf* and *set1b.vcf* are representative of VCF
output files obtained from the Michigan Imputation Server. An
information file is also included for each set, *set1a.info* and
*set1b.info*.

Example datasets *set3a.imp* and *set3b.imp* are representative of files
return by the Impute imputation software. For GEN files the subject IDs
are contained in separated files. For this example these are
*set3a.sample* and *set3b.sample*.

The VCF and GEN files contain the same data. These files are in the
extdata directory of the BinaryDosage package. These sets contain the
following:

| Set   | Number of subjects | Number of SNPS |
|-------|--------------------|----------------|
| 1a,3a | 60                 | 10             |
| 1b,3b | 40                 | 10             |

Since these files are distributed with the Binary Dosage package, it is
necessary to get the complete file name and path for use in the
following examples. The following code gets all the file names needed
for the examples.

``` r
library(BinaryDosage)

# Get the file names for the VCF and information files
vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")
vcf1ainfo <- system.file("extdata", "set1a.info", package = "BinaryDosage")
vcf1bfile <- system.file("extdata", "set1b.vcf", package = "BinaryDosage")
vcf1binfo <- system.file("extdata", "set1b.info", package = "BinaryDosage")

# Get the file names for the GEN and sample files
gen3afile <- system.file("extdata", "set3a.imp", package = "BinaryDosage")
gen3asample <- system.file("extdata", "set3a.sample", package = "BinaryDosage")
gen3bfile <- system.file("extdata", "set3b.imp", package = "BinaryDosage")
gen3bsample <- system.file("extdata", "set3b.sample", package = "BinaryDosage")
```

###### Output files

The binary dosage output files will be written to temporary files. There
needs to be only one output file per data set because the examples use
the default format value of 4. The following code creates these
temporary output files.

``` r
# The output files for set 1
bdfile1a <- tempfile()
bdfile1b <- tempfile()
mergebd1 <- tempfile()

# The output files for set 3
bdfile3a <- tempfile()
bdfile3b <- tempfile()
mergebd3 <- tempfile()
```

##### Converting VCF files to a binary dosage data set

Converting a VCF file into a binary dosage file is simple. The user
passes the names of the VCF and information files along with the name
for the binary dosage file to the
<span style="font-family:Courier">vcftobdlegacy</span> function. There
are some options available for the
<span style="font-family:Courier">vcftobdlegacy</span> functions such as
using gz compressed files vcf files. More information about these
options can be found using the help files or reading the vignette
<span style="font-family:Courier">usingvcffiles</span>.

The following commands convert VCF data sets 1a and 1b into the binary
dosage format.

``` r
vcftobdlegacy(vcffiles = c(vcf1afile, vcf1ainfo), bdfiles = bdfile1a, format = 4)
vcftobdlegacy(vcffiles = c(vcf1bfile, vcf1binfo), bdfiles = bdfile1b, format = 4)
```

##### Converting GEN files to the Binary Dosage Format and Merging into one data set

Converting GEN files to binary dosage files is a little more difficult
than converting VCF files. This is because GEN files aren’t as strictly
formatted as VCF files. The user needs to have knowledge of how the GEN
file is formatted. More information on this can be found in the help
files and the vignette
<span style="font-family:Courier">usinggenfiles</span>.

In the example GEN file, the first column contains “--” for each SNP and
the second column contains the SNP ID in the format

<span style="font-family:Courier">\<chromosome\>:\<location\>\_\<reference
allele\>\_\<alternate allele\></span>

Because of this formatting, the function
<span style="font-family:Courier">gentobd</span> requires the
<span style="font-family:Courier">snpcolumns</span> parameter to have
the value <span style="font-family:Courier">c(0L, 2L:5L)</span>. To
convert the GEN data sets to binary dosage data sets, the names of the
input and output files are passed to
<span style="font-family:Courier">gentobd</span> along with the needed
value for snpcolumns.

The following commands convert the two GEN files into binary dosage
files.

``` r
gentobd(genfiles = c(gen3afile, gen3asample), snpcolumns = c(0L, 2L:5L), bdfiles = bdfile3a)
gentobd(genfiles = c(gen3bfile, gen3bsample), snpcolumns = c(0L, 2L:5L), bdfiles = bdfile3b)
```

##### Merging binary dosage files

Merging binary dosage files is done by SNP ID. The files to merge cannot
have the same subject IDs. See the vignette
<span style="font-family:Courier">usingbdfiles</span> for more
information. In this example we are assuming two separate groups of
subjects were imputed separately to the same reference panel.

To merge files, the user calls the
<span style="font-family:Courier">bdmerge</span> function and passes the
names of the files to merge along with a file name for the merged data
set. Other options exist for bdmerge and can be found in the help files
and the vignette <span style="font-family:Courier">mergingfiles</span>.

The following code first merges the binary files bdfile1a and bdfile1b
created from the VCF files into a single file, mergedbd1, and then does
the analogous action for the binary dosage files created from the GEN
files.

``` r
bdmerge(mergefiles = mergebd1, bdfiles = c(bdfile1a, bdfile1b))
bdmerge(mergefiles = mergebd3, bdfiles = c(bdfile3a, bdfile3b))
```

##### Applying a function to all the SNPs in a data set

Once binary dosage files have been created, a function can be applied to
all the SNPs in a file.

###### Defining the function

The function applied to the SNPs in a binary dosage file must have the
following four parameters, dosage, p0, p1, and p2. These are the dosage,
Pr(*g=0*), Pr(*g=1*), and Pr(*g=2*), respectively. Other parameters can
also be passed. For more information on defining the function see the
vignette <span style="font-family:Courier">usingbdfiles</span>.

The following code defines a function to calculate the alternate allele
frequency.

``` r
calculateaaf <- function(dosage, p0, p1, p2) {
    return(mean(dosage, na.rm = TRUE)/2)
}
```

###### Applying the function

To apply the function the user needs to call bdapply and pass
information about the binary dosage file and the function. The
information about the dosage file is obtained by calling the function
getbdinfo. If the user is going to call bdapply multiple times, the user
may wish to save the results of getbdinfo.

``` r
mergebd1info <- getbdinfo(mergebd1)
aaf1 <- bdapply(mergebd1info, calculateaaf)

mergebd3info <- getbdinfo(mergebd3)
aaf3 <- bdapply(mergebd3info, calculateaaf)
```

###### Checking the results

Since the VCF and GEN files contain the same information, the alternate
allele frequencies should be the same. The following code creates a data
frame with the SNP IDs and the alternate allele frequencies for both
data sets.

``` r
aaf <- cbind(mergebd1info$snps, aaf_set1 = unlist(aaf1), aaf_set3 = unlist(aaf3))
```

Here is a table showing the results.

| chromosome | location | snpid       | reference | alternate | aaf_set1 | aaf_set3 |
|:-----------|---------:|:------------|:----------|:----------|---------:|---------:|
| 1          |    10000 | 1:10000:C:A | C         | A         |   0.3527 |   0.3527 |
| 1          |    11000 | 1:11000:T:C | T         | C         |   0.0135 |   0.0135 |
| 1          |    12000 | 1:12000:T:C | T         | C         |   0.2400 |   0.2400 |
| 1          |    13000 | 1:13000:T:C | T         | C         |   0.3375 |   0.3375 |
| 1          |    14000 | 1:14000:G:C | G         | C         |   0.1901 |   0.1901 |
| 1          |    15000 | 1:15000:A:C | A         | C         |   0.5627 |   0.5627 |
| 1          |    16000 | 1:16000:G:A | G         | A         |   0.4569 |   0.4569 |
| 1          |    17000 | 1:17000:C:A | C         | A         |   0.4578 |   0.4578 |
| 1          |    18000 | 1:18000:C:G | C         | G         |   0.2591 |   0.2591 |
| 1          |    19000 | 1:19000:T:G | T         | G         |   0.2431 |   0.2431 |

##### Extracting a SNP from the data set

After doing an analysis, the user may want to extract a SNP from the
data set for further analysis. This can be done using the getsnp
function. By default the function returns a list with the dosage values
for all the subjects. The genotype probabilities can be added to the
list by setting the dosageonly option to FALSE. See the help files or
the vignette <span style="font-family:Courier">usingbdfiles</span> for
more information.

The following code extracts the 6th SNP from both the binary dosage data
sets generated above.

``` r
# Get the dosage values for the 6th SNP
set1snp6 <- getsnp(mergebd1info, 6)
# Get the dosage values for the 6th SNP
set3snp6 <- getsnp(mergebd3info, 6)
```

The results from the above lines were merged into a data frame with the
subject IDs. Here are the first 10 lines of the data frame.

| subjectid | set1snp6 | set3snp6 |
|:----------|---------:|---------:|
| I1        |    1.000 |    1.000 |
| I2        |    1.849 |    1.849 |
| I3        |    1.000 |    1.000 |
| I4        |    2.000 |    2.000 |
| I5        |    1.046 |    1.046 |
| I6        |    1.915 |    1.915 |
| I7        |    2.000 |    2.000 |
| I8        |    2.000 |    2.000 |
| I9        |    1.000 |    1.000 |
| I10       |    1.000 |    1.000 |

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(BinaryDosage)

## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
bd1afile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
bd1bfile <- system.file("extdata", "vcf1b.bdose", package = "BinaryDosage")
bd1file <- tempfile()

bdmerge(mergefiles = bd1file, bdfiles = c(bd1afile, bd1bfile))

bd1ainfo <- getbdinfo(bd1afile)
bd1binfo <- getbdinfo(bd1bfile)
bd1info <- getbdinfo(bd1file)

nrow(bd1ainfo$samples)
nrow(bd1binfo$samples)
nrow(bd1info$samples)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, echo = FALSE------------------------------------------------------
library(BinaryDosage)

## ----setup_files, message = FALSE, warning = FALSE----------------------------
bdose_full <- file.path(tempdir(), "full.bdose")

if (requireNamespace("vcfppR", quietly = TRUE)) {
  vcftobd(vcffile    = system.file("extdata", "set1a.vcf.gz", package = "BinaryDosage"),
          bdose_file = bdose_full)
} else {
  updatebd(bdfiles    = system.file("extdata", "vcf1a.bdose", package = "BinaryDosage"),
           bdose_file = bdose_full)
}
bd_full <- getbdinfo(bdose_full)

cat("Subjects:", nrow(bd_full$samples), "\n")
cat("SNPs:    ", nrow(bd_full$snps),    "\n")

## ----subject_merge, message = FALSE, warning = FALSE--------------------------
bdose_a   <- file.path(tempdir(), "set_a.bdose")
bdose_b   <- file.path(tempdir(), "set_b.bdose")
bdose_out <- file.path(tempdir(), "merged_subjects.bdose")

sids <- bd_full$samples$sid

subsetbd(bdfiles    = bdose_full,
         bdose_file = bdose_a,
         subjectids = sids[1:30])

subsetbd(bdfiles    = bdose_full,
         bdose_file = bdose_b,
         subjectids = sids[31:60])

mergebd(bdose_files = c(bdose_a, bdose_b),
        bdose_file  = bdose_out)

bd_a   <- getbdinfo(bdose_a)
bd_b   <- getbdinfo(bdose_b)
bd_out <- getbdinfo(bdose_out)

cat("File A subjects:", nrow(bd_a$samples),   "\n")
cat("File B subjects:", nrow(bd_b$samples),   "\n")
cat("Merged subjects:", nrow(bd_out$samples), "\n")
cat("Merged SNPs:    ", nrow(bd_out$snps),    "\n")

## ----subject_order------------------------------------------------------------
knitr::kable(bd_out$samples, caption = "Subjects in merged file")

## ----snp_merge, message = FALSE, warning = FALSE------------------------------
bdose_snp_a   <- file.path(tempdir(), "snp_a.bdose")
bdose_snp_b   <- file.path(tempdir(), "snp_b.bdose")
bdose_snp_out <- file.path(tempdir(), "merged_snps.bdose")

locs <- bd_full$snps$location

subsetbd(bdfiles    = bdose_full,
         bdose_file = bdose_snp_a,
         locations  = locs[1:5])

subsetbd(bdfiles    = bdose_full,
         bdose_file = bdose_snp_b,
         locations  = locs[6:10])

mergebd(bdose_files = c(bdose_snp_a, bdose_snp_b),
        bdose_file  = bdose_snp_out)

bd_snp_a   <- getbdinfo(bdose_snp_a)
bd_snp_b   <- getbdinfo(bdose_snp_b)
bd_snp_out <- getbdinfo(bdose_snp_out)

cat("File A SNPs:    ", nrow(bd_snp_a$snps),   "\n")
cat("File B SNPs:    ", nrow(bd_snp_b$snps),   "\n")
cat("Merged SNPs:    ", nrow(bd_snp_out$snps),  "\n")
cat("Merged subjects:", nrow(bd_snp_out$samples), "\n")

## ----snp_order----------------------------------------------------------------
knitr::kable(bd_snp_out$snps, caption = "SNPs in merged file")

## ----cleanup, include = FALSE-------------------------------------------------
unlink(c(bdose_full,    paste0(bdose_full,    ".bdi"),
         bdose_a,       paste0(bdose_a,       ".bdi"),
         bdose_b,       paste0(bdose_b,       ".bdi"),
         bdose_out,     paste0(bdose_out,     ".bdi"),
         bdose_snp_a,   paste0(bdose_snp_a,   ".bdi"),
         bdose_snp_b,   paste0(bdose_snp_b,   ".bdi"),
         bdose_snp_out, paste0(bdose_snp_out, ".bdi")))

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, echo = FALSE------------------------------------------------------
library(BinaryDosage)

## ----files--------------------------------------------------------------------
vcf_file   <- system.file("extdata", "set1a.vcf.gz", package = "BinaryDosage")
bd4_file   <- system.file("extdata", "vcf1a.bdose",  package = "BinaryDosage")
bdose_file <- file.path(tempdir(), "set1a.bdose")

## ----convert, eval = TRUE, echo = TRUE, message = FALSE, warning = FALSE------
if (requireNamespace("vcfppR", quietly = TRUE)) {
  vcftobd(vcffile = vcf_file, bdose_file = bdose_file)
} else {
  updatebd(bdfiles = bd4_file, bdose_file = bdose_file)
}

## ----convert_region, eval = F, echo = T---------------------------------------
# vcftobd(vcffile    = vcf_file,
#          bdose_file = bdose_file,
#          region     = "1:10000-15000")

## ----load_info, eval = TRUE, echo = T, message = F, warning = F---------------
bd5 <- getbdinfo(bdose_file)

## ----meta, eval = TRUE, echo = T----------------------------------------------
cat("Class:          ", class(bd5), "\n")
cat("Uses family ID: ", bd5$usesfid, "\n")
cat("One chromosome: ", bd5$onechr, "\n")
cat("SNP ID format:  ", bd5$snpidformat, "\n")
cat("Format:         ", bd5$additionalinfo$format, "\n")
cat("Header size:    ", bd5$additionalinfo$headersize, "bytes\n")

## ----samples, eval = TRUE, echo = T-------------------------------------------
cat("Number of samples:", nrow(bd5$samples), "\n")
knitr::kable(head(bd5$samples, 5), caption = "First 5 samples")

## ----snps, eval = TRUE, echo = T----------------------------------------------
cat("Number of SNPs:", nrow(bd5$snps), "\n")
knitr::kable(bd5$snps, caption = "SNP table")

## ----indices, eval = TRUE, echo = T-------------------------------------------
cat("Indices (bytes):", bd5$indices, "\n")

## ----snp_by_index, eval = TRUE, echo = T, message = F, warning = F------------
snp1 <- getbd5snp(bd5info = bd5, snp = 1L)

snp1_df <- data.frame(
  SampleID = bd5$samples$sid,
  Dosage   = round(snp1$dosage, 4),
  P_00     = round(snp1$p0,     4),
  P_01     = round(snp1$p1,     4),
  P_11     = round(snp1$p2,     4)
)

knitr::kable(head(snp1_df, 10),
             caption = paste("SNP", bd5$snps$snpid[1], "— first 10 subjects"))

## ----snp_by_id, eval = TRUE, echo = T, message = F, warning = F---------------
snp_id <- "1:12000:T:C"
snp3   <- getbd5snp(bd5info = bd5, snp = snp_id)

snp3_df <- data.frame(
  SampleID = bd5$samples$sid,
  Dosage   = round(snp3$dosage, 4),
  P_00     = round(snp3$p0,     4),
  P_01     = round(snp3$p1,     4),
  P_11     = round(snp3$p2,     4)
)

knitr::kable(head(snp3_df, 10),
             caption = paste("SNP", snp_id, "— first 10 subjects"))

## ----apply_bdapply, eval = TRUE, echo = T, message = F, warning = F-----------
getaaf <- function(dosage, p0, p1, p2) mean(dosage, na.rm = TRUE) / 2

aaf_list  <- bdapply(bdinfo = bd5, func = getaaf)
aaf_table <- data.frame(snpid = bd5$snps$snpid,
                        aaf   = round(unlist(aaf_list), 4))

knitr::kable(aaf_table, caption = "Alternate allele frequencies via bdapply")

## ----apply_con, eval = TRUE, echo = T, message = F, warning = F---------------
n_snps <- nrow(bd5$snps)
n_samp <- nrow(bd5$samples)
dosage <- numeric(n_samp)
p0     <- numeric(n_samp)
p1     <- numeric(n_samp)
p2     <- numeric(n_samp)

con <- openbd5con(bd5)
aaf <- numeric(n_snps)
for (i in seq_len(n_snps)) {
  getbd5snp_con(bd5info = bd5, snp = i,
                dosage = dosage, p0 = p0, p1 = p1, p2 = p2,
                bd5con = con)
  aaf[i] <- mean(dosage, na.rm = TRUE) / 2
}
closebd5con(con)

knitr::kable(data.frame(snpid = bd5$snps$snpid, aaf = round(aaf, 4)),
             caption = "Alternate allele frequencies via getbd5snp_con")

## ----cleanup, include = FALSE, eval = TRUE------------------------------------
unlink(c(bdose_file, paste0(bdose_file, ".bdi")))

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, echo=FALSE--------------------------------------------------------
library(BinaryDosage)

## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")
bdfile1a_woinfo <- tempfile()
vcftobdlegacy(vcffiles = vcf1afile, bdfiles = bdfile1a_woinfo)


## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")
vcf1ainfo <- system.file("extdata", "set1a.info", package = "BinaryDosage")
bdfile1a_winfo <- tempfile()
vcftobdlegacy(vcffiles = c(vcf1afile, vcf1ainfo), bdfiles = bdfile1a_winfo)


## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
bdinfo1a_woinfo <- getbdinfo(bdfiles = bdfile1a_woinfo)
bdinfo1a_woinfo$snpinfo

bdinfo1a_winfo <- getbdinfo(bdfiles = bdfile1a_winfo)
knitr::kable(data.frame(bdinfo1a_winfo$snpinfo), caption = "bdinfo1a_winfo$snpinfo")

## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
vcf1afile_gz <- system.file("extdata", "set1a.vcf.gz", package = "BinaryDosage")
vcf1ainfo <- system.file("extdata", "set1a.info", package = "BinaryDosage")

bdfile1a_woinfo_gz <- tempfile()
vcftobdlegacy(vcffiles = vcf1afile_gz, bdfiles = bdfile1a_woinfo_gz)

bdfile1a_winfo_gz <- tempfile()
vcftobdlegacy(vcffiles = c(vcf1afile_gz, vcf1ainfo), bdfiles = bdfile1a_winfo_gz)


## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------

bdinfo1a_woinfo_gz <- getbdinfo(bdfiles = bdfile1a_woinfo_gz)
bdinfo1a_winfo_gz <- getbdinfo(bdfiles = bdfile1a_winfo_gz)

aaf1a_woinfo <- unlist(bdapply(bdinfo = bdinfo1a_woinfo, getaaf))
aaf1a_winfo <- unlist(bdapply(bdinfo = bdinfo1a_winfo, getaaf))
aaf1a_woinfo_gz <- unlist(bdapply(bdinfo = bdinfo1a_woinfo_gz, getaaf))
aaf1a_winfo_gz <- unlist(bdapply(bdinfo = bdinfo1a_winfo_gz, getaaf))

aaf1a <- data.frame(SNPID = bdinfo1a_woinfo$snps$snpid,
                    aaf1a_woinfo = aaf1a_woinfo,
                    aaf1a_winfo = aaf1a_winfo,
                    aaf1a_woinfo_gz = aaf1a_woinfo_gz,
                    aaf1a_winfo_gz = aaf1a_winfo_gz)

knitr::kable(aaf1a, caption = "Alternate Allele Frequencies", digits = 4)


## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
vcf2afile <- system.file("extdata", "set2a.vcf", package = "BinaryDosage")
bdfile2a <- tempfile()

vcftobdlegacy(vcffiles = vcf2afile, bdfiles = bdfile2a)

bdinfo2a <- getbdinfo(bdfiles = bdfile2a)
snp1_2a <- data.frame(getsnp(bdinfo = bdinfo2a, snp = 1L, dosageonly = FALSE))

snp1 <- cbind(SubjectID = bdinfo2a$samples$sid, snp1_2a)

knitr::kable(snp1[1:10,], caption = "Dosage and Genotype Probabilities")

## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
vcf1brsfile <- system.file("extdata", "set1b_rssnp.vcf", package = "BinaryDosage")
bdfile1b.snpid0 <- tempfile()
bdfile1b.snpid1 <- tempfile()
bdfile1b.snpid2 <- tempfile()
bdfile1b.snpidm1 <- tempfile()

vcftobdlegacy(vcffiles = vcf1brsfile, bdfiles = bdfile1b.snpid0)
vcftobdlegacy(vcffiles = vcf1brsfile, bdfiles = bdfile1b.snpid1, snpidformat = 1)
vcftobdlegacy(vcffiles = vcf1brsfile, bdfiles = bdfile1b.snpid2, snpidformat = 2)
vcftobdlegacy(vcffiles = vcf1brsfile, bdfiles = bdfile1b.snpidm1, snpidformat = -1)

bdinfo1b.snpid0 <- getbdinfo(bdfiles = bdfile1b.snpid0)
bdinfo1b.snpid1 <- getbdinfo(bdfiles = bdfile1b.snpid1)
bdinfo1b.snpid2 <- getbdinfo(bdfiles = bdfile1b.snpid2)
bdinfo1b.snpidm1 <- getbdinfo(bdfiles = bdfile1b.snpidm1)

snpnames <- data.frame(format0 = bdinfo1b.snpid0$snps$snpid,
                       format1 = bdinfo1b.snpid1$snps$snpid,
                       format2 = bdinfo1b.snpid2$snps$snpid,
                       formatm1 = bdinfo1b.snpidm1$snps$snpid)

knitr::kable(snpnames, caption = "SNP Names by Format")


## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")
bdfile1a_calcinfo <- tempfile()
vcftobdlegacy(vcffiles = vcf1afile, bdfiles = bdfile1a_calcinfo, bdoptions = c("aaf", "maf", "rsq"))

bdcalcinfo <- getbdinfo(bdfile1a_calcinfo)

snpinfo <- data.frame(aaf_info = bdinfo1a_winfo$snpinfo$aaf,
                      aaf_calc = bdcalcinfo$snpinfo$aaf,
                      maf_info = bdinfo1a_winfo$snpinfo$maf,
                      maf_calc = bdcalcinfo$snpinfo$maf,
                      rsq_info = bdinfo1a_winfo$snpinfo$rsq,
                      rsq_calc = bdcalcinfo$snpinfo$rsq)

knitr::kable(snpinfo, caption = "Information vs Calculated Information", digits = 3)


## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")
vcfinfo <- getvcfinfo(vcf1afile, index = TRUE)
aaf <- unlist(vcfapply(vcfinfo = vcfinfo, getaaf))

altallelefreq <- data.frame(SNP = vcfinfo$snps$snpid, aafinfo = aaf1a_winfo, aafcalc = aaf)
knitr::kable(altallelefreq, caption = "Information vs Calculated aaf", digits = 3)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(BinaryDosage)

## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
bd1afile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
bd1ainfo <- getbdinfo(bdfiles = bd1afile)

## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
aaf <- unlist(bdapply(bdinfo = bd1ainfo, func = getaaf))

altallelefreq <- data.frame(SNP = bd1ainfo$snps$snpid, aafcalc = aaf)
knitr::kable(altallelefreq, caption = "Information vs Calculated aaf", digits = 3)

## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
snp3 <- data.frame(getsnp(bdinfo = bd1ainfo, "1:12000:T:C", FALSE))

knitr::kable(snp3[1:20,], caption = "SNP 1:12000:T:C", digits = 3)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, echo=FALSE--------------------------------------------------------
library(BinaryDosage)

## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
gen3bchrfile <- system.file("extdata", "set3b.chr.imp", package = "BinaryDosage")
sample3bfile <- system.file("extdata", "set3b.sample", package = "BinaryDosage")
bdfile3b_chr <- tempfile()
gentobd(genfiles = c(gen3bchrfile, sample3bfile), bdfiles = bdfile3b_chr)

bdinfo3b_chr <- getbdinfo(bdfiles = bdfile3b_chr)


## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
gen1bfile <- system.file("extdata", "set1b.imp", package = "BinaryDosage")
bdfile1b <- tempfile()
gentobd(genfiles = gen1bfile,
        bdfiles = bdfile1b,
        snpcolumns = c(1L, 3L, 2L, 4L, 5L),
        header = TRUE)

bdinfo1b <- getbdinfo(bdfiles = bdfile1b)


## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
gen3bfile <- system.file("extdata", "set3b.imp", package = "BinaryDosage")
sample3bfile <- system.file("extdata", "set3b.sample", package = "BinaryDosage")
bdfile3b <- tempfile()
gentobd(genfiles = c(gen3bfile, sample3bfile),
        bdfiles = bdfile3b,
        snpcolumns = c(0L,2L:5L))

bdinfo3b <- getbdinfo(bdfiles = bdfile3b)


## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
gen4bfile <- system.file("extdata", "set4b.imp", package = "BinaryDosage")
sample4bfile <- system.file("extdata", "set4b.sample", package = "BinaryDosage")
bdfile4b <- tempfile()
gentobd(genfiles = c(gen4bfile, sample4bfile),
        bdfiles = bdfile4b,
        snpcolumns = c(1L,2L,4L,5L,6L),
        startcolumn = 7L,
        impformat = 2L)

bdinfo4b <- getbdinfo(bdfiles = bdfile4b)


## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
gen2bfile <- system.file("extdata", "set2b.imp", package = "BinaryDosage")
sample2bfile <- system.file("extdata", "set2b.sample", package = "BinaryDosage")
bdfile2b <- tempfile()
gentobd(genfiles = c(gen2bfile, sample2bfile),
        bdfiles = bdfile2b,
        snpcolumns = c(1L,3L,2L,4L,5L),
        impformat = 1L)

bdinfo2b <- getbdinfo(bdfiles = bdfile2b)


## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
gen3bfile <- system.file("extdata", "set3b.imp", package = "BinaryDosage")
sample3bfile <- system.file("extdata", "set3b.sample", package = "BinaryDosage")
bdfile3bm1 <- tempfile()
gentobd(genfiles = c(gen3bfile, sample3bfile),
        bdfiles = bdfile3bm1,
        snpcolumns = c(-1L,2L:5L),
        chromosome = "1")

bdinfo3bm1 <- getbdinfo(bdfiles = bdfile3bm1)


## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
gen3bfile <- system.file("extdata", "set3b.imp", package = "BinaryDosage")
sample3bnhfile <- system.file("extdata", "set3bnh.sample", package = "BinaryDosage")
bdfile3bnh <- tempfile()
gentobd(genfiles = c(gen3bfile, sample3bnhfile),
        bdfiles = bdfile3bnh,
        snpcolumns = c(0L,2L:5L),
        header = c(FALSE, FALSE))

bdinfo3bnh <- getbdinfo(bdfiles = bdfile3bnh)


## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
gen4bgzfile <- system.file("extdata", "set4b.imp.gz", package = "BinaryDosage")
sample4bfile <- system.file("extdata", "set4b.sample", package = "BinaryDosage")
bdfile4bgz <- tempfile()
gentobd(genfiles = c(gen4bgzfile, sample4bfile),
        bdfiles = bdfile4bgz,
        snpcolumns = c(1L,2L,4L,5L,6L),
        startcolumn = 7L,
        impformat = 2L,
        gz = TRUE)

bdinfo4bgz <- getbdinfo(bdfiles = bdfile4bgz)


## ----eval = T, echo = T, message = F, warning = F, tidy = T-------------------
gen3bchrfile <- system.file("extdata", "set3b.chr.imp", package = "BinaryDosage")
sample3bfile <- system.file("extdata", "set3b.sample", package = "BinaryDosage")
geninfo <- getgeninfo(genfiles = c(gen3bchrfile, sample3bfile), index = TRUE)

aaf <- unlist(genapply(geninfo = geninfo, getaaf))

altallelefreq <- data.frame(SNP = geninfo$snps$snpid, aafcalc = aaf)
knitr::kable(altallelefreq, caption = "Calculated aaf", digits = 3)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, echo = FALSE------------------------------------------------------
library(BinaryDosage)

## ----setup_files, message = FALSE, warning = FALSE----------------------------
bdose_file <- file.path(tempdir(), "set1a.bdose")

if (requireNamespace("vcfppR", quietly = TRUE)) {
  vcftobd(vcffile    = system.file("extdata", "set1a.vcf.gz", package = "BinaryDosage"),
          bdose_file = bdose_file)
} else {
  updatebd(bdfiles    = system.file("extdata", "vcf1a.bdose", package = "BinaryDosage"),
           bdose_file = bdose_file)
}
bd5 <- getbdinfo(bdose_file)

n_snps <- nrow(bd5$snps)
n_samp <- nrow(bd5$samples)
cat("SNPs:", n_snps, " Samples:", n_samp, "\n")

## ----getbd5snp_index----------------------------------------------------------
snp1 <- getbd5snp(bd5info = bd5, snp = 1L)

knitr::kable(
  data.frame(
    SampleID = head(bd5$samples$sid, 10),
    Dosage   = round(head(snp1$dosage, 10), 4),
    P_00     = round(head(snp1$p0,     10), 4),
    P_01     = round(head(snp1$p1,     10), 4),
    P_11     = round(head(snp1$p2,     10), 4)
  ),
  caption = paste("SNP", bd5$snps$snpid[1], "— first 10 subjects")
)

## ----getbd5snp_id-------------------------------------------------------------
snp3 <- getbd5snp(bd5info = bd5, snp = "1:12000:T:C")

knitr::kable(
  data.frame(
    SampleID = head(bd5$samples$sid, 10),
    Dosage   = round(head(snp3$dosage, 10), 4),
    P_00     = round(head(snp3$p0,     10), 4),
    P_01     = round(head(snp3$p1,     10), 4),
    P_11     = round(head(snp3$p2,     10), 4)
  ),
  caption = "SNP 1:12000:T:C — first 10 subjects"
)

## ----getbd5snp_loop-----------------------------------------------------------
aaf <- numeric(n_snps)
for (i in seq_len(n_snps)) {
  aaf[i] <- mean(getbd5snp(bd5, i)$dosage, na.rm = TRUE) / 2
}

knitr::kable(
  data.frame(snpid = bd5$snps$snpid, aaf = round(aaf, 4)),
  caption = "Alternate allele frequencies via getbd5snp"
)

## ----getbd5snp_buf_loop-------------------------------------------------------
dosage <- numeric(n_samp)
p0     <- numeric(n_samp)
p1     <- numeric(n_samp)
p2     <- numeric(n_samp)

aaf_buf <- numeric(n_snps)
for (i in seq_len(n_snps)) {
  getbd5snp_buf(bd5info = bd5, snp = i, dosage = dosage,
                p0 = p0, p1 = p1, p2 = p2)
  aaf_buf[i] <- mean(dosage, na.rm = TRUE) / 2
}

knitr::kable(
  data.frame(snpid = bd5$snps$snpid, aaf = round(aaf_buf, 4)),
  caption = "Alternate allele frequencies via getbd5snp_buf"
)

## ----getbd5snp_con_loop-------------------------------------------------------
dosage <- numeric(n_samp)
p0     <- numeric(n_samp)
p1     <- numeric(n_samp)
p2     <- numeric(n_samp)

con <- openbd5con(bd5)

aaf_con <- numeric(n_snps)
for (i in seq_len(n_snps)) {
  getbd5snp_con(bd5info = bd5, snp = i, dosage = dosage,
                p0 = p0, p1 = p1, p2 = p2, bd5con = con)
  aaf_con[i] <- mean(dosage, na.rm = TRUE) / 2
}

closebd5con(con)

knitr::kable(
  data.frame(snpid = bd5$snps$snpid, aaf = round(aaf_con, 4)),
  caption = "Alternate allele frequencies via getbd5snp_con"
)

## ----verify-------------------------------------------------------------------
all(aaf == aaf_buf)
all(aaf == aaf_con)

## ----cleanup, include = FALSE-------------------------------------------------
unlink(c(bdose_file, paste0(bdose_file, ".bdi")))

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, echo = FALSE------------------------------------------------------
library(BinaryDosage)

## ----vcf_files----------------------------------------------------------------
vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")
vcf1ainfo <- system.file("extdata", "set1a.info", package = "BinaryDosage")
vcf1bfile <- system.file("extdata", "set1b.vcf", package = "BinaryDosage")
vcf1binfo <- system.file("extdata", "set1b.info", package = "BinaryDosage")

## ----vcf_convert, message = FALSE, warning = FALSE----------------------------
bdfile1a <- tempfile()
bdfile1b <- tempfile()

# Without information file
vcftobdlegacy(vcffiles = vcf1afile, bdfiles = bdfile1a)

# With information file — embeds aaf, maf, avgcall, rsq in the header
vcftobdlegacy(vcffiles = c(vcf1bfile, vcf1binfo), bdfiles = bdfile1b)

## ----vcf_gz, message = FALSE, warning = FALSE---------------------------------
vcf1agzfile <- system.file("extdata", "set1a.vcf.gz", package = "BinaryDosage")
bdfile1a_gz <- tempfile()
vcftobdlegacy(vcffiles = vcf1agzfile, bdfiles = bdfile1a_gz, gz = TRUE)

## ----vcf_dosageonly, message = FALSE, warning = FALSE-------------------------
vcf2afile <- system.file("extdata", "set2a.vcf", package = "BinaryDosage")
bdfile2a <- tempfile()
vcftobdlegacy(vcffiles = vcf2afile, bdfiles = bdfile2a)

## ----snpidformat, message = FALSE, warning = FALSE----------------------------
vcf1brsfile <- system.file("extdata", "set1b_rssnp.vcf", package = "BinaryDosage")
bdfile_id0  <- tempfile()
bdfile_id1  <- tempfile()
bdfile_id2  <- tempfile()
bdfile_idm1 <- tempfile()

vcftobdlegacy(vcffiles = vcf1brsfile, bdfiles = bdfile_id0)
vcftobdlegacy(vcffiles = vcf1brsfile, bdfiles = bdfile_id1, snpidformat = 1L)
vcftobdlegacy(vcffiles = vcf1brsfile, bdfiles = bdfile_id2, snpidformat = 2L)
vcftobdlegacy(vcffiles = vcf1brsfile, bdfiles = bdfile_idm1, snpidformat = -1L)

snpnames <- data.frame(
  format0  = getbdinfo(bdfile_id0)$snps$snpid,
  format1  = getbdinfo(bdfile_id1)$snps$snpid,
  format2  = getbdinfo(bdfile_id2)$snps$snpid,
  formatm1 = getbdinfo(bdfile_idm1)$snps$snpid
)
knitr::kable(snpnames, caption = "SNP IDs by snpidformat")

## ----bdoptions, message = FALSE, warning = FALSE------------------------------
bdfile1a_calc <- tempfile()
vcftobdlegacy(vcffiles = vcf1afile, bdfiles = bdfile1a_calc,
              bdoptions = c("aaf", "maf", "rsq"))
bdcalcinfo <- getbdinfo(bdfile1a_calc)
knitr::kable(data.frame(bdcalcinfo$snpinfo),
             caption = "Calculated per-SNP statistics", digits = 3)

## ----gen_files----------------------------------------------------------------
gen3afile   <- system.file("extdata", "set3a.imp",    package = "BinaryDosage")
gen3asample <- system.file("extdata", "set3a.sample", package = "BinaryDosage")
gen3bfile   <- system.file("extdata", "set3b.imp",    package = "BinaryDosage")
gen3bsample <- system.file("extdata", "set3b.sample", package = "BinaryDosage")

## ----gen_convert, message = FALSE, warning = FALSE----------------------------
bdfile3a <- tempfile()
bdfile3b <- tempfile()

gentobd(genfiles = c(gen3afile, gen3asample),
        snpcolumns = c(0L, 2L:5L),
        bdfiles = bdfile3a)
gentobd(genfiles = c(gen3bfile, gen3bsample),
        snpcolumns = c(0L, 2L:5L),
        bdfiles = bdfile3b)

## ----gen_snpcolumns, message = FALSE, warning = FALSE-------------------------
gen3bchrfile <- system.file("extdata", "set3b.chr.imp", package = "BinaryDosage")
bdfile3b_chr <- tempfile()
# chromosome is in column 1, SNP ID in column 2, location in column 3, etc.
gentobd(genfiles = c(gen3bchrfile, gen3bsample), bdfiles = bdfile3b_chr)

## ----gen_impformat, message = FALSE, warning = FALSE--------------------------
gen2bfile   <- system.file("extdata", "set2b.imp",    package = "BinaryDosage")
sample2bfile <- system.file("extdata", "set2b.sample", package = "BinaryDosage")
bdfile2b <- tempfile()
gentobd(genfiles = c(gen2bfile, sample2bfile),
        snpcolumns = c(1L, 3L, 2L, 4L, 5L),
        impformat = 1L,
        bdfiles = bdfile2b)

## ----gen_gz, message = FALSE, warning = FALSE---------------------------------
gen4bgzfile  <- system.file("extdata", "set4b.imp.gz", package = "BinaryDosage")
sample4bfile <- system.file("extdata", "set4b.sample", package = "BinaryDosage")
bdfile4b_gz  <- tempfile()
gentobd(genfiles = c(gen4bgzfile, sample4bfile),
        snpcolumns = c(1L, 2L, 4L, 5L, 6L),
        startcolumn = 7L,
        impformat = 2L,
        gz = TRUE,
        bdfiles = bdfile4b_gz)

## ----merge, message = FALSE, warning = FALSE----------------------------------
bd1afile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
bd1bfile <- system.file("extdata", "vcf1b.bdose", package = "BinaryDosage")
bd1merged <- tempfile()

bdmerge(mergefiles = bd1merged, bdfiles = c(bd1afile, bd1bfile))

bd1ainfo   <- getbdinfo(bd1afile)
bd1binfo   <- getbdinfo(bd1bfile)
bd1merinfo <- getbdinfo(bd1merged)

cat("Set 1a subjects:", nrow(bd1ainfo$samples), "\n")
cat("Set 1b subjects:", nrow(bd1binfo$samples), "\n")
cat("Merged subjects:", nrow(bd1merinfo$samples), "\n")

## ----getbdinfo, message = FALSE, warning = FALSE------------------------------
bd1ainfo <- getbdinfo(bdfiles = bd1afile)

## ----bdapply, message = FALSE, warning = FALSE--------------------------------
calculateaaf <- function(dosage, p0, p1, p2) {
  mean(dosage, na.rm = TRUE) / 2
}

bd1ainfo <- getbdinfo(bd1afile)
aaf_vals <- unlist(bdapply(bdinfo = bd1ainfo, func = calculateaaf))

aaf_table <- data.frame(snpid = bd1ainfo$snps$snpid, aaf = aaf_vals)
knitr::kable(aaf_table, caption = "Alternate allele frequencies", digits = 4)

## ----bdapply_getaaf, message = FALSE, warning = FALSE-------------------------
aaf_vals2 <- unlist(bdapply(bdinfo = bd1ainfo, func = getaaf))

## ----getsnp, message = FALSE, warning = FALSE---------------------------------
snp_data <- data.frame(getsnp(bdinfo = bd1ainfo, snp = "1:12000:T:C", dosageonly = FALSE))
snp_table <- cbind(sid = bd1ainfo$samples$sid, snp_data)
knitr::kable(head(snp_table, 10), caption = "SNP 1:12000:T:C (first 10 subjects)", digits = 4)

## ----vcfinfo, message = FALSE, warning = FALSE--------------------------------
vcfinfo <- getvcfinfo(vcffiles = vcf1afile, index = TRUE)
aaf_vcf  <- unlist(vcfapply(vcfinfo = vcfinfo, func = getaaf))

aaf_compare <- data.frame(snpid = vcfinfo$snps$snpid, aaf = aaf_vcf)
knitr::kable(aaf_compare, caption = "AAF from VCF via vcfapply", digits = 4)

## ----geninfo, message = FALSE, warning = FALSE--------------------------------
geninfo <- getgeninfo(genfiles = c(gen3bchrfile, gen3bsample), index = TRUE)
aaf_gen  <- unlist(genapply(geninfo = geninfo, func = getaaf))

aaf_gen_table <- data.frame(snpid = geninfo$snps$snpid, aaf = aaf_gen)
knitr::kable(aaf_gen_table, caption = "AAF from GEN via genapply", digits = 4)

## ----full_workflow, message = FALSE, warning = FALSE--------------------------
library(BinaryDosage)

# --- Input files ---
vcf1afile   <- system.file("extdata", "set1a.vcf",    package = "BinaryDosage")
vcf1ainfo   <- system.file("extdata", "set1a.info",   package = "BinaryDosage")
vcf1bfile   <- system.file("extdata", "set1b.vcf",    package = "BinaryDosage")
vcf1binfo   <- system.file("extdata", "set1b.info",   package = "BinaryDosage")
gen3afile   <- system.file("extdata", "set3a.imp",    package = "BinaryDosage")
gen3asample <- system.file("extdata", "set3a.sample", package = "BinaryDosage")
gen3bfile   <- system.file("extdata", "set3b.imp",    package = "BinaryDosage")
gen3bsample <- system.file("extdata", "set3b.sample", package = "BinaryDosage")

# --- Convert to binary dosage ---
bdfile1a <- tempfile(); bdfile1b <- tempfile()
bdfile3a <- tempfile(); bdfile3b <- tempfile()

vcftobdlegacy(vcffiles = c(vcf1afile, vcf1ainfo), bdfiles = bdfile1a)
vcftobdlegacy(vcffiles = c(vcf1bfile, vcf1binfo), bdfiles = bdfile1b)
gentobd(genfiles = c(gen3afile, gen3asample), snpcolumns = c(0L, 2L:5L), bdfiles = bdfile3a)
gentobd(genfiles = c(gen3bfile, gen3bsample), snpcolumns = c(0L, 2L:5L), bdfiles = bdfile3b)

# --- Merge ---
mergebd1 <- tempfile(); mergebd3 <- tempfile()
bdmerge(mergefiles = mergebd1, bdfiles = c(bdfile1a, bdfile1b))
bdmerge(mergefiles = mergebd3, bdfiles = c(bdfile3a, bdfile3b))

# --- Apply function ---
mergebd1info <- getbdinfo(mergebd1)
mergebd3info <- getbdinfo(mergebd3)

aaf1 <- unlist(bdapply(mergebd1info, getaaf))
aaf3 <- unlist(bdapply(mergebd3info, getaaf))

aaf <- cbind(mergebd1info$snps[, c("chromosome", "location", "snpid")],
             aaf_vcf = aaf1, aaf_gen = aaf3)
knitr::kable(aaf, caption = "AAF: VCF vs GEN sources", digits = 4, row.names = FALSE)

## ----full_workflow_snp, message = FALSE, warning = FALSE----------------------
# --- Extract SNP 6 ---
set1snp6 <- getsnp(mergebd1info, 6L)
set3snp6 <- getsnp(mergebd3info, 6L)

snpdf <- data.frame(
  subjectid = mergebd1info$samples$sid,
  dosage_vcf = unlist(set1snp6),
  dosage_gen = unlist(set3snp6)
)
knitr::kable(head(snpdf, 10),
             caption = "SNP 6 dosage: VCF vs GEN (first 10 subjects)",
             digits = 4, row.names = FALSE)

#***************************************************************************#
#                                                                           #
#               Format 5 Binary Dosage - Write Functions                    #
#                                                                           #
# File pair:                                                                #
#   .bdose  - compressed dosage/genotype probability data                   #
#   .bdi    - SNP metadata, sample IDs, and SNP byte offsets (RDS)          #
#                                                                           #
# Encoding:                                                                 #
#   Values in [0, 2] stored as unsigned short = round(value * 10000)        #
#   Range 0-20000; 0xffff (stored as -1L) = missing                        #
#   DS and GP concatenated per SNP, then gzip-compressed                    #
#                                                                           #
#***************************************************************************#


#***************************************************************************#
#                        Internal constants                                 #
#***************************************************************************#

# Magic bytes at the start of every .bdose Format 5 file:
# byte 0: 0x01 (format major version)
# byte 1: 0x00
# byte 2: 0x05 (format 5)
# byte 3: 0x00
.bdose5_magic <- as.raw(c(0x01, 0x00, 0x05, 0x00))


#***************************************************************************#
#                        Internal helpers                                   #
#***************************************************************************#

# Convert a numeric vector (values in [0, 2], NA = missing) to a raw vector
# of little-endian unsigned 16-bit integers.
# Encoding: stored = round(x * 10000), range 0-20000.
# Missing (NA) is stored as 0xffff, written as -1L in signed int16.
vals_to_ushort_raw <- function(x) {
  stored <- as.integer(round(x * 10000))
  stored[is.na(x)] <- -1L
  writeBin(stored, raw(), size = 2L, endian = "little")
}

# Gzip-compress the DS and GP data for one SNP into a raw vector.
# ds: numeric vector, length n_samples, values in [0, 2]
# gp: numeric vector, length 3 * n_samples, interleaved P(0/0), P(0/1), P(1/1)
compress_snp_block <- function(ds, gp) {
  raw_block <- c(vals_to_ushort_raw(ds), vals_to_ushort_raw(gp))
  memCompress(raw_block, type = "gzip")
}

# Save a Format 5 .bdi file as an RDS object with class "genetic-info".
# The output path is derived as paste0(bdose_file, ".bdi").
#
# Parameters:
#   bdose_file    - path to the associated .bdose file (stored in $filename)
#   samples_sid   - character vector of sample subject IDs
#   snps_df       - data.frame with columns: chromosome, location, snpid,
#                   reference, alternate
#   indices       - numeric vector of byte offsets in the .bdose file
#   snpidformat   - integer; resolved snpidformat value to store in bdinfo
#   snpinfo_list  - optional named list of per-SNP annotation vectors
#                   (e.g. list(aaf = ..., maf = ..., rsq = ...))
write_bdinfo5 <- function(bdose_file, samples_sid,
                          snps_df, indices, snpidformat = 0L,
                          snpinfo_list = list()) {
  bdinfo_file <- paste0(bdose_file, ".bdi")
  n_samp <- length(samples_sid)
  n_snps <- nrow(snps_df)

  samples_df <- data.frame(
    fid = rep("", n_samp),
    sid = samples_sid,
    stringsAsFactors = FALSE
  )

  onechr <- length(unique(snps_df$chromosome)) == 1L

  additional <- list(
    format     = 5,
    subformat  = 1L,
    headersize = 4L,
    numgroups  = 1L,
    groups     = n_samp
  )
  class(additional) <- "bdose-info"

  bdinfo <- list(
    filename       = normalizePath(bdose_file, winslash = "/", mustWork = FALSE),
    usesfid        = FALSE,
    samples        = samples_df,
    onechr         = onechr,
    snpidformat    = snpidformat,
    snps           = snps_df,
    snpinfo        = snpinfo_list,
    additionalinfo = additional,
    datasize       = integer(0L),
    indices        = indices
  )
  class(bdinfo) <- "genetic-info"

  saveRDS(bdinfo, bdinfo_file)
}


#***************************************************************************#
#                        Read functions                                     #
#***************************************************************************#

# Validate a Format 5 .bdose file and load its .bdi metadata.
# Called internally by getbdinfo when a Format 5 magic header is detected.
#
#' @keywords internal
getbd5info <- function(bdose_file) {
  if (missing(bdose_file))
    stop("No .bdose file specified")
  if (!file.exists(bdose_file))
    stop("File not found: ", bdose_file)
  bdinfo_file <- paste0(bdose_file, ".bdi")
  if (!file.exists(bdinfo_file))
    stop("File not found: ", bdinfo_file)

  # Validate the .bdose magic header.
  con <- file(bdose_file, open = "rb")
  magic <- readBin(con, what = raw(), n = 4L)
  close(con)
  if (!identical(magic, .bdose5_magic))
    stop("File does not appear to be a Format 5 .bdose file: ", bdose_file)

  # Load and validate the .bdinfo RDS.
  bdinfo <- readRDS(bdinfo_file)
  if (!is.list(bdinfo))
    stop("Invalid .bdinfo file: not an R list")
  required <- c("snps", "samples", "indices")
  missing_fields <- setdiff(required, names(bdinfo))
  if (length(missing_fields) > 0L)
    stop("Invalid .bdinfo file: missing fields: ",
         paste(missing_fields, collapse = ", "))

  n_snps <- nrow(bdinfo$snps)
  if (length(bdinfo$indices) != n_snps)
    stop("Invalid .bdinfo: indices length (", length(bdinfo$indices),
         ") does not match SNP count (", n_snps, ")")

  # Overwrite filename with the caller-supplied (normalized) path so the
  # object remains valid even if files have been moved.
  bdinfo$filename <- normalizePath(bdose_file, winslash = "/")
  class(bdinfo) <- "genetic-info"
  bdinfo
}

# Decode a raw vector of little-endian signed int16 values back to doubles.
# -1L (stored 0xffff) -> NA; all other values -> x / 10000.
ushort_raw_to_vals <- function(raw_block, n) {
  x <- readBin(raw_block, what = integer(), n = n,
               size = 2L, signed = TRUE, endian = "little")
  result <- as.double(x) / 10000.0
  result[x == -1L] <- NA_real_
  result
}

#' Read a SNP from a Format 5 binary dosage file
#'
#' Seeks to the SNP's compressed block in the .bdose file, decompresses it,
#' and returns the dosage and genotype probabilities for all samples.
#'
#' @param bd5info  Object returned by \code{getbdinfo}.
#' @param snp  The SNP to retrieve: either a 1-based integer index or a
#'   character SNP ID matching a value in \code{bd5info$snps$snpid}.
#'
#' @return A list with four numeric vectors, each of length n_samples:
#' \describe{
#'   \item{dosage}{DS values in \[0, 2\]; NA = missing.}
#'   \item{p0}{P(g=0) values in \[0, 1\]; NA = missing.}
#'   \item{p1}{P(g=1) values in \[0, 1\]; NA = missing.}
#'   \item{p2}{P(g=2) values in \[0, 1\]; NA = missing.}
#' }
#' @export
getbd5snp <- function(bd5info, snp) {
  if (missing(bd5info))
    stop("bd5info missing")
  if (!inherits(bd5info, "genetic-info"))
    stop("bd5info must be an object returned by getbdinfo")
  if (missing(snp))
    stop("No SNP specified")
  if (length(snp) != 1L)
    stop("snp must be of length 1")

  # Resolve SNP to a 1-based integer index.
  if (is.character(snp)) {
    snp_idx <- match(snp, bd5info$snps$snpid)
    if (is.na(snp_idx))
      stop("SNP '", snp, "' not found in bd5info")
  } else {
    snp_idx <- as.integer(floor(snp))
    if (snp_idx < 1L || snp_idx > nrow(bd5info$snps))
      stop("snp index out of range: ", snp_idx)
  }

  n_snps <- nrow(bd5info$snps)
  n_samp <- nrow(bd5info$samples)
  start  <- bd5info$indices[snp_idx]

  # Compressed block size: difference between consecutive offsets, or
  # distance to end-of-file for the last SNP.
  if (snp_idx < n_snps) {
    nbytes <- as.integer(bd5info$indices[snp_idx + 1L] - start)
  } else {
    nbytes <- as.integer(file.info(bd5info$filename)$size - start)
  }

  # Read and decompress the block.
  con <- file(bd5info$filename, open = "rb")
  on.exit(close(con))
  seek(con, where = start, origin = "start")
  compressed <- readBin(con, what = raw(), n = nbytes)
  raw_block  <- memDecompress(compressed, type = "gzip")

  # Decode DS (first n_samp values) and GP (next 3*n_samp values).
  dosage <- ushort_raw_to_vals(raw_block, n_samp)

  gp_raw <- raw_block[seq(2L * n_samp + 1L, length(raw_block))]
  gp     <- ushort_raw_to_vals(gp_raw, 3L * n_samp)

  # Deinterleave GP: layout is [P00_s1, P01_s1, P11_s1, P00_s2, ...]
  idx <- seq_len(n_samp)
  p0  <- gp[3L * (idx - 1L) + 1L]
  p1  <- gp[3L * (idx - 1L) + 2L]
  p2  <- gp[3L * (idx - 1L) + 3L]

  list(dosage = dosage, p0 = p0, p1 = p1, p2 = p2)
}

#' Read a Format 5 SNP into pre-allocated vectors (buffered variant)
#'
#' Like \code{getbd5snp} but writes results into caller-supplied vectors
#' instead of allocating new ones.  Intended for tight loops where thousands
#' of SNPs are read sequentially; pre-allocating the output vectors once
#' avoids repeated memory allocation.
#'
#' The four output vectors \strong{must not} have more than one R binding at
#' the call site (no extra variables pointing to the same object); R's
#' copy-on-modify semantics would otherwise prevent in-place update.
#'
#' @param bd5info  Object returned by \code{getbdinfo}.
#' @param snp  1-based integer index or character SNP ID.
#' @param dosage  Pre-allocated \code{numeric(n_samples)} vector.
#' @param p0  Pre-allocated \code{numeric(n_samples)} vector.
#' @param p1  Pre-allocated \code{numeric(n_samples)} vector.
#' @param p2  Pre-allocated \code{numeric(n_samples)} vector.
#'
#' @return \code{NULL} invisibly.  \code{dosage}, \code{p0}, \code{p1}, and
#'   \code{p2} are updated in place.
#' @export
getbd5snp_buf <- function(bd5info, snp, dosage, p0, p1, p2) {
  if (missing(bd5info))
    stop("bd5info missing")
  if (!inherits(bd5info, "genetic-info"))
    stop("bd5info must be an object returned by getbdinfo")
  if (missing(snp))
    stop("No SNP specified")
  if (length(snp) != 1L)
    stop("snp must be of length 1")

  if (is.character(snp)) {
    snp_idx <- match(snp, bd5info$snps$snpid)
    if (is.na(snp_idx))
      stop("SNP '", snp, "' not found in bd5info")
  } else {
    snp_idx <- as.integer(floor(snp))
    if (snp_idx < 1L || snp_idx > nrow(bd5info$snps))
      stop("snp index out of range: ", snp_idx)
  }

  n_snps <- nrow(bd5info$snps)
  n_samp <- nrow(bd5info$samples)
  start  <- bd5info$indices[snp_idx]

  if (snp_idx < n_snps) {
    nbytes <- as.integer(bd5info$indices[snp_idx + 1L] - start)
  } else {
    nbytes <- as.integer(file.info(bd5info$filename)$size - start)
  }

  con <- file(bd5info$filename, open = "rb")
  on.exit(close(con))
  seek(con, where = start, origin = "start")
  compressed <- readBin(con, what = raw(), n = nbytes)
  raw_block  <- memDecompress(compressed, type = "gzip")

  DecodeFormat5BlockC(raw_block, n_samp, dosage, p0, p1, p2)
  invisible(NULL)
}

#' Open a persistent connection to a Format 5 binary dosage file
#'
#' Opens the .bdose file for reading and returns an object that holds the
#' connection open across multiple calls to \code{getbd5snp_con}.  The
#' connection is closed automatically when the object is garbage-collected
#' or when R exits; call \code{closebd5con} to close it explicitly.
#'
#' @param bd5info  Object returned by \code{getbdinfo}.
#'
#' @return An object of class \code{"bd5con"} to be passed to
#'   \code{getbd5snp_con} and \code{closebd5con}.
#' @export
openbd5con <- function(bd5info) {
  if (missing(bd5info))
    stop("bd5info missing")
  if (!inherits(bd5info, "genetic-info"))
    stop("bd5info must be an object returned by getbdinfo")
  e <- new.env(parent = emptyenv())
  e$xptr <- OpenFormat5FileC(bd5info$filename)
  reg.finalizer(e, function(e) {
    if (!is.null(e$xptr))
      CloseFormat5FileC(e$xptr)
  }, onexit = TRUE)
  class(e) <- "bd5con"
  e
}

#' Close a persistent Format 5 connection
#'
#' Explicitly closes the connection opened by \code{openbd5con}.  Calling
#' this is optional — the finalizer will close it on garbage collection or
#' R exit — but explicit close is preferred to release the file handle
#' promptly.
#'
#' @param bd5con  Object returned by \code{openbd5con}.
#'
#' @return \code{NULL} invisibly.
#' @export
closebd5con <- function(bd5con) {
  if (!inherits(bd5con, "bd5con"))
    stop("bd5con must be an object returned by openbd5con")
  if (!is.null(bd5con$xptr))
    CloseFormat5FileC(bd5con$xptr)
  bd5con$xptr <- NULL
  invisible(NULL)
}

#' Read a Format 5 SNP using a persistent open connection
#'
#' Like \code{getbd5snp_buf} but reuses an already-open file connection
#' instead of opening and closing it on every call.  Use \code{openbd5con}
#' before the loop and \code{closebd5con} (or let the finalizer handle it)
#' after.
#'
#' @param bd5info  Object returned by \code{getbdinfo}.
#' @param snp  1-based integer index or character SNP ID.
#' @param dosage  Pre-allocated \code{numeric(n_samples)} vector.
#' @param p0  Pre-allocated \code{numeric(n_samples)} vector.
#' @param p1  Pre-allocated \code{numeric(n_samples)} vector.
#' @param p2  Pre-allocated \code{numeric(n_samples)} vector.
#' @param bd5con  Object returned by \code{openbd5con}.
#'
#' @return \code{NULL} invisibly.  \code{dosage}, \code{p0}, \code{p1}, and
#'   \code{p2} are updated in place.
#' @export
getbd5snp_con <- function(bd5info, snp, dosage, p0, p1, p2, bd5con) {
  if (missing(bd5info))
    stop("bd5info missing")
  if (!inherits(bd5info, "genetic-info"))
    stop("bd5info must be an object returned by getbdinfo")
  if (missing(snp))
    stop("No SNP specified")
  if (length(snp) != 1L)
    stop("snp must be of length 1")
  if (!inherits(bd5con, "bd5con"))
    stop("bd5con must be an object returned by openbd5con")

  if (is.character(snp)) {
    snp_idx <- match(snp, bd5info$snps$snpid)
    if (is.na(snp_idx))
      stop("SNP '", snp, "' not found in bd5info")
  } else {
    snp_idx <- as.integer(floor(snp))
    if (snp_idx < 1L || snp_idx > nrow(bd5info$snps))
      stop("snp index out of range: ", snp_idx)
  }

  n_snps <- nrow(bd5info$snps)
  n_samp <- nrow(bd5info$samples)
  start  <- bd5info$indices[snp_idx]

  if (snp_idx < n_snps) {
    nbytes <- as.integer(bd5info$indices[snp_idx + 1L] - start)
  } else {
    nbytes <- as.integer(file.info(bd5info$filename)$size - start)
  }

  ReadFormat5SNPC(bd5con$xptr, start, nbytes, n_samp, dosage, p0, p1, p2)
  invisible(NULL)
}

#***************************************************************************#
#                        Main conversion function                           #
#***************************************************************************#

#' Convert a VCF file to Format 5 binary dosage files
#'
#' Reads the DS (dosage) and GP (genotype probabilities) FORMAT fields from a
#' bgzipped, tabix-indexed VCF file — as produced by imputation servers such
#' as the Michigan Imputation Server — and writes a pair of Format 5
#' BinaryDosage files.
#'
#' The .bdose file begins with a 4-byte magic number followed by one
#' gzip-compressed block per SNP.  Each block contains the DS values for all
#' samples followed by the GP values, encoded as unsigned 16-bit integers
#' (round(value * 10000); 0xffff = missing).
#'
#' The .bdi file is an RDS-serialised R list of class \code{"genetic-info"}
#' with the following elements:
#' \describe{
#'   \item{filename}{Path to the associated .bdose file.}
#'   \item{usesfid}{Logical; always FALSE for VCF-sourced files.}
#'   \item{samples}{data.frame with columns \code{fid} (empty) and
#'     \code{sid} (sample IDs).}
#'   \item{onechr}{Logical; TRUE if all SNPs are on a single chromosome.}
#'   \item{snpidformat}{Numeric; resolved SNP ID format (see \code{snpidformat}
#'     parameter).}
#'   \item{snps}{data.frame with columns chromosome, location, snpid,
#'     reference, alternate.}
#'   \item{snpinfo}{Named list of per-SNP annotations requested via
#'     \code{bdoptions}.  Each element is a numeric vector of length equal to
#'     the number of SNPs.  Values are read from the VCF INFO column when
#'     available for the first SNP (AF for aaf, MAF for maf, R2 for rsq);
#'     otherwise they are calculated from the dosage values.}
#'   \item{additionalinfo}{List of class \code{"bdose-info"} with format,
#'     subformat, headersize, numgroups, and groups.}
#'   \item{datasize}{Integer vector of length 0 (unused in Format 5).}
#'   \item{indices}{Numeric vector of byte offsets into .bdose, one per SNP.}
#' }
#'
#' @param vcffile  Path to the bgzipped, tabix-indexed VCF file.
#' @param bdose_file  Path for the output .bdose file. The companion .bdi
#'   metadata file is written to \code{paste0(bdose_file, ".bdi")}.
#' @param region  Optional genomic region string in bcftools format
#'   (e.g. \code{"chr21"} or \code{"chr21:1-5000000"}).  Requires a
#'   tabix index.  Default \code{NULL} processes the entire file.
#' @param snpidformat  Integer controlling how SNP IDs are stored.
#'   \describe{
#'     \item{-1}{Generate IDs as \code{chr:pos:ref:alt}; equivalent to 2 for
#'       Format 5.}
#'     \item{0}{Use the IDs as they appear in the VCF file (default).
#'       Auto-detects format 1 or 2 if all IDs match.}
#'     \item{1}{Store IDs as \code{chr:pos}.  An error is raised if the VCF
#'       already uses \code{chr:pos:ref:alt} format, as information would be
#'       lost.}
#'     \item{2}{Store IDs as \code{chr:pos:ref:alt}.}
#'     \item{3}{Store IDs as \code{chr:pos_ref_alt}.}
#'   }
#'
#' @param bdoptions  Character vector specifying which per-SNP statistics to
#'   store.  Any combination of \code{"aaf"} (alternate allele frequency),
#'   \code{"maf"} (minor allele frequency), and \code{"rsq"} (imputation
#'   r-squared).  For each statistic, the corresponding VCF INFO field is used
#'   when present for the first SNP (AF, MAF, R2 respectively); otherwise the
#'   value is calculated from the dosage data.  Default \code{character(0)}
#'   stores no statistics.
#'
#' @return NULL (invisibly)
#' @importFrom stats var
#' @export
vcftobd <- function(vcffile, bdose_file, region = NULL,
                     snpidformat = 0L, bdoptions = character(0)) {

  if (!requireNamespace("vcfppR", quietly = TRUE))
    stop("Package 'vcfppR' is required for vcftobd(). ",
         "Install it with install.packages(\"vcfppR\"). ",
         "vcfppR provides approximately 12x faster VCF conversion, ",
         "and Format 5 files are smaller than both compressed VCF and ",
         "legacy binary dosage files.")

  if (!file.exists(vcffile))
    stop("VCF file not found: ", vcffile)

  if (!is.null(region) && (!is.character(region) || length(region) != 1L))
    stop("region must be a single character string or NULL")

  if (!is.numeric(snpidformat) || length(snpidformat) != 1L ||
      floor(snpidformat) != snpidformat)
    stop("snpidformat must be an integer value")
  snpidformat <- as.integer(snpidformat)
  if (snpidformat < -1L || snpidformat > 3L)
    stop("snpidformat must be -1, 0, 1, 2, or 3")

  if (!is.character(bdoptions))
    stop("bdoptions must be a character vector")
  invalid_opts <- setdiff(bdoptions, c("aaf", "maf", "rsq"))
  if (length(invalid_opts) > 0L)
    stop("Invalid bdoptions value(s): ",
         paste(invalid_opts, collapse = ", "),
         ". Valid values are: aaf, maf, rsq")

  want_aaf <- "aaf" %in% bdoptions
  want_maf <- "maf" %in% bdoptions
  want_rsq <- "rsq" %in% bdoptions

  # Pre-allocate metadata vectors; will trim to actual SNP count at the end.
  # 2 million SNPs is sufficient for any single chromosome.
  MAX_SNPS    <- 2000000L
  snp_chr     <- character(MAX_SNPS)
  snp_pos     <- integer(MAX_SNPS)
  snp_id      <- character(MAX_SNPS)
  snp_ref     <- character(MAX_SNPS)
  snp_alt     <- character(MAX_SNPS)
  snp_indices <- numeric(MAX_SNPS)
  snp_count   <- 0L

  # Pre-allocate snpinfo accumulation vectors and source flags.
  # Flags (use_*_from_vcf) are set on the first SNP based on INFO availability.
  if (want_aaf) { aaf_vec <- numeric(MAX_SNPS); use_aaf_from_vcf <- FALSE }
  if (want_maf) { maf_vec <- numeric(MAX_SNPS); use_maf_from_vcf <- FALSE }
  if (want_rsq) { rsq_vec <- numeric(MAX_SNPS); use_rsq_from_vcf <- FALSE }

  # current_pos tracks the write position in the .bdose file (numeric to
  # handle files larger than 2 GB without integer overflow).
  current_pos <- 4  # bytes consumed by the 4-byte header

  # Open .bdose and write the magic header.
  bdose_con <- file(bdose_file, open = "wb")
  on.exit(close(bdose_con), add = TRUE)
  writeBin(.bdose5_magic, bdose_con)

  # Open the VCF reader and retrieve sample IDs.
  if (!is.null(region))
    reader <- vcfppR::vcfreader$new(vcffile, region)
  else
    reader <- vcfppR::vcfreader$new(vcffile)

  samples_sid <- reader$samples()

  # Stream through every variant in the VCF.
  while (reader$variant()) {
    snp_count <- snp_count + 1L

    if (snp_count > MAX_SNPS)
      stop("SNP count exceeds MAX_SNPS (", MAX_SNPS, "). ",
           "Increase MAX_SNPS in vcftobd().")

    # Validate required FORMAT tags and detect INFO field availability on
    # the first SNP.
    if (snp_count == 1L) {
      if (reader$getFormatType("DS") == 0L)
        stop("FORMAT field 'DS' not found in VCF file: ", vcffile)
      if (reader$getFormatType("GP") == 0L)
        stop("FORMAT field 'GP' not found in VCF file: ", vcffile)
      if (want_aaf) {
        val <- tryCatch(reader$infoFloat("AF"),  error = function(e) NA_real_)
        use_aaf_from_vcf <- length(val) > 0L && !is.na(val[1L])
      }
      if (want_maf) {
        val <- tryCatch(reader$infoFloat("MAF"), error = function(e) NA_real_)
        use_maf_from_vcf <- length(val) > 0L && !is.na(val[1L])
      }
      if (want_rsq) {
        val <- tryCatch(reader$infoFloat("R2"),  error = function(e) NA_real_)
        use_rsq_from_vcf <- length(val) > 0L && !is.na(val[1L])
      }
    }

    # Collect SNP metadata.
    snp_chr[snp_count] <- reader$chr()
    snp_pos[snp_count] <- reader$pos()
    snp_id[snp_count]  <- reader$id()
    snp_ref[snp_count] <- reader$ref()
    snp_alt[snp_count] <- reader$alt()

    # Read dosage and genotype probability data.
    ds <- reader$formatFloat("DS")
    gp <- reader$formatFloat("GP")

    # Accumulate snpinfo values.
    if (want_aaf) {
      aaf_vec[snp_count] <- if (use_aaf_from_vcf)
        tryCatch(reader$infoFloat("AF")[1L],  error = function(e) NA_real_)
      else
        mean(ds, na.rm = TRUE) / 2
    }
    if (want_maf) {
      maf_vec[snp_count] <- if (use_maf_from_vcf)
        tryCatch(reader$infoFloat("MAF")[1L], error = function(e) NA_real_)
      else {
        aaf_i <- if (want_aaf) aaf_vec[snp_count] else mean(ds, na.rm = TRUE) / 2
        min(aaf_i, 1 - aaf_i)
      }
    }
    if (want_rsq) {
      rsq_vec[snp_count] <- if (use_rsq_from_vcf)
        tryCatch(reader$infoFloat("R2")[1L],  error = function(e) NA_real_)
      else {
        aaf_i <- if (want_aaf) aaf_vec[snp_count] else mean(ds, na.rm = TRUE) / 2
        denom <- 2 * aaf_i * (1 - aaf_i)
        if (denom > 0) var(ds, na.rm = TRUE) / denom else NA_real_
      }
    }

    # Record the byte offset for this SNP, compress, and write.
    snp_indices[snp_count] <- current_pos
    compressed <- compress_snp_block(ds, gp)
    writeBin(compressed, bdose_con)
    current_pos <- current_pos + length(compressed)
  }

  if (snp_count == 0L)
    stop("No variants found in VCF file: ", vcffile)

  # Trim metadata vectors to actual SNP count.
  chr_vec <- snp_chr[1:snp_count]
  pos_vec <- snp_pos[1:snp_count]
  ref_vec <- snp_ref[1:snp_count]
  alt_vec <- snp_alt[1:snp_count]
  id_vec  <- snp_id[1:snp_count]

  # Apply snpidformat: resolve IDs and the stored format value.
  chrlocid          <- paste(chr_vec, pos_vec, sep = ":")
  chrlocrefaltid    <- paste(chr_vec, pos_vec, ref_vec, alt_vec, sep = ":")
  chrlocrefaltid_us <- paste(chrlocid, ref_vec, alt_vec, sep = "_")

  if (snpidformat == 0L) {
    # Auto-detect: check whether all VCF IDs already match a known format.
    if (all(id_vec == chrlocid))
      snpidformat <- 1L
    else if (all(id_vec == chrlocrefaltid))
      snpidformat <- 2L
    else if (all(id_vec == chrlocrefaltid_us))
      snpidformat <- 3L
    # else stays 0 — IDs are used verbatim
  } else if (snpidformat == 1L) {
    if (all(id_vec == chrlocrefaltid) || all(id_vec == chrlocrefaltid_us))
      stop("snpidformat 1 specified but VCF file uses a format that encodes ",
           "ref/alt; information would be lost")
    id_vec <- chrlocid
  } else if (snpidformat == 2L) {
    id_vec <- chrlocrefaltid
  } else if (snpidformat == 3L) {
    id_vec <- chrlocrefaltid_us
  } else {
    # snpidformat == -1: generate chr:pos:ref:alt, store as format 2.
    id_vec      <- chrlocrefaltid
    snpidformat <- 2L
  }

  # Build the SNP data frame.
  snps_df <- data.frame(
    chromosome = chr_vec,
    location   = pos_vec,
    snpid      = id_vec,
    reference  = ref_vec,
    alternate  = alt_vec,
    stringsAsFactors = FALSE
  )

  # Build the snpinfo list (trim vectors to actual SNP count).
  snpinfo_list <- list()
  if (want_aaf) snpinfo_list$aaf <- aaf_vec[1:snp_count]
  if (want_maf) snpinfo_list$maf <- maf_vec[1:snp_count]
  if (want_rsq) snpinfo_list$rsq <- rsq_vec[1:snp_count]

  # Save the .bdi file.
  write_bdinfo5(bdose_file   = bdose_file,
                samples_sid  = samples_sid,
                snps_df      = snps_df,
                indices      = snp_indices[1:snp_count],
                snpidformat  = snpidformat,
                snpinfo_list = snpinfo_list)

  invisible(NULL)
}

test_that("getsnp", {
  vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")
  vcfinfo <- getvcfinfo(vcffiles = vcf1afile)
  vcf1abdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
  bdinfo <- getbdinfo(bdfiles = vcf1abdfile)

  expect_error(getsnp(snp = 1),
               "bdinfo missing")
  expect_error(getsnp(bdinfo = 1,
                      snp = 1),
               "bdinfo is not of class genetic-info")
  expect_error(getsnp(bdinfo = vcfinfo,
                      snp = 1),
               "bdinfo does not contain information about a binary dosage file")

  expect_error(getsnp(bdinfo = bdinfo),
               "No SNP specified")
  expect_error(getsnp(bdinfo = bdinfo,
                      snp = 1:3),
               "snp must be of length one")
  expect_error(getsnp(bdinfo = bdinfo,
                      snp = 1.3),
               "snp must be a character or integer value")
  expect_error(getsnp(bdinfo = bdinfo,
                      snp = TRUE),
               "snp must be a character or integer value")
  expect_error(getsnp(bdinfo = bdinfo,
                      snp = "xyz"),
               "Cannot find SNP in bdinfo")
  expect_error(getsnp(bdinfo = bdinfo,
                      snp = 100),
               "snp value out or range")

  expect_error(getsnp(bdinfo = bdinfo,
                      snp = 1,
                      FALSE),
               NA)
  expect_error(getsnp(bdinfo = bdinfo,
                      snp = "1:11000:T:C",
                      TRUE),
               NA)
})

test_that("getsnp format5", {
  skip_if_not_installed("vcfppR")

  vcfgzfile  <- system.file("extdata", "set1a.vcf.gz", package = "BinaryDosage")
  bdose_file <- tempfile(fileext = ".bdose")

  expect_error(vcftobd(vcffile = vcfgzfile, bdose_file = bdose_file), NA)

  bd5info <- getbdinfo(bdose_file)
  n_snps <- nrow(bd5info$snps)
  expect_equal(n_snps, 10L)

  for (i in seq_len(n_snps)) {
    ref    <- getbd5snp(bd5info, i)
    result <- getsnp(bdinfo = bd5info, snp = i, dosageonly = FALSE)

    expect_equal(result$dosage, ref$dosage, tolerance = 5e-5,
                 label = paste0("dosage SNP ", i))
    expect_equal(result$p0, ref$p0, tolerance = 5e-5,
                 label = paste0("p0 SNP ", i))
    expect_equal(result$p1, ref$p1, tolerance = 5e-5,
                 label = paste0("p1 SNP ", i))
    expect_equal(result$p2, ref$p2, tolerance = 5e-5,
                 label = paste0("p2 SNP ", i))
  }
})

###########################################################
#                Binary Dosage                            #
###########################################################

#' Get information about a binary dosage file
#'
#' Routine to return information about a binary dosage file.
#' This information is used by other routines to
#' allow for quicker extraction of values from the
#' file.
#'
#' @param bdfiles Vector of file names. For Format 5 files, a single .bdose
#' file name; the companion .bdi metadata file is read automatically from
#' \code{paste0(bdfiles[1], ".bdi")}. For format 4, a single file name. For
#' formats 1, 2, and 3, a vector of three file names: the binary dosage file,
#' the family information file, and the SNP information file.
#'
#' @return List with information about the binary dosage file.
#' This includes family and subject IDs along with
#' a list of the SNPs in the file. Other information needed
#' to read the file is also included.
#'
#' @export
#'
#' @examples
#' vcf1abdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
#' bdinfo <- getbdinfo(bdfiles = vcf1abdfile)
getbdinfo <- function(bdfiles) {
  if (missing(bdfiles) == TRUE)
    stop("No binary dosage files specified")
  if (is.character(bdfiles) == FALSE)
    stop("bdfiles must be a character vector")
  if (length(bdfiles) != 1 & length(bdfiles) != 3)
    stop("bdfiles must be a character vector of length 1 or 3")
  if (bdfiles[1] == "")
    stop("No binary dosage file specified")
  if (length(bdfiles) == 3) {
    if (bdfiles[2] == "" & bdfiles[3] == "") {
      bdfiles <- bdfiles[1]
    } else if (bdfiles[2] == "" | bdfiles[3] == "") {
      stop("bdfiles contains empty strings")
    }
  }

  # Check for Format 5: peek at the first 4 bytes and delegate if magic matches.
  if (length(bdfiles) == 1L && file.exists(bdfiles[1L])) {
    con <- file(bdfiles[1L], open = "rb")
    magic <- readBin(con, what = raw(), n = 4L)
    close(con)
    if (identical(magic, .bdose5_magic))
      return(getbd5info(bdfiles[1L]))
  }

  headerinfo <- ReadBinaryDosageHeader(bdfiles)
  indices <- ReadBinaryDosageIndices(headerinfo)
  headerinfo$datasize <- indices$datasize
  headerinfo$indices <- indices$indices
  class(headerinfo) <- c("genetic-info")

  return (headerinfo)
}

###########################################################
#                        VCF                              #
###########################################################
summarizevcfadditionalinfo <- function(x) {
  if (length(unique(x)) != 1)
    return (x)
  if (x[1] == '.')
    return (character(0))
  return (x[1])
}

readminimacinfofile <- function(filename) {
  addinfo <- read.table(filename, header = TRUE, stringsAsFactors = FALSE)
  if (ncol(addinfo) != 13)
    stop("Error reading information file - Wrong number of columns")
  if (all(colnames(addinfo) == c("SNP", "REF.0.", "ALT.1.", "ALT_Frq", "MAF",
                                 "AvgCall", "Rsq", "Genotyped", "LooRsq",
                                 "EmpR", "EmpRsq", "Dose0", "Dose1")) == FALSE)
    stop("Error reading information file - Wrong column names")
  return(addinfo)
}

#' Get information about a vcf file
#'
#' Routine to return information about a vcf file.
#' This information is used by other routines to
#' allow for quicker extraction of values from the
#' file.
#'
#' @param vcffiles A vector of file names.
#' The first is the name of the vcf file. The
#' second is name of the file that contains information
#' about the imputation of the SNPs. This file is produced
#' by minimac 3 and 4.
#' @param gz Indicator if VCF file is compressed using gzip.
#' Default value is FALSE.
#' @param index Indicator if file should be indexed. This
#' allows for faster reading of the file. Indexing a gzipped
#' file is not supported.
#' Default value is TRUE.
#' @param snpidformat The format that the SNP ID will be saved as.
#' 0 - same as in the VCF file
#' 1 - chr:pos
#' 2 - chr:pos:ref:alt
#' If snpidformat is 1 and the VCF file uses format 2, an error is
#' generated. Default value is 0.
#'
#' @return List containing information about the VCF file
#' to include file name, subject IDs, and information about
#' the SNPs. Indices for faster reading will be included
#' if index is set to TRUE
#'
#' @export
#'
#' @examples
#' # Get file names of th vcf and infromation file
#' vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")
#' vcf1ainfo <- system.file("extdata", "set1a.info", package = "BinaryDosage")
#'
#' # Get the information about the vcf file
#' vcf1ainfo <- getvcfinfo(vcffiles = c(vcf1afile, vcf1ainfo))
getvcfinfo <- function(vcffiles,
                       gz = FALSE,
                       index = TRUE,
                       snpidformat = 0L) {
  if (missing(vcffiles) == TRUE)
    stop("No VCF file specified")
  if (is.character(vcffiles) == FALSE)
    stop("vcfiles must be a character value")
  if (length(vcffiles) != 1 & length(vcffiles) != 2)
    stop("vcffiles must be a character vector of length 1 or 2")
  filename = vcffiles[1]
  if (length(vcffiles) != 1)
    infofile <- vcffiles[2]
  else
    infofile <- ""
  if (filename == "")
    stop("No VCF file specified")

  if (is.logical(gz) == FALSE)
    stop("gz must be a logical value")
  if (length(gz) != 1)
    stop("gz must be a logical vector of length 1")

  if (is.logical(index) == FALSE)
    stop("index must be a logical value")
  if (length(index) != 1)
    stop("index must be a logical vector of length 1")
  if (gz == TRUE && index == TRUE)
    stop("Indexing gzipped files is not supported.")

  if (is.numeric(snpidformat) == FALSE)
    stop("snpidformat must be an integer value")
  if (length(snpidformat) != 1)
    stop("snpidformat must be an interger vector of length 1")
  if (floor(snpidformat) != snpidformat)
    stop("snpidformat must be an integer value")
  snpidformat <- as.integer(snpidformat)
  if (snpidformat < 0 || snpidformat > 2)
    stop("snpidformat must have a value of 0, 1, or 2")

  if (gz == FALSE) {
    con <- file(filename, "r")
  } else {
    con <- gzfile(filename, "r")
  }

  fqfilename <- normalizePath(filename, winslash = '/')

  headerlines <- 1L
  headersize <- -1L
  while (TRUE) {
    currentpos <- seek(con, origin = "current")
    line <- readLines(con, n = 1)
    if (substr(line, 1, 1) != '#') {
      close(con)
      stop("Error processing header")
    }
    if (substr(line, 2, 2) != '#') {
      x <- unlist(strsplit(line, "\t"))
      if (x[1] != "#CHROM") {
        close(con)
        stop("Error processing header")
      }
      x[1] = "CHROM"
      begindata <- seek(con, origin = "current")
      break
    }
    headerlines <- headerlines + 1L
  }
  if (index == TRUE)
    headersize <- begindata
  close(con)

  if (all(x[1:9] == c("CHROM", "POS", "ID", "REF", "ALT", "QUAL", "FILTER", "INFO", "FORMAT")) == FALSE)
    stop("Column names incorrect")

  samples = data.frame(fid = rep("", length(x) - 9L),
                       sid = x[10:length(x)],
                       stringsAsFactors = FALSE)
  usesfid = FALSE

  coltypes = c("character", "integer", rep("character", 7),
               rep("NULL", nrow(samples)))
  snps <- read.table(filename,
                     skip = headerlines,
                     colClasses = coltypes,
                     stringsAsFactors = FALSE)
  colnames(snps) <- c("chromosome", "location", "snpid",
                      "reference", "alternate", "quality",
                      "filter", "info", "format")

  vcfinfo <- as.list(snps[,6:9])
  summaryinfo <- lapply(vcfinfo, summarizevcfadditionalinfo)
  datacolumns <- data.frame(numcolumns = rep(0L, length(summaryinfo$format)),
                            dosage = rep(0L, length(summaryinfo$format)),
                            genotypeprob = rep(0L, length(summaryinfo$format)),
                            genotype = rep(0L, length(summaryinfo$format)),
                            stringsAsFactors = FALSE)
  for (i in 1:length(summaryinfo$format)) {
    formatsplit <- unlist(strsplit(summaryinfo$format[i], split = ':'))
    datacolumns$numcolumns[i] <- length(formatsplit)
    datacolumns$dosage[i] <- match("DS", formatsplit)
    datacolumns$genotypeprob[i] <- match("GP", formatsplit)
    datacolumns$genotype[i] <- match("GT", formatsplit)
  }
  additionalinfo <- list(gzipped = gz,
                         headerlines = headerlines,
                         headersize = headersize,
                         quality = summaryinfo$quality,
                         filter = summaryinfo$filter,
                         info = summaryinfo$info,
                         format = summaryinfo$format,
                         datacolumns = datacolumns)
  class(additionalinfo) <- "vcf-info"
  rm(vcfinfo)
  rm(datacolumns)

  if (infofile == "") {
    snpinfo <- list()
  } else {
    minimacinfo <- readminimacinfofile(infofile)
    if (nrow(minimacinfo) == nrow(snps)) {
      if (all(minimacinfo$SNP == snps$snpid) == TRUE &
          all(minimacinfo$REF.0. == snps$reference) == TRUE &
          all(minimacinfo$ALT.1. == snps$alternate) == TRUE) {
        snpinfo <- list(aaf = as.matrix(minimacinfo$ALT_Frq),
                        maf = as.matrix(minimacinfo$MAF),
                        avgcall = as.matrix(minimacinfo$AvgCall),
                        rsq = as.matrix(minimacinfo$Rsq))
      } else {
        stop("Infromation file does not line up with VCF file - different SNPs")
      }
    } else {
      stop("Information file does not line up with VCF file - different number of SNPs")
    }
  }

  snps <- snps[,1:5]
  chr1 <- snps$chromosome[1]
  oneChr <- all(snps$chromosome == chr1)
  chrlocid <- paste(snps$chromosome, snps$location, sep = ":")
  vcfsnpformat1 <- all(snps$snpid == chrlocid)
  chrlocrefaltid <- paste(snps$chromosome, snps$location,
                          snps$reference, snps$alternate, sep = ":")
  vcfsnpformat2 <- all(snps$snpid == chrlocrefaltid)
  if (snpidformat == 0) {
    if (vcfsnpformat1 == TRUE) {
      snps$snpid <- chrlocid
      snpidformat <- 1L
    } else if (vcfsnpformat2 == TRUE) {
      snps$snpid <- chrlocrefaltid
      snpidformat <- 2L
    }
  } else if (snpidformat == 1) {
    if (vcfsnpformat2 == TRUE)
      stop ("snpidformat 1 specified but VCF file uses snpidformat 2")
    if (vcfsnpformat1 == FALSE)
      snps$snpid <- chrlocid
  } else if (snpidformat == 2) {
    if (vcfsnpformat2 == FALSE)
      snps$snpid <- chrlocrefaltid
  }

  if (index == TRUE) {
    datasize <- integer(nrow(snps))
    indices <- numeric(nrow(snps))
    if (gz == FALSE) {
      x <- GetLineLocations(filename)
      indices <- x[(headerlines + 1):(length(x) - 1)]
      for (i in 1:length(datasize))
        datasize[i] <- x[headerlines + i + 1] - x[headerlines + i]
    }
  } else {
    datasize <- integer(0)
    indices <- numeric(0)
  }

  retval = list(filename = fqfilename,
                usesfid = usesfid,
                samples = samples,
                onechr = oneChr,
                snpidformat = snpidformat,
                snps = snps,
                snpinfo = snpinfo,
                datasize = datasize,
                indices = indices,
                additionalinfo = additionalinfo)
  class(retval) <- c("genetic-info")
  return (retval)
}

###########################################################
#                 GEN (Impute2)                           #
###########################################################

#' Get information about a gen, impute2, file
#'
#' Routine to return information about a gen file.
#' This information is used by other routines to
#' allow for quicker extraction of values from the
#' file.
#'
#' @param genfiles A vector of file names.
#' The first is the name of the gen file. The
#' second is name of the sample file that contains
#' the subject information.
#' @param snpcolumns Column numbers containing chromosome,
#' snpid, location, reference allele, alternate allele,
#' respectively. This must be an integer vector. All
#' values must be positive except for the chromosome.
#' The value for the chromosome may be -1 or -0.
#' -1 indicates that the chromosome value is passed to
#' the routine using the chromosome parameter.
#' 0 indicates that the chromosome value is in the snpid
#' and that the snpid has the format chromosome:other_data.
#' Default value is c(1L, 2L, 3L, 4L, 5L).
#' @param startcolumn Column number of first column with
#' genetic probabilities or dosages. Must
#' be an integer value. Default value is 6L.
#' @param impformat Number of genetic data values per
#' subject. 1 indicates dosage only, 2 indicates P(g=0)
#' and P(g=1) only, 3 indicates P(g=0), P(g=1), and
#' P(g=2). Default value is 3L.
#' @param chromosome Chromosome value to use if the
#' first value of the snpcolumns is equal to 0.
#' Default value is character().
#' @param header Indicators if the gen and sample files
#' have headers. If the gen file does not have a
#' header. A sample file must be included.
#' Default value is c(FALSE, TRUE).
#' @param gz Indicator if file is compressed using gzip.
#' Default value is FALSE.
#' @param index Indicator if file should be indexed. This
#' allows for faster reading of the file. Indexing a gzipped
#' file is not supported.
#' Default value is TRUE.
#' @param snpidformat Format to change the snpid to.
#' 0 indicates to use the snpid format in the file.
#' 1 indicates to change the snpid into chr:pos,
#' 2 indicates to change the snpid into chr:pos:ref:alt,
#' 3 indicates to change the snpid into chr:pos_ref_alt,
#' Default value is 0.
#' @param sep Separators used in the gen file and sample files,
#' respectively. If only value is provided it is used for both
#' files. Default value is c(`"\t"`, `"\t"`)
#'
#' @return List with information about the gen file.
#' This includes family and subject IDs along with
#' a list of the SNPs in the file. Other information needed
#' to read the file is also included.
#' @export
#'
#' @examples
#' # Get file names of th gen and sample file
#' gen3afile <- system.file("extdata", "set3a.imp", package = "BinaryDosage")
#' gen3ainfo <- system.file("extdata", "set3a.sample", package = "BinaryDosage")
#'
#' # Get the information about the gen file
#' geninfo <- getgeninfo(genfiles = c(gen3afile, gen3ainfo),
#'                       snpcolumns = c(0L, 2L:5L))
getgeninfo <- function(genfiles,
                       snpcolumns = 1L:5L,
                       startcolumn = 6L,
                       impformat = 3L,
                       chromosome = character(),
                       header = c(FALSE, TRUE),
                       gz = FALSE,
                       index = TRUE,
                       snpidformat = 0L,
                       sep = c("\t", "\t")) {
  if (missing(genfiles) == TRUE)
    stop("No gen file specified")
  if (is.character(genfiles) == FALSE)
    stop("genfiles must be a character value")
  if (length(genfiles) != 1 & length(genfiles) != 2)
    stop("genfiles must be a character vector of length 1 or 2")
  genfile <- genfiles[1]
  if (length(genfiles) == 1) {
    samplefile <- character()
  } else {
    samplefile <- genfiles[2]
    if (samplefile == "")
      samplefile <- character()
  }
  if (genfile == "")
    stop("No gen file specified")

  if (is.integer(snpcolumns) == FALSE)
    stop("snpcolumns must be an integer vector")
  if (length(snpcolumns) != 5)
    stop("snpcolumns must be an integer vector of length 5")
  if (min(snpcolumns[2:5]) < 1)
    stop("snpcolumns values other than chromosome must be positive integers")
  if (snpcolumns[1] < -1)
    stop("snpcolumns chromosome value must be -1, or a non-negative integer")

  if (is.integer(startcolumn) == FALSE)
    stop("startcolumn must be an integer value")
  if (length(startcolumn) != 1)
    stop("startcolumn must be an integer vector of length 1")
  if (startcolumn < 1)
    stop("startcolumn must be a positive integer")
  if (startcolumn <= max(snpcolumns))
    stop("startcolumn value must be larger than any value in snpcolumns")

  if (is.integer(impformat) == FALSE)
    stop("impformat must be an integer value")
  if (length(impformat) != 1)
    stop("impformat must be an integer vector of length 1")
  if (impformat < 1 | impformat > 3)
    stop("impformat must have a value of 1, 2, or 3")

  if (is.character(chromosome) == FALSE)
    stop("chromosome must be a character variable")
  if (length(chromosome) > 1)
    stop("chromosome must be a character vector of length 0 or 1")
  if (length(chromosome) == 1) {
    if (chromosome == "")
      chromosome = character()
  }
  if (length(chromosome) == 0) {
    if (snpcolumns[1] == -1)
      stop("No chromosome column or value provided")
  } else {
    if (snpcolumns[1] > -1)
      stop("Both chromosome column and chromosome value provided")
  }

  if (is.logical(header) == FALSE)
    stop("header must be a logical value")
  if (length(header) !=  1 & length(header) != 2)
    stop("header must be a logical vector of length 1 or 2")
  if (length(header) == 1)
    header = c(header, TRUE)
  if (header[1] == FALSE) {
    if (length(samplefile) == 0)
      stop("File has no header and no sample file is provided")
  } else {
    if (length(samplefile) > 0)
      stop("header = TRUE and a sample file is provided")
  }

  if (is.logical(gz) == FALSE)
    stop("gz must be a logical value")
  if (length(gz) != 1)
    stop("gz must be a logical vector of length 1")

  if (is.logical(index) == FALSE)
    stop("index must be a logical value")
  if (length(index) != 1)
    stop("index must be a logical vector of length 1")
  if (gz == TRUE & index == TRUE)
    stop("Indexing gzipped files is not supported.")

  if (is.numeric(snpidformat) == FALSE)
    stop("snpidformat must be an integer value")
  if (length(snpidformat) != 1)
    stop("snpidformat must be an interger vector of length 1")
  if (floor(snpidformat) != snpidformat)
    stop("snpidformat must be an integer value")
  snpidformat <- as.integer(snpidformat)
  if (snpidformat < 0 || snpidformat > 3)
    stop("snpidformat must have a value of 0, 1, 2, or 3")

  if (is.character(sep) == FALSE)
    stop("sep must be a character value")
  if (length(sep) != 1 & length(sep) != 2)
    stop("sep must be a character vector of length 1 or 2")
  if (length(sep) == 1)
    sep = c(sep, sep)
  if (sep[1] == "" | sep[2] == "")
    stop("sep values cannot be empty strings")


  if (header[1] == TRUE) {
    if (gz == TRUE)
      filecon <- gzfile(genfile, "r")
    else
      filecon <- file(genfile, "r")
    headerline <- readLines(filecon, 1)
    close(filecon)
    headervalues <- unlist(strsplit(headerline, sep[1]))
    if (length(headervalues) < startcolumn)
      stop("Number of values in header less than startcolumn")
    headervalues <- headervalues[startcolumn:length(headervalues)]
    if (length(headervalues) != 2 * floor(length(headervalues) / 2))
      stop("Odd number of values for family and subject ID")
    fid <- headervalues[seq(1,length(headervalues) - 1, 2)]
    iid <- headervalues[seq(2,length(headervalues), 2)]
    samples <- data.frame(fid = fid,
                          sid = iid,
                          stringsAsFactors = FALSE)
  } else {
    samples <- read.table(samplefile,
                          header = header[2],
                          sep = sep[2],
                          stringsAsFactors = FALSE)
    if (ncol(samples) == 1)
      samples[,2] <- samples[,1]
    samples <- samples[,1:2]
    samples[,1] <- as.character(samples[,1])
    samples[,2] <- as.character(samples[,2])
    colnames(samples) <- c("fid", "sid")
    if (samples[1,1] == "0" & samples[1,2] == "0")
      samples <- samples[2:nrow(samples),]
  }
  if (all(samples$fid == samples$sid)) {
    usesfid <- FALSE
    samples$fid <- ""
  } else {
    usesfid <- TRUE
  }

  coltypes = rep("NULL", impformat * nrow(samples) + (startcolumn - 1))
  if (snpcolumns[1] > 0)
    coltypes[snpcolumns[1]] <- "character"
  coltypes[snpcolumns[c(2, 4, 5)]] <- "character"
  coltypes[snpcolumns[3]] <- "integer"
  headersize <- 0
  if (header[1] == TRUE)
    headersize <- 1
  snps <- read.table(genfile,
                     skip = headersize,
                     colClasses = coltypes,
                     sep = sep[1],
                     stringsAsFactors = FALSE)
  if (snpcolumns[1] == -1) {
    snps$chromosome <- chromosome
    snpcolumns[1] <- startcolumn
  } else if (snpcolumns[1] == 0) {
    chromosomeid <- unlist(strsplit(snps[,1], ':'))
    snpidentries <- length(chromosomeid) / nrow(snps)
    snps$chromosome <- chromosomeid[seq(1, length(chromosomeid) + 1 - snpidentries, snpidentries)]
    snpcolumns[1] <- startcolumn
  }
  x <- snpcolumns[2]
  snpcolumns[2] <- snpcolumns[3]
  snpcolumns[3] <- x
  snps <- snps[,order(order(snpcolumns))]
  colnames(snps) <- c("chromosome", "location", "snpid", "reference", "alternate")

  chr1 <- snps$chromosome[1]
  onechr <- FALSE
  if (all(snps$chromosome == chr1))
    onechr <- TRUE

  if (snpidformat == 0) {
    chrlocid <- unlist(paste(snps$chromosome, snps$location, sep = ':'))
    if (all(chrlocid == snps$snpid)) {
      snpidformat <- 1
    } else {
      chrlocrefaltid <- unlist(paste(snps$chromosome, snps$location,
                                     snps$reference, snps$alternate, sep = ':'))
      if (all(chrlocrefaltid == snps$snpid)) {
        snpidformat <- 2
      } else {
        chrlocrefaltid <- unlist(paste(snps$chromosome, snps$location, sep = ':'))
        chrlocrefaltid <- unlist(paste(chrlocrefaltid, snps$reference,
                                       snps$alternate, sep = '_'))
        if (all(chrlocrefaltid == snps$snpid))
          snpidformat <- 3
      }
    }
  } else if (snpidformat == 1) {
    chrlocrefaltid <- unlist(paste(snps$chromosome, snps$location,
                                   snps$reference, snps$alternate, sep = ':'))
    if (all(chrlocrefaltid == snps$snpid)) {
      stop("snpidformat 1 specified but GEN file uses snpidformat 2")
    }
    snps$snpid <- unlist(paste(snps$chromosome, snps$location, sep = ':'))
  } else if (snpidformat == 2) {
    snps$snpid <- unlist(paste(snps$chromosome, snps$location,
                               snps$reference, snps$alternate, sep = ':'))
  } else {
    chrlocrefaltid <- unlist(paste(snps$chromosome, snps$location, sep = ':'))
    snps$snpid <- unlist(paste(chrlocrefaltid, snps$reference,
                               snps$alternate, sep = '_'))
  }
  snpinfo <- list()
  if (index == FALSE) {
    datasize <- integer(0)
    indices <- numeric(0)
  } else {
    datasize <- integer(nrow(snps))
    indices <- numeric(nrow(snps))
    if (gz == FALSE) {
      x <- GetLineLocations(genfile)
      if (header[1] == TRUE)
        headerlines <- 1
      else
        headerlines <- 0
      indices <- x[(headerlines + 1):(length(x) - 1)]
      for (i in 1:length(datasize))
        datasize[i] <- x[headerlines + i + 1] - x[headerlines + i]
    }
  }

  additionalinfo <- list(gzipped = gz,
                         headersize = headersize,
                         format = impformat,
                         startcolumn = startcolumn,
                         sep = sep[1])
  class(additionalinfo) <- "gen-info"

  retval <- list(filename = normalizePath(genfile, winslash = '/'),
                 usesfid = usesfid,
                 samples = samples,
                 onechr = onechr,
                 snpidformat = snpidformat,
                 snps = snps,
                 snpinfo = snpinfo,
                 datasize = datasize,
                 indices = indices,
                 additionalinfo = additionalinfo)
  class(retval) <- "genetic-info"

  return(retval)
}

## Helper: convert format-4 extdata file to Format 5
make_bd5 <- function(bdfile4) {
  bdose <- tempfile(fileext = ".bdose")
  updatebd(bdfiles = bdfile4, bdose_file = bdose)
  list(bdose = bdose, info = getbdinfo(bdose))
}

test_that("mergebd errors", {
  bdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
  f5     <- make_bd5(bdfile)

  expect_error(mergebd(bdose_file = "x.bdose"),
               "No input .bdose files specified")
  expect_error(mergebd(bdose_files = c(f5$bdose, f5$bdose)),
               "No output .bdose file specified")
  expect_error(mergebd(bdose_files = f5$bdose,
                       bdose_file  = "x.bdose"),
               "At least two input files")

  # Both subjects and SNPs intersect (same file twice) -> error
  expect_error(mergebd(bdose_files = c(f5$bdose, f5$bdose),
                       bdose_file  = "x.bdose"),
               "Cannot merge: subject IDs and SNP IDs both overlap")
})

test_that("mergebd subject merge (two files)", {
  bdfile   <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
  f5       <- make_bd5(bdfile)
  all_info <- f5$info
  all_sids <- all_info$samples$sid   # I1..I60

  # Split into two non-overlapping subject sets (same SNPs)
  bdose_a <- tempfile(fileext = ".bdose")
  bdose_b <- tempfile(fileext = ".bdose")
  subsetbd(bdfiles = f5$bdose, bdose_file = bdose_a, subjectids = all_sids[1:30])
  subsetbd(bdfiles = f5$bdose, bdose_file = bdose_b, subjectids = all_sids[31:60])

  bdose_out <- tempfile(fileext = ".bdose")
  expect_error(mergebd(bdose_files = c(bdose_a, bdose_b),
                       bdose_file  = bdose_out), NA)

  merged <- getbdinfo(bdose_out)

  # All 60 subjects, all 10 SNPs
  expect_equal(nrow(merged$samples), 60L)
  expect_equal(merged$samples$sid, all_sids)
  expect_equal(nrow(merged$snps),   nrow(all_info$snps))
  expect_equal(merged$snps$location, all_info$snps$location)

  # Verify dosage data for all SNPs
  for (i in seq_len(nrow(merged$snps))) {
    orig_snp   <- getsnp(all_info, i, dosageonly = TRUE)
    merged_snp <- getsnp(merged,   i, dosageonly = TRUE)
    expect_equal(merged_snp$dosage, orig_snp$dosage, tolerance = 5e-5,
                 label = paste0("subject merge dosage SNP ", i))
  }
})

test_that("mergebd subject merge (three files)", {
  bdfile   <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
  f5       <- make_bd5(bdfile)
  all_sids <- f5$info$samples$sid

  bdose_a <- tempfile(fileext = ".bdose")
  bdose_b <- tempfile(fileext = ".bdose")
  bdose_c <- tempfile(fileext = ".bdose")
  subsetbd(bdfiles = f5$bdose, bdose_file = bdose_a, subjectids = all_sids[1:20])
  subsetbd(bdfiles = f5$bdose, bdose_file = bdose_b, subjectids = all_sids[21:40])
  subsetbd(bdfiles = f5$bdose, bdose_file = bdose_c, subjectids = all_sids[41:60])

  bdose_out <- tempfile(fileext = ".bdose")
  expect_error(mergebd(bdose_files = c(bdose_a, bdose_b, bdose_c),
                       bdose_file  = bdose_out), NA)

  merged <- getbdinfo(bdose_out)
  expect_equal(nrow(merged$samples), 60L)
  expect_equal(merged$samples$sid, all_sids)
  expect_equal(nrow(merged$snps),  nrow(f5$info$snps))
})

test_that("mergebd SNP merge (two files)", {
  bdfile   <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
  f5       <- make_bd5(bdfile)
  all_info <- f5$info

  # Split into two non-overlapping SNP sets (same subjects)
  locs    <- all_info$snps$location   # 10000..19000
  bdose_a <- tempfile(fileext = ".bdose")
  bdose_b <- tempfile(fileext = ".bdose")
  subsetbd(bdfiles = f5$bdose, bdose_file = bdose_a, locations = locs[1:5])
  subsetbd(bdfiles = f5$bdose, bdose_file = bdose_b, locations = locs[6:10])

  bdose_out <- tempfile(fileext = ".bdose")
  expect_error(mergebd(bdose_files = c(bdose_a, bdose_b),
                       bdose_file  = bdose_out), NA)

  merged <- getbdinfo(bdose_out)

  # All 60 subjects, all 10 SNPs
  expect_equal(nrow(merged$samples), 60L)
  expect_equal(merged$samples$sid,   all_info$samples$sid)
  expect_equal(nrow(merged$snps),    nrow(all_info$snps))
  expect_equal(merged$snps$location, all_info$snps$location)

  # Verify dosage data for all SNPs
  for (i in seq_len(nrow(merged$snps))) {
    orig_snp   <- getsnp(all_info, i, dosageonly = TRUE)
    merged_snp <- getsnp(merged,   i, dosageonly = TRUE)
    expect_equal(merged_snp$dosage, orig_snp$dosage, tolerance = 5e-5,
                 label = paste0("SNP merge dosage SNP ", i))
  }
})

test_that("mergebd SNP merge: subject intersection", {
  bdfile   <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
  f5       <- make_bd5(bdfile)
  all_info <- f5$info
  locs     <- all_info$snps$location
  all_sids <- all_info$samples$sid

  # File A: subjects 1-40, SNPs 1-5
  # File B: subjects 11-60, SNPs 6-10
  # Common subjects: 11-40 (30 subjects)
  bdose_a <- tempfile(fileext = ".bdose")
  bdose_b <- tempfile(fileext = ".bdose")
  subsetbd(bdfiles = f5$bdose, bdose_file = bdose_a,
           locations = locs[1:5], subjectids = all_sids[1:40])
  subsetbd(bdfiles = f5$bdose, bdose_file = bdose_b,
           locations = locs[6:10], subjectids = all_sids[11:60])

  bdose_out <- tempfile(fileext = ".bdose")
  expect_error(mergebd(bdose_files = c(bdose_a, bdose_b),
                       bdose_file  = bdose_out), NA)

  merged <- getbdinfo(bdose_out)
  expect_equal(nrow(merged$samples), 30L)
  expect_equal(merged$samples$sid,   all_sids[11:40])
  expect_equal(nrow(merged$snps),    10L)

  # SNPs 1-5: check dosage against original for common subjects
  for (i in 1:5) {
    orig_snp   <- getsnp(all_info, i, dosageonly = TRUE)
    merged_snp <- getsnp(merged,   i, dosageonly = TRUE)
    expect_equal(merged_snp$dosage, orig_snp$dosage[11:40], tolerance = 5e-5,
                 label = paste0("SNP merge intersection dosage SNP ", i))
  }
})

test_that("bdapply", {
  vcf1abdfile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")
  vcfinfo <- getvcfinfo(vcffiles = vcf1abdfile,
                        index = FALSE)
  vcf1abdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
  expect_error(bdinfo <- getbdinfo(bdfiles = vcf1abdfile), NA)
  func <- getaaf
  aaf1afile <- system.file("extdata", "aaf1a.rds", package = "BinaryDosage")
  aaf1a <- readRDS(aaf1afile)

  expect_error(bdapply(),
               "No binary dosage file information specified")
  expect_error(bdapply(bdinfo = 1,
                       func = func),
               "bdinfo does not contain information about a binary dosage file")
  expect_error(bdapply(bdinfo = vcfinfo,
                       func = func),
               "bdinfo does not contain information about a binary dosage file")

  expect_error(bdapply(bdinfo = bdinfo),
               "No function specified")
  expect_error(bdapply(bdinfo = bdinfo, func = 1),
               "func is not a function")

  expect_equal(unlist(bdapply(bdinfo = bdinfo,
                              func = func)),
               aaf1a,
               tolerance = 4e-5)
})

test_that("bdapply format 5", {
  skip_if_not_installed("vcfppR")
  vcfgzfile  <- system.file("extdata", "set1a.vcf.gz", package = "BinaryDosage")
  bdose_file <- tempfile(fileext = ".bdose")
  expect_error(vcftobd(vcffile = vcfgzfile, bdose_file = bdose_file), NA)
  bd5info <- getbdinfo(bdose_file)

  aaf1afile <- system.file("extdata", "aaf1a.rds", package = "BinaryDosage")
  aaf1a <- readRDS(aaf1afile)

  expect_equal(unlist(bdapply(bdinfo = bd5info,
                              func = getaaf)),
               aaf1a,
               tolerance = 4e-5)
})

test_that("vcfapply", {
  vcf1abdfile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")
  vcfinfo <- getvcfinfo(vcffiles = vcf1abdfile)
  vcf1abdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
  bdinfo <- getbdinfo(bdfiles = vcf1abdfile)
  func <- getaaf
  aaf1afile <- system.file("extdata", "aaf1a.rds", package = "BinaryDosage")
  aaf1a <- readRDS(aaf1afile)

  expect_error(vcfapply(),
               "No vcf file information specified")
  expect_error(vcfapply(vcfinfo = 1,
                        func = func),
               "vcfinfo does not appear to contain information about a vcf file")
  expect_error(vcfapply(vcfinfo = bdinfo,
                        func = func),
               "vcfinfo does not appear to contain information about a vcf file")

  expect_error(vcfapply(vcfinfo = vcfinfo),
               "No function specified")
  expect_error(vcfapply(vcfinfo = vcfinfo, func = 1),
               "func is not a function")

  expect_equal(unlist(vcfapply(vcfinfo = vcfinfo,
                               func = func)),
               aaf1a,
               tolerance = 4e-5)

  vcf1andfile <- system.file("extdata", "set1a_nd.vcf.gz", package = "BinaryDosage")
  expect_error(vcfinfo <- getvcfinfo(vcf1andfile,
                                     gz = TRUE,
                                     index = FALSE),
               NA)
  expect_equal(unlist(vcfapply(vcfinfo = vcfinfo,
                               func = getaaf)),
               aaf1a,
               tolerance = 4e-5)
})

test_that("genapply", {
  gen1afile <- system.file("extdata", "set1a.imp", package = "BinaryDosage")
  geninfo <- getgeninfo(genfiles = gen1afile,
                        snpcolumns = c(1L, 3L, 2L, 4L, 5L),
                        header = TRUE)
  vcf1abdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
  bdinfo <- getbdinfo(bdfiles = vcf1abdfile)
  func <- getaaf
  aaf1afile <- system.file("extdata", "aaf1a.rds", package = "BinaryDosage")
  aaf1a <- readRDS(aaf1afile)

  expect_error(genapply(),
               "No gen file information specified")
  expect_error(genapply(geninfo = 1,
                        func = func),
               "geninfo does not appear to contain information about a gen file")
  expect_error(genapply(geninfo = bdinfo,
                        func = func),
               "geninfo does not appear to contain information about a gen file")

  expect_error(genapply(geninfo = geninfo),
               "No function specified")
  expect_error(genapply(geninfo = geninfo, func = 1),
               "func is not a function")

  expect_equal(unlist(genapply(geninfo = geninfo,
                               func = func)),
               aaf1a,
               tolerance = 4e-5)

  gen4afile <- system.file("extdata", "set4a.imp.gz", package = "BinaryDosage")
  gen4asample <- system.file("extdata", "set4a.sample", package = "BinaryDosage")
  expect_error(geninfo <- getgeninfo(genfiles = c(gen4afile, gen4asample),
                                     snpcolumns = c(1L, 2L, 4L:6L),
                                     startcolumn = 7L,
                                     impformat = 2L,
                                     gz = TRUE,
                                     index = FALSE),
               NA)
  expect_equal(unlist(genapply(geninfo = geninfo,
                               func = func)),
               aaf1a,
               tolerance = 4e-5)

})

test_that("subformat2", {
  aaffile <- system.file("extdata", "aaf1a.rds", package = "BinaryDosage")
  aaf <- readRDS(aaffile)
  # Testing subformat 2
  vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")

  bdfile12 <- tempfile()
  famfile12 <- tempfile()
  mapfile12 <- tempfile()
  expect_error(vcftobdlegacy(vcffiles = vcf1afile,
                       bdfiles = c(bdfile12, famfile12, mapfile12),
                       format = 1L,
                       subformat = 2L),
               NA)
  expect_error(getbdinfo(bdfile12),
               "Binary dosage file format 1, 2, and 3 require family and map files")
  expect_error(bdinfo12 <- getbdinfo(c(bdfile12, famfile12, mapfile12)),
               NA)
  expect_equal(unlist(bdapply(bdinfo12, getaaf)),
               aaf,
               tolerance = 4e-5)

  bdfile22 <- tempfile()
  famfile22 <- tempfile()
  mapfile22 <- tempfile()
  expect_error(vcftobdlegacy(vcffiles = vcf1afile,
                       bdfiles = c(bdfile22, famfile22, mapfile22),
                       format = 2L,
                       subformat = 2L),
               NA)
  expect_error(bdinfo22 <- getbdinfo(c(bdfile22, famfile22, mapfile22)),
               NA)
  expect_equal(unlist(bdapply(bdinfo22, getaaf)),
               aaf,
               tolerance = 4e-5)

  bdfile32 <- tempfile()
  famfile32 <- tempfile()
  mapfile32 <- tempfile()
  expect_error(vcftobdlegacy(vcffiles = vcf1afile,
                       bdfiles = c(bdfile32, famfile32, mapfile32),
                       format = 3L,
                       subformat = 2L),
               NA)
  expect_error(bdinfo32 <- getbdinfo(c(bdfile32, famfile32, mapfile32)),
               NA)
  expect_equal(unlist(bdapply(bdinfo32, getaaf)),
               aaf,
               tolerance = 4e-5)

  bdfile34 <- tempfile()
  famfile34 <- tempfile()
  mapfile34 <- tempfile()
  expect_error(vcftobdlegacy(vcffiles = vcf1afile,
                       bdfiles = c(bdfile34, famfile34, mapfile34),
                       format = 3L,
                       subformat = 4L),
               NA)
  expect_error(bdinfo34 <- getbdinfo(c(bdfile34, famfile34, mapfile34)),
               NA)
  expect_equal(unlist(bdapply(bdinfo34, getaaf)),
               aaf,
               tolerance = 4e-5)

  bdfile42 <- tempfile()
  expect_error(vcftobdlegacy(vcffiles = vcf1afile,
                       bdfiles = bdfile42,
                       format = 4L,
                       subformat = 2L),
               NA)
  expect_error(bdinfo42 <- getbdinfo(bdfile42),
               NA)
  expect_equal(unlist(bdapply(bdinfo42, getaaf)),
               aaf,
               tolerance = 4e-5)

  bdfile44 <- tempfile()
  expect_error(vcftobdlegacy(vcffiles = vcf1afile,
                       bdfiles = bdfile44,
                       format = 4L,
                       subformat = 4L,
                       bdoptions = c("aaf", "maf", "rsq")),
               NA)
  expect_error(bdinfo44 <- getbdinfo(bdfile44),
               NA)
  expect_equal(unlist(bdapply(bdinfo44, getaaf)),
               aaf,
               tolerance = 4e-5)

})

test_that("subformat1", {
  aaffile <- system.file("extdata", "aaf1a.rds", package = "BinaryDosage")
  aaf <- readRDS(aaffile)
  # Testing subformat 1
  vcf2afile <- system.file("extdata", "set2a.vcf", package = "BinaryDosage")

  bdfile11 <- tempfile()
  famfile11 <- tempfile()
  mapfile11 <- tempfile()
  expect_error(vcftobdlegacy(vcffiles = vcf2afile,
                       bdfiles = c(bdfile11, famfile11, mapfile11),
                       format = 1L,
                       subformat = 1L),
               NA)
  expect_error(bdinfo11 <- getbdinfo(c(bdfile11, famfile11, mapfile11)),
               NA)
  expect_equal(unlist(bdapply(bdinfo11, getaaf)),
               aaf,
               tolerance = 4e-5)

  bdfile21 <- tempfile()
  famfile21 <- tempfile()
  mapfile21 <- tempfile()
  expect_error(vcftobdlegacy(vcffiles = vcf2afile,
                       bdfiles = c(bdfile21, famfile21, mapfile21),
                       format = 2L,
                       subformat = 1L),
               NA)
  expect_error(bdinfo21 <- getbdinfo(c(bdfile21, famfile21, mapfile21)),
               NA)
  expect_equal(unlist(bdapply(bdinfo21, getaaf)),
               aaf,
               tolerance = 4e-5)

  bdfile31 <- tempfile()
  famfile31 <- tempfile()
  mapfile31 <- tempfile()
  expect_error(vcftobdlegacy(vcffiles = vcf2afile,
                       bdfiles = c(bdfile31, famfile31, mapfile31),
                       format = 3L,
                       subformat = 1L),
               NA)
  expect_error(bdinfo31 <- getbdinfo(c(bdfile31, famfile31, mapfile31)),
               NA)
  expect_equal(unlist(bdapply(bdinfo31, getaaf)),
               aaf,
               tolerance = 4e-5)

  bdfile33 <- tempfile()
  famfile33 <- tempfile()
  mapfile33 <- tempfile()
  expect_error(vcftobdlegacy(vcffiles = vcf2afile,
                       bdfiles = c(bdfile33, famfile33, mapfile33),
                       format = 3L,
                       subformat = 3L),
               NA)
  expect_error(bdinfo33 <- getbdinfo(c(bdfile33, famfile33, mapfile33)),
               NA)
  expect_equal(unlist(bdapply(bdinfo33, getaaf)),
               aaf,
               tolerance = 4e-5)

  bdfile41 <- tempfile()
  expect_error(vcftobdlegacy(vcffiles = vcf2afile,
                       bdfiles = bdfile41),
               NA)
  expect_error(bdinfo41 <- getbdinfo(bdfiles = bdfile41),
               NA)
  expect_equal(unlist(bdapply(bdinfo41, getaaf)),
               aaf,
               tolerance = 4e-5)

  bdfile43 <- tempfile()
  expect_error(vcftobdlegacy(vcffiles = vcf2afile,
                       bdfiles = bdfile43,
                       format = 4L,
                       subformat = 3L),
               NA)
  expect_error(bdinfo43 <- getbdinfo(bdfiles = bdfile43),
               NA)
  expect_equal(unlist(bdapply(bdinfo43, getaaf)),
               aaf,
               tolerance = 4e-5)
})

test_that("getbdinfo", {
  vcf1abdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")

  expect_error(getbdinfo(),
               "No binary dosage files specified")
  expect_error(getbdinfo(bdfiles = 1),
               "bdfiles must be a character vector")
  expect_error(getbdinfo(bdfiles = character()),
               "bdfiles must be a character vector of length 1 or 3")
  expect_error(getbdinfo(bdfiles = ""),
               "No binary dosage file specified")
  expect_error(getbdinfo(bdfiles = c("file1", "file2", "")),
               "bdfiles contains empty strings")
  expect_error(getbdinfo(bdfiles = c(vcf1abdfile, "file2", "file3")),
               "Binary dosage file format 4 does not use family and map files")

  vcf1abdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
  expect_error(getbdinfo(bdfiles = c(vcf1abdfile, "", "")),
               NA)
  bdtest1 <- system.file("extdata", "test1.bdose", package = "BinaryDosage")
  expect_error(getbdinfo(bdtest1),
               NA)
  bdtest2 <- system.file("extdata", "test2.bdose", package = "BinaryDosage")
  expect_error(getbdinfo(bdtest2),
               NA)
  bdtest3 <- system.file("extdata", "test3.bdose", package = "BinaryDosage")
  expect_error(getbdinfo(bdtest3),
               NA)
})

test_that("getvcfinfo", {
  expect_error(getvcfinfo(),
               "No VCF file specified")
  expect_error(getvcfinfo(vcffiles = 1),
               "vcfiles must be a character value")
  expect_error(getvcfinfo(vcffiles = c("file1", "file2", "file3")),
               "vcffiles must be a character vector of length 1 or 2")
  expect_error(getvcfinfo(vcffiles = character()),
               "vcffiles must be a character vector of length 1 or 2")
  expect_error(getvcfinfo(vcffiles = c("", "file2")),
               "No VCF file specified")

  expect_error(getvcfinfo(vcffiles = "file1",
                          gz = 1L),
               "gz must be a logical value")
  expect_error(getvcfinfo(vcffiles = "file1",
                          gz = c(TRUE, TRUE)),
               "gz must be a logical vector of length 1")
  expect_error(getvcfinfo(vcffiles = "file1",
                          index = 1L),
               "index must be a logical value")
  expect_error(getvcfinfo(vcffiles = "file1",
                          index = c(TRUE, TRUE)),
               "index must be a logical vector of length 1")
  expect_error(getvcfinfo(vcffiles = "file1",
                          gz = TRUE,
                          index = TRUE),
               "Indexing gzipped files is not supported.")

  expect_error(getvcfinfo(vcffiles = "file1",
                          snpidformat = ""),
               "snpidformat must be an integer value")
  expect_error(getvcfinfo(vcffiles = "file1",
                          snpidformat = c(1L, 2L)),
               "snpidformat must be an interger vector of length 1")
  expect_error(getvcfinfo(vcffiles = "file1",
                          snpidformat = 1.1),
               "snpidformat must be an integer value")
  expect_error(getvcfinfo(vcffiles = "file1",
                          snpidformat = 3),
               "snpidformat must have a value of 0, 1, or 2")

  vcfbad1file <- system.file("extdata", "bad1.vcf", package = "BinaryDosage")
  expect_error(vcfinfo <- getvcfinfo(vcffiles = vcfbad1file),
               "Error processing header")
  vcfbad2file <- system.file("extdata", "bad2.vcf", package = "BinaryDosage")
  expect_error(vcfinfo <- getvcfinfo(vcffiles = vcfbad2file),
               "Error processing header")
  vcfbad3file <- system.file("extdata", "bad3.vcf", package = "BinaryDosage")
  expect_error(vcfinfo <- getvcfinfo(vcffiles = vcfbad3file),
               "Column names incorrect")

  vcf5afile <- system.file("extdata", "set5a.vcf.gz", package = "BinaryDosage")
  expect_error(getvcfinfo(vcffiles = vcf5afile,
                          gz = TRUE,
                          index = FALSE),
               NA)

  vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")
  imp2asample <- system.file("extdata", "set2a.sample", package = "BinaryDosage")
  expect_error(getvcfinfo(vcffiles = c(vcf1afile, imp2asample)),
               "Error reading information file - Wrong number of columns")
  bad1info <- system.file("extdata", "bad1.info", package = "BinaryDosage")
  expect_error(getvcfinfo(vcffiles = c(vcf1afile, bad1info)),
               "Error reading information file - Wrong column names")
  vcf1bainfo <- system.file("extdata", "set1b_asnp.info", package = "BinaryDosage")
  expect_error(getvcfinfo(vcffiles = c(vcf1afile, vcf1bainfo)),
               "Information file does not line up with VCF file - different number of SNPs")
  bad2info <- system.file("extdata", "bad2.info", package = "BinaryDosage")
  expect_error(getvcfinfo(vcffiles = c(vcf1afile, bad2info)),
               "Infromation file does not line up with VCF file - different SNPs")
  vcf1ainfo <- system.file("extdata", "set1a.info", package = "BinaryDosage")
  expect_error(getvcfinfo(vcffiles = c(vcf1afile, vcf1ainfo),
                          snpidformat = 1L),
               "snpidformat 1 specified but VCF file uses snpidformat 2")
  vcf1arsfile <- system.file("extdata", "set1ars.vcf", package = "BinaryDosage")
  expect_error(getvcfinfo(vcffiles = vcf1arsfile,
                          snpidformat = 1L),
               NA)
  expect_error(getvcfinfo(vcffiles = vcf1arsfile,
                          snpidformat = 2L),
               NA)

})

test_that("getgeninfo", {
  expect_error(getgeninfo(),
               "No gen file specified")
  expect_error(getgeninfo(genfiles = 1L),
               "genfiles must be a character value")
  expect_error(getgeninfo(genfiles = c("file1", "file2", "file3")),
               "genfiles must be a character vector of length 1 or 2")
  expect_error(getgeninfo(genfiles = character()),
               "genfiles must be a character vector of length 1 or 2")
  expect_error(getgeninfo(genfiles = c("", "file2")),
               "No gen file specified")

  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          snpcolumns = ""),
               "snpcolumns must be an integer vector")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          snpcolumns = 1L:3L),
               "snpcolumns must be an integer vector of length 5")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          snpcolumns = c(1L, 0L, 2L, 3L, 4L)),
               "snpcolumns values other than chromosome must be positive integers")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          snpcolumns = c(-2L, 1L, 2L, 3L, 4L)),
               "snpcolumns chromosome value must be -1, or a non-negative integer")

  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          startcolumn = ""),
               "startcolumn must be an integer value")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          startcolumn = 1L:3L),
               "startcolumn must be an integer vector of length 1")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          startcolumn = -1L),
               "startcolumn must be a positive integer")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          startcolumn = 1L),
               "startcolumn value must be larger than any value in snpcolumns")

  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          impformat = ""),
               "impformat must be an integer value")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          impformat = 1L:3L),
               "impformat must be an integer vector of length 1")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          impformat = 4L),
               "impformat must have a value of 1, 2, or 3")

  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          chromosome = 1),
               "chromosome must be a character variable")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          chromosome = c("a", "b")),
               "chromosome must be a character vector of length 0 or 1")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          snpcolumns = c(-1L,2L:5L)),
               "No chromosome column or value provided")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          chromosome = "1"),
               "Both chromosome column and chromosome value provided")

  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          header = 1),
               "header must be a logical value")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          header = c(TRUE, TRUE, TRUE)),
               "header must be a logical vector of length 1 or 2")
  expect_error(getgeninfo(genfiles = c("file1", ""),
                          header = FALSE),
               "File has no header and no sample file is provided")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          header = TRUE),
               "header = TRUE and a sample file is provided")

  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          gz = 1L),
               "gz must be a logical value")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          gz = c(TRUE, TRUE)),
               "gz must be a logical vector of length 1")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          index = 1L),
               "index must be a logical value")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          index = c(TRUE, TRUE)),
               "index must be a logical vector of length 1")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          gz = TRUE,
                          index = TRUE),
               "Indexing gzipped files is not supported.")

  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          snpidformat = ""),
               "snpidformat must be an integer value")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          snpidformat = 1L:3L),
               "snpidformat must be an interger vector of length 1")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          snpidformat = 1.1),
               "snpidformat must be an integer value")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          snpidformat = 5L),
               "snpidformat must have a value of 0, 1, 2, or 3")

  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          sep = 1L),
               "sep must be a character value")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          sep = c("a", "b", "c")),
               "sep must be a character vector of length 1 or 2")
  expect_error(getgeninfo(genfiles = c("file1", "file2"),
                          sep = ""),
               "sep values cannot be empty strings")

  gen5afile <- system.file("extdata", "set5a.imp.gz", package = "BinaryDosage")
  expect_error(getgeninfo(genfiles = gen5afile,
                          snpcolumns = c(1L,3L,2L,4L,5L),
                          header = TRUE,
                          gz = TRUE,
                          index = FALSE,
                          snpidformat = 1L),
               "snpidformat 1 specified but GEN file uses snpidformat 2")
  gen3asample <- system.file("extdata", "set3a.sample", package = "BinaryDosage")
  expect_error(getgeninfo(genfiles = gen3asample,
                          snpcolumns = c(1L,3L,2L,4L,5L),
                          header = TRUE),
               "Number of values in header less than startcolumn")
  genbad1file <- system.file("extdata", "bad1.imp", package = "BinaryDosage")
  expect_error(getgeninfo(genfiles = genbad1file,
                          snpcolumns = c(1L,3L,2L,4L,5L),
                          header = TRUE),
               "Odd number of values for family and subject ID")
  gen3afile <- system.file("extdata", "set3a.imp", package = "BinaryDosage")
  gen3asample1 <- system.file("extdata", "set3a.sample1", package = "BinaryDosage")
  expect_error(getgeninfo(genfiles = c(gen3afile, gen3asample1),
                          snpcolumns = c(-1L, 2L:5L),
                          chromosome = "X",
                          snpidformat = 3L),
               NA)

  gen4afile <- system.file("extdata", "set4a.imp.gz", package = "BinaryDosage")
  gen4asample <- system.file("extdata", "set4a.sample", package = "BinaryDosage")
  expect_error(getgeninfo(genfiles = c(gen4afile, gen4asample),
                          snpcolumns = c(1L,2L,4L:6L),
                          startcolumn = 7L,
                          impformat = 2L,
                          gz = TRUE,
                          index = FALSE,
                          snpidformat = 1L),
               NA)
  expect_error(getgeninfo(genfiles = c(gen4afile, gen4asample),
                          snpcolumns = c(1L,2L,4L:6L),
                          startcolumn = 7L,
                          impformat = 2L,
                          gz = TRUE,
                          index = FALSE,
                          snpidformat = 2L),
               NA)
})

## Helper: compute MAF of a SNP over a subset of subject indices
snp_maf <- function(bdinfo, snp_idx, samp_idx = NULL) {
  snp <- getsnp(bdinfo, snp_idx, dosageonly = TRUE)
  ds  <- if (is.null(samp_idx)) snp$dosage else snp$dosage[samp_idx]
  aaf <- mean(ds, na.rm = TRUE) / 2.0
  min(aaf, 1.0 - aaf)
}

test_that("subsetbd errors", {
  bdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")

  expect_error(subsetbd(bdose_file = "x.bdose"),
               "No input binary dosage files specified")
  expect_error(subsetbd(bdfiles = bdfile),
               "No output .bdose file specified")
  expect_error(subsetbd(bdfiles = bdfile, bdose_file = "x.bdose"),
               "At least one of")
  expect_error(subsetbd(bdfiles = bdfile, bdose_file = "x.bdose",
                        locations = c(10000, 11000),
                        startloc = 10000),
               "locations and startloc/endloc cannot both be specified")
  expect_error(subsetbd(bdfiles = bdfile, bdose_file = "x.bdose",
                        startloc = 10000),
               "endloc must be specified")
  expect_error(subsetbd(bdfiles = bdfile, bdose_file = "x.bdose",
                        endloc = 19000),
               "startloc must be specified")
  expect_error(subsetbd(bdfiles = bdfile, bdose_file = "x.bdose",
                        startloc = 15000, endloc = 10000),
               "startloc must be less than or equal to endloc")
  expect_error(subsetbd(bdfiles = bdfile, bdose_file = "x.bdose",
                        minmaf = "high"),
               "minmaf must be a single numeric value")
  expect_error(subsetbd(bdfiles = bdfile, bdose_file = "x.bdose",
                        minmaf = 0.6),
               "minmaf must be between 0 and 0.5")
  expect_error(subsetbd(bdfiles = bdfile, bdose_file = "x.bdose",
                        subjectids = c("X1", "X2")),
               "No subjects match")
  expect_error(subsetbd(bdfiles = bdfile, bdose_file = "x.bdose",
                        minmaf = 0.5),
               "No SNPs pass")
})

test_that("subsetbd subject filter", {
  bdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
  old_info <- getbdinfo(bdfile)
  keep_ids <- old_info$samples$sid[1:30]

  bdose <- tempfile(fileext = ".bdose")
  expect_error(subsetbd(bdfiles = bdfile, bdose_file = bdose,
                        subjectids = keep_ids), NA)

  new_info <- getbdinfo(bdose)
  expect_equal(nrow(new_info$samples), 30L)
  expect_equal(new_info$samples$sid, keep_ids)
  expect_equal(nrow(new_info$snps),   nrow(old_info$snps))

  # Verify dosage values match for a sample of SNPs
  for (i in c(1L, 5L, 10L)) {
    old_snp <- getsnp(old_info, i, dosageonly = TRUE)
    new_snp <- getsnp(new_info, i, dosageonly = TRUE)
    expect_equal(new_snp$dosage, old_snp$dosage[1:30], tolerance = 5e-5,
                 label = paste0("subject filter dosage SNP ", i))
  }
})

test_that("subsetbd location vector filter", {
  bdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
  old_info <- getbdinfo(bdfile)
  keep_locs <- c(12000, 15000, 17000)   # SNPs 3, 6, 8

  bdose <- tempfile(fileext = ".bdose")
  expect_error(subsetbd(bdfiles = bdfile, bdose_file = bdose,
                        locations = keep_locs), NA)

  new_info <- getbdinfo(bdose)
  expect_equal(nrow(new_info$snps), length(keep_locs))
  expect_equal(new_info$snps$location, keep_locs)
  expect_equal(nrow(new_info$samples), nrow(old_info$samples))

  # Verify data for each kept SNP
  orig_idx <- match(keep_locs, old_info$snps$location)
  for (j in seq_along(keep_locs)) {
    old_snp <- getsnp(old_info, orig_idx[j], dosageonly = TRUE)
    new_snp <- getsnp(new_info, j, dosageonly = TRUE)
    expect_equal(new_snp$dosage, old_snp$dosage, tolerance = 5e-5,
                 label = paste0("location filter dosage SNP loc=", keep_locs[j]))
  }
})

test_that("subsetbd location range filter", {
  bdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
  old_info <- getbdinfo(bdfile)
  # startloc=13000, endloc=16000 keeps SNPs 4,5,6,7 (locs 13000-16000)
  exp_locs <- c(13000, 14000, 15000, 16000)

  bdose <- tempfile(fileext = ".bdose")
  expect_error(subsetbd(bdfiles = bdfile, bdose_file = bdose,
                        startloc = 13000, endloc = 16000), NA)

  new_info <- getbdinfo(bdose)
  expect_equal(new_info$snps$location, exp_locs)

  orig_idx <- match(exp_locs, old_info$snps$location)
  for (j in seq_along(exp_locs)) {
    old_snp <- getsnp(old_info, orig_idx[j], dosageonly = TRUE)
    new_snp <- getsnp(new_info, j,            dosageonly = TRUE)
    expect_equal(new_snp$dosage, old_snp$dosage, tolerance = 5e-5,
                 label = paste0("range filter dosage loc=", exp_locs[j]))
  }
})

test_that("subsetbd MAF filter", {
  bdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
  old_info <- getbdinfo(bdfile)
  min_maf  <- 0.25

  # Determine expected kept SNPs
  keep_snps <- which(vapply(seq_len(nrow(old_info$snps)),
                            function(i) snp_maf(old_info, i) >= min_maf,
                            logical(1L)))

  bdose <- tempfile(fileext = ".bdose")
  expect_error(subsetbd(bdfiles = bdfile, bdose_file = bdose,
                        minmaf = min_maf), NA)

  new_info <- getbdinfo(bdose)
  expect_equal(nrow(new_info$snps), length(keep_snps))
  expect_equal(new_info$snps$location,
               old_info$snps$location[keep_snps])

  for (j in seq_along(keep_snps)) {
    old_snp <- getsnp(old_info, keep_snps[j], dosageonly = TRUE)
    new_snp <- getsnp(new_info, j,             dosageonly = TRUE)
    expect_equal(new_snp$dosage, old_snp$dosage, tolerance = 5e-5,
                 label = paste0("maf filter dosage SNP ", keep_snps[j]))
  }
})

test_that("subsetbd combined filters", {
  bdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
  old_info <- getbdinfo(bdfile)

  keep_ids <- old_info$samples$sid[1:30]
  samp_idx <- 1:30
  min_maf  <- 0.25

  # Expected: range 13000-18000, maf >= 0.25 over first 30 subjects
  exp_locs <- c(13000, 14000, 15000, 16000, 17000, 18000)
  loc_snps <- match(exp_locs, old_info$snps$location)
  keep_snps <- loc_snps[vapply(loc_snps,
                               function(i) snp_maf(old_info, i, samp_idx) >= min_maf,
                               logical(1L))]

  bdose <- tempfile(fileext = ".bdose")
  expect_error(subsetbd(bdfiles    = bdfile,
                        bdose_file = bdose,
                        minmaf     = min_maf,
                        startloc   = 13000,
                        endloc     = 18000,
                        subjectids = keep_ids), NA)

  new_info <- getbdinfo(bdose)
  expect_equal(nrow(new_info$samples), 30L)
  expect_equal(new_info$samples$sid, keep_ids)
  expect_equal(new_info$snps$location, old_info$snps$location[keep_snps])

  for (j in seq_along(keep_snps)) {
    old_snp <- getsnp(old_info, keep_snps[j], dosageonly = TRUE)
    new_snp <- getsnp(new_info, j,             dosageonly = TRUE)
    expect_equal(new_snp$dosage, old_snp$dosage[samp_idx], tolerance = 5e-5,
                 label = paste0("combined filter dosage SNP ", keep_snps[j]))
  }
})

test_that("subsetbd format 5 input", {
  # Verify subsetbd also accepts format 5 as input
  bdfile4   <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
  bdose5_in <- tempfile(fileext = ".bdose")
  updatebd(bdfiles = bdfile4, bdose_file = bdose5_in)

  bdose_out <- tempfile(fileext = ".bdose")
  expect_error(subsetbd(bdfiles    = bdose5_in,
                        bdose_file = bdose_out,
                        startloc   = 12000,
                        endloc     = 14000), NA)

  new_info <- getbdinfo(bdose_out)
  expect_equal(new_info$snps$location, c(12000, 13000, 14000))
})

#***************************************************************************#
#                                                                           #
#                       Writing Binary Dosage files                         #
#                                                                           #
#***************************************************************************#


#***************************************************************************#
#             Support functions for subject and SNP data                    #
#***************************************************************************#

# Formats 1, 2, and 3 all have a separate family and map file
# This routine saves the data frames in RDS format
WriteFamilyAndMapFiles <- function(filename, samples, snps) {
  saveRDS(samples, filename[2])
  saveRDS(snps, filename[3])
  return (md5 <- c(digest(samples, "md5"), digest(snps, "md5")))
}

# Find the groups value in the genetic file info. If it doesn't
# exist it returns the number of samples
FindGroups <- function(geneticfileinfo) {
  x <- match("groups", names(geneticfileinfo$additionalinfo))
  if (is.na(x) == TRUE)
    return (nrow(geneticfileinfo$samples))
  return(geneticfileinfo$additional$groups)
}

# Create the subject and family strings to write to the binary
# dosage header for format 4
SIDandFID4 <- function(genefileinfo) {
  sid <- paste(genefileinfo$samples$sid, collapse = '\t')
  if (genefileinfo$usesfid == TRUE)
    fid <- paste(genefileinfo$samples$fid, collapse = '\t')
  else
    fid <- ""
  return (list(sid = sid, fid = fid))
}

# Find a numeric vector in the SNP info data frame or the
# bdoptions. If option is specified, a vector of zeros is
# return. If it is not found in the options, it
# returns the vector if found in the SNP info, otherwise
# it returns a numeric vector of length 0
FindSNPInfoNumeric <- function(tofind, snpinfo, numsnps, bdoptions) {
  if (is.na(match(tofind, bdoptions)) == FALSE)
    return (numeric(numsnps))
  infocol <- match(tofind, names(snpinfo))
  if (is.na(infocol) == FALSE)
    return (as.vector(snpinfo[[infocol]]))
  return (numeric(0))
}

# Find the SNP information needed for format 4
FindBDSNPInfo <- function(genefileinfo, bdoptions) {
  snpcolnames <- colnames(genefileinfo$snps)
  if (genefileinfo$snpidformat == 0)
    snpid <- genefileinfo$snps$snpid
  else
    snpid <- ""
  chr <- genefileinfo$snps$chromosome
  if (genefileinfo$onechr == TRUE)
    chr <- chr[1]
  loc <- genefileinfo$snps$location
  ref <- genefileinfo$snps$reference
  alt <- genefileinfo$snps$alternate

  aaf <- FindSNPInfoNumeric("aaf", genefileinfo$snpinfo, nrow(genefileinfo$snps), bdoptions)
  maf <- FindSNPInfoNumeric("maf", genefileinfo$snpinfo, nrow(genefileinfo$snps), bdoptions)
  # Average call cannot be calculated, therefore bdoptions are meaningless
  avgcall <- FindSNPInfoNumeric("avgcall", genefileinfo$snpinfo, nrow(genefileinfo$snps), "")
  rsq <- FindSNPInfoNumeric("rsq", genefileinfo$snpinfo, nrow(genefileinfo$snps), bdoptions)

  snpid <- paste0(snpid, collapse = '\t')
  chr <- paste0(chr, collapse = '\t')
  ref <- paste0(ref, collapse = '\t')
  alt <- paste0(alt, collapse = '\t')
  return (list(snpid = snpid,
               chromosome = chr,
               location = loc,
               reference = ref,
               alternate = alt,
               aaf = aaf,
               maf = maf,
               avgcall = avgcall,
               rsq = rsq))
}

#***************************************************************************#
#                                                                           #
#                     Writing the Binary Dosage header                      #
#                                                                           #
#***************************************************************************#


# Writes the header for the various formats of the formats
# of the binary dosage file. These vary for all the different
# formats.
WriteBinaryDosageHeader <- function(format, subformat, filename, genefileinfo, bdoptions) {
  writeHeaderFunc <- list(f1 <- c(WriteBinaryDosageHeader1, WriteBinaryDosageHeader1),
                          f2 <- c(WriteBinaryDosageHeader1, WriteBinaryDosageHeader1),
                          f3 <- c(WriteBinaryDosageHeader31, WriteBinaryDosageHeader31, WriteBinaryDosageHeader33, WriteBinaryDosageHeader34),
                          f4 <- c(WriteBinaryDosageHeader41, WriteBinaryDosageHeader41, WriteBinaryDosageHeader43, WriteBinaryDosageHeader44))
  WriteBinaryDosageBaseHeader(filename[1], format - 1, subformat - 1)
  return (writeHeaderFunc[[format]][[subformat]](filename, genefileinfo, bdoptions))
}

WriteBinaryDosageHeader1 <- function(filename, genefileinfo, bdoptions) {
  md5 <- WriteFamilyAndMapFiles(filename, genefileinfo$samples, genefileinfo$snps)
  return (0)
}

WriteBinaryDosageHeader31 <- function(filename, genefileinfo, bdoptions) {
  md5 <- WriteFamilyAndMapFiles(filename, genefileinfo$samples, genefileinfo$snps)
  WriteBinaryDosageHeader3A(filename[1], nrow(genefileinfo$samples))
  return (0)
}

WriteBinaryDosageHeader33 <- function(filename, genefileinfo, bdoptions) {
  md5 <- WriteFamilyAndMapFiles(filename, genefileinfo$samples, genefileinfo$snps)
  WriteBinaryDosageHeader3B(filename[1], md5[1], md5[2], 0)
  return (0)
}

WriteBinaryDosageHeader34 <- function(filename, genefileinfo, bdoptions) {
  md5 <- WriteFamilyAndMapFiles(filename, genefileinfo$samples, genefileinfo$snps)
  WriteBinaryDosageHeader3B(filename[1], md5[1], md5[2], nrow(genefileinfo$snps))
  return (0)
}

WriteBinaryDosageHeader4 <- function(filename, genefileinfo, bdoptions,
                                     headerEntries, offsets, numindices) {
  subInfo <- SIDandFID4(genefileinfo)
  snpInfo <- FindBDSNPInfo(genefileinfo, bdoptions)

  WriteBinaryDosageHeader4A(filename[1],
                            headerEntries,
                            nrow(genefileinfo$samples),
                            nrow(genefileinfo$snps),
                            FindGroups(genefileinfo),
                            subInfo$sid[1],
                            subInfo$fid[1],
                            snpInfo$snpid[1],
                            snpInfo$chromosome[1],
                            snpInfo$location,
                            snpInfo$reference[1],
                            snpInfo$alternate[1],
                            snpInfo$aaf,
                            snpInfo$maf,
                            snpInfo$avgcall,
                            snpInfo$rsq,
                            offsets,
                            numindices)
  return (0)
}

WriteBinaryDosageHeader41 <- function(filename, genefileinfo, bdoptions) {
  headerEntries <- 8L
  offsets <- c(seq(8L, 36L, 4L), 36L)
  return (WriteBinaryDosageHeader4(filename, genefileinfo, bdoptions,
                                   headerEntries, offsets, 0))
}

WriteBinaryDosageHeader43 <- function(filename, genefileinfo, bdoptions) {
  headerEntries <- 4
  offsets <- c(rep(-1L, 5), seq(8L, 20L, 4L))
  return (WriteBinaryDosageHeader4(filename, genefileinfo, bdoptions,
                                   headerEntries, offsets, 0))
}

WriteBinaryDosageHeader44 <- function(filename, genefileinfo, bdoptions) {
  headerEntries <- 4
  offsets <- c(rep(-1L, 5), seq(8L, 20L, 4L))
  return (WriteBinaryDosageHeader4(filename, genefileinfo, bdoptions,
                                   headerEntries, offsets, nrow(genefileinfo$snps)))
}

#***************************************************************************#
#                                                                           #
#                     Writing the Binary Dosage data                        #
#                                                                           #
#***************************************************************************#

#***************************************************************************#
#                        Allocate memory                                    #
#***************************************************************************#

# Allocates memory needed to write binary dosage files
# This is sufficient for all formats
AllocateBinaryDosageWriteMemory <- function(headerinfo) {
  filename <- headerinfo$filename
  format <- headerinfo$additionalinfo$format
  subformat <- headerinfo$additionalinfo$subformat
  headersize <- headerinfo$additionalinfo$headersize
  snpnumber <- integer(1)
  datasize <- integer(nrow(headerinfo$snps))
  us <- integer(2*nrow(headerinfo$samples))
  return(list(filename = filename,
              format = format,
              subformat = subformat,
              headersize = headersize,
              snpnumber = snpnumber,
              datasize = datasize,
              us = us))
}

#***************************************************************************#
#                        Write the data                                     #
#***************************************************************************#

# Write binary dosage data at the end of the file
# Header has already been written
# writeinfo was already created using AllocateBinaryDosageWriteMemory (see above)
WriteBinaryDosageData <- function(dosage, p0, p1, p2, writeinfo) {
  writeFunc <- list(f1 <- c(WriteBinaryDosageData1, WriteBinaryDosageData2),
                    f2 <- c(WriteBinaryDosageData3, WriteBinaryDosageData4),
                    f3 <- c(WriteBinaryDosageData3, WriteBinaryDosageData5, WriteBinaryDosageData3, WriteBinaryDosageData6),
                    f4 <- c(WriteBinaryDosageData3, WriteBinaryDosageData5, WriteBinaryDosageData3, WriteBinaryDosageData6))
  return (writeFunc[[writeinfo$format]][[writeinfo$subformat]](writeinfo, dosage, p0, p1, p2))
}

WriteBinaryDosageData1 <- function(writeinfo, dosage, p0, p1, p2) {
  return (WriteBinaryDosageDataC(writeinfo$filename, dosage, writeinfo$us, 1))
}

WriteBinaryDosageData2 <- function(writeinfo, dosage, p0, p1, p2) {
  return (WriteBinaryP1P2Data(writeinfo$filename, p1, p2, writeinfo$us,  2))
}

WriteBinaryDosageData3 <- function(writeinfo, dosage, p0, p1, p2) {
  return (WriteBinaryDosageDataC(writeinfo$filename, dosage, writeinfo$us, 3))
}

WriteBinaryDosageData4 <- function(writeinfo, dosage, p0, p1, p2) {
  return (WriteBinaryP1P2Data(writeinfo$filename, p1, p2, writeinfo$us, 3))
}

WriteBinaryDosageData5 <- function(writeinfo, dosage, p0, p1, p2) {
  snpnumber <- -1L
  return (WriteBinaryCompressed(writeinfo$filename,
                                dosage, p0, p1, p2,
                                snpnumber,
                                writeinfo$datasize,
                                writeinfo$us))
}

WriteBinaryDosageData6 <- function(writeinfo, dosage, p0, p1, p2) {
  return (WriteBinaryCompressed(writeinfo$filename,
                                dosage, p0, p1, p2,
                                writeinfo$snpnumber,
                                writeinfo$datasize,
                                writeinfo$us))
}

#***************************************************************************#
#                                                                           #
#                     Writing the Binary Dosage indices                     #
#                                                                           #
#***************************************************************************#

# Write binary dosage indices to  the file
# Header has already been written
# funcData was already created using AllocateBinaryDosageWriteMemory (see above)
WriteBinaryDosageIndices <- function(writeinfo) {
  writeFunc <- list(f1 <- c(WriteBinaryDosageIndices1, WriteBinaryDosageIndices1),
                    f2 <- c(WriteBinaryDosageIndices1, WriteBinaryDosageIndices1),
                    f3 <- c(WriteBinaryDosageIndices1, WriteBinaryDosageIndices1, WriteBinaryDosageIndices1, WriteBinaryDosageIndices2),
                    f4 <- c(WriteBinaryDosageIndices1, WriteBinaryDosageIndices1, WriteBinaryDosageIndices1, WriteBinaryDosageIndices2))
  return (writeFunc[[writeinfo$format]][[writeinfo$subformat]](writeinfo))
}

WriteBinaryDosageIndices1 <- function(writeinfo) {
  return (0)
}

WriteBinaryDosageIndices2 <- function(writeinfo) {
  return(WriteBinaryDosageIndicesC(writeinfo$filename, writeinfo$headersize, writeinfo$datasize))
}

#***************************************************************************#
#                                                                           #
#                    Update snp info, aaf, maf, and rsq                     #
#                                                                           #
#***************************************************************************#
#' Calculate alternate allele frequency
#'
#' Routine to calculate the alternate allele frequency given the dosages.
#' Missing values for dosage ignored. This function is used internally and
#' is exported for use in examples.
#'
#' @param dosage Dosage values
#' @param p0 Pr(g=0) - unused
#' @param p1 Pr(g=1) - unused
#' @param p2 Pr(g=2) - unused
#'
#' @return
#' Alternate allele frequency
#' @export
#'
#' @examples
#' # Get information about binary dosage file
#' bdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
#' bdinfo <- getbdinfo(bdfiles = bdfile)
#' snp1 <- getsnp(bdinfo = bdinfo, 1)
#' aaf <- getaaf(snp1$dosage)
getaaf <- function(dosage, p0, p1, p2) {
  return (mean(dosage, na.rm = TRUE) / 2)
}

#' Calculate minor allele frequency
#'
#' Routine to calculate the minor allele frequency given the dosages.
#' Missing values for dosage ignored. This function is used internally and
#' is exported for use in examples. Note: The minor allele in one data set
#' may be different from another data set. This can make comparing minor
#' allele frequencies between data sets nonsensical.
#'
#' @param dosage Dosage values
#' @param p0 Pr(g=0) - unused
#' @param p1 Pr(g=1) - unused
#' @param p2 Pr(g=2) - unused
#'
#' @return
#' Minor allele frequency
#' @export
#'
#' @examples
#' # Get information about binary dosage file
#' bdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
#' bdinfo <- getbdinfo(bdfiles = bdfile)
#' snp1 <- getsnp(bdinfo = bdinfo, 1)
#' maf <- getmaf(snp1$dosage)
getmaf <- function(dosage, p0, p1, p2) {
  aaf <- mean(dosage, na.rm = TRUE) / 2
  maf <- ifelse(aaf > 0.5, 1. - aaf, aaf)
  return (maf)
}

#' Calculate imputation r squared
#'
#' Routine to calculate the imputation r squared given the dosages
#' and Pr(g=2).
#' This is an estimate for the imputation r squared returned from
#' minimac and impute2. The r squared values are calculated slightly
#' differently between the programs. This estimate is based on the
#' method used by minimac. It does well for minor allele frequencies
#' above 5%. This function is used internally and is exported for
#' use in examples.
#'
#' @param dosage Dosage values
#' @param p0 Pr(g=0) - unused
#' @param p1 Pr(g=1) - unused
#' @param p2 Pr(g=2)
#'
#' @return
#' Imputation r squared
#' @export
#'
#' @examples
#' # Get information about binary dosage file
#' bdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
#' bdinfo <- getbdinfo(bdfiles = bdfile)
#' snp1 <- getsnp(bdinfo = bdinfo, 1, dosageonly = FALSE)
#' rsq <- BinaryDosage:::getrsq(snp1$dosage, p2 = snp1$p2)
getrsq <- function(dosage, p0, p1, p2) {
  q <- numeric(2 * length(dosage))
  d <- dosage * dosage - 4 * p2
  d <- ifelse(d < 0 & d > -0.01, 0., d)
  d <- sqrt(d)
  q[1:length(dosage)] <- 0.5 * (dosage - d)
  q[(length(dosage) + 1):(2*length(dosage))] <- 0.5 * (dosage + d)
  q <- q[is.na(q) == FALSE]
  mu <- mean(q)
  sigma <- mean(q*q) - mu * mu
  rsq <- sigma / (mu * (1. - mu))
  return (rsq)
}

updateaaf <- function (bdinfo) {
  if (bdinfo$additionalinfo$subformat < 3)
    headerinfo <- ReadBinaryDosageHeader4A(bdinfo$filename)
  else
    headerinfo <- ReadBinaryDosageHeader4B(bdinfo$filename)

  aaf <- unlist(bdapply(bdinfo, getaaf))

  updatesnpinfo(bdinfo$filename, headerinfo$snps$aafoffset, aaf)
}

updatemaf <- function (bdinfo) {
  if (bdinfo$additionalinfo$subformat < 3)
    headerinfo <- ReadBinaryDosageHeader4A(bdinfo$filename)
  else
    headerinfo <- ReadBinaryDosageHeader4B(bdinfo$filename)

  maf <- unlist(bdapply(bdinfo, getmaf))

  updatesnpinfo(bdinfo$filename, headerinfo$snps$mafoffset, maf)
}

updatersq <- function (bdinfo) {
  if (bdinfo$additionalinfo$subformat < 3)
    headerinfo <- ReadBinaryDosageHeader4A(bdinfo$filename)
  else
    headerinfo <- ReadBinaryDosageHeader4B(bdinfo$filename)

  rsq <- unlist(bdapply(bdinfo, getrsq))

  updatesnpinfo(bdinfo$filename, headerinfo$snps$rsqoffset, rsq)
}

#***************************************************************************#
#                                                                           #
#                       Reading Binary Dosage files                         #
#                                                                           #
#***************************************************************************#


#***************************************************************************#
#                                                                           #
#                     Reading the Binary Dosage header                      #
#                                                                           #
#***************************************************************************#

# Reads the header for the various formats of the formats
# of the binary dosage file. These vary for all the different
# formats.
ReadBinaryDosageHeader <- function(filename) {
  ReadHeaderFunc <- list(f1 <- c(ReadBinaryDosageHeader11, ReadBinaryDosageHeader12),
                         f2 <- c(ReadBinaryDosageHeader21, ReadBinaryDosageHeader22),
                         f3 <- c(ReadBinaryDosageHeader31, ReadBinaryDosageHeader32, ReadBinaryDosageHeader33, ReadBinaryDosageHeader34),
                         f4 <- c(ReadBinaryDosageHeader41, ReadBinaryDosageHeader42, ReadBinaryDosageHeader43, ReadBinaryDosageHeader44))
  bdformat <- ReadBinaryDosageBaseHeader(filename[1])
  if (is.na(match("error", names(bdformat))) == FALSE)
    stop(bdformat$error)
  if (bdformat$format == 4) {
    if (length(filename) != 1)
      stop("Binary dosage file format 4 does not use family and map files")
  } else {
    if (length(filename) != 3)
      stop("Binary dosage file format 1, 2, and 3 require family and map files")
  }
  return (ReadHeaderFunc[[bdformat$format]][[bdformat$subformat]](filename))
}

#***************************************************************************#
#             Routines to read the subject and SNP info                     #
#***************************************************************************#

# Function to read subject and map files for formats 1, 2, and 3
# @parameter filename - vector of files names for binary dosage files
# first value is the binary dosage file, the second value is the fam file
# and the third is the map file
# @parameter format - format of the binary dosage file
# @parameter subformat - subformat of the binary dosage file
# @parameter headersize - size of the binary dosage header in bytes
# @return - The subject and family data needed to create a bdinfo list
ReadFamAndMapFiles <- function(filename, format, subformat, headersize) {
  fqfilename <- normalizePath(filename[1], winslash = '/')
  samples <- readRDS(filename[2])
  if (all(samples$fid == "") == TRUE)
    usesfid <- FALSE
  else
    usesfid <- TRUE
  snps <- readRDS(filename[3])
  chr1 <- snps$chromosome[1]
  onechr <- all(snps$chromosome == chr1)
  chrlocid <- paste(snps$chromosome, snps$location, sep = ":")
  if(all(snps$snpid == chrlocid) == TRUE)
    snpidformat <- 1
  else {
    chrlocrefaltid <- paste(snps$chromosome, snps$location,
                            snps$reference, snps$alternate, sep = ":")
    if (all(snps$snpid == chrlocrefaltid) == TRUE)
      snpidformat <- 2
    else
      snpidformat <- 0
  }

  additionalinfo <- list(format = format,
                         subformat = subformat,
                         headersize = headersize,
                         numGroups = 1,
                         groups = nrow(samples))
  class(additionalinfo) <- "bdose-info"

  return (list(filename = fqfilename,
               usesfid = usesfid,
               samples = samples,
               onechr = onechr,
               snpidformat = snpidformat,
               snps = snps,
               snpinfo = list(),
               additionalinfo = additionalinfo))
}

# Function to convert the binary dosage header subject and SNP info
# into the bdinfo format
# @parameter filename - vector of file names for the binary dosage
# file
# @parameter header - header information read from the binary dosage
# file
# @paramter format - format of the binary dosage file - should be 4
# @paramter subformat - subformat of the binary dosage file
# @return - The subject and family data needed to create a bdinfo list
Convert4HeaderToBDInfo <- function(filename, header, format, subformat) {
  sid <- unlist(strsplit(header$samples$sidstring, '\t'))
  if (header$samples$fidsize == 0) {
    usesfid <- FALSE
    fid = rep("", header$numsub)
  } else {
    usesfid <- TRUE
    fid <- unlist(strsplit(header$samples$fidstring, '\t'))
  }
  samples <- data.frame(fid, sid, stringsAsFactors = FALSE)

  if (header$snps$chrsize == 0) {
    onechr <- FALSE
    chromosome <- rep("", header$numSNPs)
  } else {
    chromosome <- unlist(strsplit(header$snps$chrstring, '\t'))
    if (length(chromosome) == 1) {
      onechr <- TRUE
      chromosome <- rep(header$snps$chrstring, header$numSNPs)
    } else {
      onechr <- FALSE
    }
  }
  if (length(header$snps$location) == 0)
    location <- rep(0L, header$numSNPs)
  else
    location <- header$snps$location
  if (header$snps$refsize == 0)
    reference <- rep("", header$numSNPs)
  else
    reference <- unlist(strsplit(header$snps$refstring, '\t'))
  if (header$snps$altsize == 0)
    alternate <- rep("", header$numSNPs)
  else
    alternate <- unlist(strsplit(header$snps$altstring, '\t'))
  if (header$snps$snpsize == 0) {
    if (header$snps$refsize == 0) {
      snpid <- paste(chromosome, location, sep = ':')
      snpidformat <- 1
    } else {
      snpid <- paste(chromosome, location, reference, alternate, sep = ':')
      snpidformat <- 2
    }
  } else {
    snpid <- unlist(strsplit(header$snps$snpstring, '\t'))
    snpidformat <- 0
  }
  snps <- data.frame(chromosome, location, snpid, reference, alternate, stringsAsFactors = FALSE)

  snpinfocol <- match(c("aaf", "maf", "avgcall", "rsq"), names(header$snps))
  snpinfocol <- snpinfocol[sapply(header$snps, function(x) length(x) != 0)[snpinfocol]]
  snpinfo <- lapply(header$snps[snpinfocol], matrix, nrow = header$numSNPs, ncol = header$numgroups)

  additionalinfo <- list(format = format,
                         subformat = subformat,
                         headersize = header$dosageoffset,
                         numgroups = header$numgroups,
                         groups = header$groups)
  class(additionalinfo) <- "bdose-info"

  return (list(filename = normalizePath(filename[1], winslash = "/"),
               usesfid = usesfid,
               samples = samples,
               onechr = onechr,
               snpidformat = snpidformat,
               snps = snps,
               snpinfo = snpinfo,
               additionalinfo = additionalinfo))
}

ReadBinaryDosageHeader11 <- function(filename) {
  return (ReadFamAndMapFiles(filename, 1, 1, 8))
}

ReadBinaryDosageHeader12 <- function(filename) {
  return (ReadFamAndMapFiles(filename, 1, 2, 8))
}

ReadBinaryDosageHeader21 <- function(filename) {
  return (ReadFamAndMapFiles(filename, 2, 1, 8))
}

ReadBinaryDosageHeader22 <- function(filename) {
  return (ReadFamAndMapFiles(filename, 2, 2, 8))
}

ReadBinaryDosageHeader31 <- function(filename) {
  bdInfo <- ReadFamAndMapFiles(filename, 3, 1, 12)
  additionalInfo = ReadBinaryDosageHeader3A(filename[1])
  if (additionalInfo$numsub != nrow(bdInfo$samples))
    stop("Subject file does not line up with binary dosage file")
  return (bdInfo)
}

ReadBinaryDosageHeader32 <- function(filename) {
  bdInfo <- ReadFamAndMapFiles(filename, 3, 2, 12)
  additionalInfo = ReadBinaryDosageHeader3A(filename[1])
  if (additionalInfo$numsub != nrow(bdInfo$samples))
    stop("Subject file does not line up with binary dosage file")
  return (bdInfo)
}

ReadBinaryDosageHeader33 <- function(filename) {
  bdInfo <- ReadFamAndMapFiles(filename, 3, 3, 72)
  additionalInfo = ReadBinaryDosageHeader3B(filename[1])
  if (digest(bdInfo$samples) != additionalInfo$md5[1])
    stop("Subject file does not line up with binary dosage file")
  if (digest(bdInfo$snps) != additionalInfo$md5[2])
    stop("Map file does not line up with binary dosage file")
  return (bdInfo)
}

ReadBinaryDosageHeader34 <- function(filename) {
  bdInfo <- ReadFamAndMapFiles(filename, 3, 4, 72)
  bdInfo$additionalinfo$headersize <- bdInfo$additionalinfo$headersize + 4 * nrow(bdInfo$snps)
  additionalInfo = ReadBinaryDosageHeader3B(filename[1])
  if (digest(bdInfo$samples) != additionalInfo$md5[1])
    stop("Subject file does not line up with binary dosage file")
  if (digest(bdInfo$snps) != additionalInfo$md5[2])
    stop("Map file does not line up with binary dosage file")
  return (bdInfo)
}

ReadBinaryDosageHeader41 <- function(filename) {
  header <- ReadBinaryDosageHeader4A(filename[1])
  return (Convert4HeaderToBDInfo(filename, header, 4, 1))
}

ReadBinaryDosageHeader42 <- function(filename) {
  header <- ReadBinaryDosageHeader4A(filename[1])
  return (Convert4HeaderToBDInfo(filename, header, 4, 2))
}

ReadBinaryDosageHeader43 <- function(filename) {
  header <- ReadBinaryDosageHeader4B(filename[1])
  bdInfo <- Convert4HeaderToBDInfo(filename, header, 4, 3)
  return (bdInfo)
}

ReadBinaryDosageHeader44 <- function(filename) {
  header <- ReadBinaryDosageHeader4B(filename[1])
  bdInfo <- Convert4HeaderToBDInfo(filename, header, 4, 4)
  return (bdInfo)
}

#***************************************************************************#
#                                                                           #
#               Getting the indices for a binary dosage file                #
#                                                                           #
#***************************************************************************#

# Gets the file locations for snps in a binary dosage file
ReadBinaryDosageIndices <- function(bdInfo) {
  ReadIndicesFunc <- list(f1 <- c(ReadIndices1, ReadIndices2),
                          f2 <- c(ReadIndices1, ReadIndices2),
                          f3 <- c(ReadIndices1, ReadIndices3, ReadIndices1, ReadIndices4),
                          f4 <- c(ReadIndices1, ReadIndices3, ReadIndices1, ReadIndices4))
  return (ReadIndicesFunc[[bdInfo$additionalinfo$format]][[bdInfo$additionalinfo$subformat]](bdInfo))
}

# Function to set up the datasize and indices vectors when the
# data size is the same for each SNP.
# @parameter numsub - number of subjects in data set
# @parameter numSNPs - number of snps
FixedIndices <- function(numsub, numSNPs, firstIndex, snpsize) {
  datasize <- snpsize * numsub
  indices <- seq(firstIndex, firstIndex + datasize * (numSNPs - 1), datasize)
  return (list(datasize = datasize, indices = indices))
}

# This routine sets up the indices when only the dosages are
# in the binary dosage file. This is simple because the size
# is fixed 2 bytes per subject per SNP
ReadIndices1 <- function(bdInfo) {
  return (FixedIndices(nrow(bdInfo$samples),
                       nrow(bdInfo$snps),
                       bdInfo$additionalinfo$headersize, 2))
}

# This routine sets up the indices when reading formats 1 and 2
# when there is both dosage and genetic probability data. This
# is also simple because the size is fixed to 4 bytes per subject
# per SNP
ReadIndices2 <- function(bdInfo) {
  return (FixedIndices(nrow(bdInfo$samples),
                       nrow(bdInfo$snps),
                       bdInfo$additionalinfo$headersize, 4))
}

# This routine sets up the indices when reading formats 3 and 4,
# subformat 2. The inidices are stored at the start of each SNP's
# dosage data. This is a time consuming operation.
ReadIndices3 <- function(bdInfo) {
  return (ReadBDIndices3C(bdInfo$filename, nrow(bdInfo$snps), bdInfo$additionalinfo$headersize))
}

# This routine sets up the indices when reading formats 3 and 4,
# subformat 4. The inidices are stored in the header and are easy
# to read in.
ReadIndices4 <- function(bdInfo) {
  return (ReadBDIndices4C(bdInfo$filename, nrow(bdInfo$snps), bdInfo$additionalinfo$headersize))
}

#***************************************************************************#
#                                                                           #
#                              Read the data                                #
#                                                                           #
#***************************************************************************#

# Reads a SNP from the various formats of the binary dosage file.
# For formats 1-4, data is written into d/p0/p1/p2 via C-level mutation
# and NULL is returned invisibly.
# For format 5, returns a list(dosage, p0, p1, p2); callers must use the
# return value because pure-R functions cannot mutate caller vectors in place.
ReadBinaryDosageData <- function(bdInfo, snp, d, p0, p1, p2, us) {
  if (bdInfo$additionalinfo$format == 5L)
    return(ReadBinaryDosageDataFormat5(bdInfo, snp))
  ReadHeaderFunc <- list(f1 <- c(ReadBinaryDosageData1, ReadBinaryDosageData2),
                         f2 <- c(ReadBinaryDosageData3, ReadBinaryDosageData4),
                         f3 <- c(ReadBinaryDosageData3, ReadBinaryDosageData5, ReadBinaryDosageData3, ReadBinaryDosageData5),
                         f4 <- c(ReadBinaryDosageData3, ReadBinaryDosageData5, ReadBinaryDosageData3, ReadBinaryDosageData5))
  ReadHeaderFunc[[bdInfo$additionalinfo$format]][[bdInfo$additionalinfo$subformat]](bdInfo, snp, d, p0, p1, p2, us)
  invisible(NULL)
}

ReadBinaryDosageDataFormat5 <- function(bdInfo, snp) {
  n_samp <- nrow(bdInfo$samples)
  n_snps <- nrow(bdInfo$snps)
  start  <- bdInfo$indices[snp]

  if (snp < n_snps) {
    nbytes <- as.integer(bdInfo$indices[snp + 1L] - start)
  } else {
    nbytes <- as.integer(file.info(bdInfo$filename)$size - start)
  }

  con <- file(bdInfo$filename, open = "rb")
  on.exit(close(con))
  seek(con, where = start, origin = "start")
  compressed <- readBin(con, what = raw(), n = nbytes)
  raw_block  <- memDecompress(compressed, type = "gzip")

  dosage <- ushort_raw_to_vals(raw_block, n_samp)
  gp_raw <- raw_block[seq(2L * n_samp + 1L, length(raw_block))]
  gp     <- ushort_raw_to_vals(gp_raw, 3L * n_samp)

  idx <- seq_len(n_samp)
  list(dosage = dosage,
       p0     = gp[3L * (idx - 1L) + 1L],
       p1     = gp[3L * (idx - 1L) + 2L],
       p2     = gp[3L * (idx - 1L) + 3L])
}

ReadBinaryDosageData1 <- function(bdInfo, snp, d, p0, p1, p2, us) {
  ReadBinaryDosageDataC(filename = bdInfo$filename,
                        headersize = bdInfo$additionalinfo$headersize,
                        numsub = nrow(bdInfo$samples),
                        snp = snp,
                        dosage = d,
                        us = us,
                        base = 1)
}

ReadBinaryDosageData2 <- function(bdInfo, snp, d, p0, p1, p2, us) {
  ReadBinaryDosageDataP1P2(filename = bdInfo$filename,
                           headersize = bdInfo$additionalinfo$headersize,
                           numsub = nrow(bdInfo$samples),
                           snp = snp,
                           dosage = d,
                           p0 = p0,
                           p1 = p1,
                           p2 = p2,
                           us = us,
                           base = 2)
}

ReadBinaryDosageData3 <- function(bdInfo, snp, d, p0, p1, p2, us) {
  ReadBinaryDosageDataC(filename = bdInfo$filename,
                        headersize = bdInfo$additionalinfo$headersize,
                        numsub = nrow(bdInfo$samples),
                        snp = snp,
                        dosage = d,
                        us = us,
                        base = 3)
}

ReadBinaryDosageData4 <- function(bdInfo, snp, d, p0, p1, p2, us) {
  ReadBinaryDosageDataP1P2(filename = bdInfo$filename,
                           headersize = bdInfo$additionalinfo$headersize,
                           numsub = nrow(bdInfo$samples),
                           snp = snp,
                           dosage = d,
                           p0 = p0,
                           p1 = p1,
                           p2 = p2,
                           us = us,
                           base = 3)
}

ReadBinaryDosageData5 <- function(bdInfo, snp, d, p0, p1, p2, us) {
  return (ReadBinaryDosageDataCompressed(filename = bdInfo$filename,
                                         index = bdInfo$indices[snp],
                                         datasize = bdInfo$datasize[snp],
                                         numsub = nrow(bdInfo$samples),
                                         dosage = d,
                                         p0 = p0,
                                         p1 = p1,
                                         p2 = p2,
                                         us = us))
}

test_that("readerrors", {
  vcf1a <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")
  bdbadversion <- system.file("extdata", "badversion.bdose", package = "BinaryDosage")

  expect_error(ReadBinaryDosageHeader("junk"),
               "Unable to open binary dosage file")
  expect_error(ReadBinaryDosageHeader(vcf1a),
               "File does not appear to be a binary dosage file")
  expect_error(ReadBinaryDosageHeader(bdbadversion),
               "Unknown binary dosage file fromat")
})

###########################################################
#                Binary Dosage                            #
###########################################################

#' Apply a function to each SNP in a
#' binary dosage file
#'
#' A routine that reads in the SNP data serially
#' from a binary dosage file and applies a user
#' specified function to the data.
#'
#' @param bdinfo List with information about the
#' binary dosage file returned from getbdinfo
#' @param func A user supplied function to apply
#' to the data for each snp. The function must be
#' provide with the following parameters, dosage,
#' p0, p1, and p2, where dosage is the dosage values
#' for each subject and p0, p1, and p2 are the
#' probabilities that a subject has zero, one,
#' and two copies of the alternate allele,
#' respectively.
#' @param ... Additional parameters needed by the
#' user supplied function
#'
#' @return A list with length equal to the number
#' of SNPs in the binary dosage file. Each element
#' of the list is the value returned by the user
#' supplied function
#' @export
#' @family Iterating functions
#' @examples
#' # Get information about a binary dosage file
#'
#' vcf1abdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
#' bdinfo <- getbdinfo(bdfiles = vcf1abdfile)
#'
#' # Apply the getaaf, get alternate allele frequency, function
#' # to all the SNPs in the binary dosage file
#'
#' aaf <- bdapply(bdinfo = bdinfo,
#'                func = BinaryDosage:::getaaf)
bdapply <- function(bdinfo, func, ...) {
  if (missing(bdinfo) == TRUE)
    stop("No binary dosage file information specified")
  if (is.na(match("genetic-info", class(bdinfo))) == TRUE)
    stop("bdinfo does not contain information about a binary dosage file")
  if (is.na(match("bdose-info", class(bdinfo$additionalinfo))) == TRUE)
    stop("bdinfo does not contain information about a binary dosage file")

  if (missing(func) == TRUE)
    stop("No function specified")
  if (is.na(match("function", class(func))) == TRUE)
    stop("func is not a function")

  retval <- vector("list", nrow(bdinfo$snps))
  n_samp <- nrow(bdinfo$samples)
  dosage <- numeric(n_samp)
  p0 <- numeric(n_samp)
  p1 <- numeric(n_samp)
  p2 <- numeric(n_samp)

  if (bdinfo$additionalinfo$format == 5L) {
    con <- openbd5con(bdinfo)
    on.exit(closebd5con(con), add = TRUE)
    for (i in 1:nrow(bdinfo$snps)) {
      getbd5snp_con(bdinfo, i, dosage, p0, p1, p2, con)
      retval[[i]] <- func(dosage, p0, p1, p2, ...)
    }
  } else {
    us <- integer(2 * n_samp)
    for (i in 1:nrow(bdinfo$snps)) {
      dosage[1:n_samp] <- NA
      p0[1:n_samp] <- NA
      p1[1:n_samp] <- NA
      p2[1:n_samp] <- NA
      result <- ReadBinaryDosageData(bdinfo, i, dosage, p0, p1, p2, us)
      if (!is.null(result)) {
        dosage <- result$dosage
        p0     <- result$p0
        p1     <- result$p1
        p2     <- result$p2
      }
      retval[[i]] <- func(dosage, p0, p1, p2, ...)
    }
  }
  return (retval)
}

###########################################################
#                        VCF                              #
###########################################################

#' Apply a function to each SNP in a vcf file
#'
#' A routine that reads in the SNP data serially
#' from a vcf file and applies a user specified
#' function to the data.
#'
#' @param vcfinfo List with information about the
#' vcf file returned from getvcfinfo
#' @param func A user supplied function to apply
#' to the data for each snp. The function must be
#' provide with the following parameters, dosage,
#' p0, p1, and p2, where dosage is the dosage values
#' for each subject and p0, p1, and p2 are the
#' probabilities that a subject has zero, one,
#' and two copies of the alternate allele,
#' respectively.
#' @param ... Additional parameters needed by the
#' user supplied function
#'
#' @return A list with length equal to the number
#' of SNPs in the vcf file. Each element
#' of the list is the value returned by the user
#' supplied function
#' @export
#' @family Iterating functions
#' @examples
#' # Get information about a vcf file
#'
#' vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")
#' vcfinfo <- getvcfinfo(vcffiles = vcf1afile)
#'
#' # Apply the getaaf, get alternate allele frequency, function
#' # to all the SNPs in the vcf file
#'
#' aaf <- vcfapply(vcfinfo = vcfinfo,
#'                 func = BinaryDosage:::getaaf)
vcfapply <- function(vcfinfo, func, ...) {
  if (missing(vcfinfo) == TRUE)
    stop("No vcf file information specified")
  if (is.na(match("genetic-info", class(vcfinfo))) == TRUE)
    stop("vcfinfo does not appear to contain information about a vcf file")
  if (is.na(match("vcf-info", class(vcfinfo$additionalinfo))) == TRUE)
    stop("vcfinfo does not appear to contain information about a vcf file")

  if (missing(func) == TRUE)
    stop("No function specified")
  if (is.na(match("function", class(func))) == TRUE)
    stop("func is not a function")

  retval <- vector("list", nrow(vcfinfo$snps))
  if (vcfinfo$additionalinfo$gzipped == FALSE)
    con <- file(vcfinfo$filename, "r")
  else
    con <- gzfile(vcfinfo$filename, "r")
  line <- readLines(con, n = vcfinfo$additionalinfo$headerlines)

  dosage <- numeric(nrow(vcfinfo$samples))
  p0 <- numeric(nrow(vcfinfo$samples))
  p1 <- numeric(nrow(vcfinfo$samples))
  p2 <- numeric(nrow(vcfinfo$samples))

  for (i in 1:nrow(vcfinfo$snps)) {
    line <- readLines(con, n = 1)
    x <- unlist(strsplit(line, "\t"))
    y <- unlist(strsplit(x[10:length(x)], ":"))
    if (length(vcfinfo$additionalinfo$datacolumns$dosage) == 1) {
      dosagecol <- vcfinfo$additionalinfo$datacolumns$dosage
      gpcol <- vcfinfo$additionalinfo$datacolumns$genotypeprob
      numcolumns <- vcfinfo$additionalinfo$datacolumns$numcolumns
    } else {
      dosagecol <- vcfinfo$additionalinfo$datacolumns$dosage[i]
      gpcol <- vcfinfo$additionalinfo$datacolumns$genotypeprob[i]
      numcolumns <- vcfinfo$additionalinfo$datacolumns$numcolumns[i]
    }
    if(is.na(dosagecol) == FALSE) {
      dosage[1:nrow(vcfinfo$samples)] <- as.numeric(y[seq(dosagecol, length(y) - numcolumns + dosagecol, numcolumns)])
    }
    if(is.na(gpcol) == FALSE) {
      gpstring <- y[seq(gpcol, length(y) - numcolumns + gpcol, numcolumns)]
      z <- unlist(strsplit(gpstring, ","))
      p0[1:nrow(vcfinfo$samples)] <- as.numeric(z[seq(1, length(z) - 2, 3)])
      p1[1:nrow(vcfinfo$samples)] <- as.numeric(z[seq(2, length(z) - 1, 3)])
      p2[1:nrow(vcfinfo$samples)] <- as.numeric(z[seq(3, length(z), 3)])
    } else {
      p0[1:nrow(vcfinfo$samples)] <- NA
      p1[1:nrow(vcfinfo$samples)] <- NA
      p2[1:nrow(vcfinfo$samples)] <- NA
    }
    if(is.na(dosagecol) == FALSE) {
      dosage[1:nrow(vcfinfo$samples)] <- as.numeric(y[seq(dosagecol, length(y) - numcolumns + dosagecol, numcolumns)])
    } else {
      dosage[1:nrow(vcfinfo$samples)] <- p1 + p2 + p2
    }
    retval[[i]] <- func(dosage = dosage,
                        p0 = p0,
                        p1 = p1,
                        p2 = p2
                        , ...)
  }
  close(con)
  return (retval)
}

###########################################################
#                 GEN (Impute2)                           #
###########################################################

#' Apply a function to each SNP in a
#' gen, impute2, file
#'
#' A routine that reads in the SNP data serially
#' from a gen file and applies a user
#' specified function to the data.
#'
#' @param geninfo List with information about the
#' gen, impute2, file returned from [getgeninfo]
#' @param func A user supplied function to apply
#' to the data for each snp. The function must be
#' provide with the following parameters, dosage,
#' p0, p1, and p2, where dosage is the dosage values
#' for each subject and p0, p1, and p2 are the
#' probabilities that a subject has zero, one,
#' and two copies of the alternate allele,
#' respectively.
#' @param ... Additional parameters needed by the
#' user supplied function
#'
#' @return A list with length equal to the number
#' of SNPs in the vcf file. Each element
#' of the list is the value returned by the user
#' supplied function
#' @export
#'
#' @family Iterating functions
#' @examples
#' # Get information about a gen, impute2, file
#'
#' gen1afile <- system.file("extdata", "set1a.imp", package = "BinaryDosage")
#' geninfo <- getgeninfo(genfiles = gen1afile,
#'                       snpcolumns = c(1L, 3L, 2L, 4L, 5L),
#'                       header = TRUE)
#'
# Apply the getaaf, get alternate allele frequency, function
# to all the SNPs in the vcf file
#'
#' aaf <- genapply(geninfo = geninfo,
#'                 func = BinaryDosage:::getaaf)
genapply <- function(geninfo, func, ...) {
  if (missing(geninfo) == TRUE)
    stop("No gen file information specified")
  if (is.na(match("genetic-info", class(geninfo))) == TRUE)
    stop("geninfo does not appear to contain information about a gen file")
  if (is.na(match("gen-info", class(geninfo$additionalinfo))) == TRUE)
    stop("geninfo does not appear to contain information about a gen file")

  if (missing(func) == TRUE)
    stop("No function specified")
  if (is.na(match("function", class(func))) == TRUE)
    stop("func is not a function")

  retval <- vector("list", nrow(geninfo$snps))
  if (geninfo$additionalinfo$gzipped == FALSE)
    con <- file(geninfo$filename, "r")
  else
    con <- gzfile(geninfo$filename, "r")
  line <- readLines(con, n = geninfo$additionalinfo$headersize)

  dosage <- numeric(nrow(geninfo$samples))
  p0 <- numeric(nrow(geninfo$samples))
  p0[1:length(p0)] <- NA
  p1 <- p0
  p2 <- p0
  for (i in 1:nrow(geninfo$snps)) {
    line <- readLines(con, n = 1)
    x <- unlist(strsplit(line, geninfo$additionalinfo$sep))
    y <- x[geninfo$additionalinfo$startcolumn:length(x)]
    if (geninfo$additionalinfo$format == 1) {
      dosage[1:length(dosage)] <- as.numeric(y)
    } else if (geninfo$additionalinfo$format == 2) {
      p0[1:length(p0)] <- as.numeric(y[seq(1, length(y) - 1, 2)])
      p1[1:length(p1)] <- as.numeric(y[seq(2, length(y), 2)])
      p2[1:length(p2)] <- ifelse(p0 + p1 < 1, 1 - p0 - p1, 0.)
      dosage[1:length(dosage)] <- ifelse(p1 + p2 + p2 < 2, p1 + p2 + p2, 2.)
    } else {
      p0[1:length(p0)] <- as.numeric(y[seq(1, length(y) - 2, 3)])
      p1[1:length(p1)] <- as.numeric(y[seq(2, length(y) - 1, 3)])
      p2[1:length(p2)] <- as.numeric(y[seq(3, length(y), 3)])
      dosage[1:length(dosage)] <- ifelse(p1 + p2 + p2 < 2, p1 + p2 + p2, 2.)
    }

    retval[[i]] <- func(dosage = dosage,
                        p0 = p0,
                        p1 = p1,
                        p2 = p2
                        , ...)
  }
  close(con)
  return (retval)
}

#' Merge binary dosage files together
#'
#' Routine to merge binary dosage files together. The files
#' don't have to be in the same format. They will be merged
#' into a file with the format specified. Information about
#' the SNPs, aaf, maf, avgcall, rsq, can be maintained for
#' each file, or recalculated for the merged set.
#'
#' @param mergefiles Vector of file names for the merged binary
#' files. The first is the
#' binary dosage data containing the dosages and genetic
#' probabilities. The second file name is the family information
#' file. The third file name is the SNP information file.
#' The family and SNP information files are not used if the
#' binary dosage file is in format 4. For this format the
#' family and SNP information are in the file with the dosages
#' and genetic probabilities.
#' @param format The format of the output binary dosage file.
#' Allowed values are 1, 2, 3, and 4. The default value is 4.
#' Using the default value is recommended.
#' @param subformat The subformat of the format of the output
#' binary dosage file. A value of 1 or 3 indicates that only the
#' dosage value is saved. A value of 2 or 4 indicates
#' the dosage and genetic probabilities will be output. Values
#' of 3 or 4 are only allowed with formats 3 and 4. If a value
#' of zero if provided, and genetic probabilities are in the vcf
#' file, subformat 2 will be used for formats 1 and 2, and
#' subformat 4 will be used for formats 3 and 4. If the vcf file
#' does not contain genetic probabilities, subformat 1 will be
#' used for formats 1 and 2, and subformat 3 will be used for
#' formats 3 and 4. The default value is 0.
#' @param bdfiles Vector of binary dosage file names to be merged.
#' @param famfiles Vector of family file names that correspond to
#' the names in bdfiles. If the binary dosage files are all in format
#' 4, this may be an empty character array. Default value is character().
#' @param mapfiles Vector of map file names that correspond to the
#' names in bdfiles. If the binary dosage files are all in format
#' 4, this may be an empty character array. Default value is character().
#' @param onegroup Indicator to combine all the samples in one group.
#' If this is FALSE, the groups in each binary dosage file are
#' maintained and any binary dosage file with one group is made into
#' its own group. Default value is TRUE.
#' @param bdoptions Options indicating what information to calculate
#' and store for each SNP. These can be aaf, maf, and rsq. This option
#' is only available if format is equal to 4 and onegroup is TRUE.
#' Default value is character().
#' @param snpjoin Character value that can be either "inner" or "outer".
#' This indicates whether to do an inner or outer join of the SNPs in
#' each binary dosage file. Default value is "inner".
#'
#' @return
#' None
#' @export
#'
#' @examples
#' bdvcf1afile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
#' bdvcf1bfile <- system.file("extdata", "vcf1b.bdose", package = "BinaryDosage")
#' mergefiles <- tempfile()
#'
#' BinaryDosage:::bdmerge(mergefiles = mergefiles,
#'                        bdfiles = c(bdvcf1afile, bdvcf1bfile),
#'                        bdoptions = "maf")
#' bdinfo <- getbdinfo(mergefiles)
bdmerge <- function(mergefiles,
                    format = 4,
                    subformat = 0L,
                    bdfiles,
                    famfiles = character(),
                    mapfiles = character(),
                    onegroup = TRUE,
                    bdoptions = character(),
                    snpjoin = "inner") {
  if (is.numeric(format) == FALSE && is.integer(format) == FALSE)
    stop("format must be an integer value")
  if (length(format) != 1)
    stop("format must be an integer vector of length 1")
  if (is.numeric(format) == TRUE) {
    if (floor(format) != format)
      stop("format must be an integer")
    format = floor(format)
  }
  if (format < 1 || format > 4)
    stop("format must be an integer value from 1 to 4")

  if (is.numeric(subformat) == FALSE && is.integer(subformat) == FALSE)
    stop("subformat must be an integer value")
  if (length(subformat) != 1)
    stop("subformat must be an integer vector of length 1")
  if (is.numeric(subformat) == TRUE) {
    if (floor(subformat) != subformat)
      stop("subformat must be an integer")
    subformat = floor(subformat)
  }
  if (subformat < 0 || subformat > 4)
    stop("subformat must be an integer value from 0 to 4")
  if (format < 3 && subformat > 2)
    stop("subformat must be an integer value from 0 to 2 for formats 1 and 2")

  if (missing(mergefiles) == TRUE)
    stop("No output files specified")
  if (is.character(mergefiles) == FALSE)
    stop("Output file names must be a character values")
  if (format == 4 & length(mergefiles) != 1)
    stop("Only one file name is needed when using format 4")
  if (format < 4 & length(mergefiles) != 3)
    stop("Three file names are required when using formats 1, 2, and 3")
  if (is.na(match("", mergefiles)) == FALSE)
    stop("Output file names cannot be blank")

  if (missing(bdfiles) == TRUE)
    stop("No files specified")
  if (is.character(bdfiles) == FALSE)
    stop("bdfiles must be a character vector")
  if (length(bdfiles) < 2)
    stop("At least two binary dosage files must be specified")

  if (is.character(famfiles) == FALSE)
    stop("famfiles must be a character vector")
  if (is.character(mapfiles) == FALSE)
    stop("mapfiles must be a character vector")
  if (length(famfiles) != 0 & length(mapfiles) == 0)
    stop("If famfiles is specified, mapfiles must be specified")
  if (length(famfiles) == 0 & length(mapfiles) != 0)
    stop("If mapfiles is specified, famfiles must be specified")
  if (length(famfiles) != 0 & (length(famfiles) != length(bdfiles) | length(mapfiles) != length(bdfiles)))
    stop("If famfiles and mapfiles are specified they must have the same length as bdfiles")

  if (is.logical(onegroup) == FALSE)
    stop("onegroup must be logical value")
  if (length(onegroup) != 1)
    stop("onegroup must be a logical vector of length 1")

  if (is.character(bdoptions) == FALSE)
    stop("bdoptions must be a character vector")
  if (onegroup == FALSE & length(bdoptions) > 0)
    stop("bdoptions can only be used if onegroup is TRUE")
  if (format != 4 & length(bdoptions) > 0)
    stop("bdoptions can only be used if format = 4")
  if (length(bdoptions) > 0) {
    if (any(is.na(match(bdoptions, c("aaf", "maf", "rsq")))) == TRUE)
      stop("Only valid bdoptions are aaf, maf, and rsq")
  }
  if (snpjoin != "inner" & snpjoin != "outer")
    stop("snpjoin must have a value of either \"inner\" or \"outer\"")

  allsnps <- TRUE
  if (snpjoin == "inner")
    allsnps <- FALSE

  bdinfo <- vector("list", length(bdfiles))
  for (i in 1:length(bdfiles)) {
    if (length(famfiles) > 0)
      bdinfo[[i]] <- getbdinfo(c(bdfiles[i], famfiles[i], mapfiles[i]))
    else
      bdinfo[[i]] <- getbdinfo(bdfiles[i])
  }

  bdmergedinfo <- mergebdinfo(bdinfo = bdinfo,
                              format = format,
                              subformat = subformat,
                              onegroup = onegroup)

  mergedgeneticinfo <- mergegeneticinfo(mergefile = mergefiles[1],
                                        geneticinfo = bdinfo,
                                        allsnps = allsnps)
  mergedgeneticinfo$additionalinfo <- bdmergedinfo

  snpsbtom <- list()
  snpsmtob <- list()
  snpsbtomNA <- list()
  snpsmtobNA <- list()
  for (i in 1:length(bdinfo)) {
    snpsbtomNA[[i]] <- prodlim::row.match(bdinfo[[i]]$snps, mergedgeneticinfo$snps)
    snpsbtom[[i]] <- snpsbtomNA[[i]][!is.na(snpsbtomNA[[i]])]
    snpsmtobNA[[i]] <- prodlim::row.match(mergedgeneticinfo$snps, bdinfo[[i]]$snps)
    snpsmtob[[i]] <- snpsmtobNA[[i]][!is.na(snpsmtobNA[[i]])]
  }

  if (onegroup == FALSE) {
    mergedgeneticinfo$snpinfo <- mergesnpinfo(mergedinfo = mergedgeneticinfo,
                                              numgroups = bdmergedinfo$numgroups,
                                              geneticinfo = bdinfo,
                                              snpsbtom = snpsbtom,
                                              snpsmtob = snpsmtob)
  }

  WriteBinaryDosageHeader(format = bdmergedinfo$format,
                          subformat = bdmergedinfo$subformat,
                          filename = mergefiles,
                          genefileinfo = mergedgeneticinfo,
                          bdoptions = bdoptions)
  headerinfo <- ReadBinaryDosageHeader(filename = mergefiles)
  bdwriteinfo <- AllocateBinaryDosageWriteMemory(headerinfo = headerinfo)

  dosage <- numeric(nrow(mergedgeneticinfo$samples))
  p0 <- numeric(nrow(mergedgeneticinfo$samples))
  p1 <- numeric(nrow(mergedgeneticinfo$samples))
  p2 <- numeric(nrow(mergedgeneticinfo$samples))
  us <- integer(2 * nrow(mergedgeneticinfo$samples))
  dosaget <- numeric(nrow(mergedgeneticinfo$samples))
  p0t <- numeric(nrow(mergedgeneticinfo$samples))
  p1t <- numeric(nrow(mergedgeneticinfo$samples))
  p2t <- numeric(nrow(mergedgeneticinfo$samples))
  ust <- integer(2 * nrow(mergedgeneticinfo$samples))
  startgroup <- integer(length(bdinfo))
  endgroup <- integer(length(bdinfo))
  startgroup[1] <- 1
  endgroup[1] <- nrow(bdinfo[[1]]$samples)
  for (i in 2:length(bdinfo)) {
    startgroup[i] <- endgroup[i - 1] + 1
    endgroup[i] <- endgroup[i - 1] + nrow(bdinfo[[i]]$samples)
  }

  for (i in 1:nrow(mergedgeneticinfo$snps)) {
    for (j in 1:length(bdinfo)) {
      dosage[1:nrow(mergedgeneticinfo$samples)] <- NA
      p0[1:nrow(mergedgeneticinfo$samples)] <- NA
      p1[1:nrow(mergedgeneticinfo$samples)] <- NA
      p2[1:nrow(mergedgeneticinfo$samples)] <- NA
      if (is.na(snpsmtobNA[[j]][i]) == FALSE) {
        result <- ReadBinaryDosageData(bdinfo[[j]], snpsmtobNA[[j]][i], dosage, p0, p1, p2, us)
        if (!is.null(result)) {
          n <- nrow(bdinfo[[j]]$samples)
          dosage[1:n] <- result$dosage
          p0[1:n]     <- result$p0
          p1[1:n]     <- result$p1
          p2[1:n]     <- result$p2
        }
      }
      dosaget[startgroup[j]:endgroup[j]] <- dosage[1:nrow(bdinfo[[j]]$samples)]
      p0t[startgroup[j]:endgroup[j]] <- p0[1:nrow(bdinfo[[j]]$samples)]
      p1t[startgroup[j]:endgroup[j]] <- p1[1:nrow(bdinfo[[j]]$samples)]
      p2t[startgroup[j]:endgroup[j]] <- p2[1:nrow(bdinfo[[j]]$samples)]
    }
    WriteBinaryDosageData(dosaget, p0t, p1t, p2t, bdwriteinfo)
  }
  WriteBinaryDosageIndices(writeinfo = bdwriteinfo)
  mbdinfo <- getbdinfo(bdfiles = mergefiles)
  if (is.na(match("aaf", bdoptions)) == FALSE)
    updateaaf(mbdinfo)
  if (is.na(match("maf", bdoptions)) == FALSE)
    updatemaf(mbdinfo)
  if (is.na(match("rsq", bdoptions)) == FALSE)
    updatersq(mbdinfo)
}

mergesnpinfo <- function (mergedinfo,
                          numgroups,
                          geneticinfo,
                          snpsbtom,
                          snpsmtob) {
  hasaaf <- FALSE
  hasmaf <- FALSE
  hasavgcall <- FALSE
  hasrsq <- FALSE

  if (is.na(match("aaf", names(geneticinfo[[1]]$snpinfo))) == FALSE)
    hasaaf <- TRUE
  if (is.na(match("maf", names(geneticinfo[[1]]$snpinfo))) == FALSE)
    hasmaf <- TRUE
  if (is.na(match("avgcall", names(geneticinfo[[1]]$snpinfo))) == FALSE)
    hasavgcall <- TRUE
  if (is.na(match("rsq", names(geneticinfo[[1]]$snpinfo))) == FALSE)
    hasrsq <- TRUE
  for (i in 2:length(geneticinfo)) {
    if (is.na(match("aaf", names(geneticinfo[[i]]$snpinfo))) == FALSE)
      hasaaf <- TRUE
    if (is.na(match("maf", names(geneticinfo[[i]]$snpinfo))) == FALSE)
      hasmaf <- TRUE
    if (is.na(match("avgcall", names(geneticinfo[[i]]$snpinfo))) == FALSE)
      hasavgcall <- TRUE
    if (is.na(match("rsq", names(geneticinfo[[i]]$snpinfo))) == FALSE)
      hasrsq <- TRUE
  }

  snpinfo <- list()
  numsnpinfo <- 0
  snpinfonames <- character(0)
  if (hasaaf == TRUE) {
    numsnpinfo <- numsnpinfo + 1
    snpinfo[[numsnpinfo]] <- matrix(rep(as.numeric(NA), nrow(mergedinfo$snps) * numgroups),
                                    nrow(mergedinfo$snps), numgroups)
    snpinfonames <- c(snpinfonames, "aaf")
  }
  if (hasmaf == TRUE) {
    numsnpinfo <- numsnpinfo + 1
    snpinfo[[numsnpinfo]] <- matrix(rep(as.numeric(NA), nrow(mergedinfo$snps) * numgroups),
                                    nrow(mergedinfo$snps), numgroups)
    snpinfonames <- c(snpinfonames, "maf")
  }
  if (hasavgcall == TRUE) {
    numsnpinfo <- numsnpinfo + 1
    snpinfo[[numsnpinfo]] <- matrix(rep(as.numeric(NA), nrow(mergedinfo$snps) * numgroups),
                                    nrow(mergedinfo$snps), numgroups)
    snpinfonames <- c(snpinfonames, "avgcall")
  }
  if (hasrsq == TRUE) {
    numsnpinfo <- numsnpinfo + 1
    snpinfo[[numsnpinfo]] <- matrix(rep(as.numeric(NA), nrow(mergedinfo$snps) * numgroups),
                                    nrow(mergedinfo$snps), numgroups)
    snpinfonames <- c(snpinfonames, "rsq")
  }
  if(numsnpinfo > 0) {
    names(snpinfo) <- snpinfonames
    currentgroup <- 1L
    for (i in 1:length(geneticinfo)) {
      setgroups <- 1L
      if (inherits(geneticinfo[[i]]$additionalinfo, "bdose-info"))
        setgroups <- geneticinfo[[i]]$additionalinfo$numgroups
      for (j in 1:numsnpinfo) {
        if (is.na(match(names(snpinfo)[j], names(geneticinfo[[i]]$snpinfo))) == FALSE) {
          snpinfo[[j]][snpsbtom[[i]], currentgroup:(currentgroup + setgroups - 1)] <-
            geneticinfo[[i]]$snpinfo[[names(snpinfo)[j]]][snpsmtob[[i]],]
        }
      }
      currentgroup <- currentgroup + setgroups
    }
  }

  return(snpinfo)
}

mergegeneticinfo <- function(mergefile,
                             geneticinfo,
                             allsnps) {
  usesfid <- geneticinfo[[1]]$usesfid
  samples <- geneticinfo[[1]]$samples
  snpidformat <- geneticinfo[[1]]$snpidformat
  snps <- geneticinfo[[1]]$snps

  for (i in 2:length(geneticinfo)) {
    if (geneticinfo[[i]]$usesfid != usesfid)
      stop("Some files use FID and others do not")
    if (snpidformat != geneticinfo[[i]]$snpidformat)
      snpidformat <- 0L
    samples <- rbind(samples, geneticinfo[[i]]$samples)
    snps <- merge(snps, geneticinfo[[i]]$snps, all = allsnps)
  }

  if (nrow(unique(samples)) != nrow(samples))
    stop("There are duplicate samples in the files to merge")

  chr1 <- snps$chromosome[1]
  onechr <- all(snps$chromosome == chr1)

  return (list(filename = mergefile,
               usesfid = usesfid,
               samples = samples,
               onechr = FALSE,
               snpidformat = snpidformat,
               snps = snps,
               snpinfo = list(),
               datasize = integer(0),
               indices = numeric(0)))
}

mergebdinfo <- function(bdinfo, format, subformat, onegroup) {
  if (onegroup) {
    numgroups <- 1L
    groups <- 0L
    for (i in 1:length(bdinfo))
      groups <- groups + nrow(bdinfo[[i]]$samples)
  } else {
    groups <- integer()
    for (i in 1:length(bdinfo))
      groups <- c(groups, bdinfo[[i]]$additionalinfo$groups)
    numgroups <- length(groups)
  }
  if (subformat == 0L) {
    subformat <- 2L
    for (i in 1:length(bdinfo)) {
      if (bdinfo[[i]]$additionalinfo$subformat == 1)
        subformat <- 1L
    }
  }
  retval <- list(format = format,
                 subformat = subformat,
                 headersize = 0,
                 numgroups = numgroups,
                 groups = groups)
  class(retval) <- "bdose-info"

  return (retval)
}

#' @useDynLib BinaryDosage, .registration = TRUE
#' @importFrom Rcpp sourceCpp
#' @importFrom digest digest
#' @importFrom prodlim row.match
#' @importFrom utils read.table
NULL

validatebdinput <- function(bdfiles,
                            format,
                            subformat,
                            snpidformat,
                            bdoptions) {
  if (is.character(bdfiles) == FALSE)
    stop("bdfiles must be a vector of characters")

  if (is.numeric(format) == FALSE && is.integer(format) == FALSE)
    stop("format must be an integer value")
  if (length(format) != 1)
    stop("format must be an integer vector of length 1")
  if (is.numeric(format) == TRUE) {
    if (floor(format) != format)
      stop("format must be an integer")
    format <- floor(format)
  }
  if (format < 1 || format > 4)
    stop("format must be an integer value from 1 to 4")

  if (is.numeric(subformat) == FALSE && is.integer(subformat) == FALSE)
    stop("subformat must be an integer value")
  if (length(subformat) != 1)
    stop("subformat must be an integer vector of length 1")
  if (is.numeric(subformat) == TRUE) {
    if (floor(subformat) != subformat)
      stop("subformat must be an integer")
    subformat <- floor(subformat)
  }
  if (subformat < 0 || subformat > 4)
    stop("subformat must be an integer value from 0 to 4")
  if (format < 3 && subformat > 2)
    stop("subformat must be an integer value from 0 to 2 for formats 1 and 2")

  if (format == 4 & length(bdfiles) != 1)
    stop("Only one output file name is needed when using format 4")
  if (format < 4 & length(bdfiles) != 3)
    stop("Three output file names are required when using formats 1, 2, and 3")
  if (is.na(match("", bdfiles)) == FALSE)
    stop("Output file names cannot be blank")

  if (is.numeric(snpidformat) == FALSE && is.integer(snpidformat) == FALSE)
    stop("snpidformat must be an integer value")
  if (length(snpidformat) != 1)
    stop("snpidformat must be an integer vector of length 1")
  if (is.numeric(snpidformat) == TRUE) {
    if (floor(snpidformat) != snpidformat)
      stop("snpidformat must be an integer")
    snpidformat = floor(snpidformat)
  }
  if (snpidformat < -1 | snpidformat > 3)
    stop("snpidformat must be and integer from -1 to 3")

  if (is.character(bdoptions) == FALSE)
    stop("bdoptions must be a character array")
  if (length(bdoptions) > 0 & format != 4)
    stop("bdoptions can only be used with format 4")
  if (length(bdoptions) > 0) {
    if (any(is.na(match(bdoptions, c("aaf", "maf", "rsq")))) == TRUE)
      stop("Only valid bdoptions are aaf, maf, and rsq")
  }
  return(list(format = format,
              subformat = subformat))
}

###########################################################
#                  VCF to Binary Dosage                   #
###########################################################

#' Convert a VCF file to a binary dosage file
#'
#' Routine to read information from a VCF file and create
#' a binary dosage file. The function is designed to use
#' files return from the Michigan Imputation Server but will
#' run on other VCF files if they contain dosage and genetic
#' probabilities. Note: This routine can take a long time to
#' run if the VCF file is large.
#'
#' @param vcffiles A vector of file names.
#' The first is the name of the vcf file. The
#' second is name of the file that contains information
#' about the imputation of the SNPs. This file is produced
#' by minimac 3 and 4.
#' @param gz Indicator if VCF file is compressed using gzip.
#' Default value is FALSE.
#' @param bdfiles Vector of names of the output files.
#' The binary dosage file name is first. The family and
#' map files follow. For format 4, no family and map file
#' names are needed.
#' @param format The format of the output binary dosage file.
#' Allowed values are 1, 2, 3, and 4. The default value is 4.
#' Using the default value is recommended.
#' @param subformat The subformat of the format of the output
#' binary dosage file. A value of 1 or 3 indicates that only the
#' dosage value is saved. A value of 2 or 4 indicates
#' the dosage and genetic probabilities will be output. Values
#' of 3 or 4 are only allowed with formats 3 and 4. If a value
#' of zero if provided, and genetic probabilities are in the vcf
#' file, subformat 2 will be used for formats 1 and 2, and
#' subformat 4 will be used for formats 3 and 4. If the vcf file
#' does not contain genetic probabilities, subformat 1 will be
#' used for formats 1 and 2, and subformat 3 will be used for
#' formats 3 and 4. The default value is 0.
#' @param snpidformat The format that the SNP ID will be saved as.
#' -1 SNP ID not written
#' 0 - same as in the VCF file
#' 1 - chr:pos
#' 2 - chr:pos:ref:alt
#' If snpidformat is 1 and the VCF file uses format 2, an error is
#' generated. Default value is 0.
#' @param bdoptions Character array containing any of the following
#' value, "aaf", "maf", "rsq". The presence of any of these
#' values indicates that the specified values should be
#' calculates and stored in the binary dosage file. These values only
#' apply to format 4.
#'
#' @return
#' None
#' @export
#'
#' @examples
#' # Find the vcf file names
#' vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")
#' vcf1ainfo <- system.file("extdata", "set1a.info", package = "BinaryDosage")
#' bdfiles <- tempfile()
#' # Convert the file
#' vcftobdlegacy(vcffiles = c(vcf1afile, vcf1ainfo), bdfiles = bdfiles)
#' # Verify the file was written correctly
#' bdinfo <- getbdinfo(bdfiles)
vcftobdlegacy <- function(vcffiles,
                    gz = FALSE,
                    bdfiles,
                    format = 4L,
                    subformat = 0L,
                    snpidformat = 0,
                    bdoptions = character(0)) {
  message("vcftobdlegacy() uses legacy formats 1-4. ",
          "Consider using vcftobd() instead, which produces Format 5 files ",
          "with faster conversion and smaller file sizes.")
  if (missing(vcffiles) == TRUE)
    stop("No VCF file specified")

  if (missing(bdfiles) == TRUE)
    stop("No output files specified")

  validation <- validatebdinput(bdfiles = bdfiles,
                                format = format,
                                subformat = subformat,
                                snpidformat = snpidformat,
                                bdoptions = bdoptions)
  format <- validation$format
  subformat <- validation$subformat

  if (snpidformat == -1)
    readsnpformat = 0
  else
    readsnpformat = snpidformat
  vcfinfo <- getvcfinfo(vcffiles = vcffiles,
                        gz = gz,
                        index = FALSE,
                        snpidformat = readsnpformat)
  if (snpidformat == -1)
    vcfinfo$snpidformat = 1
  else
    vcfinfo$snpidformat = 0

  if (subformat == 0) {
    if (anyNA(vcfinfo$additionalinfo$datacolumns$genotypeprob) == TRUE)
      subformat <- 1
    else
      subformat <- 2
  }
  WriteBinaryDosageHeader(format = format,
                          subformat = subformat,
                          filename = bdfiles,
                          genefileinfo = vcfinfo,
                          bdoptions = bdoptions)
  headerinfo <- ReadBinaryDosageHeader(filename = bdfiles)
  bdwriteinfo <- AllocateBinaryDosageWriteMemory(headerinfo = headerinfo)
  vcfapply(vcfinfo = vcfinfo,
           func = WriteBinaryDosageData,
           writeinfo = bdwriteinfo)
  WriteBinaryDosageIndices(writeinfo = bdwriteinfo)
  bdinfo <- getbdinfo(bdfiles = bdfiles)
  if (is.na(match("aaf", bdoptions)) == FALSE)
    updateaaf(bdinfo)
  if (is.na(match("maf", bdoptions)) == FALSE)
    updatemaf(bdinfo)
  if (is.na(match("rsq", bdoptions)) == FALSE)
    updatersq(bdinfo)
  ##return (0)
}

###########################################################
#          Binary Dosage Format 1 to Format 5            #
###########################################################

#' Convert a format 1 binary dosage file to Format 5
#'
#' Reads a format 1 binary dosage file (subformat 1 or 2) and converts it to
#' a Format 5 file pair. If the source file does not contain genotype
#' probabilities (subformat 1), those values are stored as missing in the
#' output.
#'
#' @param bdfiles Vector of three file names for the format 1 binary dosage
#'   file: the binary dosage file, the family file, and the map file.
#' @param bdose_file Path for the output .bdose file. The companion .bdi
#'   metadata file is written to \code{paste0(bdose_file, ".bdi")}.
#'
#' @return NULL (invisibly)
#'
#' @keywords internal
bd1tobd5 <- function(bdfiles, bdose_file) {
  if (missing(bdfiles))
    stop("No binary dosage files specified")
  if (missing(bdose_file))
    stop("No output .bdose file specified")

  old_bdinfo <- getbdinfo(bdfiles)

  fmt    <- old_bdinfo$additionalinfo$format
  subfmt <- old_bdinfo$additionalinfo$subformat

  if (fmt != 1L)
    stop("bdfiles must be a format 1 binary dosage file")
  if (!subfmt %in% c(1L, 2L))
    stop("bdfiles must be format 1 subformat 1 or 2")

  has_gp      <- subfmt == 2L
  n_samp      <- nrow(old_bdinfo$samples)
  n_snps      <- nrow(old_bdinfo$snps)
  indices     <- numeric(n_snps)
  current_pos <- 4.0  # bytes written after 4-byte magic header

  bdose_con <- file(bdose_file, open = "wb")
  on.exit(close(bdose_con), add = TRUE)
  writeBin(.bdose5_magic, bdose_con)

  for (i in seq_len(n_snps)) {
    snp <- getsnp(old_bdinfo, i, dosageonly = FALSE)
    if (has_gp) {
      gp <- as.numeric(rbind(snp$p0, snp$p1, snp$p2))
    } else {
      gp <- rep(NA_real_, 3L * n_samp)
    }
    block      <- compress_snp_block(snp$dosage, gp)
    indices[i] <- current_pos
    writeBin(block, bdose_con)
    current_pos <- current_pos + length(block)
  }

  new_bdinfo <- list(
    filename       = normalizePath(bdose_file),
    usesfid        = old_bdinfo$usesfid,
    samples        = old_bdinfo$samples,
    onechr         = old_bdinfo$onechr,
    snpidformat    = old_bdinfo$snpidformat,
    snps           = old_bdinfo$snps,
    snpinfo        = old_bdinfo$snpinfo,
    additionalinfo = structure(
      list(format     = 5L,
           subformat  = 1L,
           headersize = 4L,
           numgroups  = 1L,
           groups     = n_samp),
      class = "bdose-info"
    ),
    datasize       = integer(0),
    indices        = indices
  )
  class(new_bdinfo) <- "genetic-info"
  saveRDS(new_bdinfo, paste0(bdose_file, ".bdi"))

  invisible(NULL)
}

###########################################################
#          Binary Dosage Format 2 to Format 5            #
###########################################################

#' Convert a format 2 binary dosage file to Format 5
#'
#' Reads a format 2 binary dosage file (subformat 1 or 2) and converts it to
#' a Format 5 file pair. If the source file does not contain genotype
#' probabilities (subformat 1), those values are stored as missing in the
#' output.
#'
#' @param bdfiles Vector of three file names for the format 2 binary dosage
#'   file: the binary dosage file, the family file, and the map file.
#' @param bdose_file Path for the output .bdose file. The companion .bdi
#'   metadata file is written to \code{paste0(bdose_file, ".bdi")}.
#'
#' @return NULL (invisibly)
#'
#' @keywords internal
bd2tobd5 <- function(bdfiles, bdose_file) {
  if (missing(bdfiles))
    stop("No binary dosage files specified")
  if (missing(bdose_file))
    stop("No output .bdose file specified")

  old_bdinfo <- getbdinfo(bdfiles)

  fmt    <- old_bdinfo$additionalinfo$format
  subfmt <- old_bdinfo$additionalinfo$subformat

  if (fmt != 2L)
    stop("bdfiles must be a format 2 binary dosage file")
  if (!subfmt %in% c(1L, 2L))
    stop("bdfiles must be format 2 subformat 1 or 2")

  has_gp      <- subfmt == 2L
  n_samp      <- nrow(old_bdinfo$samples)
  n_snps      <- nrow(old_bdinfo$snps)
  indices     <- numeric(n_snps)
  current_pos <- 4.0

  bdose_con <- file(bdose_file, open = "wb")
  on.exit(close(bdose_con), add = TRUE)
  writeBin(.bdose5_magic, bdose_con)

  for (i in seq_len(n_snps)) {
    snp <- getsnp(old_bdinfo, i, dosageonly = FALSE)
    if (has_gp) {
      gp <- as.numeric(rbind(snp$p0, snp$p1, snp$p2))
    } else {
      gp <- rep(NA_real_, 3L * n_samp)
    }
    block      <- compress_snp_block(snp$dosage, gp)
    indices[i] <- current_pos
    writeBin(block, bdose_con)
    current_pos <- current_pos + length(block)
  }

  new_bdinfo <- list(
    filename       = normalizePath(bdose_file),
    usesfid        = old_bdinfo$usesfid,
    samples        = old_bdinfo$samples,
    onechr         = old_bdinfo$onechr,
    snpidformat    = old_bdinfo$snpidformat,
    snps           = old_bdinfo$snps,
    snpinfo        = old_bdinfo$snpinfo,
    additionalinfo = structure(
      list(format     = 5L,
           subformat  = 1L,
           headersize = 4L,
           numgroups  = 1L,
           groups     = n_samp),
      class = "bdose-info"
    ),
    datasize       = integer(0),
    indices        = indices
  )
  class(new_bdinfo) <- "genetic-info"
  saveRDS(new_bdinfo, paste0(bdose_file, ".bdi"))

  invisible(NULL)
}

###########################################################
#          Binary Dosage Format 3 to Format 5            #
###########################################################

#' Convert a format 3 binary dosage file to Format 5
#'
#' Reads a format 3 binary dosage file (subformat 1, 2, 3, or 4) and converts
#' it to a Format 5 file pair. If the source file does not contain genotype
#' probabilities (subformats 1 and 3), those values are stored as missing in
#' the output.
#'
#' @param bdfiles Vector of three file names for the format 3 binary dosage
#'   file: the binary dosage file, the family file, and the map file.
#'
#' @param bdose_file Path for the output .bdose file. The companion .bdi
#'   metadata file is written to \code{paste0(bdose_file, ".bdi")}.
#'
#' @return NULL (invisibly)
#'
#' @keywords internal
bd3tobd5 <- function(bdfiles, bdose_file) {
  if (missing(bdfiles))
    stop("No binary dosage files specified")
  if (missing(bdose_file))
    stop("No output .bdose file specified")

  old_bdinfo <- getbdinfo(bdfiles)

  fmt    <- old_bdinfo$additionalinfo$format
  subfmt <- old_bdinfo$additionalinfo$subformat

  if (fmt != 3L)
    stop("bdfiles must be a format 3 binary dosage file")
  if (!subfmt %in% c(1L, 2L, 3L, 4L))
    stop("bdfiles must be format 3 subformat 1, 2, 3, or 4")

  has_gp      <- subfmt %in% c(2L, 4L)
  n_samp      <- nrow(old_bdinfo$samples)
  n_snps      <- nrow(old_bdinfo$snps)
  indices     <- numeric(n_snps)
  current_pos <- 4.0

  bdose_con <- file(bdose_file, open = "wb")
  on.exit(close(bdose_con), add = TRUE)
  writeBin(.bdose5_magic, bdose_con)

  for (i in seq_len(n_snps)) {
    snp <- getsnp(old_bdinfo, i, dosageonly = FALSE)
    if (has_gp) {
      gp <- as.numeric(rbind(snp$p0, snp$p1, snp$p2))
    } else {
      gp <- rep(NA_real_, 3L * n_samp)
    }
    block      <- compress_snp_block(snp$dosage, gp)
    indices[i] <- current_pos
    writeBin(block, bdose_con)
    current_pos <- current_pos + length(block)
  }

  new_bdinfo <- list(
    filename       = normalizePath(bdose_file),
    usesfid        = old_bdinfo$usesfid,
    samples        = old_bdinfo$samples,
    onechr         = old_bdinfo$onechr,
    snpidformat    = old_bdinfo$snpidformat,
    snps           = old_bdinfo$snps,
    snpinfo        = old_bdinfo$snpinfo,
    additionalinfo = structure(
      list(format     = 5L,
           subformat  = 1L,
           headersize = 4L,
           numgroups  = 1L,
           groups     = n_samp),
      class = "bdose-info"
    ),
    datasize       = integer(0),
    indices        = indices
  )
  class(new_bdinfo) <- "genetic-info"
  saveRDS(new_bdinfo, paste0(bdose_file, ".bdi"))

  invisible(NULL)
}

###########################################################
#          Binary Dosage Format 4 to Format 5            #
###########################################################

#' Convert a format 4 binary dosage file to Format 5
#'
#' Reads a format 4 binary dosage file (subformat 1, 2, 3, or 4) and converts
#' it to a Format 5 file pair. If the source file does not contain genotype
#' probabilities (subformats 1 and 3), those values are stored as missing in
#' the output.
#'
#' @param bdfile File name of the format 4 binary dosage file. Format 4 stores
#'   all data in a single file.
#'
#' @param bdose_file Path for the output .bdose file. The companion .bdi
#'   metadata file is written to \code{paste0(bdose_file, ".bdi")}.
#'
#' @return NULL (invisibly)
#'
#' @keywords internal
bd4tobd5 <- function(bdfile, bdose_file) {
  if (missing(bdfile))
    stop("No binary dosage file specified")
  if (missing(bdose_file))
    stop("No output .bdose file specified")

  old_bdinfo <- getbdinfo(bdfile)

  fmt    <- old_bdinfo$additionalinfo$format
  subfmt <- old_bdinfo$additionalinfo$subformat

  if (fmt != 4L)
    stop("bdfile must be a format 4 binary dosage file")
  if (!subfmt %in% c(1L, 2L, 3L, 4L))
    stop("bdfile must be format 4 subformat 1, 2, 3, or 4")

  has_gp      <- subfmt %in% c(2L, 4L)
  n_samp      <- nrow(old_bdinfo$samples)
  n_snps      <- nrow(old_bdinfo$snps)
  indices     <- numeric(n_snps)
  current_pos <- 4.0

  bdose_con <- file(bdose_file, open = "wb")
  on.exit(close(bdose_con), add = TRUE)
  writeBin(.bdose5_magic, bdose_con)

  for (i in seq_len(n_snps)) {
    snp <- getsnp(old_bdinfo, i, dosageonly = FALSE)
    if (has_gp) {
      gp <- as.numeric(rbind(snp$p0, snp$p1, snp$p2))
    } else {
      gp <- rep(NA_real_, 3L * n_samp)
    }
    block      <- compress_snp_block(snp$dosage, gp)
    indices[i] <- current_pos
    writeBin(block, bdose_con)
    current_pos <- current_pos + length(block)
  }

  new_bdinfo <- list(
    filename       = normalizePath(bdose_file),
    usesfid        = old_bdinfo$usesfid,
    samples        = old_bdinfo$samples,
    onechr         = old_bdinfo$onechr,
    snpidformat    = old_bdinfo$snpidformat,
    snps           = old_bdinfo$snps,
    snpinfo        = old_bdinfo$snpinfo,
    additionalinfo = structure(
      list(format     = 5L,
           subformat  = 1L,
           headersize = 4L,
           numgroups  = 1L,
           groups     = n_samp),
      class = "bdose-info"
    ),
    datasize       = integer(0),
    indices        = indices
  )
  class(new_bdinfo) <- "genetic-info"
  saveRDS(new_bdinfo, paste0(bdose_file, ".bdi"))

  invisible(NULL)
}

###########################################################
#        Update Binary Dosage Format 1-4 to Format 5     #
###########################################################

#' Update a binary dosage file to Format 5
#'
#' Reads a binary dosage file in format 1, 2, 3, or 4, detects the format
#' automatically, and converts it to a Format 5 file pair by calling the
#' appropriate conversion routine. If the source file does not contain genotype
#' probabilities, those values are stored as missing in the output.
#'
#' @param bdfiles Vector of file names for the binary dosage file. Format 4
#'   files require one file name. Formats 1, 2, and 3 require three file names:
#'   the binary dosage file, the family file, and the map file.
#'
#' @param bdose_file Path for the output .bdose file. The companion .bdi
#'   metadata file is written to \code{paste0(bdose_file, ".bdi")}.
#'
#' @return NULL (invisibly)
#' @export
#'
#' @examples
#' vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")
#' bdfile <- tempfile()
#' suppressWarnings(
#'   vcftobdlegacy(vcffiles = vcf1afile,
#'                 bdfiles = bdfile,
#'                 format = 4L)
#' )
#' bdose_file <- tempfile(fileext = ".bdose")
#' updatebd(bdfiles    = bdfile,
#'          bdose_file = bdose_file)
updatebd <- function(bdfiles, bdose_file) {
  if (missing(bdfiles))
    stop("No binary dosage files specified")
  if (missing(bdose_file))
    stop("No output .bdose file specified")

  old_bdinfo <- getbdinfo(bdfiles)
  fmt        <- old_bdinfo$additionalinfo$format

  if (fmt == 1L)
    bd1tobd5(bdfiles = bdfiles, bdose_file = bdose_file)
  else if (fmt == 2L)
    bd2tobd5(bdfiles = bdfiles, bdose_file = bdose_file)
  else if (fmt == 3L)
    bd3tobd5(bdfiles = bdfiles, bdose_file = bdose_file)
  else if (fmt == 4L)
    bd4tobd5(bdfile = bdfiles[1L], bdose_file = bdose_file)
  else
    stop("bdfiles must be a format 1, 2, 3, or 4 binary dosage file")

  invisible(NULL)
}

###########################################################
#                  Gen to Binary Dosage                   #
###########################################################

#' Convert a gen file to a binary dosage file
#'
#' Routine to read information from a gen file and create
#' a binary dosage file. Note: This routine can take a long
#' time to run if the gen file is large.
#'
#' @param genfiles A vector of file names.
#' The first is the name of the gen file. The
#' second is name of the sample file that contains
#' the subject information.
#' @param snpcolumns Column numbers containing chromosome,
#' snpid, location, reference allele, alternate allele,
#' respectively. This must be an integer vector. All
#' values must be positive except for the chromosome.
#' The value for the chromosome may be -1 or -0.
#' -1 indicates that the chromosome value is passed to
#' the routine using the chromosome parameter.
#' 0 indicates that the chromosome value is in the snpid
#' and that the snpid has the format chromosome:other_data.
#' Default value is c(1L, 2L, 3L, 4L, 5L).
#' @param startcolumn Column number of first column with
#' genetic probabilities or dosages. Must
#' be an integer value. Default value is 6L.
#' @param impformat Number of genetic data values per
#' subject. 1 indicates dosage only, 2 indicates P(g=0)
#' and P(g=1) only, 3 indicates P(g=0), P(g=1), and
#' P(g=2). Default value is 3L.
#' @param chromosome Chromosome value to use if the
#' first value of the snpcolumns is equal to 0.
#' Default value is character().
#' @param header Indicators if the gen and sample files
#' have headers. If the gen file does not have a
#' header. A sample file must be included.
#' Default value is c(FALSE, TRUE).
#' @param gz Indicator if file is compressed using gzip.
#' Default value is FALSE.
#' @param sep Separator used in the gen file. Default
#' value is `"\t"`
#' @param bdfiles Vector of names of the output files.
#' The binary dosage file name is first. The family and
#' map files follow. For format 4, no family and map file
#' names are needed.
#' @param format The format of the output binary dosage file.
#' Allowed values are 1, 2, 3, and 4. The default value is 4.
#' Using the default value is recommended.
#' @param subformat The subformat of the format of the output
#' binary dosage file. A value of 1 or 3 indicates that only the
#' dosage value is saved. A value of 2 or 4 indicates
#' the dosage and genetic probabilities will be output. Values
#' of 3 or 4 are only allowed with formats 3 and 4. If a value
#' of zero if provided, and genetic probabilities are in the vcf
#' file, subformat 2 will be used for formats 1 and 2, and
#' subformat 4 will be used for formats 3 and 4. If the vcf file
#' does not contain genetic probabilities, subformat 1 will be
#' used for formats 1 and 2, and subformat 3 will be used for
#' formats 3 and 4. The default value is 0.
#' @param snpidformat The format that the SNP ID will be saved as.
#' -1 - SNP ID not written.
#' 0 - same as in the VCF file.
#' 1 - chr:pos.
#' 2 - chr:pos:ref:alt.
#' If snpidformat is 1 and the VCF file uses format 2, an error is
#' generated. Default value is 0.
#' @param bdoptions Character array containing any of the following
#' value, "aaf", "maf", "rsq". The presence of any of these
#' values indicates that the specified values should be
#' calculates and stored in the binary dosage file. These values only
#' apply to format 4.
#'
#' @return
#' None
#' @export
#'
#' @examples
#' # Find the gen file names
#' gen3afile <- system.file("extdata", "set3a.imp", package = "BinaryDosage")
#' gen3asample <- system.file("extdata", "set3a.sample", package = "BinaryDosage")
#' # Get temporary output file name
#' bdfiles <- tempfile()
#' # Convert the file
#' gentobd(genfiles = c(gen3afile, gen3asample),
#'         snpcolumns = c(0L, 2L:5L),
#'         bdfiles = bdfiles)
#' # Verify the file was written correctly
#' bdinfo <- getbdinfo(bdfiles = bdfiles)
gentobd <- function(genfiles,
                    snpcolumns = 1L:5L,
                    startcolumn = 6L,
                    impformat = 3L,
                    chromosome = character(),
                    header = c(FALSE, TRUE),
                    gz = FALSE,
                    sep = "\t",
                    bdfiles,
                    format = 4L,
                    subformat = 0L,
                    snpidformat = 0L,
                    bdoptions = character(0)) {
  if (missing(genfiles) == TRUE)
    stop("No gen file specified")

  if (missing(bdfiles) == TRUE)
    stop("No output files specified")

  validation <- validatebdinput(bdfiles = bdfiles,
                                format = format,
                                subformat = subformat,
                                snpidformat = snpidformat,
                                bdoptions = bdoptions)
  format <- validation$format
  subformat <- validation$subformat

  if (snpidformat == -1)
    readsnpformat = 0L
  else
    readsnpformat = as.integer(snpidformat)
  geninfo <- getgeninfo(genfiles = genfiles,
                        snpcolumns = snpcolumns,
                        startcolumn = startcolumn,
                        impformat = impformat,
                        chromosome = chromosome,
                        header = header,
                        gz = gz,
                        index = FALSE,
                        snpidformat = readsnpformat,
                        sep = sep)
  if (snpidformat == -1)
    geninfo$snpidformat = 1
  else
    geninfo$snpidformat = 0

  if (subformat == 0) {
    if (geninfo$additionalinfo$format == 1L)
      subformat <- 1L
    else
      subformat <- 2L
  }
  WriteBinaryDosageHeader(format = format,
                          subformat = subformat,
                          filename = bdfiles,
                          genefileinfo = geninfo,
                          bdoptions = bdoptions)
  headerinfo <- ReadBinaryDosageHeader(filename = bdfiles)
  bdwriteinfo <- AllocateBinaryDosageWriteMemory(headerinfo = headerinfo)
  genapply(geninfo = geninfo,
           func = WriteBinaryDosageData,
           writeinfo = bdwriteinfo)
  WriteBinaryDosageIndices(writeinfo = bdwriteinfo)
  bdinfo <- getbdinfo(bdfiles = bdfiles)
  if (is.na(match("aaf", bdoptions)) == FALSE)
    updateaaf(bdinfo)
  if (is.na(match("maf", bdoptions)) == FALSE)
    updatemaf(bdinfo)
  if (is.na(match("rsq", bdoptions)) == FALSE)
    updatersq(bdinfo)
}

test_that("filemismatch12", {
  gen3afile <- system.file("extdata", "set3a.imp", package = "BinaryDosage")
  gen3asample <- system.file("extdata", "set3a.sample", package = "BinaryDosage")
  bdfile3a <- tempfile()
  bdfam3a <- tempfile()
  bdmap3a <- tempfile()
  gentobd(genfiles = c(gen3afile, gen3asample),
          snpcolumns = c(0L, 2L:5L),
          bdfiles = c(bdfile3a, bdfam3a, bdmap3a),
          format = 3)

  vcf1basnpfile <- system.file("extdata", "set1b_asnp.vcf", package = "BinaryDosage")
  bdfile1basnp <- tempfile()
  bdfam1basnp <- tempfile()
  bdmap1basnp <- tempfile()
  vcftobdlegacy(vcffiles = vcf1basnpfile,
          bdfiles = c(bdfile1basnp, bdfam1basnp, bdmap1basnp),
          format = 3)
  expect_error(getbdinfo(bdfiles = c(bdfile1basnp, bdfam3a, bdmap3a)),
               "Subject file does not line up with binary dosage file")

  vcf2bfile <- system.file("extdata", "set2b.vcf", package = "BinaryDosage")
  bdfile2b <- tempfile()
  bdfam2b <- tempfile()
  bdmap2b <- tempfile()
  vcftobdlegacy(vcffiles = vcf2bfile,
          bdfiles = c(bdfile2b, bdfam2b, bdmap2b),
          format = 3)
  expect_error(getbdinfo(bdfiles = c(bdfile2b, bdfam3a, bdmap3a)),
               "Subject file does not line up with binary dosage file")
})

test_that("filemismatch34", {
  gen3afile <- system.file("extdata", "set3a.imp", package = "BinaryDosage")
  gen3asample <- system.file("extdata", "set3a.sample", package = "BinaryDosage")
  bdfile3a <- tempfile()
  bdfam3a <- tempfile()
  bdmap3a <- tempfile()
  gentobd(genfiles = c(gen3afile, gen3asample),
          snpcolumns = c(0L, 2L:5L),
          bdfiles = c(bdfile3a, bdfam3a, bdmap3a),
          format = 3)

  vcf1basnpfile <- system.file("extdata", "set1b_asnp.vcf", package = "BinaryDosage")
  bdfile1basnp <- tempfile()
  bdfam1basnp <- tempfile()
  bdmap1basnp <- tempfile()
  vcftobdlegacy(vcffiles = vcf1basnpfile,
          bdfiles = c(bdfile1basnp, bdfam1basnp, bdmap1basnp),
          format = 3,
          subformat = 4)
  expect_error(getbdinfo(bdfiles = c(bdfile1basnp, bdfam3a, bdmap3a)),
               "Subject file does not line up with binary dosage file")
  expect_error(getbdinfo(bdfiles = c(bdfile1basnp, bdfam1basnp, bdmap3a)),
               "Map file does not line up with binary dosage file")

  vcf2bfile <- system.file("extdata", "set2b.vcf", package = "BinaryDosage")
  bdfile2b <- tempfile()
  bdfam2b <- tempfile()
  bdmap2b <- tempfile()
  vcftobdlegacy(vcffiles = vcf2bfile,
          bdfiles = c(bdfile2b, bdfam2b, bdmap2b),
          format = 3,
          subformat = 3)
  expect_error(getbdinfo(bdfiles = c(bdfile2b, bdfam3a, bdmap3a)),
               "Subject file does not line up with binary dosage file")
  expect_error(getbdinfo(bdfiles = c(bdfile2b, bdfam2b, bdmap3a)),
               "Map file does not line up with binary dosage file")
})

test_that("merge", {
  expect_error(bdmerge(bdfiles = c("file1", "file2")),
               "No output files specified")
  expect_error(bdmerge(mergefiles = 1L,
                       bdfiles = c("file1", "file2")),
               "Output file names must be a character values")
  expect_error(bdmerge(mergefiles = c("file1", "file2", "file3"),
                       bdfiles = c("file1", "file2")),
               "Only one file name is needed when using format 4")
  expect_error(bdmerge(mergefiles = "file1",
                       format = 3,
                       bdfiles = c("file1", "file2")),
               "Three file names are required when using formats 1, 2, and 3")
  expect_error(bdmerge(mergefiles = "",
                       bdfiles = c("file1", "file2")),
               "Output file names cannot be blank")

  expect_error(bdmerge(mergefiles = "file1"),
               "No files specified")
  expect_error(bdmerge(mergefiles = "file1",
                       bdfiles = 1),
               "bdfiles must be a character vector")
  expect_error(bdmerge(mergefiles = "file1",
                       bdfiles = "file1"),
               "At least two binary dosage files must be specified")

  expect_error(bdmerge(mergefiles = "file1",
                       bdfiles = c("file1", "file2"),
                       famfiles = 1,
                       mapfiles = c("file1", "file2")),
               "famfiles must be a character vector")
  expect_error(bdmerge(mergefiles = "file1",
                       bdfiles = c("file1", "file2"),
                       famfiles = c("file1", "file2"),
                       mapfiles = 1),
               "mapfiles must be a character vector")
  expect_error(bdmerge(mergefiles = "file1",
                       bdfiles = c("file1", "file2"),
                       famfiles = c("file1", "file2")),
               "If famfiles is specified, mapfiles must be specified")
  expect_error(bdmerge(mergefiles = "file1",
                       bdfiles = c("file1", "file2"),
                       mapfiles = c("file1", "file2")),
               "If mapfiles is specified, famfiles must be specified")
  expect_error(bdmerge(mergefiles = "file1",
                       bdfiles = c("file1", "file2"),
                       famfiles = c("file1", "file2"),
                       mapfiles = "file1"),
               "If famfiles and mapfiles are specified they must have the same length as bdfiles")

  expect_error(bdmerge(mergefiles = "file1",
                       format = "a",
                       bdfiles = c("file1", "file2")),
               "format must be an integer value")
  expect_error(bdmerge(mergefiles = "file1",
                       format = 1:2,
                       bdfiles = c("file1", "file2")),
               "format must be an integer vector of length 1")
  expect_error(bdmerge(mergefiles = "file1",
                       format = 1.2,
                       bdfiles = c("file1", "file2")),
               "format must be an integer")
  expect_error(bdmerge(mergefiles = "file1",
                       format = 5,
                       bdfiles = c("file1", "file2")),
               "format must be an integer value from 1 to 4")

  expect_error(bdmerge(mergefiles = "file1",
                       subformat = "a",
                       bdfiles = c("file1", "file2")),
               "subformat must be an integer value")
  expect_error(bdmerge(mergefiles = "file1",
                       subformat = 1:2,
                       bdfiles = c("file1", "file2")),
               "subformat must be an integer vector of length 1")
  expect_error(bdmerge(mergefiles = "file1",
                       subformat = 1.2,
                       bdfiles = c("file1", "file2")),
               "subformat must be an integer")
  expect_error(bdmerge(mergefiles = "file1",
                       subformat = 5,
                       bdfiles = c("file1", "file2")),
               "subformat must be an integer value from 0 to 4")
  expect_error(bdmerge(mergefiles = "file1",
                       format = 2,
                       subformat = 3,
                       bdfiles = c("file1", "file2")),
               "subformat must be an integer value from 0 to 2 for formats 1 and 2")

  expect_error(bdmerge(mergefiles = "file1",
                       bdfiles = c("file1", "file2"),
                       onegroup = 1),
               "onegroup must be logical value")
  expect_error(bdmerge(mergefiles = "file1",
                       bdfiles = c("file1", "file2"),
                       onegroup = c(TRUE, TRUE)),
               "onegroup must be a logical vector of length 1")

  expect_error(bdmerge(mergefiles = "file1",
                       bdfiles = c("file1", "file2"),
                       bdoptions = 1),
               "bdoptions must be a character vector")
  expect_error(bdmerge(mergefiles = "file1",
                       bdfiles = c("file1", "file2"),
                       onegroup = FALSE,
                       bdoptions = "aaf"),
               "bdoptions can only be used if onegroup is TRUE")
  expect_error(bdmerge(mergefiles = c("file1", "file2", "file3"),
                       bdfiles = c("file1", "file2"),
                       format = 3,
                       bdoptions = "aaf"),
               "bdoptions can only be used if format = 4")
  expect_error(bdmerge(mergefiles = "file1",
                       bdfiles = c("file1", "file2"),
                       bdoptions = "abc"),
               "Only valid bdoptions are aaf, maf, and rsq")

  expect_error(bdmerge(mergefiles = "file1",
                       bdfiles = c("file1", "file2"),
                       snpjoin = "abc"),
               "snpjoin must have a value of either \"inner\" or \"outer\"")

  bdvcf1afile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
  bdvcf1bfile <- system.file("extdata", "vcf1b.bdose", package = "BinaryDosage")
  mergefiles <- tempfile()

  expect_error(bdmerge(mergefiles = mergefiles,
                       bdfiles = c(bdvcf1afile, bdvcf1bfile),
                       bdoptions = c("aaf", "maf", "rsq")),
               NA)
  expect_error(bdinfo <- getbdinfo(mergefiles), NA)

  vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")
  vcf1ainfo <- system.file("extdata", "set1a.info", package = "BinaryDosage")
  vcf1bfile <- system.file("extdata", "set1b.vcf", package = "BinaryDosage")
  vcf1binfo <- system.file("extdata", "set1b.info", package = "BinaryDosage")
  bdfile1a <- tempfile()
  bdfile1b <- tempfile()
  vcftobdlegacy(vcffiles = c(vcf1afile, vcf1ainfo),
          bdfiles = bdfile1a)
  vcftobdlegacy(vcffiles = c(vcf1bfile, vcf1binfo),
          bdfiles = bdfile1b)

  bdinfo1a <- getbdinfo(bdfile1a)
  bdinfo1b <- getbdinfo(bdfile1b)

  mergefiles1g <- tempfile()
  expect_error(bdmerge(mergefiles = mergefiles1g,
                       bdfiles = c(bdfile1a, bdfile1b),
                       onegroup = FALSE),
               NA)
  expect_error(getbdinfo(mergefiles1g), NA)

  vcf2afile <- system.file("extdata", "set2a.vcf", package = "BinaryDosage")
  vcf2bfile <- system.file("extdata", "set2b.vcf", package = "BinaryDosage")

  bdfile2a <- tempfile()
  bdfam2a <- tempfile()
  bdmap2a <- tempfile()
  bdfile2b <- tempfile()
  bdfam2b <- tempfile()
  bdmap2b <- tempfile()
  vcftobdlegacy(vcffiles = vcf2afile,
          bdfiles = c(bdfile2a, bdfam2a, bdmap2a),
          format = 3L)
  vcftobdlegacy(vcffiles = vcf2bfile,
          bdfiles = c(bdfile2b, bdfam2b, bdmap2b),
          format = 3L)

  bdinfo2a <- getbdinfo(c(bdfile2a, bdfam2a, bdmap2a))
  bdinfo2b <- getbdinfo(c(bdfile2b, bdfam2b, bdmap2b))

  mergefiles2g <- tempfile()
  expect_error(bdmerge(mergefiles = mergefiles2g,
                       bdfiles = c(bdfile2a, bdfile2b),
                       famfiles = c(bdfam2a, bdfam2b),
                       mapfiles = c(bdmap2a, bdmap2b)),
               NA)
  expect_error(getbdinfo(mergefiles2g), NA)

  vcf1brfile <- system.file("extdata", "set1b_rsub.vcf", package = "BinaryDosage")
  bdfile1br <- tempfile()
  vcftobdlegacy(vcffiles = vcf1brfile,
          bdfiles = bdfile1br)
  bdinfo1br <- getbdinfo(bdfile1br)

  mergefile3 <- tempfile()
  expect_error(bdmerge(mergefiles = mergefile3,
                       bdfiles = c(bdfile1a, bdfile1br)),
               "There are duplicate samples in the files to merge")

  gen3bfile <- system.file("extdata", "set3b.imp", package = "BinaryDosage")
  gen3bsample <- system.file("extdata", "set3b.sample", package = "BinaryDosage")
  bdfile3b <- tempfile()
  gentobd(genfiles = c(gen3bfile, gen3bsample),
          snpcolumns = c(0L, 2L:5L),
          bdfiles = bdfile3b)
  bdinfo3b <- getbdinfo(bdfiles = bdfile3b)
  mergefile3b <- tempfile()
  expect_error(bdmerge(mergefiles = mergefile3b,
                       bdfiles = c(bdfile1a, bdfile3b)),
               "Some files use FID and others do not")

  bdvcf1afile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
  vcf1brs <- system.file("extdata", "set1b_rssnp.vcf", package = "BinaryDosage")
  bdinfo1a <- getbdinfo(bdvcf1afile)
  bdfile1brs <- tempfile()
  vcftobdlegacy(vcffiles = vcf1brs,
          bdfiles = bdfile1brs)
  bdinfo1brs <- getbdinfo(bdfiles = bdfile1brs)
  merge1brs <- tempfile()
  expect_error(bdmerge(mergefiles = merge1brs,
                       bdfiles = c(bdvcf1afile, bdfile1brs)),
               NA)
  expect_error(getbdinfo(merge1brs), NA)
})

#***************************************************************************#
#                                                                           #
#                      Merging Format 5 Binary Dosage files                 #
#                                                                           #
#***************************************************************************#

#' Merge Format 5 binary dosage files
#'
#' Merges two or more Format 5 binary dosage files into a single Format 5
#' output file. The merge type is determined automatically:
#'
#' \itemize{
#'   \item If subject IDs do not overlap across files, a \strong{subject merge}
#'         is performed: the output contains all subjects from every file and
#'         the SNPs common to all files.
#'   \item If SNP IDs do not overlap across files, a \strong{SNP merge} is
#'         performed: the output contains all SNPs from every file and the
#'         subjects common to all files.
#' }
#'
#' If both subject IDs and SNP IDs overlap across files the merge cannot be
#' performed and an error is returned.
#'
#' SNPs are identified by chromosome, position, reference allele, and alternate
#' allele.
#'
#' @param bdose_files Character vector of paths to the input .bdose files.
#'   Must contain at least two entries. The companion .bdi file for each is
#'   expected at \code{paste0(bdose_files[i], ".bdi")}.
#' @param bdose_file Path for the output .bdose file. The companion .bdi
#'   metadata file is written to \code{paste0(bdose_file, ".bdi")}.
#'
#' @return NULL (invisibly)
#' @export
#'
#' @examples
#' bdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
#' bdinfo_src <- getbdinfo(bdfile)
#'
#' # Create two format 5 files with non-overlapping subjects
#' bdose_a   <- tempfile(fileext = ".bdose")
#' bdose_b   <- tempfile(fileext = ".bdose")
#' bdose_tmp <- tempfile(fileext = ".bdose")
#' updatebd(bdfiles = bdfile, bdose_file = bdose_tmp)
#' subsetbd(bdfiles    = bdose_tmp,
#'          bdose_file = bdose_a,
#'          subjectids = bdinfo_src$samples$sid[1:30])
#' subsetbd(bdfiles    = bdose_tmp,
#'          bdose_file = bdose_b,
#'          subjectids = bdinfo_src$samples$sid[31:60])
#'
#' bdose_out <- tempfile(fileext = ".bdose")
#' mergebd(bdose_files = c(bdose_a, bdose_b),
#'         bdose_file  = bdose_out)
mergebd <- function(bdose_files, bdose_file) {

  # --- Argument validation ---
  if (missing(bdose_files))
    stop("No input .bdose files specified")
  if (missing(bdose_file))
    stop("No output .bdose file specified")
  if (!is.character(bdose_files))
    stop("bdose_files must be a character vector")
  if (length(bdose_files) < 2L)
    stop("At least two input files are required for merging")

  n_files <- length(bdose_files)

  # --- Load all bdinfo objects (getbd5info verifies Format 5) ---
  bdinfos <- vector("list", n_files)
  for (k in seq_len(n_files)) {
    bdinfos[[k]] <- getbd5info(bdose_file = bdose_files[k])
  }

  n_snps <- vapply(bdinfos, function(b) nrow(b$snps),    integer(1L))
  n_samp <- vapply(bdinfos, function(b) nrow(b$samples), integer(1L))

  # SNP key: chr:pos:ref:alt (unambiguous regardless of snpidformat)
  snp_keys <- lapply(bdinfos, function(b)
    paste(b$snps$chromosome, b$snps$location,
          b$snps$reference,  b$snps$alternate, sep = ":"))
  subj_ids <- lapply(bdinfos, function(b) b$samples$sid)

  # --- Determine merge type ---
  subjects_intersect <- anyDuplicated(unlist(subj_ids))  > 0L
  snps_intersect     <- anyDuplicated(unlist(snp_keys)) > 0L

  if (subjects_intersect && snps_intersect)
    stop("Cannot merge: subject IDs and SNP IDs both overlap across files")

  if (!subjects_intersect) {
    # ------------------------------------------------------------------ #
    #  Subject merge: unique subjects across files, SNPs must overlap     #
    # ------------------------------------------------------------------ #
    common_keys <- snp_keys[[1L]]
    for (k in seq(2L, n_files))
      common_keys <- intersect(common_keys, snp_keys[[k]])

    if (length(common_keys) == 0L)
      stop("Cannot merge: no SNPs in common across files for subject merge")

    # Index of each common SNP in each file
    snp_idx <- lapply(snp_keys, function(kv) match(common_keys, kv))

    n_snps_out <- length(common_keys)
    n_samp_out <- sum(n_samp)

    bdose_con <- file(bdose_file, open = "wb")
    on.exit(close(bdose_con), add = TRUE)
    writeBin(.bdose5_magic, bdose_con)

    indices     <- numeric(n_snps_out)
    current_pos <- 4.0

    for (i in seq_len(n_snps_out)) {
      ds_all <- numeric(0)
      gp_all <- numeric(0)

      for (k in seq_len(n_files)) {
        snp <- getsnp(bdinfos[[k]], snp_idx[[k]][i], dosageonly = FALSE)
        ds_all <- c(ds_all, snp$dosage)
        if (all(is.na(snp$p0))) {
          gp_all <- c(gp_all, rep(NA_real_, 3L * n_samp[k]))
        } else {
          gp_all <- c(gp_all,
                      as.numeric(rbind(snp$p0, snp$p1, snp$p2)))
        }
      }

      block       <- compress_snp_block(ds_all, gp_all)
      indices[i]  <- current_pos
      writeBin(block, bdose_con)
      current_pos <- current_pos + length(block)
    }

    # Output SNP info from file 1 (in common-SNP order)
    out_snps <- bdinfos[[1L]]$snps[snp_idx[[1L]], , drop = FALSE]
    rownames(out_snps) <- NULL

    # Output samples: all subjects in file order
    out_samples <- do.call(rbind, lapply(bdinfos, function(b) b$samples))
    rownames(out_samples) <- NULL

  } else {
    # ------------------------------------------------------------------ #
    #  SNP merge: unique SNPs across files, subjects must overlap         #
    # ------------------------------------------------------------------ #
    common_sids <- subj_ids[[1L]]
    for (k in seq(2L, n_files))
      common_sids <- intersect(common_sids, subj_ids[[k]])

    if (length(common_sids) == 0L)
      stop("Cannot merge: no subjects in common across files for SNP merge")

    # Index of each common subject in each file
    samp_idx <- lapply(subj_ids, function(sv) match(common_sids, sv))

    n_samp_out <- length(common_sids)

    bdose_con <- file(bdose_file, open = "wb")
    on.exit(close(bdose_con), add = TRUE)
    writeBin(.bdose5_magic, bdose_con)

    indices     <- numeric(sum(n_snps))
    current_pos <- 4.0
    snp_count   <- 0L

    for (k in seq_len(n_files)) {
      si <- samp_idx[[k]]
      for (i in seq_len(n_snps[k])) {
        snp <- getsnp(bdinfos[[k]], i, dosageonly = FALSE)
        ds  <- snp$dosage[si]
        if (all(is.na(snp$p0))) {
          gp <- rep(NA_real_, 3L * n_samp_out)
        } else {
          gp <- as.numeric(rbind(snp$p0[si], snp$p1[si], snp$p2[si]))
        }
        snp_count            <- snp_count + 1L
        block                <- compress_snp_block(ds, gp)
        indices[snp_count]   <- current_pos
        writeBin(block, bdose_con)
        current_pos <- current_pos + length(block)
      }
    }

    # Output SNPs: all SNPs from all files in file order
    out_snps <- do.call(rbind, lapply(bdinfos, function(b) b$snps))
    rownames(out_snps) <- NULL

    # Output samples: common subjects in first-file order
    out_samples <- bdinfos[[1L]]$samples[samp_idx[[1L]], , drop = FALSE]
    rownames(out_samples) <- NULL
  }

  # --- Save output bdinfo ---
  ref      <- bdinfos[[1L]]
  new_bdinfo <- list(
    filename       = normalizePath(bdose_file),
    usesfid        = ref$usesfid,
    samples        = out_samples,
    onechr         = length(unique(out_snps$chromosome)) == 1L,
    snpidformat    = ref$snpidformat,
    snps           = out_snps,
    snpinfo        = list(),
    additionalinfo = structure(
      list(format     = 5L,
           subformat  = 1L,
           headersize = 4L,
           numgroups  = 1L,
           groups     = n_samp_out),
      class = "bdose-info"
    ),
    datasize       = integer(0),
    indices        = indices
  )
  class(new_bdinfo) <- "genetic-info"
  saveRDS(new_bdinfo, paste0(bdose_file, ".bdi"))

  invisible(NULL)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

GetLineLocations <- function(filename) {
    .Call(`_BinaryDosage_GetLineLocations`, filename)
}

ReadBinaryDosageBaseHeader <- function(filename) {
    .Call(`_BinaryDosage_ReadBinaryDosageBaseHeader`, filename)
}

ReadBinaryDosageHeader3A <- function(filename) {
    .Call(`_BinaryDosage_ReadBinaryDosageHeader3A`, filename)
}

ReadBinaryDosageHeader3B <- function(filename) {
    .Call(`_BinaryDosage_ReadBinaryDosageHeader3B`, filename)
}

ReadBinaryDosageHeader4A <- function(filename) {
    .Call(`_BinaryDosage_ReadBinaryDosageHeader4A`, filename)
}

ReadBinaryDosageHeader4B <- function(filename) {
    .Call(`_BinaryDosage_ReadBinaryDosageHeader4B`, filename)
}

ReadBDIndices3C <- function(filename, numSNPs, indexStart) {
    .Call(`_BinaryDosage_ReadBDIndices3C`, filename, numSNPs, indexStart)
}

ReadBDIndices4C <- function(filename, numSNPs, headersize) {
    .Call(`_BinaryDosage_ReadBDIndices4C`, filename, numSNPs, headersize)
}

DecodeFormat5BlockC <- function(raw_block, n_samp, dosage, p0, p1, p2) {
    .Call(`_BinaryDosage_DecodeFormat5BlockC`, raw_block, n_samp, dosage, p0, p1, p2)
}

OpenFormat5FileC <- function(filename) {
    .Call(`_BinaryDosage_OpenFormat5FileC`, filename)
}

CloseFormat5FileC <- function(xptr_sexp) {
    invisible(.Call(`_BinaryDosage_CloseFormat5FileC`, xptr_sexp))
}

ReadFormat5SNPC <- function(xptr_sexp, start, nbytes, n_samp, dosage, p0, p1, p2) {
    .Call(`_BinaryDosage_ReadFormat5SNPC`, xptr_sexp, start, nbytes, n_samp, dosage, p0, p1, p2)
}

ReadBinaryDosageDataC <- function(filename, headersize, numsub, snp, dosage, us, base) {
    .Call(`_BinaryDosage_ReadBinaryDosageDataC`, filename, headersize, numsub, snp, dosage, us, base)
}

ReadBinaryDosageDataP1P2 <- function(filename, headersize, numsub, snp, dosage, p0, p1, p2, us, base) {
    .Call(`_BinaryDosage_ReadBinaryDosageDataP1P2`, filename, headersize, numsub, snp, dosage, p0, p1, p2, us, base)
}

ReadBinaryDosageDataCompressed <- function(filename, index, datasize, numsub, dosage, p0, p1, p2, us) {
    .Call(`_BinaryDosage_ReadBinaryDosageDataCompressed`, filename, index, datasize, numsub, dosage, p0, p1, p2, us)
}

WriteBinaryDosageBaseHeader <- function(filename, format, subformat) {
    .Call(`_BinaryDosage_WriteBinaryDosageBaseHeader`, filename, format, subformat)
}

WriteBinaryDosageHeader3A <- function(filename, numSubjects) {
    .Call(`_BinaryDosage_WriteBinaryDosageHeader3A`, filename, numSubjects)
}

WriteBinaryDosageHeader3B <- function(filename, md5samples, md5SNPs, numIndices) {
    .Call(`_BinaryDosage_WriteBinaryDosageHeader3B`, filename, md5samples, md5SNPs, numIndices)
}

WriteBinaryDosageHeader4A <- function(filename, headerEntries, numSubjects, numSNPs, groups, sid, fid, snpid, chromosome, location, reference, alternate, aaf, maf, avgCall, rsq, offsets, numIndices) {
    .Call(`_BinaryDosage_WriteBinaryDosageHeader4A`, filename, headerEntries, numSubjects, numSNPs, groups, sid, fid, snpid, chromosome, location, reference, alternate, aaf, maf, avgCall, rsq, offsets, numIndices)
}

WriteBinaryDosageDataC <- function(filename, dosage, us, base) {
    .Call(`_BinaryDosage_WriteBinaryDosageDataC`, filename, dosage, us, base)
}

WriteBinaryP1P2Data <- function(filename, p1, p2, us, base) {
    .Call(`_BinaryDosage_WriteBinaryP1P2Data`, filename, p1, p2, us, base)
}

WriteBinaryCompressed <- function(filename, dosage, p0, p1, p2, snpnumber, datasize, us) {
    .Call(`_BinaryDosage_WriteBinaryCompressed`, filename, dosage, p0, p1, p2, snpnumber, datasize, us)
}

WriteBinaryDosageIndicesC <- function(filename, headersize, datasize) {
    .Call(`_BinaryDosage_WriteBinaryDosageIndicesC`, filename, headersize, datasize)
}

updatesnpinfo <- function(filename, offset, value) {
    .Call(`_BinaryDosage_updatesnpinfo`, filename, offset, value)
}

test_that("vcftobdlegacy", {
  expect_error(vcftobdlegacy(bdfiles = "file1"),
               "No VCF file specified")
  expect_error(vcftobdlegacy(vcffiles = "file1"),
               "No output files specified")
  expect_error(vcftobdlegacy(vcffiles = "file1",
                       bdfiles = 1),
               "bdfiles must be a vector of characters")

  expect_error(vcftobdlegacy(vcffiles = "file1",
                       bdfiles = "file2",
                       format = ""),
               "format must be an integer value")
  expect_error(vcftobdlegacy(vcffiles = "file1",
                       bdfiles = "file2",
                       format = 1:2),
               "format must be an integer vector of length 1")
  expect_error(vcftobdlegacy(vcffiles = "file1",
                       bdfiles = "file2",
                       format = 1.2),
               "format must be an integer")
  expect_error(vcftobdlegacy(vcffiles = "file1",
                       bdfiles = "file2",
                       format = 5),
               "format must be an integer value from 1 to 4")

  expect_error(vcftobdlegacy(vcffiles = "file1",
                       bdfiles = "file2",
                       subformat = ""),
               "subformat must be an integer value")
  expect_error(vcftobdlegacy(vcffiles = "file1",
                       bdfiles = "file2",
                       subformat = 1:2),
               "subformat must be an integer vector of length 1")
  expect_error(vcftobdlegacy(vcffiles = "file1",
                       bdfiles = "file2",
                       subformat = 1.2),
               "subformat must be an integer")
  expect_error(vcftobdlegacy(vcffiles = "file1",
                       bdfiles = "file2",
                       subformat = 5),
               "subformat must be an integer value from 0 to 4")
  expect_error(vcftobdlegacy(vcffiles = "file1",
                       bdfiles = "file2",
                       format = 2,
                       subformat = 3),
               "subformat must be an integer value from 0 to 2 for formats 1 and 2")

  expect_error(vcftobdlegacy(vcffiles = "file1",
                       bdfiles = c("file2", "file3"),
                       format = 4),
               "Only one output file name is needed when using format 4")
  expect_error(vcftobdlegacy(vcffiles = "file1",
                       bdfiles = c("file2", "file3"),
                       format = 3),
               "Three output file names are required when using formats 1, 2, and 3")
  expect_error(vcftobdlegacy(vcffiles = "file1",
                       bdfiles = c("file2", "file3", ""),
                       format = 3),
               "Output file names cannot be blank")

  expect_error(vcftobdlegacy(vcffiles = "file1",
                       bdfiles = "file2",
                       snpidformat = "A"),
               "snpidformat must be an integer value")
  expect_error(vcftobdlegacy(vcffiles = "file1",
                       bdfiles = "file2",
                       snpidformat = 1:2),
               "snpidformat must be an integer vector of length 1")
  expect_error(vcftobdlegacy(vcffiles = "file1",
                       bdfiles = "file2",
                       snpidformat = 1.2),
               "snpidformat must be an integer")
  expect_error(vcftobdlegacy(vcffiles = "file1",
                       bdfiles = "file2",
                       snpidformat = 4),
               "snpidformat must be and integer from -1 to 3")

  expect_error(vcftobdlegacy(vcffiles = "file1",
                       bdfiles = "file2",
                       bdoptions = 4),
               "bdoptions must be a character array")
  expect_error(vcftobdlegacy(vcffiles = "file1",
                       bdfiles = c("file2", "file3", "file4"),
                       format = 3,
                       bdoptions = "aaf"),
               "bdoptions can only be used with format 4")
  expect_error(vcftobdlegacy(vcffiles = "file1",
                       bdfiles = "file2",
                       bdoptions = "abc"),
               "Only valid bdoptions are aaf, maf, and rsq")

  vcf1afile = system.file("extdata", "set1a.vcf", package = "BinaryDosage")
  vcf1ainfo <- system.file("extdata", "set1a.info", package = "BinaryDosage")
  bdfiles <- tempfile()
  expect_error(vcftobdlegacy(vcffiles = c(vcf1afile, vcf1ainfo),
                       bdfiles = bdfiles,
                       bdoptions = c("aaf", "maf", "rsq")),
               NA)
  expect_error(bdinfo <- getbdinfo(bdfiles = bdfiles), NA)

  vcf2afile = system.file("extdata", "set2a.vcf", package = "BinaryDosage")
  bdfiles2 <- tempfile()
  expect_error(vcftobdlegacy(vcffiles = c(vcf2afile),
                       bdfiles = bdfiles2,
                       snpidformat = -1),
               NA)
  expect_error(bdinfo <- getbdinfo(bdfiles = bdfiles2), NA)

})

test_that("bd1tobd5", {
  vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")

  # Error checks
  expect_error(bd1tobd5(bdose_file = "x.bdose"),
               "No binary dosage files specified")
  expect_error(bd1tobd5(bdfiles = "x"),
               "No output .bdose file specified")

  # Subformat 1 (dosage only) ------------------------------------------------
  bdfile1  <- tempfile()
  famfile1 <- tempfile()
  mapfile1 <- tempfile()
  expect_error(vcftobdlegacy(vcffiles = vcf1afile,
                             bdfiles  = c(bdfile1, famfile1, mapfile1),
                             format   = 1L,
                             subformat = 1L),
               NA)
  old_bdinfo1 <- getbdinfo(c(bdfile1, famfile1, mapfile1))

  bdose1 <- tempfile(fileext = ".bdose")
  expect_error(bd1tobd5(bdfiles    = c(bdfile1, famfile1, mapfile1),
                        bdose_file = bdose1),
               NA)
  bd5info1 <- getbdinfo(bdose1)

  expect_equal(nrow(bd5info1$snps), nrow(old_bdinfo1$snps))
  expect_equal(bd5info1$samples,    old_bdinfo1$samples)

  for (i in seq_len(nrow(bd5info1$snps))) {
    old_snp <- getsnp(old_bdinfo1, i, dosageonly = TRUE)
    new_snp <- getsnp(bd5info1,    i, dosageonly = FALSE)
    expect_equal(new_snp$dosage, old_snp$dosage, tolerance = 5e-5,
                 label = paste0("subformat1 dosage SNP ", i))
    expect_true(all(is.na(new_snp$p0)),
                label = paste0("subformat1 p0 NA SNP ", i))
    expect_true(all(is.na(new_snp$p1)),
                label = paste0("subformat1 p1 NA SNP ", i))
    expect_true(all(is.na(new_snp$p2)),
                label = paste0("subformat1 p2 NA SNP ", i))
  }

  # Subformat 2 (dosage + genotype probabilities) ----------------------------
  bdfile2  <- tempfile()
  famfile2 <- tempfile()
  mapfile2 <- tempfile()
  expect_error(vcftobdlegacy(vcffiles  = vcf1afile,
                             bdfiles   = c(bdfile2, famfile2, mapfile2),
                             format    = 1L,
                             subformat = 2L),
               NA)
  old_bdinfo2 <- getbdinfo(c(bdfile2, famfile2, mapfile2))

  bdose2 <- tempfile(fileext = ".bdose")
  expect_error(bd1tobd5(bdfiles    = c(bdfile2, famfile2, mapfile2),
                        bdose_file = bdose2),
               NA)
  bd5info2 <- getbdinfo(bdose2)

  expect_equal(nrow(bd5info2$snps), nrow(old_bdinfo2$snps))
  expect_equal(bd5info2$samples,    old_bdinfo2$samples)

  for (i in seq_len(nrow(bd5info2$snps))) {
    old_snp <- getsnp(old_bdinfo2, i, dosageonly = FALSE)
    new_snp <- getsnp(bd5info2,    i, dosageonly = FALSE)
    expect_equal(new_snp$dosage, old_snp$dosage, tolerance = 5e-5,
                 label = paste0("subformat2 dosage SNP ", i))
    expect_equal(new_snp$p0, old_snp$p0, tolerance = 5e-5,
                 label = paste0("subformat2 p0 SNP ", i))
    expect_equal(new_snp$p1, old_snp$p1, tolerance = 5e-5,
                 label = paste0("subformat2 p1 SNP ", i))
    expect_equal(new_snp$p2, old_snp$p2, tolerance = 5e-5,
                 label = paste0("subformat2 p2 SNP ", i))
  }
})

test_that("bd2tobd5", {
  vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")

  # Error checks
  expect_error(bd2tobd5(bdose_file = "x.bdose"),
               "No binary dosage files specified")
  expect_error(bd2tobd5(bdfiles = "x"),
               "No output .bdose file specified")

  # Subformat 1 (dosage only) ------------------------------------------------
  bdfile1  <- tempfile()
  famfile1 <- tempfile()
  mapfile1 <- tempfile()
  suppressWarnings(
    vcftobdlegacy(vcffiles  = vcf1afile,
                  bdfiles   = c(bdfile1, famfile1, mapfile1),
                  format    = 2L,
                  subformat = 1L)
  )
  old_bdinfo1 <- getbdinfo(c(bdfile1, famfile1, mapfile1))

  bdose1 <- tempfile(fileext = ".bdose")
  expect_error(bd2tobd5(bdfiles    = c(bdfile1, famfile1, mapfile1),
                        bdose_file = bdose1),
               NA)
  bd5info1 <- getbdinfo(bdose1)

  expect_equal(nrow(bd5info1$snps), nrow(old_bdinfo1$snps))
  expect_equal(bd5info1$samples,    old_bdinfo1$samples)

  for (i in seq_len(nrow(bd5info1$snps))) {
    old_snp <- getsnp(old_bdinfo1, i, dosageonly = TRUE)
    new_snp <- getsnp(bd5info1,    i, dosageonly = FALSE)
    expect_equal(new_snp$dosage, old_snp$dosage, tolerance = 5e-5,
                 label = paste0("subformat1 dosage SNP ", i))
    expect_true(all(is.na(new_snp$p0)),
                label = paste0("subformat1 p0 NA SNP ", i))
    expect_true(all(is.na(new_snp$p1)),
                label = paste0("subformat1 p1 NA SNP ", i))
    expect_true(all(is.na(new_snp$p2)),
                label = paste0("subformat1 p2 NA SNP ", i))
  }

  # Subformat 2 (dosage + genotype probabilities) ----------------------------
  bdfile2  <- tempfile()
  famfile2 <- tempfile()
  mapfile2 <- tempfile()
  suppressWarnings(
    vcftobdlegacy(vcffiles  = vcf1afile,
                  bdfiles   = c(bdfile2, famfile2, mapfile2),
                  format    = 2L,
                  subformat = 2L)
  )
  old_bdinfo2 <- getbdinfo(c(bdfile2, famfile2, mapfile2))

  bdose2 <- tempfile(fileext = ".bdose")
  expect_error(bd2tobd5(bdfiles    = c(bdfile2, famfile2, mapfile2),
                        bdose_file = bdose2),
               NA)
  bd5info2 <- getbdinfo(bdose2)

  expect_equal(nrow(bd5info2$snps), nrow(old_bdinfo2$snps))
  expect_equal(bd5info2$samples,    old_bdinfo2$samples)

  for (i in seq_len(nrow(bd5info2$snps))) {
    old_snp <- getsnp(old_bdinfo2, i, dosageonly = FALSE)
    new_snp <- getsnp(bd5info2,    i, dosageonly = FALSE)
    expect_equal(new_snp$dosage, old_snp$dosage, tolerance = 5e-5,
                 label = paste0("subformat2 dosage SNP ", i))
    expect_equal(new_snp$p0, old_snp$p0, tolerance = 5e-5,
                 label = paste0("subformat2 p0 SNP ", i))
    expect_equal(new_snp$p1, old_snp$p1, tolerance = 5e-5,
                 label = paste0("subformat2 p1 SNP ", i))
    expect_equal(new_snp$p2, old_snp$p2, tolerance = 5e-5,
                 label = paste0("subformat2 p2 SNP ", i))
  }
})

test_that("bd3tobd5", {
  vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")

  # Error checks
  expect_error(bd3tobd5(bdose_file = "x.bdose"),
               "No binary dosage files specified")
  expect_error(bd3tobd5(bdfiles = "x"),
               "No output .bdose file specified")

  # Subformat 1 (dosage only) ------------------------------------------------
  bdfile1  <- tempfile()
  famfile1 <- tempfile()
  mapfile1 <- tempfile()
  suppressWarnings(
    vcftobdlegacy(vcffiles  = vcf1afile,
                  bdfiles   = c(bdfile1, famfile1, mapfile1),
                  format    = 3L,
                  subformat = 1L)
  )
  old_bdinfo1 <- getbdinfo(c(bdfile1, famfile1, mapfile1))

  bdose1 <- tempfile(fileext = ".bdose")
  expect_error(bd3tobd5(bdfiles    = c(bdfile1, famfile1, mapfile1),
                        bdose_file = bdose1),
               NA)
  bd5info1 <- getbdinfo(bdose1)

  expect_equal(nrow(bd5info1$snps), nrow(old_bdinfo1$snps))
  expect_equal(bd5info1$samples,    old_bdinfo1$samples)

  for (i in seq_len(nrow(bd5info1$snps))) {
    old_snp <- getsnp(old_bdinfo1, i, dosageonly = TRUE)
    new_snp <- getsnp(bd5info1,    i, dosageonly = FALSE)
    expect_equal(new_snp$dosage, old_snp$dosage, tolerance = 5e-5,
                 label = paste0("subformat1 dosage SNP ", i))
    expect_true(all(is.na(new_snp$p0)),
                label = paste0("subformat1 p0 NA SNP ", i))
    expect_true(all(is.na(new_snp$p1)),
                label = paste0("subformat1 p1 NA SNP ", i))
    expect_true(all(is.na(new_snp$p2)),
                label = paste0("subformat1 p2 NA SNP ", i))
  }

  # Subformat 2 (dosage + genotype probabilities) ----------------------------
  bdfile2  <- tempfile()
  famfile2 <- tempfile()
  mapfile2 <- tempfile()
  suppressWarnings(
    vcftobdlegacy(vcffiles  = vcf1afile,
                  bdfiles   = c(bdfile2, famfile2, mapfile2),
                  format    = 3L,
                  subformat = 2L)
  )
  old_bdinfo2 <- getbdinfo(c(bdfile2, famfile2, mapfile2))

  bdose2 <- tempfile(fileext = ".bdose")
  expect_error(bd3tobd5(bdfiles    = c(bdfile2, famfile2, mapfile2),
                        bdose_file = bdose2),
               NA)
  bd5info2 <- getbdinfo(bdose2)

  expect_equal(nrow(bd5info2$snps), nrow(old_bdinfo2$snps))
  expect_equal(bd5info2$samples,    old_bdinfo2$samples)

  for (i in seq_len(nrow(bd5info2$snps))) {
    old_snp <- getsnp(old_bdinfo2, i, dosageonly = FALSE)
    new_snp <- getsnp(bd5info2,    i, dosageonly = FALSE)
    expect_equal(new_snp$dosage, old_snp$dosage, tolerance = 5e-5,
                 label = paste0("subformat2 dosage SNP ", i))
    expect_equal(new_snp$p0, old_snp$p0, tolerance = 5e-5,
                 label = paste0("subformat2 p0 SNP ", i))
    expect_equal(new_snp$p1, old_snp$p1, tolerance = 5e-5,
                 label = paste0("subformat2 p1 SNP ", i))
    expect_equal(new_snp$p2, old_snp$p2, tolerance = 5e-5,
                 label = paste0("subformat2 p2 SNP ", i))
  }
})

test_that("bd4tobd5", {
  vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")

  # Error checks
  expect_error(bd4tobd5(bdose_file = "x.bdose"),
               "No binary dosage file specified")
  expect_error(bd4tobd5(bdfile = "x"),
               "No output .bdose file specified")

  # Subformat 1 (dosage only) ------------------------------------------------
  bdfile1 <- tempfile()
  suppressWarnings(
    vcftobdlegacy(vcffiles  = vcf1afile,
                  bdfiles   = bdfile1,
                  format    = 4L,
                  subformat = 1L)
  )
  old_bdinfo1 <- getbdinfo(bdfile1)

  bdose1 <- tempfile(fileext = ".bdose")
  expect_error(bd4tobd5(bdfile     = bdfile1,
                        bdose_file = bdose1),
               NA)
  bd5info1 <- getbdinfo(bdose1)

  expect_equal(nrow(bd5info1$snps), nrow(old_bdinfo1$snps))
  expect_equal(bd5info1$samples,    old_bdinfo1$samples)

  for (i in seq_len(nrow(bd5info1$snps))) {
    old_snp <- getsnp(old_bdinfo1, i, dosageonly = TRUE)
    new_snp <- getsnp(bd5info1,    i, dosageonly = FALSE)
    expect_equal(new_snp$dosage, old_snp$dosage, tolerance = 5e-5,
                 label = paste0("subformat1 dosage SNP ", i))
    expect_true(all(is.na(new_snp$p0)),
                label = paste0("subformat1 p0 NA SNP ", i))
    expect_true(all(is.na(new_snp$p1)),
                label = paste0("subformat1 p1 NA SNP ", i))
    expect_true(all(is.na(new_snp$p2)),
                label = paste0("subformat1 p2 NA SNP ", i))
  }

  # Subformat 2 (dosage + genotype probabilities) ----------------------------
  bdfile2 <- tempfile()
  suppressWarnings(
    vcftobdlegacy(vcffiles  = vcf1afile,
                  bdfiles   = bdfile2,
                  format    = 4L,
                  subformat = 2L)
  )
  old_bdinfo2 <- getbdinfo(bdfile2)

  bdose2 <- tempfile(fileext = ".bdose")
  expect_error(bd4tobd5(bdfile     = bdfile2,
                        bdose_file = bdose2),
               NA)
  bd5info2 <- getbdinfo(bdose2)

  expect_equal(nrow(bd5info2$snps), nrow(old_bdinfo2$snps))
  expect_equal(bd5info2$samples,    old_bdinfo2$samples)

  for (i in seq_len(nrow(bd5info2$snps))) {
    old_snp <- getsnp(old_bdinfo2, i, dosageonly = FALSE)
    new_snp <- getsnp(bd5info2,    i, dosageonly = FALSE)
    expect_equal(new_snp$dosage, old_snp$dosage, tolerance = 5e-5,
                 label = paste0("subformat2 dosage SNP ", i))
    expect_equal(new_snp$p0, old_snp$p0, tolerance = 5e-5,
                 label = paste0("subformat2 p0 SNP ", i))
    expect_equal(new_snp$p1, old_snp$p1, tolerance = 5e-5,
                 label = paste0("subformat2 p1 SNP ", i))
    expect_equal(new_snp$p2, old_snp$p2, tolerance = 5e-5,
                 label = paste0("subformat2 p2 SNP ", i))
  }
})

test_that("updatebd", {
  vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")

  # Error checks
  expect_error(updatebd(bdose_file = "x.bdose"),
               "No binary dosage files specified")
  expect_error(updatebd(bdfiles = "x"),
               "No output .bdose file specified")

  # Helper: convert old bdinfo → format 5 via updatebd, return bd5info
  convert_and_check <- function(bdfiles, old_bdinfo, has_gp) {
    bdose <- tempfile(fileext = ".bdose")
    expect_error(updatebd(bdfiles = bdfiles, bdose_file = bdose), NA)
    bd5info <- getbdinfo(bdose)
    expect_equal(nrow(bd5info$snps), nrow(old_bdinfo$snps))
    expect_equal(bd5info$samples,    old_bdinfo$samples)
    for (i in seq_len(nrow(bd5info$snps))) {
      old_snp <- getsnp(old_bdinfo, i, dosageonly = !has_gp)
      new_snp <- getsnp(bd5info,    i, dosageonly = FALSE)
      expect_equal(new_snp$dosage, old_snp$dosage, tolerance = 5e-5,
                   label = paste0("dosage SNP ", i))
      if (has_gp) {
        expect_equal(new_snp$p0, old_snp$p0, tolerance = 5e-5,
                     label = paste0("p0 SNP ", i))
        expect_equal(new_snp$p1, old_snp$p1, tolerance = 5e-5,
                     label = paste0("p1 SNP ", i))
        expect_equal(new_snp$p2, old_snp$p2, tolerance = 5e-5,
                     label = paste0("p2 SNP ", i))
      } else {
        expect_true(all(is.na(new_snp$p0)), label = paste0("p0 NA SNP ", i))
        expect_true(all(is.na(new_snp$p1)), label = paste0("p1 NA SNP ", i))
        expect_true(all(is.na(new_snp$p2)), label = paste0("p2 NA SNP ", i))
      }
    }
  }

  # Format 1 (subformat 1 - dosage only) ------------------------------------
  bdfile1 <- tempfile(); fam1 <- tempfile(); map1 <- tempfile()
  suppressWarnings(
    vcftobdlegacy(vcffiles = vcf1afile, bdfiles = c(bdfile1, fam1, map1),
                  format = 1L, subformat = 1L)
  )
  convert_and_check(c(bdfile1, fam1, map1), getbdinfo(c(bdfile1, fam1, map1)), FALSE)

  # Format 2 (subformat 2 - dosage + GP) ------------------------------------
  bdfile2 <- tempfile(); fam2 <- tempfile(); map2 <- tempfile()
  suppressWarnings(
    vcftobdlegacy(vcffiles = vcf1afile, bdfiles = c(bdfile2, fam2, map2),
                  format = 2L, subformat = 2L)
  )
  convert_and_check(c(bdfile2, fam2, map2), getbdinfo(c(bdfile2, fam2, map2)), TRUE)

  # Format 3 (subformat 2 - dosage + GP) ------------------------------------
  bdfile3 <- tempfile(); fam3 <- tempfile(); map3 <- tempfile()
  suppressWarnings(
    vcftobdlegacy(vcffiles = vcf1afile, bdfiles = c(bdfile3, fam3, map3),
                  format = 3L, subformat = 2L)
  )
  convert_and_check(c(bdfile3, fam3, map3), getbdinfo(c(bdfile3, fam3, map3)), TRUE)

  # Format 4 (subformat 2 - dosage + GP) ------------------------------------
  bdfile4 <- tempfile()
  suppressWarnings(
    vcftobdlegacy(vcffiles = vcf1afile, bdfiles = bdfile4,
                  format = 4L, subformat = 2L)
  )
  convert_and_check(bdfile4, getbdinfo(bdfile4), TRUE)
})

test_that("gentobd", {
  expect_error(gentobd(bdfiles = "file1"),
               "No gen file specified")
  expect_error(gentobd(genfiles = "file1"),
               "No output files specified")

  gen3afile <- system.file("extdata", "set3a.imp", package = "BinaryDosage")
  gen3asample <- system.file("extdata", "set3a.sample", package = "BinaryDosage")
  bdfile3a <- tempfile()
  bdfam3a <- tempfile()
  bdmap3a <- tempfile()
  expect_error(getgeninfo(genfiles = c(gen3afile, gen3asample),
                          snpcolumns = c(0L, 2L:5L),
                          chromosome = "",
                          index = TRUE),
               NA)
  expect_error(gentobd(genfiles = c(gen3afile, gen3asample),
                       snpcolumns = c(0L, 2L:5L),
                       bdfiles = c(bdfile3a, bdfam3a, bdmap3a),
                       chromosome = "",
                       format = 3),
               NA)
  expect_error(getbdinfo(bdfiles = c(bdfile3a, bdfam3a, bdmap3a)), NA)

  gen2afile <- system.file("extdata", "set2a.imp", package = "BinaryDosage")
  gen2asample <- system.file("extdata", "set2a.sample", package = "BinaryDosage")
  bdfile2a <- tempfile()
  expect_error(gentobd(genfiles = c(gen2afile, gen2asample),
                       snpcolumns = c(1L, 3L, 2L, 4L, 5L),
                       impformat = 1L,
                       bdfiles = bdfile2a,
                       snpidformat = -1L,
                       bdoptions = c("aaf", "maf", "rsq")),
               NA)
  expect_error(getbdinfo(bdfiles = bdfile2a), NA)

  gen4afile <- system.file("extdata", "set4a.imp.gz", package = "BinaryDosage")
  gen4asample <- system.file("extdata", "set4a.sample", package = "BinaryDosage")
  bdfile4a <- tempfile()
  expect_error(gentobd(genfiles = c(gen4afile, gen4asample),
                       snpcolumns = c(1L:2L, 4L:6L),
                       startcolumn = 7L,
                       impformat = 2L,
                       gz = TRUE,
                       bdfile = bdfile4a),
               NA)
  expect_error(getbdinfo(bdfiles = bdfile4a), NA)
})

test_that("writecpp", {
  expect_error(WriteBinaryDosageBaseHeader(filename = "",
                                           format = 1L,
                                           subformat = 2L),
               "Unable to create output file")
  expect_error(WriteBinaryDosageHeader4A(filename = "",
                                         headerEntries = 1L,
                                         numSubjects = 1L,
                                         numSNPs = 1L,
                                         groups = 1L,
                                         fid = "",
                                         sid = "",
                                         snpid = "",
                                         chromosome = "",
                                         location = 1L,
                                         reference = "",
                                         alternate = "",
                                         aaf = 1,
                                         maf = 1,
                                         avgCall = 1,
                                         rsq = 1,
                                         offsets = 1L,
                                         numIndices = 1L),
               "Unable to open file for read/write")
  expect_error(WriteBinaryDosageIndicesC(filename = "",
                                         headersize = 0L,
                                         datasize = 1L),
               "Unable to open file for read/write")
  expect_error(updatesnpinfo(filename = "",
                             offset = 0L,
                             value = 1),
               "Unable to open file for read/write")
  expect_error(WriteBinaryDosageHeader3A(filename = "",
                                         numSubjects = 1L),
               "Unable to open file for appending")
  expect_error(WriteBinaryDosageHeader3B(filename = "",
                                         md5samples = "",
                                         md5SNPs = "",
                                         numIndices = 1L),
               "Unable to open file for appending")
  expect_error(WriteBinaryDosageDataC(filename = "",
                                      dosage = 1,
                                      us = 1L,
                                      base = 1L),
               "Unable to open file for appending")
  expect_error(WriteBinaryP1P2Data(filename = "",
                                   p1 = 1,
                                   p2 = 1,
                                   us = 1L,
                                   base = 1L),
               "Unable to open file for appending")
  expect_error(WriteBinaryCompressed(filename = "",
                                     dosage = 1,
                                     p0 = 0,
                                     p1 = 1,
                                     p2 = 0,
                                     snpnumber = -1L,
                                     datasize = 1L,
                                     us = 1L),
               "Unable to open file for appending")
  bdfile <- tempfile()
  dosage <- as.numeric(c(NA, 1))
  p0 <- as.numeric(c(NA, NA))
  p1 <- as.numeric(c(NA, NA))
  p2 <- as.numeric(c(NA, NA))
  snpnumber = 0L
  datasize = 0L
  us <- integer(4)
  expect_error(WriteBinaryCompressed(filename = bdfile,
                                     dosage = dosage,
                                     p0 = p0,
                                     p1 = p1,
                                     p2 = p2,
                                     snpnumber = snpnumber,
                                     datasize = datasize,
                                     us = us),
               NA)
  rdosage <- numeric(2)
  rp0 <- numeric(2)
  rp1 <- numeric(2)
  rp2 <- numeric(2)
  rus <- integer(4)
  expect_error(ReadBinaryDosageDataCompressed(filename = bdfile,
                                              index = 0L,
                                              datasize = datasize,
                                              numsub = 2L,
                                              dosage = rdosage,
                                              p0 = rp0,
                                              p1 = rp1,
                                              p2 = rp2,
                                              us = rus),
               NA)
  expect_true(is.na(rdosage[1]))
  expect_equal(rdosage[2], 1)
  expect_true(all(is.na(rp0)))
  expect_true(all(is.na(rp1)))
  expect_true(all(is.na(rp2)))

  bdfile <- tempfile()
  p1 <- as.numeric(c(NA, 0.0025))
  p2 <- as.numeric(c(NA, 0.999))
  us <- integer(4)
  base <- 3L
  expect_error(WriteBinaryP1P2Data(filename = bdfile,
                                   p1 = p1,
                                   p2 = p2,
                                   us = us,
                                   base = base),
               NA)

  rdosage <- numeric(2)
  rp0 <- numeric(2)
  rp1 <- numeric(2)
  rp2 <- numeric(2)
  rus <- integer(4)
  expect_error(ReadBinaryDosageDataP1P2(filename = bdfile,
                                        headersize = 0L,
                                        numsub = 2L,
                                        snp = 1L,
                                        dosage = rdosage,
                                        p0 = rp0,
                                        p1 = rp1,
                                        p2 = rp2,
                                        us = rus,
                                        base = base),
               NA)
  expect_true(is.na(rdosage[1]))
  expect_equal(rdosage[2], 2)
  expect_true(is.na(rp0[1]))
  expect_true(is.na(rp1[1]))
  expect_true(is.na(rp2[1]))
  expect_equal(rp0[2], 0)
  expect_equal(rp1[2], 0.0025)
  expect_equal(rp2[2], 0.999)
})

#***************************************************************************#
#                                                                           #
#                      Subsetting Binary Dosage files                       #
#                                                                           #
#***************************************************************************#

#' Subset a binary dosage file
#'
#' Creates a new Format 5 binary dosage file containing a subset of the SNPs
#' and/or subjects from an existing binary dosage file. The input file may be
#' in any format (1-5). At least one filtering criterion must be supplied, and
#' all supplied criteria must be met for a SNP or subject to be retained.
#'
#' @param bdfiles Vector of file names for the input binary dosage file.
#'   Format 4 files require one file name. Formats 1, 2, and 3 require three
#'   file names: the binary dosage file, the family file, and the map file.
#'   Format 5 files require two file names: the .bdose file and the .bdinfo
#'   file.
#' @param bdose_file Path for the output .bdose file. The companion .bdi
#'   metadata file is written to \code{paste0(bdose_file, ".bdi")}.
#' @param minmaf Minimum minor allele frequency. SNPs whose MAF (computed over
#'   the retained subjects) is below this value are excluded. Must be a single
#'   numeric value between 0 and 0.5.
#' @param locations Integer or numeric vector of SNP base-pair locations to
#'   retain. Cannot be used together with \code{startloc} and \code{endloc}.
#' @param startloc Start of the location range to retain (inclusive). Must be
#'   used together with \code{endloc}. Cannot be used together with
#'   \code{locations}.
#' @param endloc End of the location range to retain (inclusive). Must be used
#'   together with \code{startloc}. Cannot be used together with
#'   \code{locations}.
#' @param subjectids Character vector of subject IDs to retain.
#'
#' @return NULL (invisibly)
#' @export
#'
#' @examples
#' bdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
#' bdinfo     <- getbdinfo(bdfile)
#' bdose_file <- tempfile(fileext = ".bdose")
#' subsetbd(bdfiles    = bdfile,
#'          bdose_file = bdose_file,
#'          subjectids = bdinfo$samples$sid[1:30])
subsetbd <- function(bdfiles, bdose_file,
                     minmaf = NULL, locations = NULL,
                     startloc = NULL, endloc = NULL,
                     subjectids = NULL) {

  # --- Required argument checks ---
  if (missing(bdfiles))
    stop("No input binary dosage files specified")
  if (missing(bdose_file))
    stop("No output .bdose file specified")

  # --- At least one filter required ---
  if (is.null(minmaf) && is.null(locations) &&
      is.null(startloc) && is.null(endloc) && is.null(subjectids))
    stop("At least one of minmaf, locations, startloc/endloc, or subjectids must be specified")

  # --- Location filter mutual exclusivity ---
  if (!is.null(locations) && (!is.null(startloc) || !is.null(endloc)))
    stop("locations and startloc/endloc cannot both be specified")

  # --- startloc/endloc must be paired ---
  if (!is.null(startloc) && is.null(endloc))
    stop("endloc must be specified when startloc is provided")
  if (is.null(startloc) && !is.null(endloc))
    stop("startloc must be specified when endloc is provided")
  if (!is.null(startloc) && startloc > endloc)
    stop("startloc must be less than or equal to endloc")

  # --- minmaf validation ---
  if (!is.null(minmaf)) {
    if (!is.numeric(minmaf) || length(minmaf) != 1L)
      stop("minmaf must be a single numeric value")
    if (minmaf < 0 || minmaf > 0.5)
      stop("minmaf must be between 0 and 0.5")
  }

  # --- Load input bdinfo (formats 1-4 or format 5) ---
  bdinfo <- tryCatch(
    getbdinfo(bdfiles),
    error = function(e) {
      tryCatch(
        getbd5info(bdose_file = bdfiles[1L]),
        error = function(e2) stop(conditionMessage(e))
      )
    }
  )

  n_samp_all <- nrow(bdinfo$samples)
  n_snps_all <- nrow(bdinfo$snps)

  # --- Subject filter ---
  if (!is.null(subjectids)) {
    samp_idx <- which(bdinfo$samples$sid %in% subjectids)
    if (length(samp_idx) == 0L)
      stop("No subjects match the specified subjectids")
    unmatched <- setdiff(subjectids, bdinfo$samples$sid)
    if (length(unmatched) > 0L)
      warning(length(unmatched), " subject ID(s) not found in the file and were ignored",
              call. = FALSE)
  } else {
    samp_idx <- seq_len(n_samp_all)
  }
  n_samp_out <- length(samp_idx)

  # --- SNP location pre-filter (no data read required) ---
  snp_locs <- bdinfo$snps$location
  if (!is.null(locations)) {
    loc_keep <- snp_locs %in% locations
  } else if (!is.null(startloc)) {
    loc_keep <- snp_locs >= startloc & snp_locs <= endloc
  } else {
    loc_keep <- rep(TRUE, n_snps_all)
  }

  # --- Write output .bdose, applying SNP and MAF filters in one pass ---
  bdose_con <- file(bdose_file, open = "wb")
  on.exit(close(bdose_con), add = TRUE)
  writeBin(.bdose5_magic, bdose_con)

  snp_keep   <- logical(n_snps_all)
  indices    <- numeric(n_snps_all)
  current_pos <- 4.0
  n_kept     <- 0L

  for (i in seq_len(n_snps_all)) {
    if (!loc_keep[i]) next

    snp <- getsnp(bdinfo, i, dosageonly = FALSE)

    # MAF filter computed over retained subjects
    if (!is.null(minmaf)) {
      aaf <- mean(snp$dosage[samp_idx], na.rm = TRUE) / 2.0
      maf <- min(aaf, 1.0 - aaf)
      if (is.nan(maf) || maf < minmaf) next
    }

    n_kept       <- n_kept + 1L
    snp_keep[i]  <- TRUE

    ds <- snp$dosage[samp_idx]
    if (all(is.na(snp$p0))) {
      gp <- rep(NA_real_, 3L * n_samp_out)
    } else {
      gp <- as.numeric(rbind(snp$p0[samp_idx], snp$p1[samp_idx], snp$p2[samp_idx]))
    }

    block          <- compress_snp_block(ds, gp)
    indices[n_kept] <- current_pos
    writeBin(block, bdose_con)
    current_pos <- current_pos + length(block)
  }

  if (n_kept == 0L)
    stop("No SNPs pass the specified filters")

  # --- Build and save output bdinfo ---
  out_snps <- bdinfo$snps[snp_keep, , drop = FALSE]
  rownames(out_snps) <- NULL

  new_bdinfo <- list(
    filename       = normalizePath(bdose_file),
    usesfid        = bdinfo$usesfid,
    samples        = bdinfo$samples[samp_idx, , drop = FALSE],
    onechr         = length(unique(out_snps$chromosome)) == 1L,
    snpidformat    = bdinfo$snpidformat,
    snps           = out_snps,
    snpinfo        = lapply(bdinfo$snpinfo, function(x) x[snp_keep]),
    additionalinfo = structure(
      list(format     = 5L,
           subformat  = 1L,
           headersize = 4L,
           numgroups  = 1L,
           groups     = n_samp_out),
      class = "bdose-info"
    ),
    datasize       = integer(0),
    indices        = indices[seq_len(n_kept)]
  )
  class(new_bdinfo) <- "genetic-info"
  saveRDS(new_bdinfo, paste0(bdose_file, ".bdi"))

  invisible(NULL)
}

#' Read SNP data from a binary dosage file
#'
#' Routine to read the dosage and genetic probabilities about
#' a SNP from a binary dosage file
#'
#' @param bdinfo Information about a binary dosage file return
#' from getbdinfo
#' @param snp The SNP to read the information about. This may
#' be the SNP ID or the index of the SNP in the snps dataset in
#' the bdinfo list
#' @param dosageonly Indicator to return the dosages only or the
#' dosages allowing with the genetic probabilities.
#' Default value is TRUE
#'
#' @return
#' A list with either the dosages or the dosages and the genetic
#' probabilities.
#' @export
#'
#' @examples
#' # Get the information about the file
#' vcf1abdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
#' bdinfo <- getbdinfo(bdfiles = vcf1abdfile)
#'
#' # Read the first SNP
#' getsnp(bdinfo, 1, FALSE)
getsnp <- function(bdinfo, snp, dosageonly = TRUE) {
  if (missing(bdinfo) == TRUE)
    stop("bdinfo missing")
  if (is.na(match("genetic-info", class(bdinfo))) == TRUE)
      stop("bdinfo is not of class genetic-info")
  if (is.na(match("bdose-info", class(bdinfo$additionalinfo))) == TRUE)
    stop("bdinfo does not contain information about a binary dosage file")

  if (missing(snp) == TRUE)
    stop("No SNP specified")
  if (length(snp) != 1)
    stop("snp must be of length one")
  if (is.character(snp) == TRUE) {
    x <- match(snp, bdinfo$snps$snpid)
    if (is.na(x))
      stop("Cannot find SNP in bdinfo")
    snp <- x
  } else if (is.numeric(snp) == TRUE) {
    if (floor(snp) != snp)
      stop("snp must be a character or integer value")
    snp <- as.integer(floor(snp))
  }
  if (is.integer(snp) == FALSE)
    stop("snp must be a character or integer value")
  if (snp < 1 | snp > nrow(bdinfo$snps))
    stop("snp value out or range")

  dosage <- numeric(nrow(bdinfo$samples))
  p0 <- numeric(nrow(bdinfo$samples))
  p1 <- numeric(nrow(bdinfo$samples))
  p2 <- numeric(nrow(bdinfo$samples))
  us <- numeric(2*nrow(bdinfo$samples))
  dosage[1:nrow(bdinfo$samples)] <- NA
  p0[1:nrow(bdinfo$samples)] <- NA
  p1[1:nrow(bdinfo$samples)] <- NA
  p2[1:nrow(bdinfo$samples)] <- NA
  result <- ReadBinaryDosageData(bdinfo, snp, dosage, p0, p1, p2, us)
  if (!is.null(result)) {
    dosage <- result$dosage
    p0     <- result$p0
    p1     <- result$p1
    p2     <- result$p2
  }
  if (dosageonly == TRUE)
    return(list(dosage = dosage))
  return(list(dosage = dosage,
              p0 = p0,
              p1 = p1,
              p2 = p2))
}

---
title: "Using GEN Files"
output:
  rmarkdown::html_vignette:
    toc: true
vignette: >
  %\VignetteIndexEntry{Using GEN Files}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

**Note:** GEN files are rarely used today. They have largely been replaced by
BGEN files, a binary format that is more compact and faster to read. Support
for BGEN files in the BinaryDosage package is currently in progress.

This vignette documents the functions in the BinaryDosage package that convert GEN files to binary dosage files.

**Note:** The examples below use functions to access information in binary dosage files. Information about these functions can be found in the vignette [Using Binary Dosage Files](usingbdfiles.html). Data returned by the function <span style="font-family:Courier">getbdinfo</span> contains information about a binary dosage file. Information on the data return by <span style="font-family:Courier">getbdinfo</span> can be found in the vignette [Genetic File Information](geneticfileinfo.html).

```{r setup, echo=FALSE}
library(BinaryDosage)
```
# Introduction

GEN files are a useful way to store genetic data. They are text files and can be easily parsed. The output files returned from the imputation software [Impute2](http://mathgen.stats.ox.ac.uk/impute/impute_v2.html) are returned in this format.

Uncompressed GEN files can be very large, 100s of GB. Because of this they are quite often compressed. This makes the files much smaller but greatly increases the read time. The BinaryDosage package supports reading of gzip compressed GEN files.

There appears to have been changes to the GEN file over the years and it also appears people have created GEN-like file formats. The <span style="font-family:Courier">BinaryDosage</span> package can support many GEN-like file formats.

The BinaryDosage package has a routine to convert GEN files into a binary format that maintains the dosage, genetic and probabilities. This results in a file about 10-15% the size of the uncompressed VCF file with much faster, 200-300x, read times. In comparison, Using gzip to compress the GEN file reduces the size of the file to about 5% its original size but makes run times slower.

Routines were written to help debug the conversion routine. It was discovered these routines were quite useful for accessing data in the GEN file and are now included in the package. This document contains instructions on how to use these routines.

# GEN file formats

The GEN file may have a header. If it does have a header the format the first N entries must be the column names for the SNP information variables. The following values identify the subjects and can have either of the following formats ordered by subject

- The family ID followed by the subject ID
- The subject ID only

If the GEN file does not have a header, the subject information must be in a sample file that can be read with <span style="font-family:Courier">read.table</span>. If there is only one column the subject ID is set to this value and the family ID is set to "". Otherwise, the family ID value is set to the value of the first column and the subject ID value is set to the second column. If the first value of the subject ID and family ID are both "0", they are deleted. If family ID and subject ID are equal for all subjects, the family ID value is set to "".

**Note:** If a sample file is provided, the header is ignored.

The body GEN file must have the following format. The first N columns must contain information about the SNP. These columns must contain the following values

- SNP ID
- Location
- Alternate allele
- Reference allele

The chromosome number may also be in the first N columns.

**Note:** The first three columns of the GEN file used to be snp_id, rs_id, and position. In many cases these values got change to chromosome, snp_id, and position.


The remaining columns must contain the genotype probabilities sorted by subject. The genotype probabilities can be in any of the following formats.

- The dosage value only
- Probability subject has no alternate alleles, probability subject has one alternate allele.
- Probability subject has no alternate alleles, probability subject has one alternate allele, probability subject has two alternate allele.

**Note:** The number of genotype probabilities must agree with the number of subjects identified in the header or sample file.

# Example files

There are several sample files included with the BinaryDosage package. The file names will be found by using the <span style="font-family:Courier">system.file</span> command in the examples. This will be used many times in the examples.

The binary dosage files created will be temporary files. They will be created using the <span style="font-family:Courier">tempfile</span> command. This will also be used many times in the examples. All output files will use the default format of 4. For information on other formats see the vignette [Binary Dosage Formats](bdformats.html).

# Converting a GEN file to the Binary Dosage format

The <span style="font-family:Courier">gentobd</span> routine converts GEN files to the binary dosage format. Many different formats of GEN files by the BinaryDosage package. The following sections show how to convert GEN files in various formats to the binary dosage format.

The <span style="font-family:Courier">gentobd</span> takes the following parameters

- <span style="font-family:Courier">genfiles</span> - Name of GEN file and the optional sample file.
- <span style="font-family:Courier">snpcolumns</span> - Columns containing the values for chromosome, SNP ID, location, reference allele, and alternate allele.
- <span style="font-family:Courier">startcolumn</span> - Column where genotype probabilities start.
- <span style="font-family:Courier">impformat</span> - Number of genotype probabilities for each subject.
- <span style="font-family:Courier">chromosome</span> - Optional chromosome, provided if chromosome is not include in GEN file
- <span style="font-family:Courier">header</span> - Vector of one or two logical values indicating if GEN and sample files have headers respectively.
- <span style="font-family:Courier">gz</span> - Logical value indicating if GEN file is compressed.
- <span style="font-family:Courier">sep</span> - Separator used in GEN file.
- <span style="font-family:Courier">bdfiles</span> - Vector of character values give the names of the binary dosage file. If the binary dosage format is 3 or less there are three files names, binary dosage file, family file, and map file names. For format 4 there is only the binary dosage file name. 
- <span style="font-family:Courier">format</span> - Format of the binary dosage file.
- <span style="font-family:Courier">subformat</span> - Subformat of the binary dosage file.
- <span style="font-family:Courier">snpidformat</span> - Format to store the SNP ID in.
- <span style="font-family:Courier">bdoptions</span> - Options for calculating additional SNP data.

## Default options

The default values for the <span style="font-family:Courier">gentobd</span> require a sample file meaning there is no header in the GEN file and the first five columns be chromosome, SNP ID, location, reference allele, and alternate allele, respectively. The genotype data must have the three genotype values and the file must not be compressed.

The following code reads the gen file *set3b.chr.imp* using the *set3b.sample* sample file. These files are in the format mentioned above.

``` {r, eval = T, echo = T, message = F, warning = F, tidy = T}
gen3bchrfile <- system.file("extdata", "set3b.chr.imp", package = "BinaryDosage")
sample3bfile <- system.file("extdata", "set3b.sample", package = "BinaryDosage")
bdfile3b_chr <- tempfile()
gentobd(genfiles = c(gen3bchrfile, sample3bfile), bdfiles = bdfile3b_chr)

bdinfo3b_chr <- getbdinfo(bdfiles = bdfile3b_chr)

```

## snpcolumns

The <span style="font-family:Courier">snpcolumns</span> parameter lists the column numbers for the chromosome, SNP ID, location, reference allele, and alternate allele, respectively.

The following code reads in *set1b.imp*. This file has the SNP data in the following order, chromosome, location, SNP ID, reference allele, alternate allele. The file also has a header so there is no sample file.

``` {r, eval = T, echo = T, message = F, warning = F, tidy = T}
gen1bfile <- system.file("extdata", "set1b.imp", package = "BinaryDosage")
bdfile1b <- tempfile()
gentobd(genfiles = gen1bfile,
        bdfiles = bdfile1b,
        snpcolumns = c(1L, 3L, 2L, 4L, 5L),
        header = TRUE)

bdinfo1b <- getbdinfo(bdfiles = bdfile1b)

```

Quite often the chromosome is not part of the GEN file and first column has the value '\-\-'. In this case  the SNP ID is often in the format <span style="font-family:Courier">\<chromosome\>:\<additional SNP data\></span>. In this case, the chromosome column number (first value in snpcolumns) can be set to 0L and the <span style="font-family:Courier">gentobd</span> routine will extract the chromosome from the SNP ID value.

The following code reads in *set3b.imp*. This is in same format as *set3b.chr.imp* except that there is no column for the chromosome value. The value of 0L will be used for the chromosome column to indicate to get the chromosome value from the SNP ID value.

``` {r, eval = T, echo = T, message = F, warning = F, tidy = T}
gen3bfile <- system.file("extdata", "set3b.imp", package = "BinaryDosage")
sample3bfile <- system.file("extdata", "set3b.sample", package = "BinaryDosage")
bdfile3b <- tempfile()
gentobd(genfiles = c(gen3bfile, sample3bfile),
        bdfiles = bdfile3b,
        snpcolumns = c(0L,2L:5L))

bdinfo3b <- getbdinfo(bdfiles = bdfile3b)

```

## startcolumn

Sometimes the GEN file has more SNP information than the five values mentioned earlier. In this case the genotype probabilities start in a column number other than 6. The value of the <span style="font-family:Courier">startcolumn</span> is the column number that the genotype probabilities start

The following code reads in *set4b.imp*. It has an extra column in the SNP data in column 2. The <span style="font-family:Courier">snpcolumns</span> and <span style="font-family:Courier">startcolumn</span> has been set to handle this. The value of <span style="font-family:Courier">impformat</span> has also been set since there are only 2 genotype probabilities in the file.

``` {r, eval = T, echo = T, message = F, warning = F, tidy = T}
gen4bfile <- system.file("extdata", "set4b.imp", package = "BinaryDosage")
sample4bfile <- system.file("extdata", "set4b.sample", package = "BinaryDosage")
bdfile4b <- tempfile()
gentobd(genfiles = c(gen4bfile, sample4bfile),
        bdfiles = bdfile4b,
        snpcolumns = c(1L,2L,4L,5L,6L),
        startcolumn = 7L,
        impformat = 2L)

bdinfo4b <- getbdinfo(bdfiles = bdfile4b)

```

## impformat

The <span style="font-family:Courier">impformat</span> parameter is an integer from 1 to 3 that indicates how many genotype probabilities are in the file for each person. The value of 1 indicates that the value is the dosage value for the subject.

The following codes reads in file *set2b.imp*. This file contains only the dosage values for the subjects. The SNP information is not in the default order so the values of snpcolumns has to be specified (see above).

``` {r, eval = T, echo = T, message = F, warning = F, tidy = T}
gen2bfile <- system.file("extdata", "set2b.imp", package = "BinaryDosage")
sample2bfile <- system.file("extdata", "set2b.sample", package = "BinaryDosage")
bdfile2b <- tempfile()
gentobd(genfiles = c(gen2bfile, sample2bfile),
        bdfiles = bdfile2b,
        snpcolumns = c(1L,3L,2L,4L,5L),
        impformat = 1L)

bdinfo2b <- getbdinfo(bdfiles = bdfile2b)

```

## chromosome

The <span style="font-family:Courier">chromosome</span> parameter is a character value that used when the chromosome column value in <span style="font-family:Courier">snpcolumns</span> is set to -1L.

The following code reads in *set3b.imp*, setting the chromosome value to 1.

``` {r, eval = T, echo = T, message = F, warning = F, tidy = T}
gen3bfile <- system.file("extdata", "set3b.imp", package = "BinaryDosage")
sample3bfile <- system.file("extdata", "set3b.sample", package = "BinaryDosage")
bdfile3bm1 <- tempfile()
gentobd(genfiles = c(gen3bfile, sample3bfile),
        bdfiles = bdfile3bm1,
        snpcolumns = c(-1L,2L:5L),
        chromosome = "1")

bdinfo3bm1 <- getbdinfo(bdfiles = bdfile3bm1)

```

## header parameter 

The <span style="font-family:Courier">header</span> parameter is a character vector of length 1 or 2. These indicate if the GEN file an sample file have headers, respectively. If the first value is <span style="font-family:Courier">TRUE</span>, the second value is ignored as the subjects IDs are in the header of the GEN file.

The following code reads in *set3b.imp* using the sample file *set3bnh.sample* which has no header.

``` {r, eval = T, echo = T, message = F, warning = F, tidy = T}
gen3bfile <- system.file("extdata", "set3b.imp", package = "BinaryDosage")
sample3bnhfile <- system.file("extdata", "set3bnh.sample", package = "BinaryDosage")
bdfile3bnh <- tempfile()
gentobd(genfiles = c(gen3bfile, sample3bnhfile),
        bdfiles = bdfile3bnh,
        snpcolumns = c(0L,2L:5L),
        header = c(FALSE, FALSE))

bdinfo3bnh <- getbdinfo(bdfiles = bdfile3bnh)

```

## gz

The <span style="font-family:Courier">gz</span> parameter is a logical value that indicates if the GEN file is compressed using gzip. The sample file is always assumed to be uncompressed.

The following code reads in *set4b.imp.gz* file using the sample file *set4b.sample*.

``` {r, eval = T, echo = T, message = F, warning = F, tidy = T}
gen4bgzfile <- system.file("extdata", "set4b.imp.gz", package = "BinaryDosage")
sample4bfile <- system.file("extdata", "set4b.sample", package = "BinaryDosage")
bdfile4bgz <- tempfile()
gentobd(genfiles = c(gen4bgzfile, sample4bfile),
        bdfiles = bdfile4bgz,
        snpcolumns = c(1L,2L,4L,5L,6L),
        startcolumn = 7L,
        impformat = 2L,
        gz = TRUE)

bdinfo4bgz <- getbdinfo(bdfiles = bdfile4bgz)

```
## separator

The <span style="font-family:Courier">separator</span> parameter is a character variable. This character separates the columns in the GEN file. Multiple copies of the separator are considered to be separator.

## bdfiles

The <span style="font-family:Courier">bdfiles</span> parameter is a character vector of length 1 or 3. These are the names of the binary dosage, family, and map files. If the format of the binary dosage file is 4, the only value needed is the name of the binary dosage file.

## format

The <span style="font-family:Courier">format</span> parameter determines the format of the binary dosage files. Formats 1, 2 and 3 consist of three files, binary dosage, family, and map. Format 4 combines all of these into one file.

## subformat

The <span style="font-family:Courier">subformat</span> parameter determines what information is in the binary dosage files. All formats can have subformats 1 and 2. A <span style="font-family:Courier">subformat</span> value of 1 indicates that only the dosage values are written to the binary dosage file and a value of 2 indicates that the dosage and genotype probabilities are written to the binary dosage file. Formats 3 and 4 can also have <span style="font-family:Courier">subformat</span> values of 3 and 4. These values have the same meaning as 1 and 2 respectively but have a slightly reordered header in the binary dosage file to improve read speed.

## snpidformat

The <span style="font-family:Courier">snpidformat</span> options specifies how the SNP ID is written to the binary dosage file. The default value is 0. This tells the code to use the SNP IDs that are in the GEN file. Other values that can be supplied to the function creates a SNP ID from the chromosome, location, reference, and alternate allele values.

When the snpidformat is set to 1, the SNP ID is written in the format

<span style="font-family:Courier">Chromosome:Location</span>

When the snpidformat is set to 2, the SNP ID is written in the format

<span style="font-family:Courier">Chromosome:Location:Reference Allele:Alternate Allele</span>

When the snpidformat is set to 3, the SNP ID is written in the format

<span style="font-family:Courier">Chromosome:Location_Reference Allele_Alternate Allele</span>

When the snpidformat is set to -1, the SNP ID is not written to the binary dosage file. When the binary dosage file is read, the SNP ID is generated using the format for snpidformat equal to 2. This reduces the size of the binary dosage file.

## bdoptions

When using binary dosage format 4.x it is possible to store additional information about the SNPs in the file. This is information consists of the following values

- Alternate allele frequency
- Minor allele frequency
- Imputation r-squared

It is possible to calculate the alternate and minor allele frequency without the imputation information file. It is also possible to estimate the imputation r-squared. See the vignette [Estimating Imputed R-squares](r2estimates.html) for more information on the r-squared estimate.

The value for bdoptions is a vector of character values that can be "aaf", "maf", "rsq", or and combination of these values. The values indicate to calculate alternate allele frequency, minor allele frequency, and imputation r-squared respectively.

# Additional routines

The following routines are available for accessing information contained in VCF files

- <span style="font-family:Courier">getvcfinfo</span>
- <span style="font-family:Courier">vcfapply</span>

## getgeninfo

The <span style="font-family:Courier">getgeninfo</span> routine returns information about a GEN file. For more information about the data returned see [Genetic File Information](geneticfileinfo.html). This information needs to be passed to genapply so it can efficiently read the GEN file.

The parameters passed to <span style="font-family:Courier">getgeninfo</span> are

- <span style="font-family:Courier">genfiles</span> - Name of GEN file and the optional sample file.
- <span style="font-family:Courier">snpcolumns</span> - Columns containing the values for chromosome, SNP ID, location, reference allele, and alternate allele.
- <span style="font-family:Courier">startcolumn</span> - Column where genotype probabilities start.
- <span style="font-family:Courier">impformat</span> - Number of genotype probabilities for each subject.
- <span style="font-family:Courier">chromosome</span> - Optional chromosome, provided if chromosome is not include in GEN file
- <span style="font-family:Courier">header</span> - Vector of one or two logical values indicating if GEN and sample files have headers respectively.
- <span style="font-family:Courier">gz</span> - Logical value indicating if GEN file is compressed.
- <span style="font-family:Courier">index</span> - Logical value indicating if the GEN file is to be indexed.
- <span style="font-family:Courier">snpidformat</span> - Format to create the SNP ID in.
- <span style="font-family:Courier">sep</span> - Separator used in GEN file.

All of these parameters have the same meaning as in the <span style="font-family:Courier">gentobd</span> routine above. There is on additional parameter <span style="font-family:Courier">index</span>. This is a logical value indicating if the GEN file should be indexed for quicker reading. This is useful when using <span style="font-family:Courier">genapply</span>. However the <span style="font-family:Courier">index</span> parameter can not be TRUE if the file is compressed.

## genapply

The <span style="font-family:Courier">genapply</span> routine applies a function to all SNPs in the GEN file. The routine returns a list with length equal to the number of SNPs in the GEN file. Each element in the list is the value returned by the user supplied function. The routine takes the following parameters.

- <span style="font-family:Courier">geninfo</span> - list with information about the GEN file returned by <span style="font-family:Courier">getgeninfo</span>.
- <span style="font-family:Courier">func</span> - user supplied function to be applied to each SNP in the VCF file.
- <span style="font-family:Courier">...</span> - additional parameters needed by the user supplied function

The user supplied function must have the following parameters.

- <span style="font-family:Courier">geninfo</span> - A numeric vector with the dosage values for each subject.
- <span style="font-family:Courier">p0</span> - A numeric vector with the probabilities the subject has no alternate alleles for each subject.
- <span style="font-family:Courier">p1</span> - A numeric vector with the probabilities the subject has one alternate allele for each subject.
- <span style="font-family:Courier">p2</span> - A numeric vector with the probabilities the subject has two alternate alleles for each subject.

The user supplied function can have other parameters. These parameters need to passed to the <span style="font-family:Courier">genapply</span> routine.

There is a function in the <span style="font-family:Courier">BinaryDosage</span> package named <span style="font-family:Courier">getaaf</span> that calculates the alternate allele frequencies and is the format needed by <span style="font-family:Courier">genapply</span> routine. The following uses <span style="font-family:Courier">getaaf</span> to calculate the alternate allele frequency for each SNP in the *set3b.chr.imp* file using the <span style="font-family:Courier">vcfapply</span> routine.

``` {r, eval = T, echo = T, message = F, warning = F, tidy = T}
gen3bchrfile <- system.file("extdata", "set3b.chr.imp", package = "BinaryDosage")
sample3bfile <- system.file("extdata", "set3b.sample", package = "BinaryDosage")
geninfo <- getgeninfo(genfiles = c(gen3bchrfile, sample3bfile), index = TRUE)

aaf <- unlist(genapply(geninfo = geninfo, getaaf))

altallelefreq <- data.frame(SNP = geninfo$snps$snpid, aafcalc = aaf)
knitr::kable(altallelefreq, caption = "Calculated aaf", digits = 3)

```

---
title: "Binary Dosage Formats"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Binary Dosage Formats}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

**Note:** This vignette describes formats 1 through 4, which are legacy formats.
Format 5 is the recommended format for new work. It uses per-SNP gzip
compression and stores metadata in a companion RDS file, making it simpler and
more efficient to use. See the
[Using Format 5 Binary Dosage Files](usingformat5files.html) vignette for
details.

---

There are currently 4 legacy formats for the binary dosage file.

The first three formats consist of three files, binary dosage, family, and map. The family and maps are data frames with information about the subjects and SNPs in the binary dosage file, respectively. These are data frames saved with the <span style="font-family:Courier">saveRDS</span> command.

# Format 1

Format 1 has a header that begins with a magic word and is followed by a number indicating whether it is in format 1.1 or 1.2. It is the followed by the genotype information. The total header length is 8 bytes.

## Format 1.1

In format 1.1 the only value stored is the dosage. The dosage values stored are multiplied by $2^{15} - 1$, 0x7ffe, and stored as short integers. If a value is missing it is stored as $2^{16}$, 0xffff. Each subject needs to 2 bytes per SNP. The total size of the data section is 2 times the number of subjects times the number of SNPs bytes.

## Format 1.2

In format 1.2 the only values stored are $\Pr(g=1)$ and $\Pr(g=2)$. These values are multiplied by $2^{16} - 1$, 0xfffe, and stored as short integers. A value of $2^16$, 0xffff indicates a value is missing. The total size of the data section is 4 times the number of subjects times the number of SNPs bytes.

# Format 2

Format 2 has the same header as format 1.

## Format 2.1

The format of the data section is same as format 1.1 except the dosage values are multiplied by 20,000, 0x4e20. The missing value is still $2^{16}$, 0xffff.

## Format 2.2

The format of the data section is same as format 1.1 except the dosage values are multiplied by 20,000, 0x4e20. The missing value is still $2^{16}$, 0xffff.

**Note:** Format 2 was adopted when it was discovered that the values returned from the imputation programs were limited to 3 or 4 digits passed the decimal point. When results from fitting models were compared between using the binary dosage file and the original VCF or GEN file, there were slight but unimportant differences. It was considered desirable to be able to return the values exactly as they appear in the original imputation file.

# Format 3

Format 3.1 and 3.2 has a header similar to formats 1 and 2 but the number of subjects and SNPs were added to the header to avoid problems associating the wrong family or map file to the binary dosage file.

Format 3.3 and 3.4 has a header similar to formats 1 and 2 but the md5 hash of the family and map data frames were added to the header to avoid problems associating the wrong family or map file to the binary dosage file.

## Format 3.1 and 3.3

The data section of formats 3.1 and 3.3 are the same as format 2.1

## Format 3.2

Each SNP in the data section begins with a integer value identifying how long the section is for that SNP. The data is then stored as described below in minimizing the data stored.

## Format 3.4

Format 3.4 stores the data is a similar format as 3.2 but the section begins with the lengths of all the sections for the SNPs and then is followed by the genotype information.

# Format 4

Format 4 takes the data that is in the family and map files and moves it into the header of the binary dosage file. The first section of the header has the magic word and the format. This is followed by information on where the family, map, and genotype data are stored in the file. After the header there is the family data, followed by the map data, and then the imputation data.

## Format 4.1 and 4.3

The data section of formats 4.1 and 4.3 are the same as format 2.1

## Format 4.2 and 4.4

The data section of formats 4.2 and 4.4 are the same as format 4.2 and 4.3 respectively.


# Minimizing the data stored

A lot is known about the imputation data. We know the following
$$\Pr(g=0) + \Pr(g=1) + \Pr(g=2) = 1 $$
$$ d = \Pr(g=1) + 2\Pr(g=2)$$
where $d$ is the dosage. This means we only need to know two of the values to calculate the other two. In the <span style="font-family:Courier">BinaryDosage</span> package, the dosage and $\Pr(g=1) are used.

It is quite often the case that either $\Pr(g=0)$ or $\Pr(g=2)$ is 0. In this case, knowing the dosage is enough.

$$
\Pr(g = 1) = \left\{\begin{array}{ll}%
d & \; \Pr(g=2)=0, d \leq 1\\%
2 - d & \; \Pr(g=0) = 0, d > 1 %
\end{array}\right.
$$
Once the dosage and $\Pr(g=1)$ is know the other values can be quickly calculated.

$$\Pr(g=2) = \frac{d - \Pr(g=1)}{2}$$

$$\Pr(g=0) = 1 - \Pr(g=1) - \Pr(g=2)$$

These formulae work well but sometimes there is round off error in the imputation values. In these cases the above equations can't be used to get the exact imputation values. In these situations all four imputations values, $d$, $\Pr(g=0)$,$\Pr(g=1)$, and $\Pr(g=2)$ have to be saved. Fortunately this is not a common occurrence.

Since the values stored are short integers of 2 bytes in length, only the last 15 bits are used. This allows the 16th bit to be used as an indicator. For each SNP for each subject the first value saved is the dosage. If the 16th bit is 0, this indicates that either $\Pr(g=0)$ or $\Pr(g=2)$ is 0 and the other values can be calculated as described above. If the 16th bit is set to 1, this indicates that the value of $\Pr(g=1)$ follows. If the 16th bit is set to 0, this indicates that the above equations can be used to calculate $\Pr(g=0)$ and $\Pr(g=2)$. If the 16th bit is set to one, this indicates the next two values are $\Pr(g=0)$ and $\Pr(g=2)$ respectively.

**Note:** Usage of the this method general results in 2.2 to 2.4 bytes needed to store each SNP for each subject.

---
title: "Using Format 5 Binary Dosage Files"
output:
  rmarkdown::html_vignette:
    toc: true
vignette: >
  %\VignetteIndexEntry{Using Format 5 Binary Dosage Files}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, echo = FALSE}
library(BinaryDosage)
```

# Introduction

Format 5 is a binary dosage file format designed for reading imputed genotype
data from bgzipped, tabix-indexed VCF files — such as those returned by the
[Michigan Imputation Server](https://imputationserver.sph.umich.edu) using
[minimac](https://genome.sph.umich.edu/wiki/Minimac).

Format 5 differs from earlier formats (1–4) in two important ways.

First, it uses **per-SNP gzip compression**. Each SNP's dosage and genotype
probability values are compressed independently and written as a contiguous
block in the data file. This means any SNP can be decompressed in isolation
without reading surrounding data.

Second, the metadata — sample IDs, the SNP table, byte offsets, and file
information — is stored in a companion **RDS file** rather than embedded in a
binary header. This makes the metadata immediately accessible from R without
reading the data file.

A Format 5 dataset consists of two files.

- **`.bdose`** — the data file. Begins with a 4-byte magic number followed by
  one gzip-compressed block per SNP.
- **`.bdose.bdi`** — an RDS file containing a list of class `"genetic-info"`
  with the sample table, SNP table, byte offsets, and format metadata. The
  name is always the `.bdose` filename with `.bdi` appended (e.g.
  `set1a.bdose.bdi`). This structure is identical to the one returned by
  `getbdinfo`, `getvcfinfo`, and `getgeninfo` for other file types.

# Example files

The BinaryDosage package includes a small bgzipped VCF file, *set1a.vcf.gz*,
which contains dosage (DS) and genotype probability (GP) data for 60 subjects
and 10 SNPs on chromosome 1. This file is used in all examples below.

The output files in each example are written to a temporary directory so that
nothing is left behind after the vignette runs.

```{r files}
vcf_file   <- system.file("extdata", "set1a.vcf.gz", package = "BinaryDosage")
bd4_file   <- system.file("extdata", "vcf1a.bdose",  package = "BinaryDosage")
bdose_file <- file.path(tempdir(), "set1a.bdose")
```

# Converting a VCF file to Format 5

The `vcftobd` function converts a bgzipped VCF file to a Format 5 file pair.
The VCF file must contain the FORMAT fields `DS` (dosage) and `GP` (genotype
probabilities), as produced by minimac and the Michigan Imputation Server.

The function takes the following parameters.

- `vcffile` — path to the bgzipped VCF file.
- `bdose_file` — path for the output `.bdose` file. The companion `.bdi`
  metadata file is written automatically to `paste0(bdose_file, ".bdi")`.
- `region` — optional genomic region string in bcftools format (e.g.
  `"chr21"` or `"chr21:1-5000000"`). Requires a tabix index. Default `NULL`
  processes the entire file.

The following code converts *set1a.vcf.gz* to a Format 5 file pair.

```{r convert, eval = TRUE, echo = TRUE, message = FALSE, warning = FALSE}
if (requireNamespace("vcfppR", quietly = TRUE)) {
  vcftobd(vcffile = vcf_file, bdose_file = bdose_file)
} else {
  updatebd(bdfiles = bd4_file, bdose_file = bdose_file)
}
```

To convert only a specific region of the file, supply the `region` parameter.
The VCF file must have a tabix index (a `.tbi` file alongside it).

```{r convert_region, eval = F, echo = T}
vcftobd(vcffile    = vcf_file,
         bdose_file = bdose_file,
         region     = "1:10000-15000")
```

# Loading Format 5 file information

The `getbdinfo` function reads and validates a Format 5 file pair, returning
a list of class `"genetic-info"` that is used by `getbd5snp` to retrieve
individual SNPs. It automatically detects the Format 5 magic header and
delegates to the internal Format 5 reader.

The function takes the following parameters.

- `bdose_file` — path to the `.bdose` file. The companion `.bdi` file is
  read automatically from `paste0(bdose_file, ".bdi")`.

```{r load_info, eval = TRUE, echo = T, message = F, warning = F}
bd5 <- getbdinfo(bdose_file)
```

The object returned has the same structure as the list returned by `getbdinfo`
for older formats. For a full description of all fields see
[Genetic File Information](geneticfileinfo.html). The key fields are described
below.

## File and format metadata

```{r meta, eval = TRUE, echo = T}
cat("Class:          ", class(bd5), "\n")
cat("Uses family ID: ", bd5$usesfid, "\n")
cat("One chromosome: ", bd5$onechr, "\n")
cat("SNP ID format:  ", bd5$snpidformat, "\n")
cat("Format:         ", bd5$additionalinfo$format, "\n")
cat("Header size:    ", bd5$additionalinfo$headersize, "bytes\n")
```

## Sample information

The `samples` element is a data frame with one row per subject. The `fid`
column holds the family ID (empty for VCF files, which carry no family
information) and the `sid` column holds the subject ID.

```{r samples, eval = TRUE, echo = T}
cat("Number of samples:", nrow(bd5$samples), "\n")
knitr::kable(head(bd5$samples, 5), caption = "First 5 samples")
```

## SNP information

The `snps` element is a data frame with one row per SNP.

```{r snps, eval = TRUE, echo = T}
cat("Number of SNPs:", nrow(bd5$snps), "\n")
knitr::kable(bd5$snps, caption = "SNP table")
```

## Byte offsets

The `indices` element is a numeric vector of byte offsets into the `.bdose`
file, one per SNP, giving the start of that SNP's compressed block. These are
used internally by `getbd5snp`.

```{r indices, eval = TRUE, echo = T}
cat("Indices (bytes):", bd5$indices, "\n")
```

# Reading SNP data

The `getbd5snp` function decompresses and returns the dosage and genotype
probability data for a single SNP. The general-purpose `getsnp` function also
works with Format 5 files and can be used as an alternative.

For details on all three Format 5 SNP reader functions — including
`getbd5snp_buf` (pre-allocated output vectors) and `getbd5snp_con` (persistent
file connection for high-throughput loops) — see the
[Reading SNPs from Format 5 Files](bd5snpreaders.html) vignette.

The function takes the following parameters.

- `bd5info` — object returned by `getbdinfo`.
- `snp` — the SNP to retrieve, either as a 1-based integer index or as a
  character SNP ID matching a value in `bd5info$snps$snpid`.

The function returns a list with four numeric vectors, each of length equal
to the number of samples.

- `dosage` — DS values in \[0, 2\]; `NA` = missing.
- `p0` — $\Pr(g=0)$ values in \[0, 1\]; `NA` = missing.
- `p1` — $\Pr(g=1)$ values in \[0, 1\]; `NA` = missing.
- `p2` — $\Pr(g=2)$ values in \[0, 1\]; `NA` = missing.

## Reading a SNP by index

The following code retrieves the first SNP by its 1-based index.

```{r snp_by_index, eval = TRUE, echo = T, message = F, warning = F}
snp1 <- getbd5snp(bd5info = bd5, snp = 1L)

snp1_df <- data.frame(
  SampleID = bd5$samples$sid,
  Dosage   = round(snp1$dosage, 4),
  P_00     = round(snp1$p0,     4),
  P_01     = round(snp1$p1,     4),
  P_11     = round(snp1$p2,     4)
)

knitr::kable(head(snp1_df, 10),
             caption = paste("SNP", bd5$snps$snpid[1], "— first 10 subjects"))
```

## Reading a SNP by ID

A SNP can also be retrieved by passing its SNP ID as a character string.

```{r snp_by_id, eval = TRUE, echo = T, message = F, warning = F}
snp_id <- "1:12000:T:C"
snp3   <- getbd5snp(bd5info = bd5, snp = snp_id)

snp3_df <- data.frame(
  SampleID = bd5$samples$sid,
  Dosage   = round(snp3$dosage, 4),
  P_00     = round(snp3$p0,     4),
  P_01     = round(snp3$p1,     4),
  P_11     = round(snp3$p2,     4)
)

knitr::kable(head(snp3_df, 10),
             caption = paste("SNP", snp_id, "— first 10 subjects"))
```

# Applying a function to all SNPs

The simplest option is `bdapply`, which handles the loop internally and uses
`getbd5snp_con` automatically for Format 5 files.

```{r apply_bdapply, eval = TRUE, echo = T, message = F, warning = F}
getaaf <- function(dosage, p0, p1, p2) mean(dosage, na.rm = TRUE) / 2

aaf_list  <- bdapply(bdinfo = bd5, func = getaaf)
aaf_table <- data.frame(snpid = bd5$snps$snpid,
                        aaf   = round(unlist(aaf_list), 4))

knitr::kable(aaf_table, caption = "Alternate allele frequencies via bdapply")
```

For manual loops, `getbd5snp_buf` and `getbd5snp_con` are faster than
`getbd5snp` because they avoid allocating a new list on every call.
`getbd5snp_con` is the fastest option as it also keeps the file open across
all iterations. See the
[Reading SNPs from Format 5 Files](bd5snpreaders.html) vignette for details
and a performance comparison.

```{r apply_con, eval = TRUE, echo = T, message = F, warning = F}
n_snps <- nrow(bd5$snps)
n_samp <- nrow(bd5$samples)
dosage <- numeric(n_samp)
p0     <- numeric(n_samp)
p1     <- numeric(n_samp)
p2     <- numeric(n_samp)

con <- openbd5con(bd5)
aaf <- numeric(n_snps)
for (i in seq_len(n_snps)) {
  getbd5snp_con(bd5info = bd5, snp = i,
                dosage = dosage, p0 = p0, p1 = p1, p2 = p2,
                bd5con = con)
  aaf[i] <- mean(dosage, na.rm = TRUE) / 2
}
closebd5con(con)

knitr::kable(data.frame(snpid = bd5$snps$snpid, aaf = round(aaf, 4)),
             caption = "Alternate allele frequencies via getbd5snp_con")
```

```{r cleanup, include = FALSE, eval = TRUE}
unlink(c(bdose_file, paste0(bdose_file, ".bdi")))
```

---
title: "Reading SNPs from Format 5 Files"
output:
  rmarkdown::html_vignette:
    toc: true
vignette: >
  %\VignetteIndexEntry{Reading SNPs from Format 5 Files}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, echo = FALSE}
library(BinaryDosage)
```

# Introduction

The BinaryDosage package provides three functions for reading individual SNPs
from Format 5 binary dosage files. They return the same data but differ in how
they manage memory and file connections, which matters when reading many SNPs
in a loop.

| Function | Allocates output? | Opens file per call? |
|---|---|---|
| `getbd5snp` | Yes — returns a new list | Yes |
| `getbd5snp_buf` | No — writes into caller vectors | Yes |
| `getbd5snp_con` | No — writes into caller vectors | No — reuses open connection |

All three functions require the `"genetic-info"` list returned by `getbdinfo`.

# Setup

The examples below use the small bgzipped VCF file included with the package.
First, convert it to a Format 5 file pair and load the metadata.

```{r setup_files, message = FALSE, warning = FALSE}
bdose_file <- file.path(tempdir(), "set1a.bdose")

if (requireNamespace("vcfppR", quietly = TRUE)) {
  vcftobd(vcffile    = system.file("extdata", "set1a.vcf.gz", package = "BinaryDosage"),
          bdose_file = bdose_file)
} else {
  updatebd(bdfiles    = system.file("extdata", "vcf1a.bdose", package = "BinaryDosage"),
           bdose_file = bdose_file)
}
bd5 <- getbdinfo(bdose_file)

n_snps <- nrow(bd5$snps)
n_samp <- nrow(bd5$samples)
cat("SNPs:", n_snps, " Samples:", n_samp, "\n")
```

---

# getbd5snp

`getbd5snp` is the simplest interface. It opens the `.bdose` file, seeks to
the requested SNP's compressed block, decompresses it, decodes the values, and
returns them as a new list. The file is opened and closed on every call.

## Parameters

- `bd5info` — object returned by `getbdinfo`
- `snp` — 1-based integer index or character SNP ID

## Return value

A list with four numeric vectors of length *n_samples*:

- `dosage` — DS values in \[0, 2\]; `NA` = missing
- `p0` — Pr(*g*=0) in \[0, 1\]; `NA` = missing
- `p1` — Pr(*g*=1) in \[0, 1\]; `NA` = missing
- `p2` — Pr(*g*=2) in \[0, 1\]; `NA` = missing

## Reading a SNP by index

```{r getbd5snp_index}
snp1 <- getbd5snp(bd5info = bd5, snp = 1L)

knitr::kable(
  data.frame(
    SampleID = head(bd5$samples$sid, 10),
    Dosage   = round(head(snp1$dosage, 10), 4),
    P_00     = round(head(snp1$p0,     10), 4),
    P_01     = round(head(snp1$p1,     10), 4),
    P_11     = round(head(snp1$p2,     10), 4)
  ),
  caption = paste("SNP", bd5$snps$snpid[1], "— first 10 subjects")
)
```

## Reading a SNP by ID

```{r getbd5snp_id}
snp3 <- getbd5snp(bd5info = bd5, snp = "1:12000:T:C")

knitr::kable(
  data.frame(
    SampleID = head(bd5$samples$sid, 10),
    Dosage   = round(head(snp3$dosage, 10), 4),
    P_00     = round(head(snp3$p0,     10), 4),
    P_01     = round(head(snp3$p1,     10), 4),
    P_11     = round(head(snp3$p2,     10), 4)
  ),
  caption = "SNP 1:12000:T:C — first 10 subjects"
)
```

## Using getbd5snp in a loop

`getbd5snp` is convenient but allocates a new list and opens the file on every
call. For a small number of SNPs this is fine. The following calculates the
alternate allele frequency for every SNP.

```{r getbd5snp_loop}
aaf <- numeric(n_snps)
for (i in seq_len(n_snps)) {
  aaf[i] <- mean(getbd5snp(bd5, i)$dosage, na.rm = TRUE) / 2
}

knitr::kable(
  data.frame(snpid = bd5$snps$snpid, aaf = round(aaf, 4)),
  caption = "Alternate allele frequencies via getbd5snp"
)
```

---

# getbd5snp_buf

`getbd5snp_buf` eliminates the per-call list allocation by writing results
directly into four numeric vectors that the caller pre-allocates once before
the loop. This avoids repeated garbage collection pressure when reading many
SNPs. The file is still opened and closed on every call.

## Parameters

- `bd5info` — object returned by `getbdinfo`
- `snp` — 1-based integer index or character SNP ID
- `dosage`, `p0`, `p1`, `p2` — pre-allocated `numeric(n_samples)` vectors

## Return value

`NULL` invisibly. The four output vectors are updated in place.

## Important: R copy-on-modify semantics

The output vectors **must not have more than one R binding** at the call site.
If another variable also points to the same vector, R's copy-on-modify rule
will copy the vector before writing, so the caller's variable will not be
updated. Always use plain, freshly allocated vectors:

```r
# Correct
dosage <- numeric(n_samp)
getbd5snp_buf(bd5, 1L, dosage, p0, p1, p2)

# Wrong — dosage2 creates a second binding; the update may not be visible
dosage2 <- dosage
getbd5snp_buf(bd5, 1L, dosage, p0, p1, p2)
```

## Example

```{r getbd5snp_buf_loop}
dosage <- numeric(n_samp)
p0     <- numeric(n_samp)
p1     <- numeric(n_samp)
p2     <- numeric(n_samp)

aaf_buf <- numeric(n_snps)
for (i in seq_len(n_snps)) {
  getbd5snp_buf(bd5info = bd5, snp = i, dosage = dosage,
                p0 = p0, p1 = p1, p2 = p2)
  aaf_buf[i] <- mean(dosage, na.rm = TRUE) / 2
}

knitr::kable(
  data.frame(snpid = bd5$snps$snpid, aaf = round(aaf_buf, 4)),
  caption = "Alternate allele frequencies via getbd5snp_buf"
)
```

---

# getbd5snp_con

`getbd5snp_con` is the highest-performance option. It combines the
pre-allocated vector approach of `getbd5snp_buf` with a persistent open file
connection, so the `.bdose` file is opened once before the loop and kept open
for all reads. The C-level read uses zlib directly rather than R's `memDecompress`.

Use this when reading a large number of SNPs sequentially and minimizing
elapsed time matters.

## Workflow

1. Call `openbd5con(bd5info)` to open the file and get a `"bd5con"` object.
2. Call `getbd5snp_con` inside the loop, passing the `"bd5con"` object.
3. Call `closebd5con` when finished to release the file handle promptly.
   (The connection is also closed automatically when the `"bd5con"` object
   is garbage-collected or when R exits, so the explicit close is optional
   but recommended.)

## openbd5con

Opens a persistent connection to the `.bdose` file.

**Parameters**

- `bd5info` — object returned by `getbdinfo`

**Return value**

An object of class `"bd5con"` to be passed to `getbd5snp_con` and
`closebd5con`.

## closebd5con

Explicitly closes the connection.

**Parameters**

- `bd5con` — object returned by `openbd5con`

**Return value**

`NULL` invisibly.

## getbd5snp_con

**Parameters**

- `bd5info` — object returned by `getbdinfo`
- `snp` — 1-based integer index or character SNP ID
- `dosage`, `p0`, `p1`, `p2` — pre-allocated `numeric(n_samples)` vectors
- `bd5con` — object returned by `openbd5con`

**Return value**

`NULL` invisibly. The four output vectors are updated in place.

The same copy-on-modify constraint as `getbd5snp_buf` applies.

## Example

```{r getbd5snp_con_loop}
dosage <- numeric(n_samp)
p0     <- numeric(n_samp)
p1     <- numeric(n_samp)
p2     <- numeric(n_samp)

con <- openbd5con(bd5)

aaf_con <- numeric(n_snps)
for (i in seq_len(n_snps)) {
  getbd5snp_con(bd5info = bd5, snp = i, dosage = dosage,
                p0 = p0, p1 = p1, p2 = p2, bd5con = con)
  aaf_con[i] <- mean(dosage, na.rm = TRUE) / 2
}

closebd5con(con)

knitr::kable(
  data.frame(snpid = bd5$snps$snpid, aaf = round(aaf_con, 4)),
  caption = "Alternate allele frequencies via getbd5snp_con"
)
```

---

# Choosing a reader

- **`getbd5snp`** — use when reading a small number of SNPs, or when
  simplicity matters more than speed.
- **`getbd5snp_buf`** — use in loops where allocation overhead is measurable
  but the file seek cost is acceptable.
- **`getbd5snp_con`** — use when reading all or most SNPs sequentially and
  elapsed time is important. The persistent connection avoids both the
  allocation and the file open/close on every iteration.

All three functions produce identical results, as confirmed below.

```{r verify}
all(aaf == aaf_buf)
all(aaf == aaf_con)
```

```{r cleanup, include = FALSE}
unlink(c(bdose_file, paste0(bdose_file, ".bdi")))
```

---
title: "Estimating Imputed R-squares"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Estimating Imputed R-squares}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

A common way to measure the imputation $r^2$ is calculate the variance of the imputed alleles probabilities and divide that by the variance if the alleles were perfectly imputed. An allele is perfectly imputed if $\Pr(a_i = 1)$ equals 0 or 1 for all $i$.

The variance of the alleles when are perfectly imputed is $q(1-q)$ where $q$ is the alternate allele frequency. Given the imputation data We do not know what $q$ is the general population. However we can estimate it using the dosage values for each subject.
$$
\hat q = \sum_{i = 1}^{N}\frac{d_i}{2N}
$$
where the dosage is calculated as
$$
d_i = \frac{\Pr(g_i = 1) + 2\Pr(g_i=2)}{2}
$$
Another problem with the dosage data is we don't have the probabilities for each allele. Instead we have $\Pr(g_i=0), \Pr(g_i=1),$ and $\Pr(g_i=2)$. If we assume that a subject's two allelic probabilities, $q_1, q_2$ are independently imputed we know the following
$$
q_1(1-q2) + (1-q_1)q_2 = \Pr(g=1)
$$
and
$$
q_1 q_2 = \Pr(g=2)
$$
These equations can be solved resulting in the following values
$$
q_1 = \frac{d - \sqrt{d^2 - \Pr(g = 2)}}{2}\\%
q_2 = \frac{d + \sqrt{d^2 - \Pr(g = 2)}}{2}
$$
There can be some problems using the above equations. Sometimes the value inside the radical can be negative. This can be caused by roundoff error. If the value is negative and close to zero, the value can be set to zero.

**Note:** The documentation for minimac and Impute 2 indicate that the imputation values for the two alleles are imputed independently.

Since each subject has two alleles we can let $q_1$ to $q_N$ represent the first allele of each subject and $q_{N+1}$ to $q_{2N}$ represent the second allele. Given this we can calculate all the $q$'s as follows
$$
q_i = \left\{\begin{array}{ll}%
\frac{d_i - \sqrt{d_i^2 - 4\Pr(g_i = 2)}}{2} & \; 0<i\leq N\\%
\frac{d_i + \sqrt{d_i^2 - 4\Pr(g_i = 2)}}{2} & \; N<i\leq 2N %
\end{array}\right.
$$
Once the $q$'s have been calculated, the imputation $r^2$ can be estimated as follows

$$
\hat r^2 = \frac{\sum_{i = 1}^{2N}\frac{(q_i - \hat q)^2}{2N}}{\hat q(1 - \hat q)}
$$

---
title: "Merging Format 5 Binary Dosage Files"
output:
  rmarkdown::html_vignette:
    toc: true
vignette: >
  %\VignetteIndexEntry{Merging Format 5 Binary Dosage Files}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, echo = FALSE}
library(BinaryDosage)
```

# Introduction

The `mergebd` function merges two or more Format 5 binary dosage files into a
single Format 5 output file. The merge type is detected automatically from the
input files.

- **Subject merge** — subject IDs do not overlap across files. The output
  contains all subjects from every input file and the SNPs common to all files.
- **SNP merge** — SNP IDs do not overlap across files. The output contains all
  SNPs from every input file and the subjects common to all files.

If both subject IDs and SNP IDs overlap across files an error is returned,
since the merge type is ambiguous.

SNPs are identified by chromosome, position, reference allele, and alternate
allele, regardless of the SNP ID format stored in each file.

The function takes the following parameters.

- `bdose_files` — character vector of paths to the input `.bdose` files (at
  least two). The companion `.bdi` file for each is expected at
  `paste0(bdose_files[i], ".bdi")`.
- `bdose_file` — path for the output `.bdose` file. The companion `.bdi` file
  is written automatically to `paste0(bdose_file, ".bdi")`.

# Setup

The examples below use the bgzipped VCF file included with the package,
*set1a.vcf.gz*, which contains data for 60 subjects and 10 SNPs on chromosome
1. All output files are written to a temporary directory.

```{r setup_files, message = FALSE, warning = FALSE}
bdose_full <- file.path(tempdir(), "full.bdose")

if (requireNamespace("vcfppR", quietly = TRUE)) {
  vcftobd(vcffile    = system.file("extdata", "set1a.vcf.gz", package = "BinaryDosage"),
          bdose_file = bdose_full)
} else {
  updatebd(bdfiles    = system.file("extdata", "vcf1a.bdose", package = "BinaryDosage"),
           bdose_file = bdose_full)
}
bd_full <- getbdinfo(bdose_full)

cat("Subjects:", nrow(bd_full$samples), "\n")
cat("SNPs:    ", nrow(bd_full$snps),    "\n")
```

# Subject merge

A subject merge combines files that cover different subjects but the same (or
overlapping) set of SNPs. The output contains all subjects and the SNPs common
to every input file.

The example splits the 60-subject file into two 30-subject files using
`subsetbd`, then merges them back together.

```{r subject_merge, message = FALSE, warning = FALSE}
bdose_a   <- file.path(tempdir(), "set_a.bdose")
bdose_b   <- file.path(tempdir(), "set_b.bdose")
bdose_out <- file.path(tempdir(), "merged_subjects.bdose")

sids <- bd_full$samples$sid

subsetbd(bdfiles    = bdose_full,
         bdose_file = bdose_a,
         subjectids = sids[1:30])

subsetbd(bdfiles    = bdose_full,
         bdose_file = bdose_b,
         subjectids = sids[31:60])

mergebd(bdose_files = c(bdose_a, bdose_b),
        bdose_file  = bdose_out)

bd_a   <- getbdinfo(bdose_a)
bd_b   <- getbdinfo(bdose_b)
bd_out <- getbdinfo(bdose_out)

cat("File A subjects:", nrow(bd_a$samples),   "\n")
cat("File B subjects:", nrow(bd_b$samples),   "\n")
cat("Merged subjects:", nrow(bd_out$samples), "\n")
cat("Merged SNPs:    ", nrow(bd_out$snps),    "\n")
```

The merged file contains all 60 subjects and all 10 SNPs.

## Verifying subject order

The subjects in the merged file appear in input-file order: all subjects from
the first file followed by all subjects from the second file.

```{r subject_order}
knitr::kable(bd_out$samples, caption = "Subjects in merged file")
```

# SNP merge

A SNP merge combines files that cover different SNPs but the same (or
overlapping) set of subjects. The output contains all SNPs and the subjects
common to every input file.

The example splits the 10-SNP file into two 5-SNP files using `subsetbd`, then
merges them back together.

```{r snp_merge, message = FALSE, warning = FALSE}
bdose_snp_a   <- file.path(tempdir(), "snp_a.bdose")
bdose_snp_b   <- file.path(tempdir(), "snp_b.bdose")
bdose_snp_out <- file.path(tempdir(), "merged_snps.bdose")

locs <- bd_full$snps$location

subsetbd(bdfiles    = bdose_full,
         bdose_file = bdose_snp_a,
         locations  = locs[1:5])

subsetbd(bdfiles    = bdose_full,
         bdose_file = bdose_snp_b,
         locations  = locs[6:10])

mergebd(bdose_files = c(bdose_snp_a, bdose_snp_b),
        bdose_file  = bdose_snp_out)

bd_snp_a   <- getbdinfo(bdose_snp_a)
bd_snp_b   <- getbdinfo(bdose_snp_b)
bd_snp_out <- getbdinfo(bdose_snp_out)

cat("File A SNPs:    ", nrow(bd_snp_a$snps),   "\n")
cat("File B SNPs:    ", nrow(bd_snp_b$snps),   "\n")
cat("Merged SNPs:    ", nrow(bd_snp_out$snps),  "\n")
cat("Merged subjects:", nrow(bd_snp_out$samples), "\n")
```

## Verifying SNP order

SNPs appear in input-file order: all SNPs from the first file followed by all
SNPs from the second file.

```{r snp_order}
knitr::kable(bd_snp_out$snps, caption = "SNPs in merged file")
```

```{r cleanup, include = FALSE}
unlink(c(bdose_full,    paste0(bdose_full,    ".bdi"),
         bdose_a,       paste0(bdose_a,       ".bdi"),
         bdose_b,       paste0(bdose_b,       ".bdi"),
         bdose_out,     paste0(bdose_out,     ".bdi"),
         bdose_snp_a,   paste0(bdose_snp_a,   ".bdi"),
         bdose_snp_b,   paste0(bdose_snp_b,   ".bdi"),
         bdose_snp_out, paste0(bdose_snp_out, ".bdi")))
```

---
title: "Legacy Binary Dosage Formats (Formats 1-4)"
output:
  rmarkdown::html_vignette:
    toc: true
vignette: >
  %\VignetteIndexEntry{Legacy Binary Dosage Formats (Formats 1-4)}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, echo = FALSE}
library(BinaryDosage)
```

# Introduction

The BinaryDosage package was originally developed to convert VCF and GEN files produced by imputation servers into a compact binary format suitable for genome-wide association and gene-environment interaction analyses. This document covers the original binary dosage formats (formats 1–4), the functions used to create and read them, and the workflow for applying analyses across all SNPs in a file.

**Current recommendation:** New work should use Format 5 via `vcftobd`. Format 5 uses per-SNP gzip compression and stores all metadata in a companion `.bdinfo` file, making it better suited for large modern datasets. See the vignette *Using Format 5 Binary Dosage Files* for details.

The legacy formats remain fully supported and all files created with earlier versions of BinaryDosage are compatible with the current package.

---

# Legacy File Formats

Each binary dosage data set stores the following information:

- **Sample information**: family ID and subject ID
- **SNP information**: chromosome, SNP ID, genomic location, reference allele, alternate allele
- **Genetic data**: dosage values and genotype probabilities Pr(*g*=0), Pr(*g*=1), Pr(*g*=2)

## Formats 1, 2, and 3 (three-file layout)

Formats 1, 2, and 3 split data across three files:

- **Binary dosage file** — the genetic data
- **Family file** — subject IDs (saved with `saveRDS`)
- **Map file** — SNP information (saved with `saveRDS`)

### Format 1

Format 1 has an 8-byte header consisting of a magic word followed by a subformat indicator.

- **Subformat 1.1** stores only dosage. Values are multiplied by 0x7ffe (32,766) and stored as unsigned 16-bit integers. Missing = 0xffff.
- **Subformat 1.2** stores Pr(*g*=1) and Pr(*g*=2). Values are multiplied by 0xfffe (65,534). Missing = 0xffff.

### Format 2

Format 2 uses the same 8-byte header as Format 1 but multiplies values by 20,000 (0x4e20) rather than 0x7ffe. This was adopted when it was found that imputation programs report values to only 3–4 decimal places and it was desirable to reproduce those values exactly. Missing = 0xffff.

- **Subformat 2.1** stores only dosage.
- **Subformat 2.2** stores Pr(*g*=1) and Pr(*g*=2).

### Format 3

Format 3 extends the Format 2 header to include validation information.

- **Subformats 3.1 and 3.2** add the number of subjects and SNPs to the header to prevent mismatched family/map files.
- **Subformats 3.3 and 3.4** add MD5 hashes of the family and map data frames for the same purpose.
- **Subformats 3.2 and 3.4** use the compressed storage scheme described below.

## Format 4 (single-file layout)

Format 4 embeds the family and map data directly in the binary dosage file header, eliminating the need for separate files. The header is followed by the family data, then the map data, and then the genotype data. Format 4 is the recommended legacy format and is the default for all conversion functions.

- **Subformats 4.1 and 4.3** store one dosage value per subject per SNP (same encoding as 2.1).
- **Subformats 4.2 and 4.4** store dosage and genotype probabilities.

Format 4 can also store optional per-SNP statistics in the header:

- Alternate allele frequency
- Minor allele frequency
- Average call rate
- Imputation r-squared

## Compressed storage (subformats x.2 and x.4)

Knowing that Pr(*g*=0) + Pr(*g*=1) + Pr(*g*=2) = 1 and *d* = Pr(*g*=1) + 2·Pr(*g*=2), only two values are needed to recover all four. For most subjects, either Pr(*g*=0) or Pr(*g*=2) is zero, so the dosage alone is sufficient:

$$
\Pr(g=1) = \begin{cases} d & \Pr(g=2)=0,\ d \le 1 \\ 2-d & \Pr(g=0)=0,\ d > 1 \end{cases}
$$

The 16th bit of each stored dosage value is used as a flag: 0 means the above formula applies; 1 means an explicit Pr(*g*=1) value follows. If both Pr(*g*=0) and Pr(*g*=2) are non-zero, three values (dosage, Pr(*g*=0), Pr(*g*=2)) are written. This approach typically requires 2.2–2.4 bytes per subject per SNP.

---

# Functions Overview

| Function | Purpose |
|---|---|
| `vcftobdlegacy` | Convert a VCF file to a legacy binary dosage file |
| `gentobd` | Convert a GEN (Impute2) file to a legacy binary dosage file |
| `bdmerge` | Merge multiple legacy binary dosage files by subject |
| `getbdinfo` | Load metadata for a legacy binary dosage file |
| `getvcfinfo` | Load metadata for a VCF file |
| `getgeninfo` | Load metadata for a GEN file |
| `bdapply` | Apply a function to every SNP in a legacy binary dosage file |
| `vcfapply` | Apply a function to every SNP in a VCF file |
| `genapply` | Apply a function to every SNP in a GEN file |
| `getsnp` | Extract dosage and genotype probabilities for one SNP |

---

# Converting VCF Files

The `vcftobdlegacy` function converts VCF files to the legacy binary dosage format. The default output format is 4 (single file, all metadata embedded).

## Example files

The package includes representative VCF files modelled on output from the Michigan Imputation Server (minimac). The following code locates them.

```{r vcf_files}
vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")
vcf1ainfo <- system.file("extdata", "set1a.info", package = "BinaryDosage")
vcf1bfile <- system.file("extdata", "set1b.vcf", package = "BinaryDosage")
vcf1binfo <- system.file("extdata", "set1b.info", package = "BinaryDosage")
```

These sets contain:

| Set | Subjects | SNPs |
|-----|----------|------|
| 1a  | 60       | 10   |
| 1b  | 40       | 10   |

## Basic conversion

Pass the VCF file name to `vcftobdlegacy`. An optional imputation information file (e.g., from minimac) can be appended to the `vcffiles` vector to embed per-SNP statistics.

```{r vcf_convert, message = FALSE, warning = FALSE}
bdfile1a <- tempfile()
bdfile1b <- tempfile()

# Without information file
vcftobdlegacy(vcffiles = vcf1afile, bdfiles = bdfile1a)

# With information file — embeds aaf, maf, avgcall, rsq in the header
vcftobdlegacy(vcffiles = c(vcf1bfile, vcf1binfo), bdfiles = bdfile1b)
```

## Compressed VCF files

Add `gz = TRUE` for gzip-compressed VCF files. The information file, if provided, must remain uncompressed.

```{r vcf_gz, message = FALSE, warning = FALSE}
vcf1agzfile <- system.file("extdata", "set1a.vcf.gz", package = "BinaryDosage")
bdfile1a_gz <- tempfile()
vcftobdlegacy(vcffiles = vcf1agzfile, bdfiles = bdfile1a_gz, gz = TRUE)
```

## Other VCF formats

`vcftobdlegacy` reads the FORMAT field of each SNP and looks for the DS (dosage) and GP (genotype probabilities) entries. Files that contain only one of these are handled automatically.

```{r vcf_dosageonly, message = FALSE, warning = FALSE}
vcf2afile <- system.file("extdata", "set2a.vcf", package = "BinaryDosage")
bdfile2a <- tempfile()
vcftobdlegacy(vcffiles = vcf2afile, bdfiles = bdfile2a)
```

## snpidformat option

The `snpidformat` parameter controls how SNP IDs are stored.

| Value | Stored ID |
|-------|-----------|
| 0 (default) | ID from the VCF file |
| 1 | `chr:pos` |
| 2 | `chr:pos:ref:alt` |
| -1 | Not stored; reconstructed as `chr:pos:ref:alt` on read |

Setting `snpidformat = -1` reduces file size when SNP IDs are not needed.

```{r snpidformat, message = FALSE, warning = FALSE}
vcf1brsfile <- system.file("extdata", "set1b_rssnp.vcf", package = "BinaryDosage")
bdfile_id0  <- tempfile()
bdfile_id1  <- tempfile()
bdfile_id2  <- tempfile()
bdfile_idm1 <- tempfile()

vcftobdlegacy(vcffiles = vcf1brsfile, bdfiles = bdfile_id0)
vcftobdlegacy(vcffiles = vcf1brsfile, bdfiles = bdfile_id1, snpidformat = 1L)
vcftobdlegacy(vcffiles = vcf1brsfile, bdfiles = bdfile_id2, snpidformat = 2L)
vcftobdlegacy(vcffiles = vcf1brsfile, bdfiles = bdfile_idm1, snpidformat = -1L)

snpnames <- data.frame(
  format0  = getbdinfo(bdfile_id0)$snps$snpid,
  format1  = getbdinfo(bdfile_id1)$snps$snpid,
  format2  = getbdinfo(bdfile_id2)$snps$snpid,
  formatm1 = getbdinfo(bdfile_idm1)$snps$snpid
)
knitr::kable(snpnames, caption = "SNP IDs by snpidformat")
```

## bdoptions — calculating per-SNP statistics

When converting without an information file, `bdoptions` can trigger on-the-fly calculation of alternate allele frequency (`"aaf"`), minor allele frequency (`"maf"`), and an estimated imputation r-squared (`"rsq"`).

```{r bdoptions, message = FALSE, warning = FALSE}
bdfile1a_calc <- tempfile()
vcftobdlegacy(vcffiles = vcf1afile, bdfiles = bdfile1a_calc,
              bdoptions = c("aaf", "maf", "rsq"))
bdcalcinfo <- getbdinfo(bdfile1a_calc)
knitr::kable(data.frame(bdcalcinfo$snpinfo),
             caption = "Calculated per-SNP statistics", digits = 3)
```

---

# Converting GEN Files

GEN files are produced by the Impute2 imputation software. They are text files with more variable formatting than VCF, so `gentobd` exposes several parameters to handle different layouts.

## Parameters

| Parameter | Description |
|-----------|-------------|
| `genfiles` | GEN file name and optional sample file name |
| `snpcolumns` | Column positions for chromosome, SNP ID, location, reference, alternate |
| `startcolumn` | Column where genotype probabilities begin (default 6) |
| `impformat` | Number of values per subject: 1=dosage only, 2=Pr(g=0)+Pr(g=1), 3=all three |
| `chromosome` | Chromosome value when not present in the file |
| `header` | Logical vector indicating whether GEN and sample files have headers |
| `gz` | TRUE if the GEN file is gzip-compressed |
| `sep` | Column separator |
| `bdfiles` | Output binary dosage file name(s) |
| `format` | Output binary dosage format (default 4) |
| `subformat` | Output subformat |
| `snpidformat` | How to store the SNP ID |
| `bdoptions` | Additional per-SNP statistics to calculate |

## Example files

```{r gen_files}
gen3afile   <- system.file("extdata", "set3a.imp",    package = "BinaryDosage")
gen3asample <- system.file("extdata", "set3a.sample", package = "BinaryDosage")
gen3bfile   <- system.file("extdata", "set3b.imp",    package = "BinaryDosage")
gen3bsample <- system.file("extdata", "set3b.sample", package = "BinaryDosage")
```

The example GEN files have no chromosome column; the first column is `--` and the SNP ID encodes the chromosome as `chr:pos:ref:alt`. Setting `snpcolumns = c(0L, 2L:5L)` tells `gentobd` to extract the chromosome from the SNP ID.

```{r gen_convert, message = FALSE, warning = FALSE}
bdfile3a <- tempfile()
bdfile3b <- tempfile()

gentobd(genfiles = c(gen3afile, gen3asample),
        snpcolumns = c(0L, 2L:5L),
        bdfiles = bdfile3a)
gentobd(genfiles = c(gen3bfile, gen3bsample),
        snpcolumns = c(0L, 2L:5L),
        bdfiles = bdfile3b)
```

## snpcolumns

The five values in `snpcolumns` give the column indices for chromosome, SNP ID, location, reference, and alternate allele respectively.

- Use `0L` for chromosome to extract it from the SNP ID.
- Use `-1L` for chromosome and supply `chromosome = "1"` to assign a fixed value.

```{r gen_snpcolumns, message = FALSE, warning = FALSE}
gen3bchrfile <- system.file("extdata", "set3b.chr.imp", package = "BinaryDosage")
bdfile3b_chr <- tempfile()
# chromosome is in column 1, SNP ID in column 2, location in column 3, etc.
gentobd(genfiles = c(gen3bchrfile, gen3bsample), bdfiles = bdfile3b_chr)
```

## impformat

Use `impformat` when the file stores fewer than three probability values per subject.

```{r gen_impformat, message = FALSE, warning = FALSE}
gen2bfile   <- system.file("extdata", "set2b.imp",    package = "BinaryDosage")
sample2bfile <- system.file("extdata", "set2b.sample", package = "BinaryDosage")
bdfile2b <- tempfile()
gentobd(genfiles = c(gen2bfile, sample2bfile),
        snpcolumns = c(1L, 3L, 2L, 4L, 5L),
        impformat = 1L,
        bdfiles = bdfile2b)
```

## gz

Add `gz = TRUE` for compressed GEN files. The sample file must not be compressed.

```{r gen_gz, message = FALSE, warning = FALSE}
gen4bgzfile  <- system.file("extdata", "set4b.imp.gz", package = "BinaryDosage")
sample4bfile <- system.file("extdata", "set4b.sample", package = "BinaryDosage")
bdfile4b_gz  <- tempfile()
gentobd(genfiles = c(gen4bgzfile, sample4bfile),
        snpcolumns = c(1L, 2L, 4L, 5L, 6L),
        startcolumn = 7L,
        impformat = 2L,
        gz = TRUE,
        bdfiles = bdfile4b_gz)
```

---

# Merging Legacy Binary Dosage Files

`bdmerge` combines multiple legacy binary dosage files that share the same SNP set but have different subjects. Matching is done by SNP ID.

## Parameters

| Parameter | Description |
|-----------|-------------|
| `mergefiles` | Output file name(s) |
| `bdfiles` | Character vector of input binary dosage file names |
| `format` | Output binary dosage format |
| `subformat` | Output subformat |
| `onegroup` | If FALSE, per-source-file SNP statistics are retained in the header |
| `bdoptions` | Per-SNP statistics to calculate for the merged file |
| `snpjoin` | `"inner"` or `"outer"` join on SNPs (default inner) |

## Example

```{r merge, message = FALSE, warning = FALSE}
bd1afile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
bd1bfile <- system.file("extdata", "vcf1b.bdose", package = "BinaryDosage")
bd1merged <- tempfile()

bdmerge(mergefiles = bd1merged, bdfiles = c(bd1afile, bd1bfile))

bd1ainfo   <- getbdinfo(bd1afile)
bd1binfo   <- getbdinfo(bd1bfile)
bd1merinfo <- getbdinfo(bd1merged)

cat("Set 1a subjects:", nrow(bd1ainfo$samples), "\n")
cat("Set 1b subjects:", nrow(bd1binfo$samples), "\n")
cat("Merged subjects:", nrow(bd1merinfo$samples), "\n")
```

---

# Reading Legacy Binary Dosage Files

## getbdinfo

`getbdinfo` reads the header of a legacy binary dosage file and returns a list of class `"genetic-info"` that is required by `bdapply` and `getsnp`.

For Format 4 files, pass the single file name. For Formats 1–3, pass a character vector of length 3: binary dosage file, family file, and map file.

```{r getbdinfo, message = FALSE, warning = FALSE}
bd1ainfo <- getbdinfo(bdfiles = bd1afile)
```

The returned list contains:

| Element | Description |
|---------|-------------|
| `filename` | Full path to the binary dosage file |
| `usesfid` | Whether family IDs are present |
| `samples` | Data frame with columns `fid` and `sid` |
| `onechr` | TRUE if all SNPs are on one chromosome |
| `snpidformat` | Integer encoding the SNP ID format (0–3) |
| `snps` | Data frame: `chromosome`, `location`, `snpid`, `reference`, `alternate` |
| `snpinfo` | Named list: `aaf`, `maf`, `avgcall`, `rsq` (may be empty) |
| `datasize` | Byte sizes of per-SNP data blocks |
| `indices` | Byte offsets of per-SNP data blocks |
| `additionalinfo` | Format-specific metadata (class `"bdose-info"`) |

The `additionalinfo` element contains:

| Element | Description |
|---------|-------------|
| `format` | Binary dosage format (1–4) |
| `subformat` | Binary dosage subformat |
| `headersize` | Header size in bytes |
| `numgroups` | Number of merged source files |
| `groups` | Subject count per source file |

## bdapply

`bdapply` applies a user-supplied function to every SNP in a legacy binary dosage file. It returns a list of length equal to the number of SNPs.

The user function must accept `dosage`, `p0`, `p1`, and `p2` as its first four parameters. Additional parameters can be passed through `...`.

```{r bdapply, message = FALSE, warning = FALSE}
calculateaaf <- function(dosage, p0, p1, p2) {
  mean(dosage, na.rm = TRUE) / 2
}

bd1ainfo <- getbdinfo(bd1afile)
aaf_vals <- unlist(bdapply(bdinfo = bd1ainfo, func = calculateaaf))

aaf_table <- data.frame(snpid = bd1ainfo$snps$snpid, aaf = aaf_vals)
knitr::kable(aaf_table, caption = "Alternate allele frequencies", digits = 4)
```

The built-in function `getaaf` provides this calculation in the form expected by `bdapply`.

```{r bdapply_getaaf, message = FALSE, warning = FALSE}
aaf_vals2 <- unlist(bdapply(bdinfo = bd1ainfo, func = getaaf))
```

## getsnp

`getsnp` returns the dosage and genotype probabilities for a single SNP, identified by 1-based index or SNP ID string.

```{r getsnp, message = FALSE, warning = FALSE}
snp_data <- data.frame(getsnp(bdinfo = bd1ainfo, snp = "1:12000:T:C", dosageonly = FALSE))
snp_table <- cbind(sid = bd1ainfo$samples$sid, snp_data)
knitr::kable(head(snp_table, 10), caption = "SNP 1:12000:T:C (first 10 subjects)", digits = 4)
```

Set `dosageonly = TRUE` (the default) to return only the dosage vector.

---

# Working Directly with VCF and GEN Files

For cases where conversion to binary dosage is not needed, `getvcfinfo` / `vcfapply` and `getgeninfo` / `genapply` allow functions to be applied directly to source files.

## VCF files

`getvcfinfo` returns a `"genetic-info"` list for a VCF file. The `index` parameter pre-indexes the file for faster sequential reading (not available for compressed files).

```{r vcfinfo, message = FALSE, warning = FALSE}
vcfinfo <- getvcfinfo(vcffiles = vcf1afile, index = TRUE)
aaf_vcf  <- unlist(vcfapply(vcfinfo = vcfinfo, func = getaaf))

aaf_compare <- data.frame(snpid = vcfinfo$snps$snpid, aaf = aaf_vcf)
knitr::kable(aaf_compare, caption = "AAF from VCF via vcfapply", digits = 4)
```

The additional information in a `"vcf-info"` list includes:

| Element | Description |
|---------|-------------|
| `gzipped` | Whether the file is gzip-compressed |
| `headerlines` | Number of header lines |
| `headersize` | Header size in bytes |
| `quality`, `filter`, `info`, `format` | Per-SNP VCF metadata columns |
| `datacolumns` | Data frame describing FORMAT field structure |

## GEN files

`getgeninfo` and `genapply` work analogously for GEN files. The parameters mirror those of `gentobd`.

```{r geninfo, message = FALSE, warning = FALSE}
geninfo <- getgeninfo(genfiles = c(gen3bchrfile, gen3bsample), index = TRUE)
aaf_gen  <- unlist(genapply(geninfo = geninfo, func = getaaf))

aaf_gen_table <- data.frame(snpid = geninfo$snps$snpid, aaf = aaf_gen)
knitr::kable(aaf_gen_table, caption = "AAF from GEN via genapply", digits = 4)
```

The additional information in a `"gen-info"` list includes:

| Element | Description |
|---------|-------------|
| `gzipped` | Whether the file is gzip-compressed |
| `headersize` | Header size in bytes |
| `format` | Number of genotype probabilities per subject (1, 2, or 3) |
| `startcolumn` | Column where genotype data begins |
| `sep` | Column separator |

---

# Full Workflow Example

The following example reproduces the complete legacy workflow: convert two VCF files and two GEN files to binary dosage, merge each pair, apply a function to both merged files, extract one SNP, and verify that the VCF and GEN results agree.

```{r full_workflow, message = FALSE, warning = FALSE}
library(BinaryDosage)

# --- Input files ---
vcf1afile   <- system.file("extdata", "set1a.vcf",    package = "BinaryDosage")
vcf1ainfo   <- system.file("extdata", "set1a.info",   package = "BinaryDosage")
vcf1bfile   <- system.file("extdata", "set1b.vcf",    package = "BinaryDosage")
vcf1binfo   <- system.file("extdata", "set1b.info",   package = "BinaryDosage")
gen3afile   <- system.file("extdata", "set3a.imp",    package = "BinaryDosage")
gen3asample <- system.file("extdata", "set3a.sample", package = "BinaryDosage")
gen3bfile   <- system.file("extdata", "set3b.imp",    package = "BinaryDosage")
gen3bsample <- system.file("extdata", "set3b.sample", package = "BinaryDosage")

# --- Convert to binary dosage ---
bdfile1a <- tempfile(); bdfile1b <- tempfile()
bdfile3a <- tempfile(); bdfile3b <- tempfile()

vcftobdlegacy(vcffiles = c(vcf1afile, vcf1ainfo), bdfiles = bdfile1a)
vcftobdlegacy(vcffiles = c(vcf1bfile, vcf1binfo), bdfiles = bdfile1b)
gentobd(genfiles = c(gen3afile, gen3asample), snpcolumns = c(0L, 2L:5L), bdfiles = bdfile3a)
gentobd(genfiles = c(gen3bfile, gen3bsample), snpcolumns = c(0L, 2L:5L), bdfiles = bdfile3b)

# --- Merge ---
mergebd1 <- tempfile(); mergebd3 <- tempfile()
bdmerge(mergefiles = mergebd1, bdfiles = c(bdfile1a, bdfile1b))
bdmerge(mergefiles = mergebd3, bdfiles = c(bdfile3a, bdfile3b))

# --- Apply function ---
mergebd1info <- getbdinfo(mergebd1)
mergebd3info <- getbdinfo(mergebd3)

aaf1 <- unlist(bdapply(mergebd1info, getaaf))
aaf3 <- unlist(bdapply(mergebd3info, getaaf))

aaf <- cbind(mergebd1info$snps[, c("chromosome", "location", "snpid")],
             aaf_vcf = aaf1, aaf_gen = aaf3)
knitr::kable(aaf, caption = "AAF: VCF vs GEN sources", digits = 4, row.names = FALSE)
```

```{r full_workflow_snp, message = FALSE, warning = FALSE}
# --- Extract SNP 6 ---
set1snp6 <- getsnp(mergebd1info, 6L)
set3snp6 <- getsnp(mergebd3info, 6L)

snpdf <- data.frame(
  subjectid = mergebd1info$samples$sid,
  dosage_vcf = unlist(set1snp6),
  dosage_gen = unlist(set3snp6)
)
knitr::kable(head(snpdf, 10),
             caption = "SNP 6 dosage: VCF vs GEN (first 10 subjects)",
             digits = 4, row.names = FALSE)
```

The alternate allele frequencies and dosage values are identical between the VCF-derived and GEN-derived files, confirming the two sources contain the same underlying data.

---
title: "Merging Files"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Merging Files}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(BinaryDosage)
```

**Note:** `bdmerge` is a legacy function that operates on formats 1–4 only.
For Format 5 files, use `mergebd` instead. See the
[Merging Format 5 Binary Dosage Files](mergingbd5files.html) vignette for
details.

---

Quite often subjects have their genotypes imputed in batches. The files returned by these imputation can be converted into binary dosage files. These binary files can be merged into a single file if they have the same SNPs and different subjects using the bdmerge routine.

## bdmerge

The bdmerge routine takes the following parameters

- mergefiles - A character vector of the binary dosage file, family file, and map file names
- format - Integer value indicating which format of the binary dosage file should be used for the merged files
- subformat - Integer value indicating which subformat should be used for the merged files
- bdfiles - A character vector of the binary dosage files to merge
- famfiles - Character vector of the family files associated with the binary dosage files to merge
- mapfiles - Character vector of the map files associated with the binary dosage files to merge
- onegroup - Logical value indicating if the binary dosage saves SNP summary information about each merged file
- bdoptions - Character vector indicating on which SNP information should be evaluated for the merged files. This cannot be used if onegroup is set to FALSE
- snpjoin - Character value indicating if an inner or outer join is done for the SNPs

The following code merges *vcf1a.bdose* and *vcf1b.bdose* into one binary dosage file. It then displays the number of subjects in each file.

``` {r, eval = T, echo = T, message = F, warning = F, tidy = T}
bd1afile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
bd1bfile <- system.file("extdata", "vcf1b.bdose", package = "BinaryDosage")
bd1file <- tempfile()

bdmerge(mergefiles = bd1file, bdfiles = c(bd1afile, bd1bfile))

bd1ainfo <- getbdinfo(bd1afile)
bd1binfo <- getbdinfo(bd1bfile)
bd1info <- getbdinfo(bd1file)

nrow(bd1ainfo$samples)
nrow(bd1binfo$samples)
nrow(bd1info$samples)
```

---
title: "Using Binary Dosage files"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Using Binary Dosage files}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(BinaryDosage)
```
The following routines are available for accessing information contained in binary dosage files

- <span style="font-family:Courier">getbdinfo</span>
- <span style="font-family:Courier">bdapply</span>
- <span style="font-family:Courier">getsnp</span>

## getbdinfo

The <span style="font-family:Courier">getbdinfo</span> routine returns information about a binary dosage file. For more information about the data returned see [Genetic File Information](geneticfileinfo.html). This information needs to be passed to <span style="font-family:Courier">bdapply</span and <span style="font-family:Courier">getsnp</span> routines so it can read the binary dosage file.

The only parameter used by <span style="font-family:Courier">getbdinfo</span> is <span style="font-family:Courier">bdfiles</span>. This parameter is a character vector. If the format of the binary dosage file is 1, 2, or 3, this must be a character vector of length 3 with the following values, binary dosage file name, family file name, and map file name. If the format of the binary dosage file is 4 then the parameter value is a single character value with the name of the binary dosage file.

The following code gets the information about the binary dosage file *vcf1a.bdose*.

``` {r, eval = T, echo = T, message = F, warning = F, tidy = T}
bd1afile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
bd1ainfo <- getbdinfo(bdfiles = bd1afile)
```

## bdapply

The <span style="font-family:Courier">bdapply</span> routine applies a function to all SNPs in the binary dosage file. The routine returns a list with length equal to the number of SNPs in the binary dosage file. Each element in the list is the value returned by the user supplied function. The routine takes the following parameters.

- <span style="font-family:Courier">bdinfo</span> - list with information about the binary dosage file returned by <span style="font-family:Courier">getbdinfo</span>.
- <span style="font-family:Courier">func</span> - user supplied function to be applied to each SNP in the VCF file.
- <span style="font-family:Courier">...</span> - additional parameters needed by the user supplied function

The user supplied function must have the following parameters.

- <span style="font-family:Courier">dosage</span> - A numeric vector with the dosage values for each subject.
- <span style="font-family:Courier">p0</span> - A numeric vector with the probabilities the subject has no alternate alleles for each subject.
- <span style="font-family:Courier">p1</span> - A numeric vector with the probabilities the subject has one alternate allele for each subject.
- <span style="font-family:Courier">p2</span> - A numeric vector with the probabilities the subject has two alternate alleles for each subject.

The user supplied function can have other parameters. These parameters need to passed to the <span style="font-family:Courier">bdapply</span> routine.

There is a function in the <span style="font-family:Courier">BinaryDosage</span> package named <span style="font-family:Courier">getaaf</span> that calculates the alternate allele frequencies and is the format needed by <span style="font-family:Courier">vcfapply</span> routine. The following uses <span style="font-family:Courier">getaaf</span> to calculate the alternate allele frequency for each SNP in the *vcf1a.bdose* file using the <span style="font-family:Courier">bdapply</span> routine.

``` {r, eval = T, echo = T, message = F, warning = F, tidy = T}
aaf <- unlist(bdapply(bdinfo = bd1ainfo, func = getaaf))

altallelefreq <- data.frame(SNP = bd1ainfo$snps$snpid, aafcalc = aaf)
knitr::kable(altallelefreq, caption = "Information vs Calculated aaf", digits = 3)
```

## getsnp

The <span style="font-family:Courier">getsnp</span> routine return the dosage and genotype probabilities for each subject for a given SNP in a binary dosage file.

The routine takes the following parameters.

- <span style="font-family:Courier">bdinfo</span> - list with information about the binary dosage file returned by <span style="font-family:Courier">getbdinfo</span>.
- <span style="font-family:Courier">snp</span> - the SNP to return information about. This can either be the index of the SNP in the binary dosage file or its SNP ID.
- <span style="font-family:Courier">dosageonly</span> - a logical value indicating if only the dosage values are returned without the genotype probabilities. The default value is TRUE indicating that only the dosage values are returned.

The following code returns the dosage values and the genotype probabilities for SNP 1:12000:T:C from the *vcf1a.bdose"
binary dosage file.

``` {r, eval = T, echo = T, message = F, warning = F, tidy = T}
snp3 <- data.frame(getsnp(bdinfo = bd1ainfo, "1:12000:T:C", FALSE))

knitr::kable(snp3[1:20,], caption = "SNP 1:12000:T:C", digits = 3)
```

---
title: "Genetic File Information"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Genetic File Information}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(BinaryDosage)
```

The routines <span style="font-family:Courier">getbdinfo</span>, <span style="font-family:Courier">getvcfinfo</span>, and <span style="font-family:Courier">getgeninfo</span> return a list with information about the data in the files. The list returned by each of these routines a section common to them all and a list <span style="font-family:Courier">additionalinfo</span> that is specific to the file type.

## Common section

The common section has the following elements

- filename - Character value with the complete path and file name of the file with the genetic data
- usesfid - Logical value indicating if the subject data has family IDs.
- samples - Data frame containing the following information about the subjects
  + fid - Character value with family IDs
  + sid - Character value with the individual IDs
- onchr - Logical value indicating if all the SNPs are on the same chromosome
- snpidformat - Integer indicating the format of the SNP IDs as follows
  + 0 - Unknown for VCF and GEN files or user specified for binary dosage files
  + 1 - chr:pos
  + 2 - chr:pos:ref:alt
  + 3 - chr:pos_ref_alt
- snps - Data frame containing the following values
  + chromosome - Character value indicating what chromosome the SNP is on
  + location - Integer value with the location of the SNP on the chromosome
  + snpid - Character value with the ID of the SNP
  + reference - Character value of the reference allele
  + alternate - Character value of the alternate allele
- snpinfo - List that contain the following information
  + aaf - numeric vector with the alternate allele frequencies
  + maf - numeric vector with the minor allele frequencies
  + avgcall - Numeric vector with the imputation average call
  + rsq - Numeric vector with the imputation r squared value
- datasize - Numeric vector indicating the size of data in the file for each SNP
- indices - Numeric vector indicating the starting location in the file for each SNP

The list returned has its class value set to "genetic-info".

The <span style="font-family:Courier">datasize</span> and <span style="font-family:Courier">indices</span> values are only returned if the parameter <span style="font-family:Courier">index</span> is set equal to <span style="font-family:Courier">TRUE</span>

## Binary Dosage Additional Information

The additional information returned for binary dosage files contains the following information.

- format - numeric value with the format of the binary dosage file
- subformat - numeric value with the subformat of the binary dosage file
- headersize - integer value with the size of the header in the binary dosage file
- numgroups - integer value of the number of groups of subjects in the binary dosage file. This is usually the number of binary dosage files merged together to form the file
- groups - integer vector with size of each of the groups

This list has its class value set to "bdose-info".

## VCF File Additional Information

The additional information returned for VCF files contains the following information.

- gzipped - Logical value indicating if the file has been compressed using gzip
- headerlines - Integer value indicating the number of lines in the header
- headersize - Numeric value indicating the size of the header in bytes
- quality - Character vector containing the values in QUALITY column
- filter - Character vector containing the values in the FILTER column
- info - Character vector containing the values in the INFO column
- format - Character vector containing the values in the FORMAT column
- datacolumns - Data frame summarizing the entries in the FORMAT value containing the following information
  + numcolumns - Integer value indicating the number of values in the FORMAT value
  + dosage - Integer value indicating the column containing the dosage value
  + genotypeprob - Integer value indicating the column containing the genotype probabilities
  + genotype - Integer value indicating the column containing the genotype call

This list has its class value set to "vcf-info".

The values for quality, filter, info, and format can have a length of 0 if all the values are missing. They will have a length of 1 if all the values are equal. The number of rows in the datacolumns data frame will be equal to the length of the format value.

## GEN File Additional Information

The additional information returned for GEN files contains the following information.

- gzipped - Logical value indicating if the GEN file is compressed using gz
- headersize - Integer value indicating the size of the header in bytes
- format - Integer value indicating the number of genotype probabilities for each subject with the following meanings
  + 1 - Dosage only
  + 2 - $\Pr(g=0)$ and $\Pr(g=1)$
  + 3 - $\Pr(g=0)$, $\Pr(g=1)$, and $\Pr(g=2)$
- startcolumn - Integer value indicating in which column the genetic data starts
- sep - Character value indicating what value separates the columns

$g$ indicates the number of alternate alleles the subject has.

This list has its class value set to "gen-info".

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/WriteBinaryDosage.R
\name{getmaf}
\alias{getmaf}
\title{Calculate minor allele frequency}
\usage{
getmaf(dosage, p0, p1, p2)
}
\arguments{
\item{dosage}{Dosage values}

\item{p0}{Pr(g=0) - unused}

\item{p1}{Pr(g=1) - unused}

\item{p2}{Pr(g=2) - unused}
}
\value{
Minor allele frequency
}
\description{
Routine to calculate the minor allele frequency given the dosages.
Missing values for dosage ignored. This function is used internally and
is exported for use in examples. Note: The minor allele in one data set
may be different from another data set. This can make comparing minor
allele frequencies between data sets nonsensical.
}
\examples{
# Get information about binary dosage file
bdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
bdinfo <- getbdinfo(bdfiles = bdfile)
snp1 <- getsnp(bdinfo = bdinfo, 1)
maf <- getmaf(snp1$dosage)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getinfo.R
\name{getbdinfo}
\alias{getbdinfo}
\title{Get information about a binary dosage file}
\usage{
getbdinfo(bdfiles)
}
\arguments{
\item{bdfiles}{Vector of file names. For Format 5 files, a single .bdose
file name; the companion .bdi metadata file is read automatically from
\code{paste0(bdfiles[1], ".bdi")}. For format 4, a single file name. For
formats 1, 2, and 3, a vector of three file names: the binary dosage file,
the family information file, and the SNP information file.}
}
\value{
List with information about the binary dosage file.
This includes family and subject IDs along with
a list of the SNPs in the file. Other information needed
to read the file is also included.
}
\description{
Routine to return information about a binary dosage file.
This information is used by other routines to
allow for quicker extraction of values from the
file.
}
\examples{
vcf1abdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
bdinfo <- getbdinfo(bdfiles = vcf1abdfile)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format5.R
\name{getbd5snp}
\alias{getbd5snp}
\title{Read a SNP from a Format 5 binary dosage file}
\usage{
getbd5snp(bd5info, snp)
}
\arguments{
\item{bd5info}{Object returned by \code{getbdinfo}.}

\item{snp}{The SNP to retrieve: either a 1-based integer index or a
character SNP ID matching a value in \code{bd5info$snps$snpid}.}
}
\value{
A list with four numeric vectors, each of length n_samples:
\describe{
\item{dosage}{DS values in [0, 2]; NA = missing.}
\item{p0}{P(g=0) values in [0, 1]; NA = missing.}
\item{p1}{P(g=1) values in [0, 1]; NA = missing.}
\item{p2}{P(g=2) values in [0, 1]; NA = missing.}
}
}
\description{
Seeks to the SNP's compressed block in the .bdose file, decompresses it,
and returns the dosage and genotype probabilities for all samples.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Merge.R
\name{bdmerge}
\alias{bdmerge}
\title{Merge binary dosage files together}
\usage{
bdmerge(
  mergefiles,
  format = 4,
  subformat = 0L,
  bdfiles,
  famfiles = character(),
  mapfiles = character(),
  onegroup = TRUE,
  bdoptions = character(),
  snpjoin = "inner"
)
}
\arguments{
\item{mergefiles}{Vector of file names for the merged binary
files. The first is the
binary dosage data containing the dosages and genetic
probabilities. The second file name is the family information
file. The third file name is the SNP information file.
The family and SNP information files are not used if the
binary dosage file is in format 4. For this format the
family and SNP information are in the file with the dosages
and genetic probabilities.}

\item{format}{The format of the output binary dosage file.
Allowed values are 1, 2, 3, and 4. The default value is 4.
Using the default value is recommended.}

\item{subformat}{The subformat of the format of the output
binary dosage file. A value of 1 or 3 indicates that only the
dosage value is saved. A value of 2 or 4 indicates
the dosage and genetic probabilities will be output. Values
of 3 or 4 are only allowed with formats 3 and 4. If a value
of zero if provided, and genetic probabilities are in the vcf
file, subformat 2 will be used for formats 1 and 2, and
subformat 4 will be used for formats 3 and 4. If the vcf file
does not contain genetic probabilities, subformat 1 will be
used for formats 1 and 2, and subformat 3 will be used for
formats 3 and 4. The default value is 0.}

\item{bdfiles}{Vector of binary dosage file names to be merged.}

\item{famfiles}{Vector of family file names that correspond to
the names in bdfiles. If the binary dosage files are all in format
4, this may be an empty character array. Default value is character().}

\item{mapfiles}{Vector of map file names that correspond to the
names in bdfiles. If the binary dosage files are all in format
4, this may be an empty character array. Default value is character().}

\item{onegroup}{Indicator to combine all the samples in one group.
If this is FALSE, the groups in each binary dosage file are
maintained and any binary dosage file with one group is made into
its own group. Default value is TRUE.}

\item{bdoptions}{Options indicating what information to calculate
and store for each SNP. These can be aaf, maf, and rsq. This option
is only available if format is equal to 4 and onegroup is TRUE.
Default value is character().}

\item{snpjoin}{Character value that can be either "inner" or "outer".
This indicates whether to do an inner or outer join of the SNPs in
each binary dosage file. Default value is "inner".}
}
\value{
None
}
\description{
Routine to merge binary dosage files together. The files
don't have to be in the same format. They will be merged
into a file with the format specified. Information about
the SNPs, aaf, maf, avgcall, rsq, can be maintained for
each file, or recalculated for the merged set.
}
\examples{
bdvcf1afile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
bdvcf1bfile <- system.file("extdata", "vcf1b.bdose", package = "BinaryDosage")
mergefiles <- tempfile()

BinaryDosage:::bdmerge(mergefiles = mergefiles,
                       bdfiles = c(bdvcf1afile, bdvcf1bfile),
                       bdoptions = "maf")
bdinfo <- getbdinfo(mergefiles)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/WriteBinaryDosage.R
\name{getrsq}
\alias{getrsq}
\title{Calculate imputation r squared}
\usage{
getrsq(dosage, p0, p1, p2)
}
\arguments{
\item{dosage}{Dosage values}

\item{p0}{Pr(g=0) - unused}

\item{p1}{Pr(g=1) - unused}

\item{p2}{Pr(g=2)}
}
\value{
Imputation r squared
}
\description{
Routine to calculate the imputation r squared given the dosages
and Pr(g=2).
This is an estimate for the imputation r squared returned from
minimac and impute2. The r squared values are calculated slightly
differently between the programs. This estimate is based on the
method used by minimac. It does well for minor allele frequencies
above 5\%. This function is used internally and is exported for
use in examples.
}
\examples{
# Get information about binary dosage file
bdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
bdinfo <- getbdinfo(bdfiles = bdfile)
snp1 <- getsnp(bdinfo = bdinfo, 1, dosageonly = FALSE)
rsq <- BinaryDosage:::getrsq(snp1$dosage, p2 = snp1$p2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert.R
\name{updatebd}
\alias{updatebd}
\title{Update a binary dosage file to Format 5}
\usage{
updatebd(bdfiles, bdose_file)
}
\arguments{
\item{bdfiles}{Vector of file names for the binary dosage file. Format 4
files require one file name. Formats 1, 2, and 3 require three file names:
the binary dosage file, the family file, and the map file.}

\item{bdose_file}{Path for the output .bdose file. The companion .bdi
metadata file is written to \code{paste0(bdose_file, ".bdi")}.}
}
\value{
NULL (invisibly)
}
\description{
Reads a binary dosage file in format 1, 2, 3, or 4, detects the format
automatically, and converts it to a Format 5 file pair by calling the
appropriate conversion routine. If the source file does not contain genotype
probabilities, those values are stored as missing in the output.
}
\examples{
vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")
bdfile <- tempfile()
suppressWarnings(
  vcftobdlegacy(vcffiles = vcf1afile,
                bdfiles = bdfile,
                format = 4L)
)
bdose_file <- tempfile(fileext = ".bdose")
updatebd(bdfiles    = bdfile,
         bdose_file = bdose_file)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getinfo.R
\name{getgeninfo}
\alias{getgeninfo}
\title{Get information about a gen, impute2, file}
\usage{
getgeninfo(
  genfiles,
  snpcolumns = 1L:5L,
  startcolumn = 6L,
  impformat = 3L,
  chromosome = character(),
  header = c(FALSE, TRUE),
  gz = FALSE,
  index = TRUE,
  snpidformat = 0L,
  sep = c("\\t", "\\t")
)
}
\arguments{
\item{genfiles}{A vector of file names.
The first is the name of the gen file. The
second is name of the sample file that contains
the subject information.}

\item{snpcolumns}{Column numbers containing chromosome,
snpid, location, reference allele, alternate allele,
respectively. This must be an integer vector. All
values must be positive except for the chromosome.
The value for the chromosome may be -1 or -0.
-1 indicates that the chromosome value is passed to
the routine using the chromosome parameter.
0 indicates that the chromosome value is in the snpid
and that the snpid has the format chromosome:other_data.
Default value is c(1L, 2L, 3L, 4L, 5L).}

\item{startcolumn}{Column number of first column with
genetic probabilities or dosages. Must
be an integer value. Default value is 6L.}

\item{impformat}{Number of genetic data values per
subject. 1 indicates dosage only, 2 indicates P(g=0)
and P(g=1) only, 3 indicates P(g=0), P(g=1), and
P(g=2). Default value is 3L.}

\item{chromosome}{Chromosome value to use if the
first value of the snpcolumns is equal to 0.
Default value is character().}

\item{header}{Indicators if the gen and sample files
have headers. If the gen file does not have a
header. A sample file must be included.
Default value is c(FALSE, TRUE).}

\item{gz}{Indicator if file is compressed using gzip.
Default value is FALSE.}

\item{index}{Indicator if file should be indexed. This
allows for faster reading of the file. Indexing a gzipped
file is not supported.
Default value is TRUE.}

\item{snpidformat}{Format to change the snpid to.
0 indicates to use the snpid format in the file.
1 indicates to change the snpid into chr:pos,
2 indicates to change the snpid into chr:pos:ref:alt,
3 indicates to change the snpid into chr:pos_ref_alt,
Default value is 0.}

\item{sep}{Separators used in the gen file and sample files,
respectively. If only value is provided it is used for both
files. Default value is c(\code{"\\t"}, \code{"\\t"})}
}
\value{
List with information about the gen file.
This includes family and subject IDs along with
a list of the SNPs in the file. Other information needed
to read the file is also included.
}
\description{
Routine to return information about a gen file.
This information is used by other routines to
allow for quicker extraction of values from the
file.
}
\examples{
# Get file names of th gen and sample file
gen3afile <- system.file("extdata", "set3a.imp", package = "BinaryDosage")
gen3ainfo <- system.file("extdata", "set3a.sample", package = "BinaryDosage")

# Get the information about the gen file
geninfo <- getgeninfo(genfiles = c(gen3afile, gen3ainfo),
                      snpcolumns = c(0L, 2L:5L))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert.R
\name{bd1tobd5}
\alias{bd1tobd5}
\title{Convert a format 1 binary dosage file to Format 5}
\usage{
bd1tobd5(bdfiles, bdose_file)
}
\arguments{
\item{bdfiles}{Vector of three file names for the format 1 binary dosage
file: the binary dosage file, the family file, and the map file.}

\item{bdose_file}{Path for the output .bdose file. The companion .bdi
metadata file is written to \code{paste0(bdose_file, ".bdi")}.}
}
\value{
NULL (invisibly)
}
\description{
Reads a format 1 binary dosage file (subformat 1 or 2) and converts it to
a Format 5 file pair. If the source file does not contain genotype
probabilities (subformat 1), those values are stored as missing in the
output.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getsnp.R
\name{getsnp}
\alias{getsnp}
\title{Read SNP data from a binary dosage file}
\usage{
getsnp(bdinfo, snp, dosageonly = TRUE)
}
\arguments{
\item{bdinfo}{Information about a binary dosage file return
from getbdinfo}

\item{snp}{The SNP to read the information about. This may
be the SNP ID or the index of the SNP in the snps dataset in
the bdinfo list}

\item{dosageonly}{Indicator to return the dosages only or the
dosages allowing with the genetic probabilities.
Default value is TRUE}
}
\value{
A list with either the dosages or the dosages and the genetic
probabilities.
}
\description{
Routine to read the dosage and genetic probabilities about
a SNP from a binary dosage file
}
\examples{
# Get the information about the file
vcf1abdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
bdinfo <- getbdinfo(bdfiles = vcf1abdfile)

# Read the first SNP
getsnp(bdinfo, 1, FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert.R
\name{vcftobdlegacy}
\alias{vcftobdlegacy}
\title{Convert a VCF file to a binary dosage file}
\usage{
vcftobdlegacy(
  vcffiles,
  gz = FALSE,
  bdfiles,
  format = 4L,
  subformat = 0L,
  snpidformat = 0,
  bdoptions = character(0)
)
}
\arguments{
\item{vcffiles}{A vector of file names.
The first is the name of the vcf file. The
second is name of the file that contains information
about the imputation of the SNPs. This file is produced
by minimac 3 and 4.}

\item{gz}{Indicator if VCF file is compressed using gzip.
Default value is FALSE.}

\item{bdfiles}{Vector of names of the output files.
The binary dosage file name is first. The family and
map files follow. For format 4, no family and map file
names are needed.}

\item{format}{The format of the output binary dosage file.
Allowed values are 1, 2, 3, and 4. The default value is 4.
Using the default value is recommended.}

\item{subformat}{The subformat of the format of the output
binary dosage file. A value of 1 or 3 indicates that only the
dosage value is saved. A value of 2 or 4 indicates
the dosage and genetic probabilities will be output. Values
of 3 or 4 are only allowed with formats 3 and 4. If a value
of zero if provided, and genetic probabilities are in the vcf
file, subformat 2 will be used for formats 1 and 2, and
subformat 4 will be used for formats 3 and 4. If the vcf file
does not contain genetic probabilities, subformat 1 will be
used for formats 1 and 2, and subformat 3 will be used for
formats 3 and 4. The default value is 0.}

\item{snpidformat}{The format that the SNP ID will be saved as.
-1 SNP ID not written
0 - same as in the VCF file
1 - chr:pos
2 - chr:pos:ref:alt
If snpidformat is 1 and the VCF file uses format 2, an error is
generated. Default value is 0.}

\item{bdoptions}{Character array containing any of the following
value, "aaf", "maf", "rsq". The presence of any of these
values indicates that the specified values should be
calculates and stored in the binary dosage file. These values only
apply to format 4.}
}
\value{
None
}
\description{
Routine to read information from a VCF file and create
a binary dosage file. The function is designed to use
files return from the Michigan Imputation Server but will
run on other VCF files if they contain dosage and genetic
probabilities. Note: This routine can take a long time to
run if the VCF file is large.
}
\examples{
# Find the vcf file names
vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")
vcf1ainfo <- system.file("extdata", "set1a.info", package = "BinaryDosage")
bdfiles <- tempfile()
# Convert the file
vcftobdlegacy(vcffiles = c(vcf1afile, vcf1ainfo), bdfiles = bdfiles)
# Verify the file was written correctly
bdinfo <- getbdinfo(bdfiles)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format5.R
\name{getbd5snp_con}
\alias{getbd5snp_con}
\title{Read a Format 5 SNP using a persistent open connection}
\usage{
getbd5snp_con(bd5info, snp, dosage, p0, p1, p2, bd5con)
}
\arguments{
\item{bd5info}{Object returned by \code{getbdinfo}.}

\item{snp}{1-based integer index or character SNP ID.}

\item{dosage}{Pre-allocated \code{numeric(n_samples)} vector.}

\item{p0}{Pre-allocated \code{numeric(n_samples)} vector.}

\item{p1}{Pre-allocated \code{numeric(n_samples)} vector.}

\item{p2}{Pre-allocated \code{numeric(n_samples)} vector.}

\item{bd5con}{Object returned by \code{openbd5con}.}
}
\value{
\code{NULL} invisibly.  \code{dosage}, \code{p0}, \code{p1}, and
\code{p2} are updated in place.
}
\description{
Like \code{getbd5snp_buf} but reuses an already-open file connection
instead of opening and closing it on every call.  Use \code{openbd5con}
before the loop and \code{closebd5con} (or let the finalizer handle it)
after.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format5.R
\name{getbd5snp_buf}
\alias{getbd5snp_buf}
\title{Read a Format 5 SNP into pre-allocated vectors (buffered variant)}
\usage{
getbd5snp_buf(bd5info, snp, dosage, p0, p1, p2)
}
\arguments{
\item{bd5info}{Object returned by \code{getbdinfo}.}

\item{snp}{1-based integer index or character SNP ID.}

\item{dosage}{Pre-allocated \code{numeric(n_samples)} vector.}

\item{p0}{Pre-allocated \code{numeric(n_samples)} vector.}

\item{p1}{Pre-allocated \code{numeric(n_samples)} vector.}

\item{p2}{Pre-allocated \code{numeric(n_samples)} vector.}
}
\value{
\code{NULL} invisibly.  \code{dosage}, \code{p0}, \code{p1}, and
\code{p2} are updated in place.
}
\description{
Like \code{getbd5snp} but writes results into caller-supplied vectors
instead of allocating new ones.  Intended for tight loops where thousands
of SNPs are read sequentially; pre-allocating the output vectors once
avoids repeated memory allocation.
}
\details{
The four output vectors \strong{must not} have more than one R binding at
the call site (no extra variables pointing to the same object); R's
copy-on-modify semantics would otherwise prevent in-place update.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format5.R
\name{closebd5con}
\alias{closebd5con}
\title{Close a persistent Format 5 connection}
\usage{
closebd5con(bd5con)
}
\arguments{
\item{bd5con}{Object returned by \code{openbd5con}.}
}
\value{
\code{NULL} invisibly.
}
\description{
Explicitly closes the connection opened by \code{openbd5con}.  Calling
this is optional — the finalizer will close it on garbage collection or
R exit — but explicit close is preferred to release the file handle
promptly.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getinfo.R
\name{getvcfinfo}
\alias{getvcfinfo}
\title{Get information about a vcf file}
\usage{
getvcfinfo(vcffiles, gz = FALSE, index = TRUE, snpidformat = 0L)
}
\arguments{
\item{vcffiles}{A vector of file names.
The first is the name of the vcf file. The
second is name of the file that contains information
about the imputation of the SNPs. This file is produced
by minimac 3 and 4.}

\item{gz}{Indicator if VCF file is compressed using gzip.
Default value is FALSE.}

\item{index}{Indicator if file should be indexed. This
allows for faster reading of the file. Indexing a gzipped
file is not supported.
Default value is TRUE.}

\item{snpidformat}{The format that the SNP ID will be saved as.
0 - same as in the VCF file
1 - chr:pos
2 - chr:pos:ref:alt
If snpidformat is 1 and the VCF file uses format 2, an error is
generated. Default value is 0.}
}
\value{
List containing information about the VCF file
to include file name, subject IDs, and information about
the SNPs. Indices for faster reading will be included
if index is set to TRUE
}
\description{
Routine to return information about a vcf file.
This information is used by other routines to
allow for quicker extraction of values from the
file.
}
\examples{
# Get file names of th vcf and infromation file
vcf1afile <- system.file("extdata", "set1a.vcf", package = "BinaryDosage")
vcf1ainfo <- system.file("extdata", "set1a.info", package = "BinaryDosage")

# Get the information about the vcf file
vcf1ainfo <- getvcfinfo(vcffiles = c(vcf1afile, vcf1ainfo))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subset.R
\name{subsetbd}
\alias{subsetbd}
\title{Subset a binary dosage file}
\usage{
subsetbd(
  bdfiles,
  bdose_file,
  minmaf = NULL,
  locations = NULL,
  startloc = NULL,
  endloc = NULL,
  subjectids = NULL
)
}
\arguments{
\item{bdfiles}{Vector of file names for the input binary dosage file.
Format 4 files require one file name. Formats 1, 2, and 3 require three
file names: the binary dosage file, the family file, and the map file.
Format 5 files require two file names: the .bdose file and the .bdinfo
file.}

\item{bdose_file}{Path for the output .bdose file. The companion .bdi
metadata file is written to \code{paste0(bdose_file, ".bdi")}.}

\item{minmaf}{Minimum minor allele frequency. SNPs whose MAF (computed over
the retained subjects) is below this value are excluded. Must be a single
numeric value between 0 and 0.5.}

\item{locations}{Integer or numeric vector of SNP base-pair locations to
retain. Cannot be used together with \code{startloc} and \code{endloc}.}

\item{startloc}{Start of the location range to retain (inclusive). Must be
used together with \code{endloc}. Cannot be used together with
\code{locations}.}

\item{endloc}{End of the location range to retain (inclusive). Must be used
together with \code{startloc}. Cannot be used together with
\code{locations}.}

\item{subjectids}{Character vector of subject IDs to retain.}
}
\value{
NULL (invisibly)
}
\description{
Creates a new Format 5 binary dosage file containing a subset of the SNPs
and/or subjects from an existing binary dosage file. The input file may be
in any format (1-5). At least one filtering criterion must be supplied, and
all supplied criteria must be met for a SNP or subject to be retained.
}
\examples{
bdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
bdinfo     <- getbdinfo(bdfile)
bdose_file <- tempfile(fileext = ".bdose")
subsetbd(bdfiles    = bdfile,
         bdose_file = bdose_file,
         subjectids = bdinfo$samples$sid[1:30])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mergebd.R
\name{mergebd}
\alias{mergebd}
\title{Merge Format 5 binary dosage files}
\usage{
mergebd(bdose_files, bdose_file)
}
\arguments{
\item{bdose_files}{Character vector of paths to the input .bdose files.
Must contain at least two entries. The companion .bdi file for each is
expected at \code{paste0(bdose_files[i], ".bdi")}.}

\item{bdose_file}{Path for the output .bdose file. The companion .bdi
metadata file is written to \code{paste0(bdose_file, ".bdi")}.}
}
\value{
NULL (invisibly)
}
\description{
Merges two or more Format 5 binary dosage files into a single Format 5
output file. The merge type is determined automatically:
}
\details{
\itemize{
\item If subject IDs do not overlap across files, a \strong{subject merge}
is performed: the output contains all subjects from every file and
the SNPs common to all files.
\item If SNP IDs do not overlap across files, a \strong{SNP merge} is
performed: the output contains all SNPs from every file and the
subjects common to all files.
}

If both subject IDs and SNP IDs overlap across files the merge cannot be
performed and an error is returned.

SNPs are identified by chromosome, position, reference allele, and alternate
allele.
}
\examples{
bdfile <- system.file("extdata", "vcf1a.bdose", package = "BinaryDosage")
bdinfo_src <- getbdinfo(bdfile)

# Create two format 5 files with non-overlapping subjects
bdose_a   <- tempfile(fileext = ".bdose")
bdose_b   <- tempfile(fileext = ".bdose")
bdose_tmp <- tempfile(fileext = ".bdose")
updatebd(bdfiles = bdfile, bdose_file = bdose_tmp)
subsetbd(bdfiles    = bdose_tmp,
         bdose_file = bdose_a,
         subjectids = bdinfo_src$samples$sid[1:30])
subsetbd(bdfiles    = bdose_tmp,
         bdose_file = bdose_b,
         subjectids = bdinfo_src$samples$sid[31:60])

bdose_out <- tempfile(fileext = ".bdose")
mergebd(bdose_files = c(bdose_a, bdose_b),
        bdose_file  = bdose_out)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format5.R
\name{vcftobd}
\alias{vcftobd}
\title{Convert a VCF file to Format 5 binary dosage files}
\usage{
vcftobd(
  vcffile,
  bdose_file,
  region = NULL,
  snpidformat = 0L,
  bdoptions = character(0)
)
}
\arguments{
\item{vcffile}{Path to the bgzipped, tabix-indexed VCF file.}

\item{bdose_file}{Path for the output .bdose file. The companion .bdi
metadata file is written to \code{paste0(bdose_file, ".bdi")}.}

\item{region}{Optional genomic region string in bcftools format
(e.g. \code{"chr21"} or \code{"chr21:1-5000000"}).  Requires a
tabix index.  Default \code{NULL} processes the entire file.}

\item{snpidformat}{Integer controlling how SNP IDs are stored.
\describe{
\item{-1}{Generate IDs as \code{chr:pos:ref:alt}; equivalent to 2 for
Format 5.}
\item{0}{Use the IDs as they appear in the VCF file (default).
Auto-detects format 1 or 2 if all IDs match.}
\item{1}{Store IDs as \code{chr:pos}.  An error is raised if the VCF
already uses \code{chr:pos:ref:alt} format, as information would be
lost.}
\item{2}{Store IDs as \code{chr:pos:ref:alt}.}
\item{3}{Store IDs as \code{chr:pos_ref_alt}.}
}}

\item{bdoptions}{Character vector specifying which per-SNP statistics to
store.  Any combination of \code{"aaf"} (alternate allele frequency),
\code{"maf"} (minor allele frequency), and \code{"rsq"} (imputation
r-squared).  For each statistic, the corresponding VCF INFO field is used
when present for the first SNP (AF, MAF, R2 respectively); otherwise the
value is calculated from the dosage data.  Default \code{character(0)}
stores no statistics.}
}
\value{
NULL (invisibly)
}
\description{
Reads the DS (dosage) and GP (genotype probabilities) FORMAT fields from a
bgzipped, tabix-indexed VCF file — as produced by imputation servers such
as the Michigan Imputation Server — and writes a pair of Format 5
BinaryDosage files.
}
\details{
The .bdose file begins with a 4-byte magic number followed by one
gzip-compressed block per SNP.  Each block contains the DS values for all
samples followed by the GP values, encoded as unsigned 16-bit integers
(round(value * 10000); 0xffff = missing).

The .bdi file is an RDS-serialised R list of class \code{"genetic-info"}
with the following elements:
\describe{
\item{filename}{Path to the associated .bdose file.}
\item{usesfid}{Logical; always FALSE for VCF-sourced files.}
\item{samples}{data.frame with columns \code{fid} (empty) and
\code{sid} (sample IDs).}
\item{onechr}{Logical; TRUE if all SNPs are on a single chromosome.}
\item{snpidformat}{Numeric; resolved SNP ID format (see \code{snpidformat}
parameter).}
\item{snps}{data.frame with columns chromosome, location, snpid,
reference, alternate.}
\item{snpinfo}{Named list of per-SNP annotations requested via
\code{bdoptions}.  Each element is a numeric vector of length equal to
the number of SNPs.  Values are read from the VCF INFO column when
available for the first SNP (AF for aaf, MAF for maf, R2 for rsq);
otherwise they are calculated from the dosage values.}
\item{additionalinfo}{List of class \code{"bdose-info"} with format,
subformat, headersize, numgroups, and groups.}
\item{datasize}{Integer vector of length 0 (unused in Format 5).}
\item{indices}{Numeric vector of byte offsets into .bdose, one per SNP.}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert.R
\name{gentobd}
\alias{gentobd}
\title{Convert a gen file to a binary dosage file}
\usage{
gentobd(
  genfiles,
  snpcolumns = 1L:5L,
  startcolumn = 6L,
  impformat = 3L,
  chromosome = character(),
  header = c(FALSE, TRUE),
  gz = FALSE,
  sep = "\\t",
  bdfiles,
  format = 4L,
  subformat = 0L,
  snpidformat = 0L,
  bdoptions = character(0)
)
}
\arguments{
\item{genfiles}{A vector of file names.
The first is the name of the gen file. The
second is name of the sample file that contains
the subject information.}

\item{snpcolumns}{Column numbers containing chromosome,
snpid, location, reference allele, alternate allele,
respectively. This must be an integer vector. All
values must be positive except for the chromosome.
The value for the chromosome may be -1 or -0.
-1 indicates that the chromosome value is passed to
the routine using the chromosome parameter.
0 indicates that the chromosome value is in the snpid
and that the snpid has the format chromosome:other_data.
Default value is c(1L, 2L, 3L, 4L, 5L).}

\item{startcolumn}{Column number of first column with
genetic probabilities or dosages. Must
be an integer value. Default value is 6L.}

\item{impformat}{Number of genetic data values per
subject. 1 indicates dosage only, 2 indicates P(g=0)
and P(g=1) only, 3 indicates P(g=0), P(g=1), and
P(g=2). Default value is 3L.}

\item{chromosome}{Chromosome value to use if the
first value of the snpcolumns is equal to 0.
Default value is character().}

\item{header}{Indicators if the gen and sample files
have headers. If the gen file does not have a
header. A sample file must be included.
Default value is c(FALSE, TRUE).}

\item{gz}{Indicator if file is compressed using gzip.
Default value is FALSE.}

\item{sep}{Separator used in the gen file. Default
value is \code{"\\t"}}

\item{bdfiles}{Vector of names of the output files.
The binary dosage file name is first. The family and
map files follow. For format 4, no family and map file
names are needed.}

\item{format}{The format of the output binary dosage file.
Allowed values are 1, 2, 3, and 4. The default value is 4.
Using the default value is recommended.}

\item{subformat}{The subformat of the format of the output
binary dosage file. A value of 1 or 3 indicates that only the
dosage value is saved. A value of 2 or 4 indicates
the dosage and genetic probabilities will be output. Values
of 3 or 4 are only allowed with formats 3 and 4. If a value
of zero if provided, and genetic probabilities are in the vcf
file, subformat 2 will be used for formats 1 and 2, and
subformat 4 will be used for formats 3 and 4. If the vcf file
does not contain genetic probabilities, subformat 1 will be
used for formats 1 and 2, and subformat 3 will be used for
formats 3 and 4. The default value is 0.}

\item{snpidformat}{The format that the SNP ID will be saved as.
-1 - SNP ID not written.
0 - same as in the VCF file.
1 - chr:pos.
2 - chr:pos:ref:alt.
If snpidformat is 1 and the VCF file uses format 2, an error is
generated. Default value is 0.}

\item{bdoptions}{Character array containing any of the following
value, "aaf", "maf", "rsq". The presence of any of these
values indicates that the specified values should be
calculates and stored in the binary dosage file. These values only
apply to format 4.}
}
\value{
None
}
\description{
Routine to read information from a gen file and create
a binary dosage file. Note: This routine can take a long
time to run if the gen file is large.
}
\examples{
# Find the gen file names
gen3afile <- system.file("extdata", "set3a.imp", package = "BinaryDosage")
gen3asample <- system.file("extdata", "set3a.sample", package = "BinaryDosage")
# Get temporary output file name
bdfiles <- tempfile()
# Convert the file
gentobd(genfiles = c(gen3afile, gen3asample),
        snpcolumns = c(0L, 2L:5L),
        bdfiles = bdfiles)
# Verify the file was written correctly
bdinfo <- getbdinfo(bdfiles = bdfiles)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/apply.R
\name{genapply}
\alias{genapply}
\title{Apply a function to each SNP in a
gen, impute2, file}
\usage{
genapply(geninfo, func, ...)
}
\arguments{
\item{geninfo}{List with information about the
gen, impute2, file returned from \link{getgeninfo}}

\item{func}{A user supplied function to apply
to the data for each snp. The function must be
provide with the following parameters, dosage,
p0, p1, and p2, where dosage is the dosage values
for each subject and p0, p1, and p2 are the
probabilities that a subject has zero, one,
and two copies of the alternate allele,
respectively.}

\item{...}{Additional parameters needed by the
user supplied function}
}
\value{
A list with length equal to the number
of SNPs in the vcf file. Each element
of the list is the value returned by the user
supplied function
}
\description{
A routine that reads in the SNP data serially
from a gen file and applies a user
specified function to the data.
}
\examples{
# Get information about a gen, impute2, file

gen1afile <- system.file("extdata", "set1a.imp", package = "BinaryDosage")
geninfo <- getgeninfo(genfiles = gen1afile,
                      snpcolumns = c(1L, 3L, 2L, 4L, 5L),
                      header = TRUE)


aaf <- genapply(geninfo = geninfo,
                func = BinaryDosage:::getaaf)
}
\seealso{
Other Iterating functions: 
\code{\link{bdapply}()},
\code{\link{vcfapply}()}
}
\concept{Iterating functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format5.R
\name{openbd5con}
\alias{openbd5con}
\title{Open a persistent connection to a Format 5 binary dosage file}
\usage{
openbd5con(bd5info)
}
\arguments{
\item{bd5info}{Object returned by \code{getbdinfo}.}
}
\value{
An object of class \code{"bd5con"} to be passed to
\code{getbd5snp_con} and \code{closebd5con}.
}
\description{
Opens the .bdose file for reading and returns an object that holds the
connection open across multiple calls to \code{getbd5snp_con}.  The
connection is closed automatically when the object is garbage-collected
or when R exits; call \code{closebd5con} to close it explicitly.
}
